/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef __BaseFactory_hh__
#define __BaseFactory_hh__

#include <bias_config.h>


#include <Base/Debug/Debug.hh>
#include <Base/Debug/Error.hh>
#include <Base/Debug/Exception.hh>

#include <map>
#include <vector>
#include <list>
#include <algorithm>
#include <string>
#include <iostream>


#define ALLOCATOR_IMPL(BASE_NAME, CLASS_NAME)\
inline static BASE_NAME *Allocator()\
{ return dynamic_cast<BASE_NAME *>(new CLASS_NAME); }

namespace BIAS {

#define FACTORY_INVALID_TAG ""

#define D_BF_ADD  0x00000001
#define D_BF_GEN  0x00000002

  template<class BaseClassType> class BaseFactory;

  /** @brief prints Map_ to stream
      @relates BaseFactory()
      @author woelk */
  template<class BaseClassType>
  std::ostream &operator<<(std::ostream& os, 
                           const BaseFactory<BaseClassType>& bf);

  typedef std::string TagType;

  ///////////////////////////////////////////////////////////////////////////
  // BaseFactory
  ///////////////////////////////////////////////////////////////////////////

  /** @class BaseFactory
      @brief basic factory class
       
      Instantiate this class with the desired base class as template argument.
      The base class must provide the functions 
        - std::string BaseClassType::GetTag() const
      The instantiation will fail otherwise.

      /code 
      class Base : public BaseFactoryClass
      {
      public:
        Base(const string& name)
          : Name_(name) {}
        void Who() { cout << "I am "<<Name_<<endl; }
        inline std::string GetTag() const { return Name_; }
        std::string Name_;
      };

      class A : public Base
      {
      public:
        A() : Base("a") {};
        ALLOCATOR_IMPL(Base, A);
      };


      class B : public Base
      {
      public:
        B() : Base("b") {};
        ALLOCATOR_IMPL(Base, B);
      };

      class MyFactory
      : public Factory<Base>
      {
      public:
        MyFactory(): Factory<Base>()
        {
          AddType((&A::Allocator));
          AddType((&B::Allocator));
        }
      };

      MyFactory fac;
      Base *p = fac.Generate("a");

      /endcode

      @author woelk 02/2006 */
  template<class BaseClassType>
  class BaseFactory
    : public Debug
  {
  public:
    /// typedef for function pointer to class allocator function
    typedef BaseClassType* (*AllocatorFunctionPointer)();

    /// struct for storing class details in memeber map Map_
    struct MapEntry {
      /// pointer to the static allocator function of this class
      AllocatorFunctionPointer AllocatorFunction_;
    };

    /// constructor
    BaseFactory();

    /// destructor
    virtual ~BaseFactory();

    /** @brief function for generic class generation from
        @returns pointer to class on success, NULL on error */
    BaseClassType *Generate(const TagType& tag) const;

    void GetNearTag(const TagType& tag, TagType& t1, TagType& t2);

    /** @returns number of tags on success, negative value on error */
    int GetKnownTags(std::vector<TagType>& tags) const;

    /** @brief add a new class entry to map
        @returns 0 on success, negative value on error */
    virtual void AddType(AllocatorFunctionPointer allocFunc);

    /** @brief empties the map */
    inline void Clear()
    { Map_.clear(); }

    friend std::ostream &operator<<<>(std::ostream& os, 
                                      const BaseFactory<BaseClassType>& bf);

  protected:

    /// @brief checks if tag is already in map
    inline bool IsInMap_(const TagType& tag) const
    { return (Map_.find(tag) != Map_.end()); }

  protected:
    //private: // private to avoid DLL export 
    /// tag - class correspondences are stored in this map
    std::map<TagType, MapEntry> Map_;

  }; // class BaseFactory

  ///////////////////////////////////////////////////////////////////////////
  // implementation
  ///////////////////////////////////////////////////////////////////////////

  template<class BaseClassType>
  BaseFactory<BaseClassType>::BaseFactory()
    : Debug()
  {
    //AddDebugLevel(D_BF_GEN);
    //AddDebugLevel(D_BF_ADD);
  }

  template<class BaseClassType>
  BaseFactory<BaseClassType>::~BaseFactory()
  {
    Map_.clear();
  }

  template<class BaseClassType>
  BaseClassType *BaseFactory<BaseClassType>::Generate(const TagType& tag) const
  {
    BaseClassType *pointer=NULL;
    if (tag == ""){
      BIASERR("class has invaltag tag "<<tag);
    } else {
      typename std::map<TagType, MapEntry>::const_iterator el = Map_.find(tag);
      if (el != Map_.end()){
        // call the allocator function
        pointer = (*(el->second.AllocatorFunction_))();
      } else {
        BIASCDOUT(D_BF_GEN, *this << std::endl <<" cannot generate "<<tag<<std::endl);
        //BEXCEPTION("unknown class tag "<<tag);
      }
    }
    return pointer;
  }

  template<class BaseClassType>
  void BaseFactory<BaseClassType>::GetNearTag(const TagType& tag,
                                              TagType& tagupper,
                                              TagType& taglower) {
    tagupper = taglower = tag;
    // safety check if there is anything in the map
    if (Map_.empty()) return;
    typename std::map<TagType, MapEntry>::const_iterator el = Map_.find(tag);
    if (el != Map_.end()){
      // exact match !
      tagupper = taglower = tag;
    } else {
      // init with first
      tagupper =  Map_.begin()->first;
      taglower =  Map_.rbegin()->first;
      // generate element
      Map_[tag];
      // find neighbors
      typename std::map<TagType, MapEntry>::iterator it = Map_.find(tag);
      it++;
      if (it != Map_.end()) taglower = it->first;
      it--;
      if (it != Map_.begin()) {
        it--;
        tagupper = it->first;
        it++;
      }
      // cleanup synthetic element
      Map_.erase(it);
    }   
  }


  template<class BaseClassType>
  int BaseFactory<BaseClassType>::
  GetKnownTags(std::vector<TagType>& tags) const
  {
    tags.clear();
  
    typename std::map<TagType, MapEntry>::const_iterator it;
    for (it=Map_.begin(); it!=Map_.end(); it++){
      tags.push_back(it->first);
    }

    return (int)tags.size();
  }


  template<class BaseClassType>
  void BaseFactory<BaseClassType>::
  AddType(AllocatorFunctionPointer allocFunc)
  {
    BaseClassType *cl = (*allocFunc)();
    if (!cl){
      BEXCEPTION("allocator function is not working!");
    }
    if (cl->GetTag()==FACTORY_INVALID_TAG){
      BIASERR("tag corresponds to FACTORY_INVALID_TAG, class type not added "
              <<cl->GetTag());
      BEXCEPTION("cannot add FACTORY_INVALID_TAG");
    }
    if (IsInMap_(cl->GetTag())){
      BIASERR("factory: \n"<< *this <<std::endl);
      BIASERR("tag already in factory class "<<cl->GetTag());
      BEXCEPTION("tag already in factory class "<<cl->GetTag());
    } 
    BIASCDOUT(D_BF_ADD, "BaseFactory: adding "<<cl->GetTag());
    struct MapEntry me = { allocFunc };
    Map_.insert(make_pair(cl->GetTag(), me));
    
    delete cl;
  }


  template<class BaseClassType>
  std::ostream &operator<<(std::ostream& os,
                           const BaseFactory<BaseClassType>& bf)
  {
    typename std::map<TagType, typename BaseFactory<BaseClassType>::MapEntry>::
      const_iterator it;
    for (it=bf.Map_.begin(); it!=bf.Map_.end(); it++){
      os <<"tag :"<<std::setw(4)<<(it->first)<<std::endl;
    }
    return os;
  }

}

#endif // __BaseFactory_hh__
