/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIASDEBUG_HH__
#define __BIASDEBUG_HH__

#include <bias_config.h>
#include <Base/Common/BIASpragmaStart.hh>

#include <iostream>
#include <iomanip>
#include <string>
#include <stdio.h>




/** \ingroup g_utils
    \file Debug.hh
    \brief some macros, classes and varaibles for debugging purposes:

    The class variables _liDebugLevel and _zDebugStream 
      must exist for the macros to operate. Usually they are inherited
      from the Debug class defined below

    _liDebuglevel is a bitfield. It is set to 0 in the constr. of Debug
      The bits should be addressed via a #define macros from the header file 
      of the acording class.

    _zDebugStream is a ostream. It is set to cerr in the constructor of 
      Debug 

    BIASDOUT(flag, stream)
       prints stream to cerr if flag is set in _liDebuglevel

    BIASDOUT_THREAD is the aequivalent to BIASDOUT for the use within a thread

    BIASGDOUT(level, stream)
       prints out debug information if BIASGlobalDebugLevel is greater than 
       level. It can only be used together with the global debug level.

    The global debuglevel variable 'BIASGlobalDebugLevel' is defined in
      MainDebug.hh. It should be included only in the application.

    The global debug level 'BIASGlobalDebugLevel' should be defined in every
      executable via the include of <Debug/MainDebug.hh>.

    A function ResetDebugLevelFromGlobal() can be implemented in classes that
      need it. See Examples...

    An exhaustive example exists.

    \author Felix Woelk
**/



#ifdef BIAS_DEBUG


#define BIASDOUT(flag, arg)\
    { if ((this->_liDebugLevel & flag)!=0){\
        std::ostream _cods_(std::cout.rdbuf()); \
        this->GetDebugStream(_cods_); _cods_ << #flag << ", " << __FILE__ \
                      << ", " << __LINE__ << " : " << arg << std::endl; } }

#define BIASFLUSHDOUT \
    { std::ostream _cods_(std::cout.rdbuf()); \
    this->GetDebugStream(_cods_); _cods_.flush(); }

#define BIASDOUT_THREAD(debuglevel, debugstream, flag, arg)\
    if ((debuglevel & flag)!=0)\
        debugstream << #flag << ", " << __FILE__ << ", " << __LINE__ << " : "\
                    << arg << std::endl;

#define BIASGDOUT(level, arg)\
	if (Debug::GlobalDebugLevel >= level)\
       std::cerr << arg << std::endl;

#define BIASDOUT_CLEAN(flag, arg)\
    { if ((this->_liDebugLevel & flag)!=0){\
        std::ostream _cods_(std::cout.rdbuf()); \
        this->GetDebugStream(_cods_); _cods_ << arg << std::endl; } }

#define BIASCDOUT(flag, arg)\
    { if ((this->_liDebugLevel & flag)!=0){\
        std::ostream _cods_(std::cout.rdbuf()); \
        this->GetDebugStream(_cods_); _cods_ << arg; } }


#else
#define BIASDOUT_CLEAN(flag, arg) {}
#define BIASDOUT(flag, arg) {}
#define BIASFLUSHDOUT {}
#define BIASDOUT_THREAD(debuglevel, debugstream, flag, arg) {}
#define BIASGDOUT(level, arg) {}
#define BIASCDOUT(flag, arg) {}



#endif // BIAS_DEBUG


// TODO: Please use BIASABORT instead! 
#ifdef WIN32
#define ABORT cout << "Please hit enter to terminate program!" << endl << flush; \
        getchar(); \
        exit(-1);
#else
#define ABORT abort();
#endif


/// this returns the type as a string, e.g. PRINTTYPE(int) -> "int"
/// useful when debugging templates
/// on WIN32 thsi can be done more pretty with typeid( *ptr ).name() (jw)
#define PRINTTYPE(thetype)  (\
  (typeid(thetype)==typeid(char))? "char": \
  (typeid(thetype)==typeid(signed char))?"signed char":\
  (typeid(thetype)==typeid(unsigned char))?"unsigned char":\
  (typeid(thetype)==typeid(short))? "short": \
  (typeid(thetype)==typeid(signed short))? "signed short": \
  (typeid(thetype)==typeid(unsigned short))?"unsigned short":\
  (typeid(thetype)==typeid(int))? "int":\
  (typeid(thetype)==typeid(unsigned int))?"unsigned int":\
  (typeid(thetype)==typeid(signed int))?"signed int":\
  (typeid(thetype)==typeid(double))?"double":\
  (typeid(thetype)==typeid(float))?"float":\
  (typeid(thetype)==typeid(long))?"long":\
  "UnknownStorageType")


namespace BIAS{

  /* this variable os the global debuglevel and should be defined in 
     every executable via include of <Debug/MainDebug.hh> */
  //extern long int GlobalDebugLevel;

  
  /** \class Debug
      \ingroup g_utils
      \file Debug.hh
      \brief some macros, classes and varaibles for debugging purposes
           if a class is derived from Debug, then use BIASDOUT() instead of
           cerr/cout for debugging purposes

      father of many classed to includes flags based debug information 

      The class variables _liDebugLevel and _zDebugStream 
      must exist for the macros to operate. Usually they are inherited
      from the Debug class defined below

      _liDebuglevel is a bitfield. It is set to 0 in the constr. of Debug
      The bits should be addressed via a #define macros from the header file 
      of the acording class.

      _zDebugStream is a ostream. It is set to cerr in the constructor of 
      Debug 

      - BIASDOUT(flag, stream)
        prints stream to cerr if flag is set in _liDebuglevel

      - BIASDOUT_THREAD is the aequivalent to BIASDOUT for the use within a thread
      - BIASGDOUT(level, stream)
        prints out debug information if BIASGlobalDebugLevel is greater than 
        level. It can only be used together with the global debug level.

      - BIASDOUT_CLEAN(flag, stream)
        is similar to BIASDOUT except that no flag,file and line is printed
   
        
      The global debuglevel variable 'BIASGlobalDebugLevel' is defined in
      Debug.cpp.

      A function ResetDebugLevelFromGlobal() can be implemented in classes that
      need it. See Examples...

      An exhaustive example exists.

      \author Felix Woelk
  **/

  class BIASDebug_EXPORT Debug{

  public:
    inline Debug() 
    { 
#ifdef BIAS_DEBUG
      _liDebugLevel = 0;
#endif
    };

    inline Debug(Debug& deb)
    {
#ifdef BIAS_DEBUG
      _liDebugLevel = deb.GetDebugLevel();
#endif
    }

    inline ~Debug() {};

    inline void SetDebugLevel(long int lv)
    { 
#ifdef BIAS_DEBUG
      _liDebugLevel = lv; 
#endif     
    };
    
    inline int GetDebugLevel() const
    { 
#ifdef BIAS_DEBUG     
      return _liDebugLevel; 
#else
      return 0;
#endif
    }; 

    inline bool DebugLevelIsSet(long int lv) const
    {
#ifdef BIAS_DEBUG     
      return ((_liDebugLevel & lv)!=0); 
#else
      return false;
#endif
    };       
    
    inline void AddDebugLevel(long int lv)
    {
#ifdef BIAS_DEBUG
      _liDebugLevel = _liDebugLevel | lv; 
#endif     
    };

    inline void RemoveDebugLevel(long int lv)
    {
#ifdef BIAS_DEBUG
      _liDebugLevel = _liDebugLevel & (~lv); 
#endif     
    };

    inline void PrintDebugLevel(std::ostream& os=std::cout) const
    {
#ifdef BIAS_DEBUG
#define WS sizeof(long int)*8
      os << "0x" << std::hex << std::setw(sizeof(long int)<<1) 
         << std::setfill('0') << GetDebugLevel() << std::dec << " = ";
      for (unsigned i=1; i<=WS; i++){
        os << !!((1<<(WS-i))&GetDebugLevel());
        if ((i%4)==0) os << " ";
        if ((i%8)==0) os << " ";
      }
      os << std::endl;
#endif     
    }

    inline void SetDebugStream(std::ostream& os)
    {
#ifdef BIAS_DEBUG
      _zDebugStream.rdbuf(os.rdbuf());
#endif
    }

    inline std::ostream& GetDebugStream() 
    {
#ifdef BIAS_DEBUG
      return _zDebugStream;
#else
      return std::cout;
#endif
    }

    inline void GetDebugStream(std::ostream& os) const
    {
#ifdef BIAS_DEBUG
      
      os.rdbuf(_zDebugStream.rdbuf());
#else
      os.rdbuf(std::cout.rdbuf());
#endif
    }

    inline static void SetGlobalDebugLevel(long int lev)
    {
#ifdef BIAS_DEBUG
      GlobalDebugLevel = lev;
#endif
    }

    inline static long int GetGlobalDebugLevel()
    {
#ifdef BIAS_DEBUG
      return GlobalDebugLevel;
#else
      return 0;
#endif
    }
#ifdef BIAS_DEBUG
    inline Debug& operator=(const Debug& d)
    { 
      _liDebugLevel = d._liDebugLevel; 
      _zDebugStream.rdbuf(d._zDebugStream.rdbuf());
      return *this;
    }
#endif
  protected:
#ifdef BIAS_DEBUG   
    long int _liDebugLevel;
    static std::ostream _zDebugStream;

    static long int GlobalDebugLevel;
#endif

  };
} // namespace

#include <Base/Common/BIASpragmaEnd.hh>

#endif // __BIASDEBUG_HH__
