#include "PoseParametrizationCovariance.hh"
#include <Base/Debug/Exception.hh>
#include <Base/Common/CompareFloatingPoint.hh>
#include <Base/Geometry/PoseParametrization.hh>
#include <Base/Math/Matrix4x4.hh>

using namespace BIAS;

PoseParametrizationCovariance::
PoseParametrizationCovariance()
  : Matrix<POSECOVARIANCE_TYPE>(7,7)
{}

PoseParametrizationCovariance::
~PoseParametrizationCovariance()
{}

void PoseParametrizationCovariance::
SetFromEulerAngles(const Matrix<double>& cov_pose,
                   const Vector<double>& pose)
{
  if (cov_pose.num_cols() != 6 || cov_pose.num_rows() != 6){
    BEXCEPTION("invalid argument "<<cov_pose);
  }
  if (pose.size() != 6 ){
    BEXCEPTION("invalid argument "<<pose);
  }
  //BIASERR("this is only correct when the trace of the rotation matrix is "
  //          "greater than zero");
  RMatrixBase R;
  R.SetXYZ(pose[3], pose[4], pose[5]);
  if (R.Trace()<0.0){
    BEXCEPTION("trace of rotation matrix must > 0.0");
  }

  Matrix<double> J(7, 6, MatrixIdentity);

  /** matlab code:
      alpha = sym('alpha');
      beta = sym('beta');
      gamma = sym('gamma');
      % compute rotation matrix from euler angles
      R_gamma = [ cos(gamma) -sin(gamma) 0 
      sin(gamma) cos(gamma) 0
      0 0 1];
      R_beta = [ cos(beta) 0 sin(beta)
      0 1 0
      -sin(beta) 0 cos(beta) ];
      R_alpha = [ 1 0 0
      0 cos(alpha) -sin(alpha)
      0 sin(alpha) cos(alpha) ];
      R = R_gamma * R_beta * R_alpha;
      % compute quaternion q from rotation matrix
      q = [ 1/2 * sqrt(R(1,1) + R(2,2) + R(3,3) + 1)
      (R(3,2)-R(2,3))/(2*sqrt(R(1,1) + R(2,2) + R(3,3) + 1))
      (R(1,3)-R(3,1))/(2*sqrt(R(1,1) + R(2,2) + R(3,3) + 1))
      (R(2,1)-R(1,2))/(2*sqrt(R(1,1) + R(2,2) + R(3,3) + 1)) ];
      % compute jacobian
      j = jacobian(q, [alpha, beta, gamma]);
      disp('J[0][0] = ');
      disp(j(1,1));
      disp('J[1][0] = ');
      disp(j(2,1));
      disp('J[2][0] = ');
      disp(j(3,1));
      disp('J[3][0] = ');
      disp(j(4,1));
      disp('J[0][1] = ');
      disp(j(1,2));
      disp('J[1][1] = ');
      disp(j(2,2));
      disp('J[2][1] = ');
      disp(j(3,2));
      disp('J[3][1] = ');
      disp(j(4,2));
      disp('J[0][2] = ');
      disp(j(1,3));
      disp('J[1][2] = ');
      disp(j(2,3));
      disp('J[2][2] = ');
      disp(j(3,3));
      disp('J[3][2] = ');
      disp(j(4,3)); */
  double alpha = pose[3], beta = pose[4], gamma = pose[5];
  // I guess this should be optimized
  J[0][0] = 
    1/4/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)*(-cos(gamma)*sin(alpha)+sin(gamma)*sin(beta)*cos(alpha)-cos(beta)*sin(alpha));
  J[1][0] = 
    1/2*(cos(beta)*cos(alpha)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(cos(beta)*sin(alpha)+cos(gamma)*sin(alpha)-sin(gamma)*sin(beta)*cos(alpha))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1),(3.0/2.0))*(-cos(gamma)*sin(alpha)+sin(gamma)*sin(beta)*cos(alpha)-cos(beta)*sin(alpha));
  J[2][0] = 
    1/2*(sin(gamma)*cos(alpha)-cos(gamma)*sin(beta)*sin(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(sin(gamma)*sin(alpha)+cos(gamma)*sin(beta)*cos(alpha)+sin(beta))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1),(3.0/2.0))*(-cos(gamma)*sin(alpha)+sin(gamma)*sin(beta)*cos(alpha)-cos(beta)*sin(alpha));
  J[3][0] = 
    1/2*(-sin(gamma)*sin(alpha)-cos(gamma)*sin(beta)*cos(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(sin(gamma)*cos(beta)+sin(gamma)*cos(alpha)-cos(gamma)*sin(beta)*sin(alpha))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1),(3.0/2.0))*(-cos(gamma)*sin(alpha)+sin(gamma)*sin(beta)*cos(alpha)-cos(beta)*sin(alpha));
  J[0][1] = 
    1/4/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)*(-cos(gamma)*sin(beta)+sin(gamma)*cos(beta)*sin(alpha)-sin(beta)*cos(alpha));
  J[1][1] = 
    1/2*(-sin(beta)*sin(alpha)-sin(gamma)*cos(beta)*cos(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(cos(beta)*sin(alpha)+cos(gamma)*sin(alpha)-sin(gamma)*sin(beta)*cos(alpha))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1),(3.0/2.0))*(-cos(gamma)*sin(beta)+sin(gamma)*cos(beta)*sin(alpha)-sin(beta)*cos(alpha));
  J[2][1] = 
    1/2*(cos(gamma)*cos(beta)*cos(alpha)+cos(beta))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(sin(gamma)*sin(alpha)+cos(gamma)*sin(beta)*cos(alpha)+sin(beta))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1), (3.0/2.0))*(-cos(gamma)*sin(beta)+sin(gamma)*cos(beta)*sin(alpha)-sin(beta)*cos(alpha));
  J[3][1] = 
    1/2*(-sin(gamma)*sin(beta)-cos(gamma)*cos(beta)*sin(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(sin(gamma)*cos(beta)+sin(gamma)*cos(alpha)-cos(gamma)*sin(beta)*sin(alpha))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1), (3.0/2.0))*(-cos(gamma)*sin(beta)+sin(gamma)*cos(beta)*sin(alpha)-sin(beta)*cos(alpha));
  J[0][2] = 
    1/4/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)*(-sin(gamma)*cos(beta)-sin(gamma)*cos(alpha)+cos(gamma)*sin(beta)*sin(alpha));
  J[1][2] = 
    1/2*(-sin(gamma)*sin(alpha)-cos(gamma)*sin(beta)*cos(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(cos(beta)*sin(alpha)+cos(gamma)*sin(alpha)-sin(gamma)*sin(beta)*cos(alpha))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1),(3.0/2.0))*(-sin(gamma)*cos(beta)-sin(gamma)*cos(alpha)+cos(gamma)*sin(beta)*sin(alpha));
  J[2][2] = 
    1/2*(cos(gamma)*sin(alpha)-sin(gamma)*sin(beta)*cos(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(sin(gamma)*sin(alpha)+cos(gamma)*sin(beta)*cos(alpha)+sin(beta))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1),(3.0/2.0))*(-sin(gamma)*cos(beta)-sin(gamma)*cos(alpha)+cos(gamma)*sin(beta)*sin(alpha));
  J[3][2] = 
    1/2*(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha))/sqrt(cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1)-1/4*(sin(gamma)*cos(beta)+sin(gamma)*cos(alpha)-cos(gamma)*sin(beta)*sin(alpha))/pow((cos(gamma)*cos(beta)+cos(gamma)*cos(alpha)+sin(gamma)*sin(beta)*sin(alpha)+cos(beta)*cos(alpha)+1),(3.0/2.0))*(-sin(gamma)*cos(beta)-sin(gamma)*cos(alpha)+cos(gamma)*sin(beta)*sin(alpha));

  *this = J * cov_pose * J.Transpose();
  
}


void PoseParametrizationCovariance::
Transform(const Matrix4x4<double>& transf,
          const PoseParametrization& pose)
{
  Matrix4x4<double> T = transf;
  // normalize the transformation such that the last row reads 0 0 0 1
  //if ( !Equal(T[3][3], 1.0) ) T /= T[3][3];
  if ( !Equal(T[3][0], 0.0) || !Equal(T[3][1], 0.0) || 
       !Equal(T[3][2], 0.0) || !Equal(T[3][3], 1.0) ) {
    BEXCEPTION("unsupported transformation format (only affine transformations"
               " supported): "<<T<<"\n"<<transf);
  }

  /** The seven dimensional pose p is transformed nonlinearily as follows
      f_(0..2) (p) = T[0..2][0..2] * p[0..2]
      f_3 (p) = 0
      f_(4..6) (p) = T[0..2][0..2] * 1/sin(acos(p[3])) * p[4..6]
      explanation:
      The pose is composed of the position (p[0..2]) and the orientation 
      parametrized as a quaternion (p[3..6]). The transformation of the 
      position is straight forward, i.e. by multiplying with T. The Jacobian of
      this transformation is aequivalent to the upper left 3x3 matrix of T.
      The union quaternion representing a rotation about the axis [x, y, z]' 
      with angle a can be interpreted as 
      q = [ cos(a/2), x* sin(*a/2), y*sin(a/2), z*sin(a/2)]' 
      and the transformation of the quaternion is aequivalent to the transf. 
      of the axis [x, y, z] = 1/(sin(acos(q[0])) * q[1..3])
      hint: d/dx (1/(sin(acos(x))) = x/(1-x^2)^3/2 */
  Matrix<double> J(7, 7, MatrixZero);
  double q0 = pose.GetOrientation()[0];
  for (int y=0; y<3; y++){
    for (int x=0; x<3; x++){
      J[y][x] = J[y+4][x+4] = T[y][x];
    }
    J[y+4][3] = q0 / pow( (1.0 - q0*q0), 1.5);
  }
  *this = J * *this * J.Transpose();
}
