/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_Random_hh__
#define __BIAS_Random_hh__

#include <Base/Common/BIASpragmaStart.hh>

#include <ctime>
#include <vector>

#include <Base/Debug/Error.hh>
#include <Base/Debug/Debug.hh>
#include "Vector2.hh"
#include "Matrix2x2.hh"


namespace BIAS{
/** @class Random
    @ingroup g_mathalgo
    @brief class for producing random numbers from different distributions
       
    IMPORTANT: Problems arise when multiple of these instances are used
         since every call to Reset() results in a call to srand() (from 
         stdlib) with a specific seed. Calling srand() is a global action 
         that will affect all other instances as well 
         
    @author woelk 2002 */
  class BIASMathBase_EXPORT Random : public Debug
  {  
  public:
    /** initializes vars/seed with deterministic values, NOT calling reset */
    Random();

    /** calls Reset(seed) and therefore srand() */
    Random(long int seed);
    
    ~Random();

    /** calls srand() with a seed generated from system call time, 
       also initializes some other variables */
    void Reset();

    /** calls srand with seed, also initializes some other variables */
    void Reset(long int seed);

    /** on succesive calls return normal distributed random variable
        with mean and standard deviation sigma */
    inline double GetNormalDistributed(double mean, double sigma)
    { return (mean + (gasdevf_(&LongIdum_) * sigma)); };

    /** calculates num normal ditributed random variables and stores
        them in res */
    void GetNormalDistributed(double mean, double sigma, 
                              int num, double *res); 

    /** on succesive calls return uniform distributed random variable
        between min and max */
    inline double GetUniformDistributed(double min, double max)
    { return (min + (ran2_(&LongIdum_) * (max-min))); };

    /** calculates num uniform ditributed random variables and stores
        them in res */
    void GetUniformDistributed(double min, double max, 
                               int num, double *res); 

    /** get uniform distributed random variable 
        including min/max */
    int GetUniformDistributedInt(int min, int max);

    /** get uniform distributed random variable 
        including min/max */
    unsigned int GetUniformDistributedInt(unsigned int min, unsigned int max);

    /** See GetNormalDistributed().
        @author woelk 07/2005 */
    int GetNormalDistributed(const Vector2<double>& mean, 
                             const Matrix2x2<double>& cov,
                             Vector2<double>& ran);

    /** Generates a vector of length num containing 2D (dependent) normal 
        distributed random variables. The 2D distribution is described by the 
        covariance matrix cov.

        @returns 0 on success , negative on error
        @author woelk 07/2005 */
    int GetNormalDistributed(const Vector2<double>& mean, 
                             const Matrix2x2<double>& cov,
                             int num, std::vector<Vector2<double> >& ran);
    
  protected:
    unsigned int Seed_;
    int Idum_;
    long LongIdum_;
    /// produces uniform random variable between 0 and 1 exclusive
    double ran1_(int *idum); // from nr pdf
    double ran1c_(int *idum); // from D_RAN1.c
    double ran2_(long *idum); // from D_RAN2.c
    /// produces normal distributed random variable with unit variance and
    /// zero mean
    double gasdev_(int *idum);
    // from makenois.c
    double gasdev3_(int *idum);
    // selbst gebastelt
    double gasdevf_(long *idum);
  };


} // namespace

#include <Base/Common/BIASpragmaEnd.hh>

#endif // __Random_hh__

