/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef WIN32
#  include "Base/Common/W32Compat.hh"
#endif

#include <algorithm>
#include "CornerDetectorGradient.hh"
#include <Base/Image/ImageConvert.hh>

#ifdef BIAS_DEBUG
#include <Base/Image/ImageIO.hh> 
#endif

  //#define TIME_MEASURE
#ifdef TIME_MEASURE
#include <Base/Debug/TimeMeasure.hh>
#endif

#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;

//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////


template <class StorageType, class CalculationType>
CornerDetectorGradient<StorageType, CalculationType>::
CornerDetectorGradient()
  : CornerDetectorBase<StorageType>(),
    _st(),
    _gauss()
{
  _FeatureMap=NULL;
  _MinDistance= new int;
  *_MinDistance = 10;
  _MinCornerness= new double;
  *_MinCornerness = 1.0;
  _DoRefinement = new bool;
  *_DoRefinement = false;
}

template <class StorageType, class CalculationType>
CornerDetectorGradient<StorageType, CalculationType>::
CornerDetectorGradient(Param& para, const std::string& postfix)
  : CornerDetectorBase<StorageType>(para, postfix),
    _st(para, "-CornerDetectorGradient"+postfix),
    _gauss(para,"-CornerDetectorGradient"+postfix)
{
  _FeatureMap=NULL;
  int grp=para.GetFreeGroupID();
  _AddParam(para, grp, "-CornerDetectorGradient"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp,"CornerDetectorGradient");
}

template <class StorageType, class CalculationType>
CornerDetectorGradient<StorageType, CalculationType>::
CornerDetectorGradient(Param& para, const int grp, const std::string& postfix)
  :CornerDetectorBase<StorageType>(para,grp,postfix),
   _st(para, grp, "-CornerDetectorGradient"+postfix),
   _gauss(para, grp, "-CornerDetectorGradient"+postfix)
{
  _FeatureMap=NULL;
  _AddParam(para, grp, "-CornerDetectorGradient"+postfix);
}

template <class StorageType, class CalculationType>
CornerDetectorGradient<StorageType, CalculationType>::~CornerDetectorGradient()
{_DeleteInternalMem();
 if(this->ParamsAllocated_){ 
   delete _MinDistance;   _MinDistance   = NULL;
   delete _MinCornerness; _MinCornerness = NULL;
   delete _DoRefinement;  _DoRefinement  = NULL;
 }
}

template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
Detect(const Image<StorageType>& image, std::vector<HomgPoint2D>& p, 
       std::vector<QUAL>& quality, std::vector<Matrix2x2<double> > *cov) {  
  Image<StorageType> fim;
  _gauss.Filter(image, fim);
#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_CD_WRITE_DEBUG_IM)){
    Image<StorageType> im2=image;
    ImageIO::Write("float-input.mip", im2);
    ImageIO::Write("gauss-filtered.mip", fim);
  }
#endif

  Image<CalculationType> *kim = NULL;
  if (typeid(StorageType)==typeid(CalculationType)) {
    kim = (Image<CalculationType> *)&fim;
  } else {
    kim = new Image<CalculationType>(fim.GetWidth(), fim.GetHeight(),
                                     fim.GetChannelCount());
    ImageConvert::ConvertST(fim, *kim, kim->GetStorageType());
  }
  int sres=this->_st.CalcStructureTensor(*kim, this->_sgxx, this->_sgxy, 
                                         this->_sgyy);
  if (typeid(StorageType)!=typeid(CalculationType)) delete kim;

  if (this->_sgxx.GetWidth()>this->_Cornerness.GetWidth() ||
      this->_sgxx.GetHeight()>this->_Cornerness.GetHeight()){
    _AllocInternalMem(this->_sgxx.GetWidth(), this->_sgxx.GetHeight());
  }

  int mres = _ComputeCornerness(this->_Cornerness);
  this->_ZeroFeatureMap();
  // do we have old corners where no new ones are desired ?
  if (p.size()>0) this->_FillFeatureMap(p);
  int eres = this->_EnforceMinimumDistance(p, quality);
  
  // compute the covariances
  if (cov){
    GetCov_(p, *cov);
  }

  return (sres==0 && mres==0 && eres==0)?(0):-1;
}

template <class StorageType, class CalculationType>
void CornerDetectorGradient<StorageType, CalculationType>::
GetCov_(const vector<HomgPoint2D>& p, vector<Matrix2x2<double> >& cov,
        const double noise) const
{
  cov.resize(p.size());
  const int num = (int)p.size();
  int x, y, ires;
  const double square_noise = noise * noise;
  Matrix2x2<double> c;
  for (int i=0; i<num; i++){
    x = (int)rint(p[i][0]);
    y = (int)rint(p[i][1]);
    BIASASSERT(x>=0 && y>=0 && 
               x<(int)_sgxx.GetWidth() && y<(int)_sgxx.GetHeight());
    c[0][0] = _sgxx.GetImageDataArray()[y][x];
    c[1][1] = _sgyy.GetImageDataArray()[y][x];
    c[1][0] = c[0][1] = _sgxy.GetImageDataArray()[y][x];
    ires = c.Invert(cov[i]);
    cov[i] *= square_noise;
    if (ires!=0) { ABORT; }
  }
}

template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
Detect(const Image<CalculationType>& gradx, 
       const Image<CalculationType>& grady, 
       std::vector<HomgPoint2D>& p,std::vector<QUAL>& quality, 
       std::vector<Matrix2x2<double> > *cov)
{
#ifdef TIME_MEASURE
  TimeMeasure timer2;
  timer2.Start();
#endif  

#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_CD_WRITE_DEBUG_IM)){
    Image<CalculationType> im2=gradx;
    ImageIO::Write("cd-gradx.mip", im2);
    im2=grady;
    ImageIO::Write("cd-grady.mip", im2);
  }
  BIASASSERT(gradx.SamePixelAndChannelCount(grady));
#endif


#ifdef TIME_MEASURE
  timer2.Stop();
  cerr << "allocating  took "<<timer2.GetRealTime()<<" us\n";
  timer2.Reset();
  timer2.Start();
#endif  

  int sres = 0;
  // create instances of gradienttype images, so we can compare against 
  // storagetype
  sres=this->_st.CalcStructureTensor(gradx, grady, this->_sgxx, 
                                     this->_sgxy, this->_sgyy);

#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_CD_WRITE_DEBUG_IM)){
    ImageIO::Write("cd-sgxx.mip", _sgxx);
    ImageIO::Write("cd-sgxy.mip", _sgxy);
    ImageIO::Write("cd-sgyy.mip", _sgyy);
  }
#endif

#ifdef TIME_MEASURE
  timer2.Stop();
  cerr << "CalcStructureTensor  took "<<timer2.GetRealTime()<<" us\n";
  timer2.Reset();
  timer2.Start();
#endif  

  if (this->_sgxx.GetWidth()>this->_Cornerness.GetWidth() ||
      this->_sgxx.GetHeight()>this->_Cornerness.GetHeight()){
    _AllocInternalMem(this->_sgxx.GetWidth(), this->_sgxx.GetHeight());
  }
  int mres = _ComputeCornerness(this->_Cornerness);

#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_CD_WRITE_DEBUG_IM)){
    ImageIO::Write("cd-cornerness.mip", _Cornerness);
  }
#endif

#ifdef TIME_MEASURE
  timer2.Stop();
  cerr << "_ComputeCornerness  took "<<timer2.GetRealTime()<<" us\n";
  timer2.Reset();
  timer2.Start();
#endif  

  this->_ZeroFeatureMap();
  if (p.size()>0) this->_FillFeatureMap(p);
  int eres = this->_EnforceMinimumDistance(p, quality);

#ifdef TIME_MEASURE
  timer2.Stop();
  cerr << "EnforceMinimumDistance  took "<<timer2.GetRealTime()<<" us\n";
  timer2.Reset();
  timer2.Start();
#endif  

  // compute the covariances
  if (cov){
    GetCov_(p, *cov);
  }

  return (sres==0 && mres==0 && eres==0)?(0):-1;
}

template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
Cornerness(const Image<CalculationType>& gradx, 
           const Image<CalculationType>& grady, 
           Image<CalculationType>& cornerness) 
{
  int ires, meres;
  if (!gradx.SamePixelAndChannelCount(this->_sgxx)){
    _AllocInternalMem(gradx.GetWidth(), gradx.GetHeight());
  }

  if ((ires=this->_st.CalcStructureTensor(gradx, grady, this->_sgxx, 
                                         this->_sgxy, this->_sgyy))!=0){
    BIASERR("error calculating structure tensor");
  }

  if (!cornerness.SamePixelAndChannelCount(gradx)){
    if (!cornerness.IsEmpty()) cornerness.Release();
    cornerness.Init(gradx.GetWidth(), gradx.GetHeight(), 1);
  }

  
  meres = _ComputeCornerness(cornerness);

  return (ires==0 && meres==0)?0:-1;
}

template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
DetectFromCornerness(const Image<CalculationType>& cornerness, 
                     std::vector<HomgPoint2D>& p, std::vector<QUAL>& quality)
{
  // fill feature map from cornerness
  int tlx, tly, brx, bry;
  cornerness.GetROI()->GetCorners(tlx, tly, brx, bry);
  const int minx=tlx;
  const int miny=tly;
  const int maxx=brx;
  const int maxy=bry;
  Feat fp; this->_FeatList.clear(); 
  this->_FeatList.reserve((brx-tlx)*(bry-tly));
  register CalculationType **ime=cornerness.GetImageDataArray();
  register const double me=(double)*this->_MinCornerness;
  double ev;

  register int x, y;
  for (y=miny; y<maxy; y++){
    for (x=minx; x<maxx; x++){
      ev=ime[y][x];
      if (ev>me){
        fp.x=x;
        fp.y=y;
        fp.val=ev;
        this->_FeatList.push_back(fp);
      }
    }
  }
  // now do non maximum supression
  this->_ZeroFeatureMap();
  if (p.size()>0) this->_FillFeatureMap(p);
  int res = this->_EnforceMinimumDistance(p, quality);

  return res;
}


template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
DetectFromCornerness(const Image<CalculationType> &cornerness, 
		     const std::vector<Vector2<int> > &region,
		     std::vector<HomgPoint2D> &p, 
		     std::vector<QUAL> &quality)
{
#ifdef BIAS_DEBUG
  BIASASSERT( p.size() == quality.size() );
#endif

  CalculationType **imagedatacorn = cornerness.GetImageDataArray();
  const double mincorn = (double)*this->_MinCornerness;

  // setup feature list
  this->_FeatList.clear(); 
  this->_FeatList.reserve(region.size());

  // fill feature list with features of min cornerness
  Feat fp; 
  unsigned i;
  for (i = 0; i < region.size(); i++) {
    const int x = region[i][0];
    const int y = region[i][1];
    const double corn = imagedatacorn[y][x];
    
    if ( corn > mincorn ) {
      fp.x   = x;
      fp.y   = y;
      fp.val = corn;
      
      this->_FeatList.push_back(fp);
    }    
  }

  // init feature map, mask out given features by 'p'
  this->_ZeroFeatureMap();
  if ( p.size() > 0) 
    this->_FillFeatureMap(p);

  int res = this->_EnforceMinimumDistance(p, quality);

  return res;
}

//////////////////////////////////////////////////////////////////////////
// protected functions
//////////////////////////////////////////////////////////////////////////

template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
_EnforceMinimumDistance(vector<HomgPoint2D>& points, vector<QUAL>& qual)
{
#ifdef TIME_MEASURE
  TimeMeasure timer;
  timer.Start();
#endif

  //for (int i=0; i<50; i++) cout << i<< ": "<< _FeatList[i] << endl;
  //int maxnum=(_InternalMemWidth*_InternalMemHeight)/8;
  //partial_sort(_FeatList.begin(), _FeatList.begin()+maxnum, _FeatList.end());
  sort(_FeatList.begin(), _FeatList.end());
  //  for (int i=0; i<50; i++) cout << i<< ": "<< _FeatList[i] << endl;

#ifdef TIME_MEASURE
  timer.Stop();
  cerr << "Quicksort of "<<_FeatList.size()<<" took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif
  
  // clear points and qual
  points.clear();
  qual.clear();

  int minx, maxx, miny, maxy, mx, my;
  register int x, y;
  register double mineig;
  register vector<class Feat>::iterator p=_FeatList.begin();
  const int md=*_MinDistance-1;
  const int width=_Cornerness.GetWidth();
  const int height=_Cornerness.GetHeight();
  const int mwidth=width-md;
  const int mheight=height-md;
  HomgPoint2D hp;

  // 1 point in every md by md pixel patch
  int num=width*height/(md*md);
  points.reserve(num);
  qual.reserve(num);

  //BIASASSERT(md>=0);

  // fill _FeatureMap and points
  vector<class Feat>::iterator it;
  for (it=_FeatList.begin(); it!=_FeatList.end(); it++){
    x=it->x;
    y=it->y;
    mineig=it->val; 
    if (!_FeatureMap[y][x]) {
      // fill feature map
      miny=(y>md)?(y-md):(0);
      maxy=(y>=mheight)?(height-1):(y+md);
      minx=(x>md)?(x-md):(0);
      maxx=(x>=mwidth)?(width-1):(x+md);
      for (my=miny; my<=maxy; my++){
        for (mx=minx; mx<=maxx; mx++){
          _FeatureMap[my][mx]=true;
        }
      }
      // add point to vector
      hp.Set((double)(x), (double)(y), 1.0);
#ifdef BIAS_DEBUG
      BIASASSERT(!isnan(hp[0]) && !isinf(hp[0]));
      BIASASSERT(!isnan(hp[1]) && !isinf(hp[1]));
      if (this->DebugLevelIsSet(D_CD_FEATURES) && points.size()<15)
        cerr << points.size() << " : ("<<x<<", "<<y<<" ) qual "<<mineig<<endl;
#endif
      points.push_back(hp);
      qual.push_back((QUAL)mineig);
      
      // check if we have found enough corners:
      if ((++_NumberFeaturesSoFar >= (*(this->_MaxNum))) && 
          (*this->_MaxNum>0)) break;
    }
  } 

  // refine all selected corners
  if (*_DoRefinement) _RefineCorners(points, qual);
  
#ifdef BIAS_DEBUG
  for (unsigned i=0; i<points.size(); i++){
    BIASASSERT(!isnan(points[i][0]) && !isinf(points[i][0]));
    BIASASSERT(!isnan(points[i][1]) && !isinf(points[i][1]));
  }
#endif  


#ifdef TIME_MEASURE
  timer.Stop();
  cerr << "EnforceMinimumDistance without Quicksort took "
       <<timer.GetRealTime()<<" us\n";
#endif
  return 0;
}


template <class StorageType, class CalculationType>
bool CornerDetectorGradient<StorageType, CalculationType>::
_RefineCornerPosition(int col, int row, 
                     double &x, double &y, QUAL& NewCornerness) {
  CalculationType **cornerness = _Cornerness.GetImageDataArray();
  // cornerness values of 8-neighbourhood
  const register CalculationType& DR_P11 = cornerness[row-1][col-1];
  const register CalculationType& DR_P12 = cornerness[row-1][col  ];
  const register CalculationType& DR_P13 = cornerness[row-1][col+1];
  const register CalculationType& DR_P21 = cornerness[row  ][col-1];
  const register CalculationType& DR_P22 = cornerness[row  ][col  ];
  const register CalculationType& DR_P23 = cornerness[row  ][col+1];
  const register CalculationType& DR_P31 = cornerness[row+1][col-1];
  const register CalculationType& DR_P32 = cornerness[row+1][col  ];
  const register CalculationType& DR_P33 = cornerness[row+1][col+1];
  NewCornerness = (float)DR_P22;

  // Create weighting coefficients for the quadratic fit. They are obtained by
  // applying these masks to the 3x3 window centered on pixel (row,col) :
  float b,c,d,e,f;

  // 1/9 * [ -1 +2 -1 ]
  //       [ +2 +5 +2 ]
  //       [ -1 +2 -1 ]
  // dont need "a" here

  // 1/6 * [ -1  0 +1 ]
  //       [ -1  0 +1 ]
  //       [ -1  0 +1 ]
  b = (-DR_P11 +  DR_P13 -  DR_P21 +  
       DR_P23 -  DR_P31 +  DR_P33)/6.0f;

  // 1/6 * [ -1 -1 -1 ]
  //       [  0  0  0 ]
  //       [ +1 +1 +1 ]
  c = (-DR_P11 - DR_P12 - DR_P13 +  
       DR_P31 + DR_P32 + DR_P33)/6.0f;

  // 1/3 * [ +1 -2 +1 ]
  //       [ +1 -2 +1 ]
  //       [ +1 -2 +1 ]
  d = (DR_P11 + DR_P13 + 
       DR_P21 + DR_P23 +
       DR_P31 + DR_P33
       -2.0f*(DR_P12+DR_P22+DR_P32))/3.0f;

  // 1/4 * [ +1  0 -1 ]
  //       [  0  0  0 ]
  //       [ -1  0 +1 ]
  e = (DR_P11 - DR_P31 -  
       DR_P13 + DR_P33)/4.0f;

  // 1/3 * [ +1 +1 +1 ]
  //       [ -2 -2 -2 ]
  //       [ +1 +1 +1 ]
  f = (   DR_P11 +  DR_P12 +  DR_P13 
          -2.0f*(DR_P21+DR_P22+DR_P23)
          +DR_P31 +  DR_P32 +  DR_P33)/3.0f;

  // Next find the extremum of the fitted surface by setting its
  // partial derivatives to zero. We need to solve the following linear system:
  //
  //  [ d e ] [ off_x ] + [ b ] = [ 0 ]      (dS/dx = 0)
  //  [ e f ] [ off_y ]   [ c ]   [ 0 ]      (dS/dy = 0)
  //
  // This implies that the fitted surface is
  // S(x,y) = a + b x + c y + 1/2 d x^2 + e x y + 1/2 f y^2
  // where we don't need to know the value of a.
  
  float det = d*f - e*e;
  if (det > 0.0) {
    float off_x = (c*e - b*f) / det;
    float off_y = (b*e - c*d) / det;

    // more than one pixel away
    if (fabs (off_x) > 1.0 || fabs (off_y) > 1.0)
      return false;
    else {
      //cout<<"Succesfully refined point. old was "<<x<<" "<<y<<"with q="
      //    <<NewCornerness;
      x = col+off_x;
      y = row+off_y;
      // now compute a because we need it for the cornerness value:
      // 1/9 * [ -1 +2 -1 ]
      //       [ +2 +5 +2 ]
      //       [ -1 +2 -1 ]
      float a = (-     DR_P11 +2.0f *DR_P12 -     DR_P13
                 +2.0f*DR_P21 +5.0f *DR_P22 +2.0f*DR_P23
                 -     DR_P31 +2.0f *DR_P32 -     DR_P33)/9.0f;
      
      NewCornerness = float( a + b*off_x + c*off_y + 0.5*d*off_x*off_x 
        + e*off_x*off_y + 0.5*f*off_y*off_y );
      //cout<<" New is  "<<x<<" "<<y<<"with q=" <<NewCornerness <<endl;
      return true;
    }
  }
  // it's a saddle surface, but the corner may be good anyway :
  else if (det < 0.0) {
    x = col;
    y = row;
    return true;
  }
  
  // det is zero !
  return false;
}


template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
_ComputeCornerness(Image<CalculationType>& i) {
  BIASERR("Make instances of derived class, dont call this !!!");
  BIASABORT;
  return 0;
}

template <class StorageType, class CalculationType>
void CornerDetectorGradient<StorageType, CalculationType>::
_FillFeatureMap(vector<HomgPoint2D>& p)
{
  vector<HomgPoint2D>::iterator pit;
  const int md=*_MinDistance-1;
  const int width=_Cornerness.GetWidth();
  const int height=_Cornerness.GetHeight();
  const int mwidth=width-md;
  const int mheight=height-md;
  int minx, maxx, miny, maxy, mx, my, x, y;
  for (pit=p.begin(); pit!=p.end(); pit++){
    if (!pit->IsAtInfinity()){ 
      _NumberFeaturesSoFar++;
      x=(int)rint((*pit)[0]);
      y=(int)rint((*pit)[1]);
      // fill feature map
      miny=(y>md)?(y-md):(0);
      maxy=(y>=mheight)?(height-1):(y+md);
      minx=(x>md)?(x-md):(0);
      maxx=(x>=mwidth)?(width-1):(x+md);
      for (my=miny; my<=maxy; my++){
        for (mx=minx; mx<=maxx; mx++){
          _FeatureMap[my][mx]=true;
        }
      }
    }
  }
}

template <class StorageType, class CalculationType>
void CornerDetectorGradient<StorageType, CalculationType>::
_DeleteInternalMem()
{
  if (!_Cornerness.IsEmpty()) _Cornerness.Release(); 
  
  if (_FeatureMap!=NULL) { 
    delete[] _FeatureMap[0]; _FeatureMap[0]=NULL;
    delete[] _FeatureMap; _FeatureMap=NULL; 
  }
}

template <class StorageType, class CalculationType>
void CornerDetectorGradient<StorageType, CalculationType>::
_AllocInternalMem(const int width, const int height)
{
  _DeleteInternalMem();
  int size=width*height;
  _Cornerness.Init(width, height, 1);
//   _sgxx.Init(width, height, 1);
//   _sgxy.Init(width, height, 1);
//   _sgyy.Init(width, height, 1);
  _FeatureMap = new bool*[height];
  _FeatureMap[0] = new bool[size];
  for (int i=1; i<height; i++) {
    _FeatureMap[i]=_FeatureMap[i-1]+width;
  }
  _FeatList.reserve(size);
}



template <class StorageType, class CalculationType>
void CornerDetectorGradient<StorageType, CalculationType>::
_AddParam(Param& para, const int grp, const std::string& postfix) {
  if (para.CheckParam("MinDistance"+postfix)) { 
    _MinDistance=para.GetParamInt("MinDistance"+postfix);
  } else { 
    _MinDistance=para.AddParamInt("MinDistance"+postfix, 
                                  "minimum distance between corners", 
                                  10, 0, INT_MAX, 0, grp);
  }

  if (para.CheckParam("MinCornerness"+postfix)){ 
    _MinCornerness=para.GetParamDouble("MinCornerness"+postfix);
  } else { 
    _MinCornerness=para.AddParamDouble("MinCornerness"+postfix, 
                                       "threshold to accept point as corner", 
                                       1.0, 0.0, DBL_MAX, 0, grp);
  }

  if (para.CheckParam("SubpixelAccuracy"+postfix)){ 
    _DoRefinement = para.GetParamBool("SubpixelAccuracy"+postfix);
  } else { 
    _DoRefinement =
      para.AddParamBool("SubpixelAccuracy"+postfix, 
                        "refine (int) pixel position by quad. surface fitting",
                        false, 0, grp);
  }
}



template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
_CalcCornerness(const Image<StorageType>& image, 
                Image<CalculationType>& cornerness)
{
  Image<StorageType> fim;
  _gauss.Filter(image, fim);
#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_CD_WRITE_DEBUG_IM)){
    Image<StorageType> im2=image;
    ImageIO::Write("float-input.mip", im2);
    ImageIO::Write("gauss-filtered.mip", fim);
  }
#endif


  Image<CalculationType> *kim = NULL;
  if (typeid(StorageType)==typeid(CalculationType)) {
    kim = (Image<CalculationType> *) &fim;
  } else {
    kim = new Image<CalculationType>(fim.GetWidth(), fim.GetHeight(),
                                     fim.GetChannelCount());
    ImageConvert::ConvertST(fim, *kim, kim->GetStorageType());
  }
  int sres=this->_st.CalcStructureTensor(*kim, this->_sgxx, this->_sgxy, 
                                         this->_sgyy);
  if (typeid(StorageType)!=typeid(CalculationType)) delete kim;

  if (this->_sgxx.GetWidth()>this->_Cornerness.GetWidth() ||
      this->_sgxx.GetHeight()>this->_Cornerness.GetHeight()){
    _AllocInternalMem(this->_sgxx.GetWidth(), this->_sgxx.GetHeight());
  }


  int mres = _ComputeCornerness(cornerness);
  return (sres==0 && mres==0)?0:-1;
}



template <class StorageType, class CalculationType>
int CornerDetectorGradient<StorageType, CalculationType>::
_CalcCornerness(const Image<CalculationType>& gradx, 
                const Image<CalculationType>& grady, 
                Image<CalculationType>& cornerness) {
  int ires, meres;
  if ((ires=this->_st.CalcStructureTensorValid(gradx, grady, 
                                               this->_sgxx, this->_sgxy, 
                                               this->_sgyy))!=0){
    BIASERR("error calculatimng structure tensor");
  }
  if (!cornerness.SamePixelAndChannelCount(gradx)){
    if (!cornerness.IsEmpty()) cornerness.Release();
    cornerness.Init(gradx.GetWidth(), gradx.GetHeight(), 1);
  }
  meres = _ComputeCornerness(cornerness);

  return (ires==0 && meres==0)?0:-1;
}

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////


template class CornerDetectorGradient<unsigned char, float>; 
template class CornerDetectorGradient<float, float>; 

// fill in instances as required
#ifdef BUILD_IMAGE_INT
template class CornerDetectorGradient<unsigned char, int>; 
#endif

#ifdef BUILD_IMAGE_CHAR
template class CornerDetectorGradient<unsigned char, char>; 
#endif

#ifdef BUILD_IMAGE_SHORT
#endif

#ifdef BUILD_IMAGE_USHORT
#ifdef BUILD_IMAGE_INT
template class CornerDetectorGradient<unsigned short, int>; 
#endif
template class CornerDetectorGradient<unsigned short, float>; 
#endif

#ifdef BUILD_IMAGE_UINT
#endif

#ifdef BUILD_IMAGE_DOUBLE
#endif
