/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef _BIAS_MOPRH_DILATION_HH_
#define _BIAS_MOPRH_DILATION_HH_
#include "./Morphology.hh"

namespace BIAS {

/** @class Dilation
    @ingroup g_filter
    @brief Dilation operator for binary images (black and white)
      
    Dilation sets all pixel to 255, which have a neighboring pixel !=0
    It is also possible to dilate the image with a real grey-value image with
    Dilate3Fast().
    Fastest ist TBH_valid and kernelsize=3 
    or call Dilate3Fast() directly. 
    @author grest, Oct. 2004
 */

template <class InputStorageType, class OutputStorageType> 
class BIASFilter_EXPORT Dilation : public Morphology<InputStorageType, OutputStorageType>
{
 public:
  Dilation() {};
  virtual ~Dilation() {};
  
  /** Does the dilation. ROI ignored, but set if TBH_valid
   */
  virtual int Filter(const Image<InputStorageType>& src, 
                     Image<OutputStorageType>& dst);

   /** dilate with square kernel filled with 255
        @author Felix Woelk (22/04/2002)
        @status alpha */
  int Dilate(const Image<InputStorageType>& Source,
             Image<OutputStorageType>& Destination, 
             int KernelSize = 3);

  /** Very fast dilate with 3x3 mask, all values, which are not
      not zero, are treated as foreground. 
      The src img is dilated with pixels from orig.
      Therefore src and orig MUST NOT be the same,
      set orig to a 255 image if u have no original grey image.
      The border of one pixel is always deleted (set to zero).
      default is 8-neighborhood.
      src and dest must NOT be the same!
      @author Daniel Grest, Sept. 2002
      @status tested */
  int Dilate3Fast(const Image<InputStorageType>& src, 
                  const Image<InputStorageType>& orig,
                  Image<OutputStorageType>& dest,
                  bool Neighbor4=false);
  
  /** Very fast dilate with 3x3 mask, all values, which are not
      not zero, are treated as foreground. 
      src and dest must NOT be the same!
      @author woelk, grest 04/2006 */
  int Dilate3Fast(const Image<InputStorageType>& src, 
                  Image<OutputStorageType>& dest,
                  bool Neighbor4 = false);

 protected:
  virtual void GetBordersValid_(int& border_x, int& border_y) const;

  using Morphology<InputStorageType, OutputStorageType>::_FilterBorderHandling;
  using Morphology<InputStorageType, OutputStorageType>::TBH_valid;
  using Morphology<InputStorageType, OutputStorageType>::TBH_same;
  using Morphology<InputStorageType, OutputStorageType>::TBH_full;
  using Morphology<InputStorageType, OutputStorageType>::kernelSize_;

};

} // namespace
#endif

