/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2004 (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel

  

BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "FilterBase.hh"

#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;



//#define DEBUG_CALLSTACK

template <class InputStorageType, class OutputStorageType>
FilterBase<InputStorageType, OutputStorageType>::FilterBase()
  : Debug()
{
  _FilterBorderHandling = new int;
  *_FilterBorderHandling = TBH_valid;
  _ParamsAllocated = true;
#ifdef DEBUG_CALLSTACK
  AddDebugLevel(D_FILTERBASE_CALLSTACK);
#endif
}

template <class InputStorageType, class OutputStorageType>
FilterBase<InputStorageType, OutputStorageType>::
FilterBase(Param& para, const string& postfix)
  : Debug()
{
  _ParamsAllocated=false;
  int grp = para.GetFreeGroupID();
  _AddParam(para, grp, "-FilterBase"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp, "FilterBase");
#ifdef DEBUG_CALLSTACK
  AddDebugLevel(D_FILTERBASE_CALLSTACK);
#endif
}


template <class InputStorageType, class OutputStorageType>
FilterBase<InputStorageType, OutputStorageType>::
FilterBase(Param& para, const int grp, const string& postfix)
  : Debug()
{
  _ParamsAllocated=false;
  _AddParam(para, grp, "-FilterBase"+postfix);
#ifdef DEBUG_CALLSTACK
  AddDebugLevel(D_FILTERBASE_CALLSTACK);
#endif
}


template <class InputStorageType, class OutputStorageType>
FilterBase<InputStorageType, OutputStorageType>::~FilterBase()
{
  if (_ParamsAllocated) {
    delete _FilterBorderHandling;
    _FilterBorderHandling = NULL;
  }
}


template <class InputStorageType, class OutputStorageType>
void FilterBase<InputStorageType, OutputStorageType>::
_AddParam(Param& para, const int grp, const string& postfix)
{
  vector<string> tt;
  vector<int> tti;
  tt.push_back("TBH_valid"); tti.push_back(TBH_valid);
  tt.push_back("TBH_same"); tti.push_back(TBH_same);
  tt.push_back("TBH_full"); tti.push_back(TBH_full);

  if (para.CheckParam("BorderHandling"+postfix)){ 
    _FilterBorderHandling=para.GetParamEnum("BorderHandling"+postfix);
  } else { 
    _FilterBorderHandling=
      para.AddParamEnum("BorderHandling"+postfix, 
                        "border handling type of filter", 
                        tt, TBH_valid, &tti, 0, grp); 
  }

  para.SetEnhancedFlag("BorderHandling"+postfix, true);
}


template <class InputStorageType, class OutputStorageType>
void FilterBase<InputStorageType, OutputStorageType>::
GetBorders(int& border_x, int& border_y) const
{
  switch(*_FilterBorderHandling){
  case TBH_same:
    border_x=0;
    border_y=0;
    break;
  case TBH_full:
    BIASERR("unfinished");
    BIASABORT;
    break;
  case TBH_valid:
    GetBordersValid_(border_x, border_y);
    break;
  default:
    BIASASSERT("invalid boder handling type" && false);
    BIASABORT;
    break;
  }
}



int *BIAS::AddGradientType(Param& para, const string& name, const int group_id)
{
  vector<string> tt;
  vector<int> tti;
  tt.push_back("GT_Sobel3x3"); 
  tti.push_back(GT_Sobel3x3);
  tt.push_back("GT_Gauss"); 
  tti.push_back(GT_Gauss);
  tt.push_back("GT_GaussAsymmetric"); 
  tti.push_back(GT_GaussAsymmetric);
  //  tt.push_back(""); tti.push_back(GradientBase<StorageType,KernelType>::);
  int *res;

  if (para.CheckParam(name)){ 
    res=para.GetParamEnum(name);
  } else { 
    res=para.AddParamEnum(name, "type of gradient filter to use", tt, 
                          GT_Gauss, &tti, 0, group_id); 
  }
  return res;
}


int *BIAS::AddLowPassType(Param& para, const string& name,  const int group_id)
{

  vector<string> tt;
  vector<int> tti;
  tt.push_back("LPT_Mean"); 
  tti.push_back(LPT_Mean);
  tt.push_back("LPT_Gauss"); 
  tti.push_back(LPT_Gauss);
  tt.push_back("LPT_Binomial"); 
  tti.push_back(LPT_Binomial);
  //  tt.push_back(""); tti.push_back(Gradient<StorageType,KernelType>::);
  int *res;

  if (para.CheckParam(name)){ 
    res=para.GetParamEnum(name);
  } else { 
    res=para.AddParamEnum(name, 
                          "type of mean filter to use", tt, 
                          LPT_Gauss, &tti, 0, group_id); 
  }
  return res;
}

#define FILTER_INSTANTIATION_CLASS FilterBase
#include "Filterinst.hh"
