/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "GradientSimple.hh"
#include "FilterMask.hh"

using namespace BIAS;
using namespace std;

namespace BIAS {

#define POLAR_TABLE_OFFSET 128
#define POLAR_TABLE_SIZE 257
//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////
template <class InputStorageType, class OutputStorageType> void
GradientSimple<InputStorageType, OutputStorageType>::
InitKernel_() 
{
  // this is a filtermask in the sense of convolution
  // it is reflected at its symmetry center before being "dropped"
  // onto the image !
  Vector<FM_INT> h(3), v(1);
  if (useDirectNeighbors_) {
    h[0] = 0;
    h[1] = 1;
    h[2] = -1;
  } else {
    h[0] = 1;
    h[1] = 0;
    h[2] = -1;
  }
  v[0] = 1;
  FilterMask f(h, v, useDirectNeighbors_?0:1, 0);
  f.CreateFloatFilter();
  _Conv.SetKernel(f);
}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
GradientSimple(bool BuildPCM)
  : FilterNTo2N<InputStorageType, OutputStorageType>() ,
    pRadius_(NULL), pAngle_(NULL), useDirectNeighbors_(false)
{
  InitKernel_();
  if (BuildPCM) BuildPolarCoordinateMap_();
}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
GradientSimple(Param & para, const std::string& postfix, bool BuildPCM)
  : FilterNTo2N<InputStorageType, OutputStorageType>
(para, "-GradientSimple"+postfix) ,
    pRadius_(NULL), pAngle_(NULL), useDirectNeighbors_(false)
{
  InitKernel_();
  if (BuildPCM) BuildPolarCoordinateMap_();
}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
GradientSimple(Param & para, const int grp, const std::string& postfix, 
               bool BuildPCM)
  : FilterNTo2N<InputStorageType, OutputStorageType>
(para, grp, "-GradientSimple"+postfix),
    pRadius_(NULL), pAngle_(NULL), useDirectNeighbors_(false)
{
  InitKernel_();
  if (BuildPCM) BuildPolarCoordinateMap_();
}

template <class InputStorageType, class OutputStorageType>
GradientSimple<InputStorageType, OutputStorageType>::
~GradientSimple()
{
  if (pRadius_ != NULL) {
    pRadius_ -= POLAR_TABLE_OFFSET;
    for (int i=0; i<POLAR_TABLE_SIZE; i++) 
      delete [] (pRadius_[i]-POLAR_TABLE_OFFSET);
    delete [] pRadius_;
    pRadius_ = NULL;
  }
  if (pAngle_ != NULL) {
    pAngle_ -= POLAR_TABLE_OFFSET;
    for (int i=0; i<POLAR_TABLE_SIZE; i++) 
      delete [] (pAngle_[i]-POLAR_TABLE_OFFSET);
    delete [] pAngle_;
    pAngle_ = NULL;
  }
}


template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& grad) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::Filter(src, dst)\n");
  Image<OutputStorageType> gy;
  int res=Filter(src, grad, gy);
  grad.AppendChannel(gy);
  return res;
}

template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& gx, 
       Image<OutputStorageType>& gy)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::Filter(src, gx, gy)\n");
  int res=0;
  if (!gx.SamePixelAndChannelCount(src)){
    gx.Release();
    gx.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
  if (!gy.SamePixelAndChannelCount(src)){
    gy.Release();
    gy.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
 
  if (src.GetChannelCount()==1) {
    if (_Conv.CalculationInFloat()){
      if (useDirectNeighbors_)
        res = DirectNeighborsGreyValidFloat_(src, gx, gy);
      else 
        res = SimpleGreyValidFloat_(src, gx, gy);
    } else {
      if (useDirectNeighbors_)
        res = DirectNeighborsGreyValidInt_(src, gx, gy);
      else 
        res = SimpleGreyValidInt_(src, gx, gy);
    }
  } else {
    // grest
    BIASERR("not for multiple channel images");
    return -5;
  }
  return res;
}

template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType,OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& gx, 
       Image<OutputStorageType>& gy, Image<OutputStorageType>& absg) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, 
            "GradientSimple::Filter(src, gx, gy, abs)\n");
  if (!src.SamePixelAndChannelCount(absg)){
    if (!absg.IsEmpty()) absg.Release();
    absg.Init(src.GetWidth(), src.GetHeight(), src.GetChannelCount());
  }
  
  int res, absres=0;
  res = Filter(src, gx, gy);

  absres = VecLen(gx, gy, absg);
  
  return (res==0 && absres==0)?0:-1;
}




template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType, OutputStorageType>::
SimpleGreyValidFloat_(const Image<InputStorageType>& src,
                      Image<OutputStorageType>& gWE,
                      Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::SimpleGreyValidFloat_\n");
 
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
  const int nwidth = - width;
  
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(_Conv.CalculationInFloat());
#endif

  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *ph;
  
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
 
 
  int step=width-maxx+minx;

  offset=minx+1+(miny+1)*width;
  ph = src.GetImageData() + offset;
  pgx=gx+offset;
  pgy=gy+offset;
  step+=2;
  end2=gx+maxx+width*(maxy-2)-1;
  lend2=gx+maxx+width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      *pgx++ = (OutputStorageType)(ph[1] - ph[-1]);
      *pgy++ = (OutputStorageType)(ph[width] - ph[nwidth]);
      ph++;
    }
    ph+=step;
    pgx+=step;
    pgy+=step;
    lend2+=width;
  }
  return 0;
}


template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType, OutputStorageType>::
SimpleGreyValidInt_(const Image<InputStorageType>& src,
                      Image<OutputStorageType>& gWE,
                      Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::SimpleGreyValidInt_\n");
 
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
  const int nwidth = - width;
  
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(!_Conv.CalculationInFloat());
#endif



  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *ph;
 
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
  
 
  int step=width-maxx+minx;

   offset=minx+1+(miny+1)*width;
  ph = src.GetImageData()+offset;
 
  pgx=gx+offset;
  pgy=gy+offset;
  step+=2;
  end2=gx+maxx+width*(maxy-2)-1;
  lend2=gx+maxx+width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      *pgx++ = (OutputStorageType)(ph[1] - ph[-1]);
      *pgy++ = (OutputStorageType)(ph[width] - ph[nwidth]);
      ph++;
    }
    ph+=step;
    pgx+=step;
    pgy+=step;
    lend2+=width;
  }

  return 0;
}


template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType, OutputStorageType>::
DirectNeighborsGreyValidFloat_(const Image<InputStorageType>& src,
                               Image<OutputStorageType>& gWE,
                               Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::SimpleGreyValidFloat_\n");
 
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
  
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(_Conv.CalculationInFloat());
#endif

  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *ph;
  
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
 
 
  int step=width-maxx+minx;

  offset=minx+1+(miny+1)*width;
  ph = src.GetImageData() + offset;
  pgx=gx+offset;
  pgy=gy+offset;
  step+=2;
  end2=gx+maxx+width*(maxy-2)-1;
  lend2=gx+maxx+width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      *pgx++ = (OutputStorageType)(ph[1] - ph[0]);
      *pgy++ = (OutputStorageType)(ph[width] - ph[0]);
      ph++;
    }
    ph+=step;
    pgx+=step;
    pgy+=step;
    lend2+=width;
  }
  return 0;
}




template <class InputStorageType, class OutputStorageType>
int GradientSimple<InputStorageType, OutputStorageType>::
DirectNeighborsGreyValidInt_(const Image<InputStorageType>& src,
                             Image<OutputStorageType>& gWE,
                             Image<OutputStorageType>& gNS)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "GradientSimple::SimpleGreyValidInt_\n");
 
  OutputStorageType *gx=gWE.GetImageData();
  OutputStorageType *gy=gNS.GetImageData();
  const int width=(int)src.GetWidth();
    
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.SamePixelAndChannelCount(gWE));
  BIASASSERT(src.SamePixelAndChannelCount(gNS));
  BIASASSERT(gx!=gy);
  BIASASSERT(!_Conv.CalculationInFloat());
#endif



  int minx, miny, maxx, maxy;
  src.GetROI()->GetCorners(minx, miny, maxx, maxy);
  gWE.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);
  gNS.GetROI()->SetCorners(minx+1, miny+1, maxx-1, maxy-1);

  InputStorageType *ph;
 
  int offset;
  OutputStorageType *pgx, *pgy, *end2, *lend2;
  
 
  int step=width-maxx+minx;

  offset = minx+1+(miny+1)*width;
  ph = src.GetImageData()+offset;
 
  // can compute one or two lines more at the border, not yet implemented
  pgx  = gx+offset;
  pgy  = gy+offset;
  step += 2;
  end2  = gx + maxx + width*(maxy-2)-1;
  lend2 = gx + maxx + width*(miny+1)-1;
  while (pgx<end2){
    while (pgx<lend2){
      *pgx++ = (OutputStorageType)(ph[1] - ph[0]);
      *pgy++ = (OutputStorageType)(ph[width] - ph[0]);
      ph++;
    }
    ph   += step;
    pgx  += step;
    pgy  += step;
    lend2+= width;
  }

  return 0;
}











template <class InputStorageType, class OutputStorageType>
void GradientSimple<InputStorageType, OutputStorageType>::
BuildPolarCoordinateMap_()
{
  BIASASSERT(pRadius_ == NULL);
  BIASASSERT(pAngle_ == NULL);

  pRadius_ = new float* [POLAR_TABLE_SIZE];
  for (int i=0; i<POLAR_TABLE_SIZE; i++)
    pRadius_[i] = new float [POLAR_TABLE_SIZE] + POLAR_TABLE_OFFSET;
  pRadius_ += POLAR_TABLE_OFFSET;
    
  pAngle_ = new float* [POLAR_TABLE_SIZE];
  for (int i=0; i<POLAR_TABLE_SIZE; i++) 
    pAngle_[i] = new float [POLAR_TABLE_SIZE] + POLAR_TABLE_OFFSET;
  pAngle_ += POLAR_TABLE_OFFSET;

  for (int x=-POLAR_TABLE_OFFSET; x<=128; x++) {
    for (int y=-POLAR_TABLE_OFFSET; y<=128; y++) {
      pRadius_[x][y] = sqrtf(float(x*x+y*y));

      pAngle_[x][y]  = -atan2f( float(y), float(x) );
      if (pAngle_[x][y]  < 0.0f)
        (pAngle_[x][y]) += 2.0f * (float)M_PI;
    }
  }   
}

template<>
void GradientSimple<unsigned char, char>::
PolarTransform(Image<char>& gx, Image<char>& gy, 
               Image<float>& magnitude, Image<float>& phase) {
  if (pRadius_ == NULL) BuildPolarCoordinateMap_();
  BIASASSERT(pRadius_ != NULL);
  if (!magnitude.SamePixelAndChannelCount(gx)) {
    if (!magnitude.IsEmpty()) magnitude.Release();
    magnitude.Init(gx.GetWidth(), gx.GetHeight(), 1);
  }
  if (!phase.SamePixelAndChannelCount(gx) ) {
    if (!phase.IsEmpty()) phase.Release();
    phase.Init(gx.GetWidth(), gx.GetHeight(), 1);
  }
  char *pgx = gx.GetImageData(), *pgy = gy.GetImageData();
  float *pm = magnitude.GetImageData(), *pp = phase.GetImageData();
  char* const pEnd = pgx + gx.GetPixelCount();
  while (pgx<pEnd) {
    *pp++ = pAngle_[*pgx][*pgy];
    *pm++ = pRadius_[*pgx++][*pgy++];
  }
}

template <class InputStorageType, class OutputStorageType>
void GradientSimple<InputStorageType,OutputStorageType>::
PolarTransform(Image<OutputStorageType>& gx, Image<OutputStorageType>& gy, 
               Image<float>& magnitude, Image<float>& phase) {
  if (!magnitude.SamePixelAndChannelCount(gx)) {
    if (!magnitude.IsEmpty()) magnitude.Release();
    magnitude.Init(gx.GetWidth(), gx.GetHeight(), 1);
  }
  if (!phase.SamePixelAndChannelCount(gx) ) {
    if (!phase.IsEmpty()) phase.Release();
    phase.Init(gx.GetWidth(), gx.GetHeight(), 1);
  }
  OutputStorageType *pgx = gx.GetImageData(), *pgy = gy.GetImageData();
  float *pm = magnitude.GetImageData(), *pp = phase.GetImageData();
  OutputStorageType* const pEnd = pgx + gx.GetPixelCount();
  while (pgx<pEnd) {
    *pp++ = -atan2f( float(*pgy), float(*pgy) );
    *pm++ = sqrtf(float(*pgx) * float(*pgx) + float(*pgy) * float(*pgy));
    pgx++;
    pgy++;
  }
}


template <class InputStorageType, class OutputStorageType>
void GradientSimple<InputStorageType, OutputStorageType>::
GetBordersValid_(int &border_x, int &border_y) const
{
  _Conv.GetBorders(border_x, border_y);
}

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
#define FILTER_INSTANTIATION_CLASS GradientSimple
#define FILTER_INSTANTIATION_NO_UNSIGNED_OUTPUT
#include "Filterinst.hh"

} // namespace BIAS
