/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __HMatrixEstimation_hh__
#define __HMatrixEstimation_hh__
#include "bias_config.h" 

#include <Geometry/HMatrix.hh>
#include <Base/Geometry/Normalization.hh>

namespace BIAS {

  /** @class HMatrixEstimation
      @ingroup g_geometry
      @brief estimate 3x3 Matrix (homography, affine transformation, ...)

      code was originally in HMatrix, moved and extended by koeser 12/2005
  */
  class BIASGeometry_EXPORT HMatrixEstimation
  {
  public:
    
    explicit HMatrixEstimation(bool normalizehartley=true) {
      normalizeHartley_ = normalizehartley; }

    /** calculatinges H using the basic linear transformation algorithm
        (see Hartley, Zisserman "Multiple View Geometry", p.73)
        p'=Hp -> (homg. coo) [p']x Hp = 0 
        3 row linear system, but rank 2 -> take only first two rows
        p=1...numpoints -> 2*numpoints equations
        solve A*h=0 with SVD

        @author Woelk/Frahm */
    int Compute(const std::vector<BIAS::HomgPoint2D>& frompoints, 
                const std::vector<BIAS::HomgPoint2D>& topoints, 
                HMatrix& H);

    /** @brief optimize a given HMatrix regarding n 2d point correspondences

        minimize an error function on H matrix using iterative 
        nonlinear optimization techniques such as Levenberg/Marquardt
        @attention susceptable to local minima, provide good starting point H
        @param H matrix which is optimized
        @param p1 (homogenized!) points in image 1
        @param p2 corresponding (homogenized!) points in image 2
        @return 0=success, !=0: error
        @author koeser 06/2006 */
    int Optimize(BIAS::HMatrix& H,
                 const std::vector<BIAS::HomgPoint2D> &p1, 
                 const std::vector<BIAS::HomgPoint2D> &p2);







    /** @brief compute affine transformation (not using DLT !)
        @author koeser 12/2005 */
    int ComputeAffine(const std::vector<BIAS::HomgPoint2D>& frompoints, 
                      const std::vector<BIAS::HomgPoint2D>& topoints, 
                      HMatrix& H);

    /** returns the average squared error of the estimation applying the H
        to all correspondences
    */
    double GetError();
  
    /** same as above, different interface, calls Compute(HomgPoints)
        @author grest, May 2006
     */
    int Compute(const std::vector<BIAS::Vector2<double> >& frompoints, 
                const std::vector<BIAS::Vector2<double> >& topoints, 
                HMatrix &H);

    /// pointer to vector of points in image1 (hack for Optimize)
    std::vector<BIAS::HomgPoint2D> const *p1_;
    /// pointer to vector of points in image2 (hack for Optimize)
    std::vector<BIAS::HomgPoint2D> const *p2_;

  protected:

    // exports for dll (cmenk)
    BIAS_EXPORT_STL_VECTOR( BIASGeometry_EXPORT, BIAS::HomgPoint2D )

    /// flag indicating whether we should normalize the given points before
		/// and denormalize the fmatrix after computation, so that it refers to
		/// the original points
		bool normalizeHartley_;
	
		/// the normalization object
		Normalization normalization_;

    HMatrix H_; // the estimated H matrix, needed in GetError()
    std::vector<BIAS::HomgPoint2D> pointsFrom_;
    std::vector<BIAS::HomgPoint2D> pointsTo_;
   
  };


}// namespace

#endif // __HMatrixEstimation_hh__
