/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Geometry/Projection.hh>
#include <Geometry/ProjectionParametersIO.hh>
#include <Geometry/ProjectionParametersPerspective.hh>
#include <Geometry/ProjectionParametersProjective.hh>
#include <Geometry/ProjectionParametersSpherical.hh>
#include <Geometry/ProjectionParametersZoom.hh>
#include <iostream>


using namespace BIAS;
using namespace std;


namespace BIAS {

#ifdef BIAS_HAVE_XML2
int Projection::ReadFromCamParamFile(const string& filename) {

  // read datafile
  CameraData cd, cd2;
  vector<CameraData> vcd;
  if(ProjectionParametersIO::ReadRigData(filename, vcd) != 0)
    return -1;
 
  // initialize parameter vector
  vecParameters_.clear();

  for (unsigned int i=0; i<vcd.size(); i++) {
    ProjectionParametersBase* ppb;
    if (vcd[i].LensIsSpherical_) {
      //initialize fisheye camera
      ProjectionParametersSpherical* pps = new ProjectionParametersSpherical();
      pps->SetRadius(vcd[i].radius_);
      pps->SetUndistortion(vcd[i].AngleCorrX_, vcd[i].AngleCorrY_, 
                           vcd[i].radius_);
      ppb = (ProjectionParametersBase*) pps;
    } else {
      if (vcd[i].zoomCamera_){
        // initialize zoom camera
        ProjectionParametersZoom* ppz = 
          new ProjectionParametersZoom();
        std::vector<CPDiscreteParam>::iterator it;
        ppz->Clear();
        for (unsigned int j=0; j<vcd[i].knownparams_vect_.size(); j++) {
           ppz->AddZoomStep(vcd[i].knownparams_vect_[j].zoom, 
                           vcd[i].knownparams_vect_[j].focallength,
                           vcd[i].knownparams_vect_[j].kc1, 
                           vcd[i].knownparams_vect_[j].kc2, 
                           vcd[i].knownparams_vect_[j].kc3, 
                           vcd[i].knownparams_vect_[j].kc4);
        }
        ppb = (ProjectionParametersBase*) ppz;
      } else {
        // initialize perspective camera
        ProjectionParametersPerspective* ppp = 
          new ProjectionParametersPerspective();
        // convert fl to meter
        ppp->SetFocalLengthAndAspect( vcd[i].focallengthDef_, 
                                      vcd[i].aspectratio_);
        ppp->SetUndistortion(vcd[i].kc1Def_,vcd[i].kc2Def_,
                             vcd[i].kc3Def_,vcd[i].kc4Def_);
        ppb = (ProjectionParametersBase*) ppp;        
      }
    }
    // Set base parameters
    ppb->SetImageSize(vcd[i].width_, vcd[i].height_);
    ppb->SetPrincipal(vcd[i].principalX_, vcd[i].principalY_);
    double eps = 1e-8;
    if ((vcd[i].aspectratio_ < vcd[i].cellSizeX_/vcd[i].cellSizeY_-eps) ||
        (vcd[i].aspectratio_ > vcd[i].cellSizeX_/vcd[i].cellSizeY_+eps)) {
      BIASERR("Aspect ratio and cellsizes do not match!");
      BIASABORT;
    }
    ppb->SetAspectratio(vcd[i].aspectratio_);
    ppb->SetC(vcd[i].PoseInRigCenter_);
    RMatrix R;
    Quaternion<double> Q;
    R.SetXYZ(vcd[i].PoseInRigPitch_, vcd[i].PoseInRigYaw_, 
             vcd[i].PoseInRigRoll_);
    R.GetQuaternion(Q);
    ppb->SetQ(Q);

    vecParameters_.push_back(ppb);
  }

  return 0;
}
#endif // BIAS_HAVE_XML2


void Projection::CreatePerspective(const Pose& pose, 
                                   const KMatrix& K,
                                   int width, int height,
                                   const Vector4<double>& 
                                   UndistortionCoefficients) {

  ProjectionParametersPerspective* ppp = 
    new ProjectionParametersPerspective();

  ppp->SetFocalLengthAndAspect(K.GetFx(), K.GetFy()/K.GetFx());
  ppp->SetUndistortion(UndistortionCoefficients[0],
                       UndistortionCoefficients[1],
                       UndistortionCoefficients[2],
                       UndistortionCoefficients[3]);  
  ppp->SetImageSize(width, height);
  ppp->SetPrincipal(K.GetHx(), K.GetHy());
  ppp->SetPose(pose);
  vecParameters_.push_back((ProjectionParametersBase*)ppp);       
}


HomgPoint2D Projection::Project(const HomgPoint3D &X, unsigned int cam,
                                bool IgnoreDistortion) const
{
  if (cam==0) {
    Vector3<double> x;
    vecParameters_[cam]->GetExternals().Mult(X, x);
    return vecParameters_[cam]->ProjectLocal(x, IgnoreDistortion);
  } 
  BIASERR("To be implemented");
  BIASABORT;
  HomgPoint2D h(0,0,0);
  return h;
}


bool Projection::
DoesPointProjectIntoImage(const HomgPoint3D& X, 
                          HomgPoint2D& x, 
                          unsigned int cam,
                          bool IgnoreDistortion) const
{
  if(cam!=0) {
    BIASERR("Projection check only implemented for cam=0\n");
    return false;
  }
  
  return vecParameters_[cam]->DoesPointProjectIntoImage(X, x, 
                                                        IgnoreDistortion);
}



Vector3<double> Projection::UnProjectToRay(const HomgPoint2D& pos,
                                           unsigned int cam,
                                           bool IgnoreDistortion) const {
  if (cam==0) {
    Vector3<double> x, rx;
    x = vecParameters_[cam]->UnProjectLocal(pos, IgnoreDistortion);
    RMatrix R = vecParameters_[cam]->GetR();
    R.Mult(x, rx);
    return rx;
  } 

  BIASERR("To be implemented");
  BIASABORT;
  HomgPoint2D h(0,0,0);
  return h;
}

   
Vector3<double> Projection::UnProjectToPoint(const HomgPoint2D& pos,
                                             double depth,
                                             unsigned int cam,
                                             bool IgnoreDistortion) const
{
  Vector3<double> opticalray = UnProjectToRay(pos, cam, IgnoreDistortion);
  if(opticalray.NormL2()==0.0) {
    BIASERR("returned ray has zero length!");
    return Vector3<double> (0,0,0);
  }
  opticalray.Normalize();
  return (GetC() + depth*opticalray);
}

Vector3<double> Projection::UnProjectToPointLocal(const HomgPoint2D& pos,
                                                  const double& depth,
                                                  unsigned int camSystem,
                                                  unsigned int cam,
                                                  bool IgnoreDistortion)const {
  if (camSystem != cam) {
    BIASERR("Not implemented.");
    BIASABORT;
  }
  Vector3<double> 
    x(vecParameters_[cam]->UnProjectToPointLocal(pos, depth,
                                                 IgnoreDistortion));
  return x;
}


Matrix3x3<double> Projection::
UnProjectToCovLocal(const HomgPoint2D& pos, const Matrix3x3<double>& cov2D, 
                    const double& /*depth*/, unsigned int camSystem,
                    unsigned int cam, bool IgnoreDistortion) const 
{
  if (camSystem != cam) {
    BIASERR("Not implemented.");
    BIASABORT;
  }
  return vecParameters_[cam]->UnProjectLocal(pos, cov2D, IgnoreDistortion);
}



Vector3<double> Projection::UnProjectToRayLocal(const HomgPoint2D& pos,
                                                unsigned int camSystem,
                                                unsigned int cam,
                                                bool IgnoreDistortion) const {
  if(cam!=0) {
    BIASERR("Not implemented.");
    BIASABORT;
  }
  if (camSystem != cam) {
    BIASERR("Not implemented.");
    BIASABORT;
  }
  Vector3<double> x = vecParameters_[cam]->UnProjectLocal(pos, 
                                                          IgnoreDistortion);  
  return x;
}

#ifdef BIAS_HAVE_XML2

int Projection::XMLGetClassName(string& TopLevelTag,
                                double& Version) const{
  TopLevelTag = "Projection";
  Version = 0.1;
  return 0;
}
   
int Projection::XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const{

  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  if(identifier_.size()>0) {
    theNode = XMLObject.addChildNode(Node, "Identifier");
    XMLObject.addContent(theNode, identifier_);
  }
  std::vector<ProjectionParametersBase* >::const_iterator it;
  for (it=vecParameters_.begin(); it!=vecParameters_.end(); it++) {
    (**it).XMLGetClassName(TopLevelName, Version);
    theNode = XMLObject.addChildNode(Node, TopLevelName);
    XMLObject.addAttribute(theNode, "Version", Version);
    (*it)->XMLOut(theNode, XMLObject);
  }

  return 0;
}

 
int Projection::XMLIn(const xmlNodePtr Node, XMLIO& XMLObject){
  
  //delete old cameras
  for (unsigned int i=0; i<vecParameters_.size(); i++) 
    delete vecParameters_[i];
  vecParameters_.clear();	

  string TopLevelName;
  xmlNodePtr childNode; 
  childNode = XMLObject.getFirstChild(Node);
  while (childNode!=NULL) {
    string nodename = XMLObject.getNodeName(childNode);
    if (nodename.compare("ProjectionParametersPerspective")==0) {
      ProjectionParametersPerspective* pp =  
        new ProjectionParametersPerspective();
      if(pp->XMLIn(childNode, XMLObject)!=0) 
        return -1;
      vecParameters_.push_back(pp);
    }
    if (nodename.compare("ProjectionParametersSpherical")==0) {
      ProjectionParametersSpherical* pp = 
        new ProjectionParametersSpherical();
      if(pp->XMLIn(childNode, XMLObject)!=0)
	return -1;
      vecParameters_.push_back(pp);
    }
    if (nodename.compare("ProjectionParametersProjective")==0) {
      ProjectionParametersProjective* pp = 
        new ProjectionParametersProjective();
      if(pp->XMLIn(childNode, XMLObject)!=0)
	return -1;
      vecParameters_.push_back(pp);
    }
    if (nodename.compare("ProjectionParametersZoom")==0) {
      ProjectionParametersZoom* pp = new ProjectionParametersZoom();
      if(pp->XMLIn(childNode, XMLObject)!=0)
	return -1;
      vecParameters_.push_back(pp);
    }
    if (nodename.compare("Identifier")==0) {
      identifier_ = XMLObject.getNodeContentString(childNode);
    }
    childNode = XMLObject.getNextChild(childNode);
  }
  if (vecParameters_.size()>0) {
    return 0;
  } else {
    return -1;
  }
}

#endif // BIAS_HAVE_XML2


BIAS::Quaternion<double> 
Projection::GetQ(unsigned int cam) const 
{
  BIASASSERT(cam<vecParameters_.size());
  if (cam==0) 
    return vecParameters_[cam]->GetQ();
  else {
    BIASERR("not yet defined behaviour !");
    BIASEXIT("not yet defined behaviour ");
    return Quaternion<double>();
  }
}


} // namespace BIAS
