/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_Projection_hh__
#define __BIAS_Projection_hh__
#include <bias_config.h>

#include <Base/Debug/Debug.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Base/Math/Vector4.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Geometry/Pose.hh>
#include <Geometry/PMatrix.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <Geometry/ProjectionParametersPerspective.hh>


#ifdef BIAS_HAVE_XML2
#  include <Base/Common/XMLBase.hh>
#endif // BIAS_HAVE_XML2


namespace BIAS {


  /** @class Projection

  This class hides the underlying projection model, like projection matrix,
  spherical camera, lens distortion, ...

  */
  class BIASGeometry_EXPORT Projection
#ifdef BIAS_HAVE_XML2
    : public XMLBase
#endif // BIAS_HAVE_XML2
  {
  public:
    Projection() { 
      vecParameters_.clear();
    }


    Projection(const ProjectionParametersBase& p) {
      vecParameters_.clear();
      vecParameters_.push_back(p.Clone());
    }

    Projection(const Projection& p) {
      (*this) = p;
    }

    const Projection& operator=(const Projection& p) {
      // delete old paramters if any
      for (unsigned int i=0; i<vecParameters_.size(); i++)
        delete vecParameters_[i];
      // copy new parameters
      vecParameters_.resize(p.vecParameters_.size());
      for (unsigned int i=0; i<p.vecParameters_.size(); i++) 
        vecParameters_[i] = p.vecParameters_[i]->Clone();
      return *this;
    }

    void CreatePerspective(const BIAS::Pose& pose, 
                           const BIAS::KMatrix& K,
                           int width, int height,
                           const BIAS::Vector4<double>&
                           UndistortionCoefficients = BIAS::Vector4<double>());

#ifdef BIAS_HAVE_XML2
    /** 
     *  @brief Reads a camera data XML file and constructs the 
     *  ProjectionParametersBase-pointer, using the ProjectionParametersIO
     *  and the CameraParameter file format.
     *  @param filename Name of the XML parameter file
     */
    int ReadFromCamParamFile(const std::string& filename);
#endif // BIAS_HAVE_XML2

    bool IsEmpty() const { return (vecParameters_.size()==0);};


    /** @brief parameter read function */
    const ProjectionParametersBase* GetParameters(unsigned int cam = 0) const {
      return vecParameters_[cam];
    } 

    /** @brief parameter write function has to invalidate cache ! */
    ProjectionParametersBase* GetParameters(unsigned int cam = 0) {
      return vecParameters_[cam];
    }

    /** @brief return Center of camera with index cam 
        
    if cam>0 this is a relative pose in the coordinate system of cam 0 */
    inline Vector3<double> GetC(unsigned int cam = 0) const {
      if (cam==0) 
        return vecParameters_[cam]->GetC();
      else {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
        return Vector3<double>(0,0,0);
      }
    };
    
    /** @brief return RMatrix of camera with index cam 
    if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    inline RMatrix GetR(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size());
      if (cam==0) return vecParameters_[cam]->GetR();
      else {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
        return RMatrix();
      }
    };
 

    /** @brief return rotation quaternion of camera with index cam 
        if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    BIAS::Quaternion<double> GetQ(unsigned int cam = 0) const;
   

    /** @brief return const ref to Pose
        @author koeser 02/2006
    */
    inline Matrix3x4<double> GetExternals(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size());
      if (cam==0) { 
        return vecParameters_[cam]->GetExternals();
      } else {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
        return vecParameters_[0]->GetExternals();
      }
    }

    /** @brief Get the Pose covariance
    */
    virtual Matrix<double> GetCov(unsigned int cam = 0) const {
      if (cam==0) return vecParameters_[cam]->GetCov();
      else {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
        return Matrix<double>(7,7);
      }
    }

    /** @brief return complete pose object */
    virtual const Pose& 
    GetPose(unsigned int cam = 0) const {
      if (cam!=0) {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
      }
      return vecParameters_[cam]->GetPose();
    }
  
    /** @brief Set the Pose covariance
    */
    virtual void SetCov(const Matrix<double>& Cov, unsigned int cam = 0) {
      if (cam==0) vecParameters_[cam]->SetCov(Cov);
      else {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
      }
    }

    void SetC(const Vector3<double>& C, unsigned int cam = 0) {
      if (cam==0) vecParameters_[cam]->SetC(C);
      else {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
      }
    }
    /** @brief Set the Pose 
    */
    virtual void SetPose(const BIAS::Pose pose, unsigned int cam = 0) {
      if (cam==0) vecParameters_[cam]->SetPose(pose);
      else {
        BIASERR("not yet defined behaviour !");
        BIASABORT;
      }
    }

    
    /** @brief returns the 2d projection of X in camera cam,
        where X is given in the global coordinate frame (not the rig)    
        (need HomgPoint3D to map points at infinity) 
    */
    HomgPoint2D Project(const HomgPoint3D &X, unsigned int cam = 0,
                        bool IgnoreDistortion = false) const;

    /** Checks if 3D point projects into specified image and returns
     *  belonging 2D image point.
     *  @param X assumes homogenized point!
     *  @author bartczak
     */
    bool DoesPointProjectIntoImage(const BIAS::HomgPoint3D& X, 
                                   BIAS::HomgPoint2D& x, 
                                   unsigned int cam = 0,
                                   bool IgnoreDistortion = false) const;


    /** @brief calculates the viewing ray in the global coordinate frame 
      (not the rig) from the camera center which belongs to the given image 
      position. The resulting vector3 may have any length. 
    */
    Vector3<double> UnProjectToRay(const HomgPoint2D& pos,
                                   unsigned int cam = 0,
                                   bool IgnoreDistortion = false) const;

    /** @brief calculates a 3D point in
        the global (not the rig) coordinate system, which 
        belongs to the image position pos with distance depth to the
        camera center.
    */
    Vector3<double> UnProjectToPoint(const HomgPoint2D& pos,
                                     double depth,
                                     unsigned int cam = 0,
                                     bool IgnoreDistortion = false) const;

    /** @brief calculates a 3D point in
        a local camera coordinate system specified by camSystem, which 
        belongs to the image position pos in cam with distance depth to the
        camera center cam.                
        @author koeser  */
    Vector3<double> UnProjectToPointLocal(const HomgPoint2D& pos,
                                          const double& depth,
                                          unsigned int camSystem = 0,
                                          unsigned int cam = 0,
                                          bool IgnoreDistortion = false) const;


    /** @brief calculate 3D point and associated covariance by unprojection
        @author woelk 03/2006 */
    Matrix3x3<double> UnProjectToCovLocal(const HomgPoint2D& pos,
                                          const Matrix3x3<double>& cov2D, 
                                          const double& depth,
                                          unsigned int camSystem = 0,
                                          unsigned int cam = 0,
                                          bool IgnoreDistortion = false) const;


    

    /** @brief calculates a 3D ray in
        a local camera coordinate system specified by camSystem, which 
        belongs to the image position pos in cam.                
        */
    Vector3<double> UnProjectToRayLocal(const HomgPoint2D& pos,
                                        unsigned int camSystem = 0,
                                        unsigned int cam = 0,
                                        bool IgnoreDistortion = false) const;
    
    virtual ~Projection() {
      for (unsigned int i=0; i<vecParameters_.size(); i++)
        delete vecParameters_[i];
    }

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif // BIAS_HAVE_XML2
    
    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
    */
    inline void Rescale(float ratio, unsigned int cam = 0) {
      vecParameters_[cam]->Rescale(ratio);
    }


    inline void SetIdentifier(const std::string& identString) {
      identifier_ = identString;
    }
    
    inline std::string GetIdentifier() const {
      return identifier_;
    }

    friend std::ostream& operator<<(std::ostream &os, const Projection& p);
    

  protected:

    // exports for dll (cmenk)
    BIAS_EXPORT_STL_VECTOR( BIASGeometry_EXPORT, ProjectionParametersBase* )

    /// the intrinsics (like K, ...) of the camera usually only one
    /// only in rig mode there may be many, in that case vecIntrinsics[0]
    /// holds the external pose while the others hold the internal pose 
    /// (pose in the coordinate system of the first camera)
    std::vector<ProjectionParametersBase* > vecParameters_; 

    /** Multifunctional identifier.
     */
    std::string identifier_;

  };

  inline std::ostream& operator<<(std::ostream &os, const Projection& p) {
    os <<"Identifier: """<<p.identifier_<<"""";
    os <<"vecParameters_.size(): "<<p.vecParameters_.size()<<std::endl;
    std::vector<ProjectionParametersBase* >::const_iterator it;
    for (it = p.vecParameters_.begin(); it!=p.vecParameters_.end(); it++){
      os<< **it;
      if (ProjectionParametersPerspective *pp 
	  = dynamic_cast<ProjectionParametersPerspective *>(*it)) os <<*pp;
    }
    return os;
  }


} // end namespace 


#endif //__BIAS_Projection_hh__

