/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sstream>
#include "Geometry/ProjectionParametersBase.hh" 


using namespace BIAS;
using namespace std;
#define STREAM_PRECISION 12



RMatrix ProjectionParametersBase::GetR() const
{
  RMatrix R;
 // unconst, because RMatrix::SetFromQ() in not const enough
  Quaternion<double> qtmp = GetQ();
  R.SetFromQuaternion(qtmp);
  return R;  
}


#ifdef BIAS_HAVE_XML2

int ProjectionParametersBase::
XMLGetClassName(std::string& TopLevelTag,
                            double& Version) const {
  TopLevelTag = "ProjectionParametersBase";
  Version = 0.2;
  return 0;
}

int ProjectionParametersBase::
XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const {

  xmlNodePtr childNode;
  childNode = XMLObject.addChildNode(Node,"ImageSize");
  XMLObject.addAttribute(childNode, "width", (int)width_);
  XMLObject.addAttribute(childNode, "height", (int)height_);

  childNode = XMLObject.addChildNode(Node,"PrincipalPoint");
  XMLObject.addAttribute(childNode, "x", principalX_);
  XMLObject.addAttribute(childNode, "y", principalY_);

  childNode = XMLObject.addChildNode(Node,"AspectRatio");
  XMLObject.addAttribute(childNode, "val", aspectratio_);

  // R
  Quaternion<double> Q = GetQ();
  stringstream streamR;
  childNode = XMLObject.addChildNode(Node,"Rotation");
  streamR << std::setprecision(STREAM_PRECISION);
  for (unsigned int i=0; i<4; i++) {
    streamR << (double) Q.GetData()[i] <<" ";
  }
  XMLObject.addContent(childNode, streamR.str());

  // C
  Vector3<double> C = GetC();
  childNode = XMLObject.addChildNode(Node,"Center");
  XMLObject.addAttribute(childNode, "x", C[0]);
  XMLObject.addAttribute(childNode, "y", C[1]);
  XMLObject.addAttribute(childNode, "z", C[2]);

  return 0;
}

int ProjectionParametersBase::
XMLIn(const xmlNodePtr Node, XMLIO& XMLObject) {

  xmlNodePtr childNode;
  if ((childNode = XMLObject.getChild(Node, "ImageSize"))!=NULL) {
    width_ = XMLObject.getAttributeValueInt(childNode, "width");
    height_ = XMLObject.getAttributeValueInt(childNode, "height");
  }
  if ((childNode = XMLObject.getChild(Node, "PrincipalPoint"))!=NULL) {
    principalX_ = XMLObject.getAttributeValueDouble(childNode, "x");
    principalY_ = XMLObject.getAttributeValueDouble(childNode, "y");
  }
  if ((childNode = XMLObject.getChild(Node, "CellSize"))!=NULL) {
    double cellSizeX, cellSizeY;
    cellSizeX = XMLObject.getAttributeValueDouble(childNode, "x");
    cellSizeY = XMLObject.getAttributeValueDouble(childNode, "y");
    aspectratio_ = cellSizeX / cellSizeY;
  }
  if ((childNode = XMLObject.getChild(Node, "AspectRatio"))!=NULL) {
    aspectratio_ = XMLObject.getAttributeValueDouble(childNode, "val");
  }
  
  Quaternion<double> Q;
  // R
  if ((childNode = XMLObject.getChild(Node, "Rotation"))!=NULL) {
    // get data fields 
    stringstream ssRotation(XMLObject.getNodeContentString(childNode));
    unsigned int CurrentData = 0;
    while (ssRotation.good() && CurrentData<4) {
      ssRotation >> (Q.GetData()[CurrentData++]);
    }
    if (CurrentData<4) {
      // reached end of stream before expected end of data
      BIASERR("error in xml structure, rotation matrix incomplete.");
      return -1;
    } 
  } else BIASERR("No Q found !!!!");
  // C
  Vector3<double> C;
  if ((childNode = XMLObject.getChild(Node, "Center"))!=NULL) {
    C[0] = XMLObject.getAttributeValueDouble(childNode, "x");
    C[1] = XMLObject.getAttributeValueDouble(childNode, "y");
    C[2] = XMLObject.getAttributeValueDouble(childNode, "z");
  } else BIASERR("No C found !!!!");
  Pose_.Set(Q,C);
  return 0;
}
#endif
  
double ProjectionParametersBase::
ViewDifference(const ProjectionParametersBase*) const {
  // there should be an implmentation in each derived class
  BIASERR("ViewDifference not implemented in derived class");
  BIASABORT;
  // visual studio needs this
  return 0.0;
}

Vector3<double> ProjectionParametersBase
::UnProjectToPointLocal(const HomgPoint2D& pos,
                        const double& depth, bool IgnoreDistortion) const {
  Vector3<double> x(UnProjectLocal(pos, IgnoreDistortion) * depth);
  if (fabs(x.NormL2()-depth)>1e-10) {
    BIASERR("Unprojection failed:" << x << "  Position:"<< pos 
            <<"  Norm:" << setw(10) << double(x.NormL2()));
    abort();
  }
  return x;
}


void ProjectionParametersBase::Transform_(const Vector<double>& src,
                                          Vector<double>& dst) const {
  BIASASSERT(src.size()==3);BIASASSERT(dst.size()==3);
  HomgPoint2D xsrc(src[0], src[1], src[2]), xdst;
  if (ustTransformIntoImage_) {
    xdst = ProjectLocal(xsrc, ustIgnoreDistortion_);
  } else {
    xdst = UnProjectLocal(xsrc, ustIgnoreDistortion_);
  }
  // we do a homogenize here, for the spherical case the covariance
  // may be interesting on the normalized point maybe.
  // For < 160 degrees, this should be no problem.
  xdst.Homogenize();
  dst[0] = xdst[0]; dst[1] = xdst[1]; dst[2] = xdst[2];
}


Matrix3x3<double> ProjectionParametersBase::
UnProjectLocal(const HomgPoint2D& pos, const Matrix3x3<double>& cov,
               bool IgnoreDistortion)
{
  // set transformation parameters using cass members
  ustIgnoreDistortion_ = IgnoreDistortion;
  ustTransformIntoImage_ = false;
  // convert into general data structs
  Matrix3x3<double> rescov(MatrixZero);
  Vector<double> srcPoint(3), resPoint(3);
  srcPoint[0] = pos[0]; srcPoint[1] = pos[1]; srcPoint[2] = pos[2];
  // use unscented transform to approximate covariance
  Transform(srcPoint, cov, resPoint, rescov);
  //cout<<"Cov_orig="<<cov<<" around "<<pos<<" transformed to "
  //    <<rescov<<" around "<<resPoint<<endl;
  return rescov; 
}
