/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersBase_hh__
#define __BIAS_ProjectionParametersBase_hh__
#include "bias_config.h" 

#include <Geometry/Pose.hh>
#include <Base/Common/BIASpragmaStart.hh>
#include <Base/Common/CompareFloatingPoint.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Math/Matrix3x4.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Geometry/PMatrix.hh>
#include <MathAlgo/UnscentedTransform.hh>

#include <Base/Debug/Debug.hh>
#ifdef BIAS_HAVE_XML2
#  include <Base/Common/XMLBase.hh>
#endif

#include <Geometry/RMatrix.hh>

namespace BIAS {

 /** @class ProjectionParametersBase
     @brief camera parameters which define the mapping between rays in the 
     camera coordinate system and pixels in the image as well as external pose
     @author koeser/evers/streckel
 */
  class BIASGeometry_EXPORT ProjectionParametersBase : 
    protected UnscentedTransform
#ifdef BIAS_HAVE_XML2
    , public XMLBase
#endif
  {
  public:

    inline ProjectionParametersBase(){
      width_=0;
      height_=0;
      principalX_=0;
      principalY_=0;
      aspectratio_=1;
      // init with some value:
      ustTransformIntoImage_ = false;
      ustIgnoreDistortion_ = false;
    }
    
    virtual ProjectionParametersBase& operator=(const 
						ProjectionParametersBase& p) {
      Pose_ = p.Pose_;
      width_= p.width_;
      height_= p.height_;
      principalX_= p.principalX_;
      principalY_= p.principalY_;
      aspectratio_= p.aspectratio_;
      return *this;
    }

    virtual ~ProjectionParametersBase(){};

    /** Checks if 3D point projects into specified image and returns
     *  belonging 2D image point.
     *  @param X assumes homogenized point!
     *  @author bartczak
     */
    bool DoesPointProjectIntoImage(const HomgPoint3D& X, 
                                   HomgPoint2D& x, 
                                   bool IgnoreDistortion = false) const
    {
      HomgPoint3D tempX = GetPose().GlobalToLocal(X);
      BIAS::Vector3<double> localX;
      tempX.GetEuclidian(localX);
      return DoesPointProjectIntoImageLocal(localX, x, IgnoreDistortion);
    }
    /** Checks if 3D point projects into specified image and returns
     *  belonging 2D image point.
     *  @param X assumes homogenized point!
     *  @author bartczak
     */
    virtual bool DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                                        HomgPoint2D& x,                       
                                        bool IgnoreDistortion = false) const
    {
      BIASERR("to be implemented in specialized child class!");
      return false;
    }


    /** @brief calculates the projection of a point 
        in the world coordinate system  
        to a pixel in the image plane of the camera 
        Was in old calss Projection before
        @author woelk 05/2006 */
    inline HomgPoint2D Project(const HomgPoint3D &X,
                               bool IgnoreDistortion = false) const {
      Vector3<double> x;
      GetExternals().Mult(X, x);
      return ProjectLocal(x, IgnoreDistortion);
    }

    /** @brief calculates the projection of a point 
        in the local camera coordinate system  
        to a pixel in the image plane of the camera
        
        In the simplest case perspective pinhole projection x = K * point  
        where point is transformed of X using  point = (R^T | -R^T C) X */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const = 0;
    
    /** @brief calculates the unit size viewing ray from the camera center 
        (in the camera coordinate system) which belongs to the image position
        pos e.g. pos=principal point results in (0,0,1)=optical axis */
    virtual Vector3<double> UnProjectLocal(const HomgPoint2D& pos,
                                           bool IgnoreDistortion = false)
      const = 0;
        
    /** @brief unprojects the covariance matrices to K=Identity camera
        
    This uses the unscented transform to propagate covariance. Use a faster
    or better specialization in derived class if you want.
        @author koeser 06/2006 */
    virtual Matrix3x3<double> 
    UnProjectLocal(const HomgPoint2D& pos, const Matrix3x3<double>& cov2D,
                   bool IgnoreDistortion = false);

    /** @brief calculates the viewing ray in the global coordinate frame 
        from the camera center which belongs to the given image 
        position. The resulting vector3 may have any length. 
    
        @author woelk 05/2006 */
    inline Vector3<double> UnProjectToRay(const HomgPoint2D& pos,
                                          bool IgnoreDistortion=false) const {
      Vector3<double> x, rx;
      x = UnProjectLocal(pos, IgnoreDistortion);
      RMatrix R = GetR();
      R.Mult(x, rx);
      return rx;
    }

    /** @brief calculates a 3D point in
        the local camera coordinate system, which 
        belongs to the image position pos in cam with distance depth to the
        camera center cam.                
        @author koeser  */
    virtual Vector3<double> 
    UnProjectToPointLocal(const HomgPoint2D& pos, const double& depth,
                          bool IgnoreDistortion = false) const;


    // ---- Getter methods  ------------------------------------------------
      
    /** @brief get projection center */
    virtual Vector3<double> GetC() const { return Pose_.GetC(); };

    /** @brief return Q of rotation*/
    virtual Quaternion<double> GetQ() const { return Pose_.GetQ(); };

    /** @brief return rotation matrix R*/
    virtual BIAS::RMatrix GetR() const;
    
    /** @brief return pose covariance as 7x7 matrix C,Q */
    virtual Matrix<POSE_TYPE> GetCov() const { return Pose_.GetCov(); }

    /** @brief return complete pose object */
    virtual const Pose& GetPose() const {
      return Pose_;
    }

    inline PoseParametrization GetPoseParametrization() const 
    { return Pose_.GetPoseParameters(); }
    
    inline void SetPoseParametrization(const PoseParametrization& pp)
    { Pose_.Set(pp); }

    /** @brief return cached 3x4 RC representation */
    virtual Matrix3x4<double> GetExternals() const { 
      return Pose_.GetMatrix3x4();}

    /** @brief return image dimensions */
    inline int GetImageSize(unsigned int& Width, unsigned int& Height) const {
      Width  = width_;
      Height = height_;
      return 0;
    }

    /** @brief return aspectratio = cellSizeX_ / cellSizeY_;*/
    inline double GetAspectratio() const {
      return aspectratio_;
    }

    /** @brief get principal point, x in pixels relative to top left corner */
    inline int GetPrincipal(double& PrincipalX, double& PrincipalY) const {
      PrincipalX = principalX_;
      PrincipalY = principalY_;
      return 0;
    }

    // ---- Setter methods  ------------------------------------------------

    
    /** @brief Set CCDaspect ratio = cellSizeX_ / cellSizeY_. 
        @author streckel */
    virtual inline void SetAspectratio(const double AspectRatio) {
      aspectratio_ = AspectRatio;
    }

    /** @brief Set principal point in pixels relative to top left corner
        @author streckel */
    virtual inline void SetPrincipal(const double x, const double y) {
      principalX_ = x;
      principalY_ = y;
    }

    virtual inline void SetImageSize(const unsigned int w, 
                                     const unsigned int h) {
      width_ = w; 
      height_ = h; 
    }

    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
        /attention only delivers correct results if ratio
        is chosen to rescale image size to integral image size 
        in both directions. Image size is rounded!
    */
    virtual inline void Rescale(float ratio) {
      double widthTemp = (double)width_/ratio;
      double heightTemp = (double)height_/ratio;
      width_  = (int)rint(widthTemp);
      height_ =  (int)rint(heightTemp);
      if( (!BIAS::Equal(widthTemp, (double)width_)) || 
          (!BIAS::Equal(heightTemp, (double)height_))) { 
        BIASERR("image size has integer type."<<width_<<"x"<<height_
                <<" Passed rescale factor "<< ratio <<" does not conform!\n");
      }
      principalX_ /= ratio;
      principalY_ /= ratio;
    }

    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersBase* Clone() const = 0;

    virtual void SetQ(const BIAS::Quaternion<double>& Q) 
    { Pose_.SetQ(Q); }
    virtual void SetC(const BIAS::Vector3<double>& C)
    { Pose_.SetC(C); }
    virtual void SetQC(const BIAS::Quaternion<double>& Q,
                     const BIAS::Vector3<double>& C)
    { Pose_.Set(Q,C); }

    virtual void SetPose(const BIAS::Pose pose)
    { Pose_=pose; }
    
    virtual void SetR(const BIAS::RMatrix& R) 
    { Quaternion<double> Q;
      R.GetQuaternion(Q); 
      Pose_.SetQ(Q);
    }

    virtual void SetCov(const Matrix<POSE_TYPE>& Cov) 
    { Pose_.SetCovariance(Cov); }

    /** @brief returns difference of two projections like | C1-C2 | 

    This function can be used for inter camera distance and ordering purposes
    @author koeser 03/2006 */
    virtual double ViewDifference(const ProjectionParametersBase*) const;

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif


    /** unscented transform worker function 
        @author koeser */
    virtual void Transform_(const Vector<double>& src,
                            Vector<double>& dst) const;

    friend std::ostream& operator<<(std::ostream &os, 
                                    ProjectionParametersBase& p);


  protected:
    Pose Pose_;

    /// width of image in pixels
    unsigned int width_;
    /// height of image in pixels
    unsigned int height_;

    /// Principal Point Pixel Coordinates (one for all zoom settings)
    double principalX_, principalY_;
    
    /// aspectratio of the camera CCD
    double aspectratio_;

    /// unscented transform worker variables
    bool ustTransformIntoImage_;
    bool ustIgnoreDistortion_;
  };

  inline std::ostream& operator<<(std::ostream &os, 
                                  ProjectionParametersBase& p) 
  {
    os <<"R:"<<p.GetR()<<"C:"<<p.GetC()<<std::endl;
    os<<"ImageSize: "<<p.width_<<" x "<<p.height_<<std::endl;
    os<<"Principal: "<<p.principalX_<<" x "<<p.principalY_<<std::endl;
    return os;
  }


 
} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

