/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersIO_hh__
#define __BIAS_ProjectionParametersIO_hh__
#include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Geometry/HomgPoint2D.hh>

#include <Base/Debug/Debug.hh>
#ifdef BIAS_HAVE_XML2
#  include <Base/Common/XMLBase.hh>
#endif

#include <Base/Geometry/KMatrix.hh>

#define CAMERA_PARAMETER_CURRENT_XML_VERSION 2.0
#define RIG_PARAMETER_CURRENT_XML_VERSION 1.0

namespace BIAS {

#ifndef __BIAS_CPDiscreteParam__
#define __BIAS_CPDiscreteParam__

  struct BIASGeometry_EXPORT CPDiscreteParam {
    double zoom;
    double focallength;
    // radial/tangential distortion parameters.
    // 4th order model of brown.
    double kc1,kc2,kc3,kc4;
  };
#endif

  struct BIASGeometry_EXPORT SensorData {

   // center and rotation relative to a common rig pose
    BIAS::Vector3<double> PoseInRigCenter_;
    double PoseInRigYaw_;
    double PoseInRigPitch_;
    double PoseInRigRoll_;

  };

  struct BIASGeometry_EXPORT CameraData {
    
    // Which camera if used?
    std::string CamModel_;
    int CamID_;
    std::string LensModel_;

    // do we have zoom or fixed optics? 
    bool zoomCamera_;

    // Lens can be Spherical or Perspective
    bool LensIsSpherical_;

    // width and height of the camera image
    unsigned int width_, height_;

    // filled by inheriting camera model
    double aspectratio_;

    // size of the single CCD-sensor-cells in meter
    double cellSizeX_, cellSizeY_;

    // Principal Point Pixel Coordinates (one for all zoom settings)
    double principalX_, principalY_;

    // image radius if image is spherical
    double radius_;
    
    // Maximum image angle of the fisheye camera
    double MaxCamAngle_;

    // only for perspective lenses
    // standard settings for non zoom lenses or
    // for zoom default state
    // focal length is stored in pixel for cellSizeX
    double focallengthDef_;
    double kc1Def_,kc2Def_,kc3Def_,kc4Def_;

    // array of camera calibration constants for zoom cameras
    std::vector<CPDiscreteParam> knownparams_vect_;

    // illumination vignette correction
    // x is distance in pixel from principal point
    // y is illumination correction factor
    std::vector<double> IlluCorrX_;
    std::vector<double> IlluCorrY_;

    // Only for sphere images:
    // Calibration measurements for angle correction 
    std::vector<double> AngleCorrX_;
    std::vector<double> AngleCorrY_;  
   
    // center and rotation relative to a common rig pose
    BIAS::Vector3<double> PoseInRigCenter_;
    double PoseInRigYaw_;
    double PoseInRigPitch_;
    double PoseInRigRoll_;
  
  };
 
  class BIASGeometry_EXPORT ProjectionParametersIO{
        
  public:   

    /** @brief read jpeg EXIF header and set camera parameter 
        @return 0 on success 
        @author koeser 07/2005 */
    static int ReadFromEXIF(const std::string& filename,
                            CameraData& cd);
    
    /** @brief pass K and image dims to construct a simple parameter file */
    static int SetFromKMatrix(const BIAS::KMatrix& K, 
                              const int w, const int h,
                              CameraData& cd);

#ifdef BIAS_HAVE_XML2

    /** @brief Write all data camera data and camera positions in the rig
        to a rig parameter file in XML-notation
        @param Filename Name of the XML-rig-data-file, 
        document tag will be to be "Rig"
        @author streckel 06/2004 */
    static int WriteRigData(const std::string &filename, 
                            const std::vector<CameraData>& vcd,
                            const std::vector<SensorData>& vsd=
                            std::vector<SensorData>(0,SensorData()));

    /** @brief Write all data to a camera data parameter file )
        @param Filename Name of the XML-camera-data-file, 
        document tag will be to be "Camera"
        @author streckel 06/2004 */
    static int WriteCameraData(const std::string& Filename,
                               const CameraData& cd);
      
    /** @brief Read a rig parameter file and store the data in (*this)
        the Camera-, Center-, RMatrix-, and KMatrix-Vectors are 
        reset from the file.
        @param Filename Name of the XML-camera-data-file, 
        document tag has to be "Rig"
        @author streckel 06/2004 */
    static int ReadRigData(const std::string &filename,
                           std::vector<CameraData>& vcd);
    static int ReadRigData(const std::string &filename,
                           std::vector<CameraData>& vcd,
                           std::vector<SensorData>& vsd);
      
    /** @brief Read a camera data parameter file and store the data in (*this)
        @param Filename Name of the XML-camera-data-file, 
        document tag has to be "Camera"
        @author streckel 06/2004 */
    static int ReadCameraData(const std::string& Filename, 
                              CameraData& cd);
   
    
#endif //BIAS_HAVE_XML2

    // Settings for specific cameras (only for testing)
    static void SetData_EVID31_147094(CameraData& cd);
    static void SetData_OMNITECH_ORIFL_190_3(CameraData& cd); 
    static void SetData_DummyRig(std::vector<CameraData>& vcd);    
    
protected:
    
  #ifdef BIAS_HAVE_XML2

    /** @brief Read a camera data form a precreated XML-tree and store 
        the data in (*this)  - called from the rig parameter class.
        @param myXML points to the XML-Tree
        @param cameraNode points to a "Camera"-node to evaluate
        @author streckel 06/2004 */
    static int ReadCameraData(XMLIO &myXML, xmlNodePtr &cameraNode,
                              CameraData& cd);

    /** @brief Add a camera sub-tree-structure to an allready created 
        XML-tree and write the data from (*this) into the tree - 
        called from the rig parameter class.
        @param myXML points to the XML-Tree
        @param cameraDataNode points to a "CameraData"-node to append the 
        "Camera"-node
        @author streckel 06/2004 */
    static int AddCameraDataToRig(XMLIO &myXML, xmlNodePtr &cameraDataNode, 
                                 const CameraData &cd);

    // called by WriteCameraData and AddCameraDataToRig to add 
    //the data of a certain camera to a node
    static int AddCameraDataToNode(XMLIO &myXML,  xmlNodePtr &rootNode, 
                                  const CameraData &cd);
#endif // BIAS_HAVE_XML2

    /** @brief default zero initialization for a single camera */
    static void Initialize(CameraData &cd);

    /** @brief default zero initialization for a single sensor */
    static void InitializeSensor(SensorData &sd);
    
    /** @brief Reading old file versions */
    static int ReadRigDataV0_8(const std::string &filename,
                               std::vector<CameraData>& vcd);
#ifdef BIAS_HAVE_XML2

    static int ReadCameraDataV1_0(XMLIO &myXML, xmlNodePtr &cameraNode,
                                  CameraData& cd);
    static int ReadCameraDataV0_8(XMLIO &myXML, xmlNodePtr &cameraNode,
                                  CameraData& cd);
#endif // BIAS_HAVE_XML2


  };
 
} // end namespace 

  /** @relates ProjectionParametersIO
      @brief standard ostream operator for dumping ProjectionParametersIO
      to screen */
BIASGeometry_EXPORT std::ostream& operator<<(std::ostream &os, std::vector<BIAS::CameraData>& vcd);
BIASGeometry_EXPORT std::ostream& operator<<(std::ostream &os, std::vector<BIAS::SensorData>& vsd);
BIASGeometry_EXPORT std::ostream& operator<<(std::ostream &os, BIAS::CameraData& cd);
BIASGeometry_EXPORT std::ostream& operator<<(std::ostream &os, BIAS::SensorData& sd);

#include <Base/Common/BIASpragmaEnd.hh>

#endif 

