/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Geometry/ProjectionParametersPerspective.hh" 


using namespace BIAS;
using namespace std;

bool ProjectionParametersPerspective::
DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                               HomgPoint2D& x,                       
                               bool IgnoreDistortion) const
{
  if(localX[2] > 0.0) {
    x = ProjectLocal(localX, IgnoreDistortion);
    x.Homogenize(); //neccessary ?
    return (x[0]>=0.0 && x[1]>=0.0 && 
            x[0]<=(double)(width_-1) && x[1]<=(double)(height_-1) );
  } 
  return false;
}


HomgPoint2D ProjectionParametersPerspective::
ProjectLocal(const Vector3<double>& point, bool IgnoreDistortion) const {
  // in-plane affine coordinate transformation with KMatrix
  HomgPoint2D pos;
  K_.Mult(point, pos);
  // radial distortion here 
  if (!IgnoreDistortion) Distort(pos);
  return pos.Homogenize();
}

Vector3<double>  ProjectionParametersPerspective::
UnProjectLocal(const HomgPoint2D& cPos, bool IgnoreDistortion) const {
  HomgPoint2D mpos(cPos); // de-const pos
  // radial undistortion here 
  if (!IgnoreDistortion) Undistort(mpos);
  // in-plane affine coordinate transformation with KMatrix
  Vector3<double> point;
  invK_.Mult(mpos, point);
  return point.Normalize();
}

/*
Matrix3x3<double> ProjectionParametersPerspective::
UnProjectLocal(const HomgPoint2D& pos, const Matrix3x3<double>& cov2D,
               bool IgnoreDistortion) 
{

 
  Matrix3x3<double> cov_unp(cov2D);
  if (!IgnoreDistortion) Undistort(pos, cov_unp);// not implemented !!
   
  // covariance matrices transform by multiplying with the jacobian and 
  // its transposed
  return invK_ * cov_unp * invK_.Transpose();
  
}

void ProjectionParametersPerspective::
Undistort(const HomgPoint2D& point2d, Matrix3x3<double>& cov) const 
{
  // unfinished, unclear if it is really necessary
  static int alreadywarned = 0;
  if (alreadywarned ==0) {
    BIASERR("not implemented");
    alreadywarned=1;
  }
}

*/

void ProjectionParametersPerspective::Distort(HomgPoint2D& point2d) const {
  
  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;

  // copy of distorted point
  double distx,disty;
  double x,y;
  double twoxy;
  
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  point2d.Homogenize();
  
  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = (point2d[0]-principalX_)/focallength_;
  y = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);
      
  twoxy = 2 * x*y;
  rsquared = x*x + y*y;
    
  // radial distortion:
  raddist = 1 + kc1_*rsquared + kc2_*rsquared*rsquared;
  // tangential distortion:
  tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
  tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);
  
  // apply distortion:
  distx = raddist * x + tandistx;
  disty = raddist * y + tandisty;
   
  // invert Bougouet-normalization
  point2d[0] = distx*focallength_ + principalX_;
  point2d[1] = disty*GetAspectratio()*focallength_ + principalY_;
}


void ProjectionParametersPerspective::Undistort(HomgPoint2D& point2d) const {

  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;

// copy of distorted point
  double distx,disty;
  double x,y;
  double twoxy;
  
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  point2d.Homogenize();

  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = point2d[0] = (point2d[0]-principalX_)/focallength_;
  y = point2d[1] = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);
  
  // 5 iterations are enough
  for (unsigned int i=0;i<5;i++){
        
    twoxy = 2 * x*y;
    rsquared = x*x + y*y;
    
    // radial distortion:
    raddist = 1 + kc1_*rsquared + kc2_*rsquared*rsquared;
    // tangential distortion:
    tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
    tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);
    
    // apply distortion:
    distx = raddist * x + tandistx;
    disty = raddist * y + tandisty;
    
    // subtract distortion from distorted coordinates
    x = point2d[0] - (distx-x);
    y = point2d[1] - (disty-y);
    
  }

  // invert Bougouet-normalization
  point2d[0] = x*focallength_ + principalX_;
  point2d[1] = y*GetAspectratio()*focallength_ + principalY_;
}

#ifdef BIAS_HAVE_XML2
 
int ProjectionParametersPerspective::XMLGetClassName(std::string& TopLevelTag,
                                                     double& Version) const {
  TopLevelTag = "ProjectionParametersPerspective";
  Version = 0.1;
  return 0;
}

int ProjectionParametersPerspective::XMLOut(const xmlNodePtr Node, 
                                            XMLIO& XMLObject) const {
  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);
  theNode = XMLObject.addChildNode(Node, TopLevelName);
  XMLObject.addAttribute(theNode, "Version", Version);
  ProjectionParametersBase::XMLOut(theNode, XMLObject);

  xmlNodePtr childNode;
  XMLObject.addAttribute(Node,"Focallength", focallength_);
  childNode = XMLObject.addChildNode(Node,"Distortion");
  XMLObject.addAttribute(childNode, "k1", kc1_);
  XMLObject.addAttribute(childNode, "k2", kc2_);
  XMLObject.addAttribute(childNode, "p1", kc3_);
  XMLObject.addAttribute(childNode, "p2", kc4_);
  return 0; 
 }

 
int ProjectionParametersPerspective::XMLIn(const xmlNodePtr Node, 
                                           XMLIO& XMLObject) {

  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);

  xmlNodePtr childNode; 
  if ((childNode = XMLObject.getChild(Node, TopLevelName))==NULL) {
    BIASERR("Error in xml, no tag" << TopLevelName);
    return -1;
  }
  if (ProjectionParametersBase::XMLIn(childNode, XMLObject)!=0) return -1;

  focallength_  = XMLObject.getAttributeValueDouble(Node, "Focallength");
  if ((childNode = XMLObject.getChild(Node, "Distortion"))!=NULL) {
    kc1_ = XMLObject.getAttributeValueDouble(childNode, "k1");
    kc2_ = XMLObject.getAttributeValueDouble(childNode, "k2");
    kc3_ = XMLObject.getAttributeValueDouble(childNode, "p1");
    kc4_ = XMLObject.getAttributeValueDouble(childNode, "p2");
  }
  ComputeK_();
  return 0;
}

#endif // BIAS_HAVE_XML2


void ProjectionParametersPerspective::ComputeK_() {
  K_.SetIdentity(); 
  K_[0][0] = focallength_;
  K_[1][1] = aspectratio_ * focallength_;
  K_[0][1] = 0; //skew_;
  K_[0][2] = principalX_;
  K_[1][2] = principalY_;
  invK_ = K_.Invert();

}

double ProjectionParametersPerspective::
ViewDifference(const ProjectionParametersBase* pPPB) const {
  const ProjectionParametersPerspective *pPPP = 
     dynamic_cast<const ProjectionParametersPerspective* >(pPPB);
  // different projection model has a very different view
  if (pPPP==NULL) {
    BIASERR("View difference for different projection models requested !");
    return DBL_MAX;
  }

  Vector3<double> C0 = pPPP->GetC(), A0, A;
  
  RMatrix R = GetR(), R0 = pPPP->GetR();
  
  A[0] = R[0][2]; A[1] = R[1][2]; A[2] = R[2][2];
  A.Normalize();
  
  A0[0] = R0[0][2]; A0[1] = R0[1][2]; A0[2] = R0[2][2];
  A0.Normalize();
  
  // compute field of view, assuming const K !
  KMatrix K0 =  GetK();
  double f = 0.5*(K0[0][0] + K0[1][1]);
  double halfw = double(width_ + height_) / 4.0;
  double coshalffieldofview = cos(atan(halfw/f));
  
  
  double newness = (GetC()-C0).NormL2();
  newness += 1.0/(1.0-coshalffieldofview)*(1.0-fabs(A.ScalarProduct(A0)));
  return newness;
}


