/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersPerspective_hh__
#define __BIAS_ProjectionParametersPerspective_hh__
#include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Geometry/PMatrix.hh>

#define LUT_RES 10
namespace BIAS {

 /** @class ProjectionParametersPerspective
     @brief camera parameters which define the mapping between rays in the 
     camera coordinate system and pixels in the image as well as external pose
     @author koeser/evers/streckel
 */
  class BIASGeometry_EXPORT ProjectionParametersPerspective
    : public  ProjectionParametersBase {
  public:

    ProjectionParametersPerspective()
      : ProjectionParametersBase() 
    {      
      K_.SetIdentity();
      invK_.SetIdentity();
      focallength_=1.0;
      kc1_=0;
      kc2_=0;
      kc3_=0;
      kc4_=0;      
    };

    ProjectionParametersPerspective(const ProjectionParametersPerspective & P){
      *this = P;
    };

    ProjectionParametersPerspective& 
    operator=(const ProjectionParametersPerspective & P){
      ProjectionParametersBase::operator=(P);
      K_ = P.K_;
      invK_ = P.invK_;
      focallength_= P.focallength_;
      kc1_= P.kc1_;
      kc2_= P.kc2_;
      kc3_= P.kc3_;
      kc4_= P.kc4_;      
      return *this;
    };

    virtual ~ProjectionParametersPerspective(){};

    bool DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                                        HomgPoint2D& x,                       
                                        bool IgnoreDistortion = false) const;
    

    /** @brief calculates the projection of a point 
        in the camera coordinate system  
        to a pixel in the image plane of the camera
        with lens distortion
        In the simplest case perspective pinhole projection x = K * y  
        where y is the projection of X using  y = (R^T | -R^T C) X 
    */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const;
   
    /** @brief calculates the viewing ray from the camera center (in the 
        camera coordinate system) which belongs to the image position pos
        with lens distortion
        e.g. (0,0,1) means optical axis
    */
    virtual Vector3<double> UnProjectLocal(const HomgPoint2D& pos,
                                           bool IgnoreDistortion = false)const;
      
    /** @brief Using the Matlab camera calibration toolbox parameters 
        for an undistortion of distorted coordinates.
        TODO: faster implementation via Lookup-table
        @author streckel 06/2004 */
    void Undistort(BIAS::HomgPoint2D& point2d) const;
 
    /** @brief Using the Matlab camera calibration toolbox parameters 
        for a distortion of undistorted coordinates. Inverse to the 
        undistort function.
        TODO: faster implementation via Lookup-table
        @author streckel 06/2004 */
    void Distort(BIAS::HomgPoint2D& point2d) const;
 
    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersPerspective* Clone() const {
      return new ProjectionParametersPerspective(*this);
    } 

    /** @brief Get the current camera focal length.
        @author buck */
    inline int GetFocalLength(double& f) const {
      f = focallength_;
      return 0;
    }

    inline void GetUndistortion(double& kc1, double& kc2,
                                double& kc3, double& kc4) const 
    {
      kc1 = kc1_;
      kc2 = kc2_;
      kc3 = kc3_;
      kc4 = kc4_;
    }

    /** @brief Set the current camera focal length in pixel and the a
        spect ratio */
    inline void SetFocalLengthAndAspect(double f, 
                                        double AspectRatio) {
      focallength_ = f;
      ProjectionParametersBase::SetAspectratio(AspectRatio);
      ComputeK_();
    }

    /** @brief Set principal point in pixels relative to top left corner,
        virtual overload to recalculate K_
        @author streckel */
    virtual inline void SetPrincipal(const double x, const double y) {
      ProjectionParametersBase::SetPrincipal(x, y);
      ComputeK_();      
    }

    /** @brief Set CCD cell-size in meter, virtual overload to 
        recalculate K_.
        @author streckel */
    virtual inline void SetAspectratio(const double AspectRatio) {
      ProjectionParametersBase::SetAspectratio(AspectRatio);
      ComputeK_();
    }

   /** @brief Set the lens undistortion parameters*/
    inline void SetUndistortion(double kc1, double kc2,
                                double kc3, double kc4) 
    {
      kc1_ = kc1;
      kc2_ = kc2;
      kc3_ = kc3;
      kc4_ = kc4;
    }
    
    /* @brief get the PMatrix composed from the projection object
     */
    inline virtual BIAS::PMatrix GetP() const {
      return PMatrix(K_, GetR(), GetC());
    }

    /* @brief get cached KMatrix
     */
    inline virtual BIAS::KMatrix GetK() const { 
      return K_;
    };

    /* @brief get cached inverted KMatrix
     */
    inline virtual BIAS::KMatrix GetKinv() const { 
      return invK_;
    };

    /** @brief sets the internal parameters from a given KMatrix
        and updates the cached K and its inverse */
    virtual void SetK(const KMatrix& K) {
      BIASASSERT(K[0][0]!=0);
      focallength_  = K[0][0];
      ProjectionParametersBase::SetPrincipal(K[0][2],K[1][2]);
      double aspectratio = K[1][1]/K[0][0];
      ProjectionParametersBase::SetAspectratio(aspectratio);
      K_ = K; invK_ = K.Invert();
    };

    /** @brief set from P  */
    virtual void SetP(const PMatrix& P) {
      PMatrix p(P); // de-const
      SetK(p.GetK());
      SetC(p.GetC());
      SetR(p.GetR());
      ComputeK_(); 
    };


    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
    */
    virtual inline void Rescale(float ratio) {
      ProjectionParametersBase::Rescale(ratio);
      focallength_ /= ratio;
      ComputeK_();
      // dont have to update radial/tangential distortion coefficients
      // because we measure in normalize space
    }
    
    virtual double ViewDifference(const ProjectionParametersBase* pPPB) const;

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif

    friend std::ostream& operator<<(std::ostream &os, 
                                    ProjectionParametersPerspective& p);


    
  protected:
    /** @brief computes the cached KMatrix and its inverse
        call after every parameter change */
    void ComputeK_();

    // focal length, aspect ratio, principal point in K_, inverse in invK_
    KMatrix K_, invK_;

    // focal length is stored in pixel for cellSizeX
    double focallength_;
    // lens undistortion parameters after Bouquet
    double kc1_,kc2_,kc3_,kc4_;

  };

  inline std::ostream& operator<<(std::ostream &os, 
                                  ProjectionParametersPerspective& p) 
  {
    os <<"R:"<<p.GetR()<<"C:"<<p.GetC()<<std::endl;
    os<<"ImageSize: "<<p.width_<<" x "<<p.height_<<std::endl;
    os<<"Principal: "<<p.principalX_<<" x "<<p.principalY_<<std::endl;
    //operator<<(os, (ProjectionParametersBase&)p);
    std::cout <<"K:"<<p.GetK()<<std::endl;
    return os;
  }


 
} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

