/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersProjective_hh__
#define __BIAS_ProjectionParametersProjective_hh__
#include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <Geometry/CoordinateTransform3D.hh>
#include <Geometry/PMatrix.hh>

namespace BIAS {

 /** @class ProjectionParametersProjective
     @brief 12 parametric p for completely projective reconstruction
     scenarios, unknown focal lengths, self calibration, ...

     The handling of the "pose" in this class is special because we cannot
     decompose P to get the center and the rotation. We pretend that it
     has R=id and C=0 and encode the external parameters also within
     the projection matrix. There may be some cases of infinite points, camera
     center at infinity and so on,  which are not covered here, however this
     interface is a compromise with respect to the performance of the "normal"
     calibrated camera case's interface.
     @author koeser 01/2006
 */
  class BIASGeometry_EXPORT ProjectionParametersProjective: 
    public  ProjectionParametersBase {
  public:

    ProjectionParametersProjective()
      : ProjectionParametersBase() {};
    ProjectionParametersProjective(const ProjectionParametersProjective & P){
      *this = P;
    };

    ProjectionParametersProjective & 
    operator=(const ProjectionParametersProjective & P){
      BIASERR("still not implemented");
      BIASABORT;
      return *this;
    };

    virtual ~ProjectionParametersProjective(){};

    /** @brief calculates the projection of a point 
        in the camera coordinate system  
        to a pixel in the image plane of the camera
        
        In the simplest case perspective pinhole projection x = K * y  
        where y is the projection of X using  y = (R^T | -R^T C) X 
    */
    virtual HomgPoint2D  ProjectLocal(const Vector3<double>& point,
                                      bool IgnoreDistortion = false) const {
      HomgPoint3D X_G(point);
      HomgPoint3D X_L = Pose_.GlobalToLocal(X_G);
      X_L.Homogenize();
      HomgPoint2D x_l(X_L[0], X_L[1], X_L[2]);
      return x_l;
    }

    /** @brief calculates the viewing ray from the camera center (in the 
        camera coordinate system) which belongs to the image position pos
        e.g. (0,0,1) means optical axis
    */
    virtual Vector3<double> UnProjectLocal(const HomgPoint2D& pos,
                                           bool IgnoreDistortion = false) const
    {
      HomgPoint3D X_L((Vector3<double>) pos);
      HomgPoint3D X_G = Pose_.LocalToGlobal(X_L);
      X_G.Homogenize();
      Vector3<double> x_g(X_G[0], X_G[1], X_G[2]);
      return x_g;
    }

    virtual Matrix3x3<double> 
    UnProjectLocal(const HomgPoint2D& pos,
                   const Matrix3x3<double>& cov,
                   bool IgnoreDistortion = false) const;

    /** @brief get projection center */
    virtual Vector3<double> GetC() const { return Vector3<double>(0,0,0); };

    /** @brief return R (R^T which is in front part of PMatrix ) */
    virtual RMatrix GetR() const { return RMatrix(MatrixIdentity); };


    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersProjective* Clone() const {
      return new ProjectionParametersProjective(*this);
    };

    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
    */
    virtual inline void Rescale(float ratio) {
      BIASERR("unimplemented");
      BIASABORT;
    }

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif
    /** @brief set 12 entries if projective P using a PMatrix */
    void SetP(const Matrix3x4<double>& P) { Pose_.SetFromMatrix3x4(P);}

    void SetR(const RMatrix& R) {BIASERR("dont call this for projective !");}
    void SetC(const Vector3<double>& C) {
      BIASERR("dont call this for projective !");}

    /** @brief return the complete pmatrix (internals+externals / projective)
        used for projection */
    inline virtual PMatrix GetP() const { 
      return PMatrix(Pose_.GetMatrix3x4()); 
    }
  };


 
} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

