/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Geometry/ProjectionParametersSpherical.hh" 


using namespace BIAS;
using namespace std;

bool ProjectionParametersSpherical::
DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                               HomgPoint2D& x,                       
                               bool IgnoreDistortion) const
{
  x = ProjectLocal(localX, IgnoreDistortion);
  double radius = 
    sqrt( (x[0]-principalX_) * (x[0]-principalX_) + 
          (x[1]-principalY_) * (x[1]-principalY_) );
    
  return (x[0]>=0.0 && x[1]>=0.0 && 
          x[0]<=(double)(width_-1) && x[1]<=(double)(height_-1) &&
          radius<= radius_); 
  
}


HomgPoint2D ProjectionParametersSpherical::
ProjectLocal(const Vector3<double>& point, bool IgnoreDistortion) const {
  // transform to spherical coordinates
  Vector3<double> rPhiTheta = point.CoordEuclidianToSphere();
  // projection point is independent of the radius
  HomgPoint2D ret;  
  if (!IgnoreDistortion) {
    if (TransfCoordSphere2Image_(rPhiTheta[2], rPhiTheta[1], ret)<0){
      ret.SetZero();
    } 
  } else {
    if (TransfCoordSphere2CalibImage_(rPhiTheta[2], rPhiTheta[1], ret)<0){
      ret.SetZero();
    } 
  }
  return ret; 
}
    
Vector3<double>  ProjectionParametersSpherical::
UnProjectLocal(const HomgPoint2D& pos, bool IgnoreDistortion) const {
  // radius = 1;
  Vector3<double> rPhiTheta(1,0,0);
  Vector3<double> ret;
  if (!IgnoreDistortion) {
    if (TransfCoordImage2Sphere_(pos, rPhiTheta[2], rPhiTheta[1])<0){
      ret.SetZero();
    } else {
      ret = rPhiTheta.CoordSphereToEuclidian();
    }
  } else {
    if (TransfCoordCalibImage2Sphere_(pos, rPhiTheta[2], rPhiTheta[1])<0){
      ret.SetZero();
    } else {
      ret = rPhiTheta.CoordSphereToEuclidian();
    }
  }
  return ret;
}

Vector3<double> ProjectionParametersSpherical::
UnProjectToPointLocal(const HomgPoint2D& pos,
                      const double& depth,
                      bool IgnoreDistortion) const {
  Vector3<double> x(UnProjectLocal(pos, IgnoreDistortion) * depth);
  return x;
}


int ProjectionParametersSpherical::
SetUndistortion(const std::vector<double>& AngleCorrX, 
                const std::vector<double>& AngleCorrY,
                const double radius) {
   
  radius_ = radius;
  
  if (AngleCorrX.size()<2 || AngleCorrY.size()<2 || 
      AngleCorrX.size()!=AngleCorrY.size()) {
    BIASERR("Initialization failed, sizes of angle correction vectors are" 
            <<  AngleCorrX.size() << AngleCorrY.size());
    return -1;
  }
  
  double dum;
  interpolatorAngle_.SetKnotPoints(AngleCorrX);  //t
  interpolatorAngle_.SetControlPoints(AngleCorrY); //y = f(t)
  interpolatorAngle_.InitSpline(); 
  interpolatorAngle_.Spline(dum, AngleCorrX[0],4);  
  MaxCamAngle_ = AngleCorrX[AngleCorrX.size()-1];
  int lutsamples = int(radius_*100);
  interpolatorAngle_.PrepareLuTSpline(AngleCorrX[0],MaxCamAngle_,
				      lutsamples,4 );

  interpolatorAngleBack_.SetKnotPoints(AngleCorrY);  
  interpolatorAngleBack_.SetControlPoints(AngleCorrX);  
  interpolatorAngleBack_.InitSpline(); 
  interpolatorAngleBack_.Spline(dum, AngleCorrY[0],4);  
  interpolatorAngleBack_.PrepareLuTSpline(AngleCorrY[0],
					  AngleCorrY[AngleCorrY.size()-1], 
					  lutsamples,4 );

  return 0;

}


int ProjectionParametersSpherical::
TransfCoordImage2Sphere_ (const HomgPoint2D& Source, 
                          double &theta, double &phi) const {

  HomgPoint2D SourceH = Source;
  SourceH.Homogenize();

  double rSrc = sqrt((SourceH[0]-principalX_)*(SourceH[0]-principalX_)+
                     (SourceH[1]-principalY_)*(SourceH[1]-principalY_));
  if (rSrc>radius_) {
    //point not in circle image!
    return -1;
  }

  phi = atan2(SourceH[1]-principalY_,SourceH[0]-principalX_);
  
  double CalMaxCamAngle=0;
  // calculate calibrated maximum angle
  interpolatorAngle_.SplineFromLuT(CalMaxCamAngle, MaxCamAngle_);  

  // calculate calibrated maximum radius
  double RadiusCalMax = MaxCamAngle_/CalMaxCamAngle*radius_;
  // calculate calibrated theta and inverse
  double CalTheta = rSrc/RadiusCalMax*MaxCamAngle_;
  //  interpolatorAngleBack_.Spline(theta, CalTheta,4); 
  interpolatorAngleBack_.SplineFromLuT(theta, CalTheta); 

  return 0;
}


int ProjectionParametersSpherical::
TransfCoordRotateSphere2Image_(double theta, double phi,
                               double viewCenterX, 
                               double viewCenterY,
                               HomgPoint2D &Source) const
{
  double viewTheta, viewPhi;
  HomgPoint2D viewCenter(viewCenterX,viewCenterY,1);
  if (TransfCoordImage2Sphere_(viewCenter, viewTheta, viewPhi)!=0) { 
    return -1;
  }
  // avoid z-rotation by changing the rotation axes
  Vector3<double> vecPoint(1.0,phi,theta);
  Vector3<double> vecView(1.0,viewPhi,viewTheta);
  //Vector3<double> vecCenter(0,0,1);
  Vector3<double> vecAxis(cos(viewPhi+M_PI/2),sin(viewPhi+M_PI/2),0);
  Vector3<double> vecRot(0,0,0);
  vecPoint = vecPoint.CoordSphereToEuclidian();
  vecView = vecView.CoordSphereToEuclidian();
  /*
  // With 1 rotation and moving axis
  RMatrix rMat;
	rMat.Set(vecAxis, viewTheta);
  double x,y,z;
  rMat.GetRotationAnglesZYX(x,y,z);
  vecRot = rMat*vecPoint;
  //cout << x << "," << y << "," << z << endl;
  */

  //With Quaternions
  Quaternion<double> rot;
  rot.SetValueAsAxisRad (vecAxis, viewTheta);
  rot.MultVec(vecPoint, vecRot);
  
  /* 
  //With 2 rotations alpha, beta (wrong rotation!!)
  // viewAlpha: angle between x-z-Plane and vecView
  double viewAlpha = -1 * atan2(vecView[1],vecView[2]);
  // viewBeta: angle between y-z-Plane and vecView
  double viewBeta = atan2(vecView[0],vecView[2]); 
  RMatrix rMat;
  rMat.SetZYX(viewAlpha,viewBeta,0);
  vecRot = rMat*vecPoint;
  */

  vecRot = vecRot.CoordEuclidianToSphere();
  phi = vecRot[1];
  theta = vecRot[2];

  if (theta > MaxCamAngle_) return -1;
  if (phi>M_PI) phi-=2*M_PI;

  double CalMaxCamAngle=0;
  double CalTheta=0;
  // calculate calibrated angles
  interpolatorAngle_.SplineFromLuT(CalMaxCamAngle, MaxCamAngle_);  
  interpolatorAngle_.SplineFromLuT(CalTheta, theta);  
  // apply calculate calibrated radius from calibrated angles
  double rSrc = radius_ * CalTheta/CalMaxCamAngle;
 
  Source[0] = rSrc * cos(phi) + principalX_;
  Source[1] = rSrc * sin(phi) + principalY_;
  Source[2] = 1;
  return 0;
}


int ProjectionParametersSpherical::
TransfCoordSphere2Image_(double theta, double phi, 
                         HomgPoint2D &Source) const {
  
  if (EnforceNormalRange_(theta, phi)<0) return -1;
  
  double CalMaxCamAngle=0;
  double CalTheta=0;
  // calculate calibrated angles
  interpolatorAngle_.SplineFromLuT(CalMaxCamAngle, MaxCamAngle_);  
  interpolatorAngle_.SplineFromLuT(CalTheta, theta);  
  // apply calculate calibrated radius from calibrated angles
  double rSrc = radius_ * CalTheta/CalMaxCamAngle;

  Source[0] = rSrc * cos(phi) + principalX_;
  Source[1] = rSrc * sin(phi) + principalY_;
  Source[2] = 1;

  return 0;
}

int ProjectionParametersSpherical::
TransfCoordSphere2CalibImage_(double theta, double phi,
                              HomgPoint2D &Source) const {
  
  if (EnforceNormalRange_(theta, phi)<0) return -1;
  
  // apply radius from calibrated angles
  double rSrc = radius_ * theta/MaxCamAngle_;

  Source[0] = rSrc * cos(phi) + principalX_;
  Source[1] = rSrc * sin(phi) + principalY_;
  Source[2] = 1;

  return 0;

}

int ProjectionParametersSpherical::
TransfCoordCalibImage2Sphere_ (const HomgPoint2D& Source, 
                               double &theta, double &phi) const {

  HomgPoint2D SourceH = Source;
  SourceH.Homogenize();

  double rSrc = sqrt((SourceH[0]-principalX_)*(SourceH[0]-principalX_)+
                     (SourceH[1]-principalY_)*(SourceH[1]-principalY_));

  phi = atan2(SourceH[1]-principalY_,SourceH[0]-principalX_);
  theta = rSrc/radius_*MaxCamAngle_;
  return 0;
}

int ProjectionParametersSpherical::
EnforceNormalRange_(double& theta, double& phi) const {

  // theta and phi are periodic in 2*M_PI, if they are out of a suitable
  // range, try to find best equivalent angle by adding 2*l*M_PI
  double ophi=phi;
  double otheta=theta;
  if (theta<0.0) {
    const int count = (int) ((double)theta / (M_PI*(-2.0)));
    theta += ((double)count+1.0) * M_PI*2.0;
    if (theta>MaxCamAngle_) {
      BIASERR("Angle theta out of range!"<<otheta);
      return -1;
    } 
  } else if (theta>MaxCamAngle_) {
    const int count = (int) ((double)theta / (M_PI*(2.0)));
    theta -= ((double)count+1.0) * M_PI*2.0;
    if (theta<0.0) {
      //if (otheta-MaxCamAngle_>0.08) {
      //  BIASERR("Angle theta out of range: "<<otheta<< "," << otheta-MaxCamAngle_);
      //}
      return -2;
    } 
  }
 
  if (phi<-M_PI) {
    const int count = (int) ((double)phi / (M_PI*(-2.0)));
    phi += ((double)count+1.0) * M_PI*2.0;
    if (phi>M_PI) {
      BIASERR("Angle phi out of range:"<<ophi);
      return -3;
    } 
  } else if (phi>M_PI) {
    const int count = (int)((double)phi / (M_PI*(2.0)));
    phi -= ((double)count+1.0) * M_PI*2.0;
    if (phi<-M_PI) {
      BIASERR("Angle phi out of range: "<<ophi);
      return -4;
    } 
  }
  return 0;
}


double ProjectionParametersSpherical::
ViewDifference(const ProjectionParametersBase* pPPB) const {
  const ProjectionParametersSpherical *pPPS = 
     dynamic_cast<const ProjectionParametersSpherical* >(pPPB);
  // different projection model has a very different view
  if (pPPS==NULL) {
    BIASERR("View difference for different projection models requested !");
    return DBL_MAX;
  }

  Vector3<double> C0 = pPPS->GetC();
  // Optical axes of the 2 cameras
  Vector3<double>  A0, A;
  
  RMatrix R = GetR(), R0 = pPPS->GetR();
  
  A[0] = R[0][2]; A[1] = R[1][2]; A[2] = R[2][2];
  A.Normalize();
  
  A0[0] = R0[0][2]; A0[1] = R0[1][2]; A0[2] = R0[2][2];
  A0.Normalize();
  
  // compute field of view, assuming const K !
  double coshalffieldofview = cos(MaxCamAngle_);
  
  double newness = (GetC()-C0).NormL2();
  newness += 1.0/(1.0-coshalffieldofview)*(1.0-fabs(A.ScalarProduct(A0)));
  return newness;
}

KMatrix ProjectionParametersSpherical::GetFakeKMatrix(int resolution) const { 
  double dum;
  return GetFakeKMatrix(dum, resolution);
}	 
 
KMatrix ProjectionParametersSpherical::GetFakeKMatrix(double& imgsize, 
                                                      int resolution) const { 
  	 
  // maximum FoV in degree
  double FoV = 160;
  imgsize=0;

  double ang;
  ang = (FoV/2.0)*M_PI/180.0;
  if (ang>GetMaxCamAngle()) ang = GetMaxCamAngle()-0.01;

  Vector3<double> ray1, ray2;
  HomgPoint2D imgCoord1, imgCoord2;
  switch (resolution) {
  case 1:
    // measure 1 deg angle in camera center
    ray1.Set(sin(ang-M_PI/180.0),0.0,cos(ang-M_PI/180.0));     
    ray2.Set(sin(ang),0.0,cos(ang));     
    imgCoord1 = ProjectLocal(ray1);
    imgCoord2 = ProjectLocal(ray2);
    imgsize=(imgCoord2[0]-imgCoord1[0])*tan(ang)/tan(M_PI/180.0);
    break;
  case 2:
    // measure 1 deg angle near max angle
    ray1.Set(sin(ang-M_PI/180.0),0.0,cos(ang-M_PI/180.0));     
    ray2.Set(sin(ang),0.0,cos(ang));     
    imgCoord1 = ProjectLocal(ray1);
    imgCoord2 = ProjectLocal(ray2);
    imgsize=(imgCoord2[0]-imgCoord1[0])/
      (tan(ang)-tan(ang-M_PI/180.0))*tan(ang);
    break;
  case 3:
    // measure 1 deg angle in camera center
    ray1.Set(sin(ang-M_PI/180.0),0.0,cos(ang-M_PI/180.0));     
    ray2.Set(sin(ang),0.0,cos(ang));     
    imgCoord1 = ProjectLocal(ray1);
    imgCoord2 = ProjectLocal(ray2);
    imgsize=(imgCoord2[0]-imgCoord1[0])*tan(ang)/tan(M_PI/180.0);
    imgsize*=2;
    break;
  default:
    ray1.Set(sin(ang),0.0,cos(ang)); 
    imgCoord1 = ProjectLocal(ray1);
    imgsize = imgCoord1[0]-principalX_;  
  }
  KMatrix K; 	 
  K.SetIdentity();
  K[0][0] = imgsize/tan(ang);
  K[1][1] = imgsize/tan(ang);
  K[0][2] = imgsize; 	 
  K[1][2] = imgsize;
  imgsize *= 2;
  return K; 	 
}


#ifdef BIAS_HAVE_XML2
 
int ProjectionParametersSpherical::XMLGetClassName(std::string& TopLevelTag,
                                                     double& Version) const {
  TopLevelTag = "ProjectionParametersSpherical";
  Version = 0.1;
  return 0;
}

 
int ProjectionParametersSpherical::XMLOut(const xmlNodePtr Node, 
                                            XMLIO& XMLObject) const {
  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);
  theNode = XMLObject.addChildNode(Node, TopLevelName);
  XMLObject.addAttribute(theNode, "Version", Version);
  ProjectionParametersBase::XMLOut(theNode, XMLObject);

  xmlNodePtr childNode, childNode2;
  XMLObject.addAttribute(Node,"Radius", radius_);
  childNode = XMLObject.addChildNode(Node,"Distortion");
  vector<double> controlpoints, knotpoints;
  interpolatorAngle_.GetControlPoints(controlpoints);
  interpolatorAngle_.GetKnotPoints(knotpoints);
  for (unsigned int i=0; i<controlpoints.size(); i++) {
    childNode2= XMLObject.addChildNode(childNode,"MeasPoint");
    XMLObject.addAttribute(childNode2, "AngRay", knotpoints[i]);
    XMLObject.addAttribute(childNode2, "AngDeformed", controlpoints[i]);
  }
  return 0; 
}

 
int ProjectionParametersSpherical::XMLIn(const xmlNodePtr Node, 
                                           XMLIO& XMLObject) {
  string TopLevelName;
  double Version;
  xmlNodePtr childNode, childNode2; 

  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);
  if ((childNode = XMLObject.getChild(Node, TopLevelName))==NULL) {
    BIASERR("Error in xml, no tag" << TopLevelName);
    return -1;
  }
  if (ProjectionParametersBase::XMLIn(childNode, XMLObject)!=0) return -1;
  
  double radius = XMLObject.getAttributeValueDouble(Node, "Radius");
  vector<double> AngleCorrX, AngleCorrY;
  if ((childNode = XMLObject.getChild(Node, "Distortion"))!=NULL) {
    childNode2 = XMLObject.getFirstChild(childNode);
    while (childNode2!=NULL) {
      string nodename = XMLObject.getNodeName(childNode2);
      if (nodename.compare("MeasPoint")==0) {
        double x = XMLObject.getAttributeValueDouble(childNode2, "AngRay");
        double y = XMLObject.getAttributeValueDouble(childNode2, 
                                                     "AngDeformed");
        AngleCorrX.push_back(x);
        AngleCorrY.push_back(y);
      }
      childNode2 = XMLObject.getNextChild(childNode2);
    }
    SetUndistortion(AngleCorrX, AngleCorrY, radius);
  }
  return 0;
}
#endif


 
