/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersSpherical_hh__
#define __BIAS_ProjectionParametersSpherical_hh__
#include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <MathAlgo/Interpolator.hh>

namespace BIAS {

 /** @class ProjectionParametersSpherical
     @brief camera parameters which define the mapping between rays in the 
     camera coordinate system and pixels in the image as well as external pose
     @author koeser/evers/streckel
 */
  class BIASGeometry_EXPORT ProjectionParametersSpherical: 
    public  ProjectionParametersBase {
  public:

    ProjectionParametersSpherical()
      : ProjectionParametersBase(){
      radius_=0;
      MaxCamAngle_=0;
    };

    ProjectionParametersSpherical(const ProjectionParametersSpherical & P){
      *this = P;
    };

    ProjectionParametersSpherical& 
    operator=(const ProjectionParametersSpherical & P){
      ProjectionParametersBase::operator=(P);
      radius_=P.radius_;
      MaxCamAngle_=P.MaxCamAngle_;
      interpolatorAngle_ = P.interpolatorAngle_;
      interpolatorAngleBack_ = P.interpolatorAngleBack_;
      return *this;
    };

    virtual ~ProjectionParametersSpherical(){};

    bool DoesPointProjectIntoImageLocal(const BIAS::Vector3<double>& localX, 
                                        HomgPoint2D& x,                       
                                        bool IgnoreDistortion = false) const;

    /** @brief calculates the projection of a point 
        in the camera coordinate system  
        to a pixel in the image plane of the camera
        
        In the simplest case perspective pinhole projection x = K * y  
        where y is the projection of X using  y = (R^T | -R^T C) X 
    */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const;

    /** @brief calculates the viewing ray from the camera center (in the 
        camera coordinate system) which belongs to the image position pos
        e.g. (0,0,1) means optical axis
    */
    virtual Vector3<double> 
    UnProjectLocal(const HomgPoint2D& pos,
                   bool IgnoreDistortion = false) const;

    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersSpherical* Clone() const {
      return new ProjectionParametersSpherical(*this);
    };

   inline double GetRadius() const{
     return radius_;
    }

    inline void SetRadius(double r) {
      radius_ = r;
    }
    
    inline double GetMaxCamAngle() const {
      return MaxCamAngle_;
    }

    int SetUndistortion(const std::vector<double>& AngleCorrX, 
                        const std::vector<double>& AngleCorrY,
                        const double radius);
      
    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
    */
    virtual inline void Rescale(float ratio) {
      ProjectionParametersBase::Rescale(ratio);
      radius_ /= ratio;
    }

    virtual double ViewDifference(const ProjectionParametersBase* pPPB) const;

    /** @brief Returns a fake KMatrix for the fisheye camera.
        The FoV of the KMatrix is 160deg or the max fisheye angle,
        if this is <160deg.
        @param resolution The resolution parameter influences the 
        Resolution of the perspective image that is represented by the camera.
        0 - The perspective image has the same size as the fisheye image
        1 - The perspective image has the same resolution as the maximum
        fisheye angular resolution up to 160deg.
        2 - The perspective image has the minimum resolution of the 
        fisheye image.
        @param imgsize returns the image size
        @author streckel 05/2006
    */
    BIAS::KMatrix GetFakeKMatrix(double& imgsize, int resolution=0) const;
    BIAS::KMatrix GetFakeKMatrix(int resolution=0) const;

    /** @brief calculates a 3D point in
        a local camera coordinate system specified by camSystem, which 
        belongs to the image position pos in cam with distance depth to the
        camera center cam. Overload to habe return value (0,0,0)
        for undefined or out of image radius pixels.
        @author streckel  */
    virtual Vector3<double> UnProjectToPointLocal(const HomgPoint2D& pos,
                                                  const double& depth,
                                                  bool IgnoreDistortion=false)
      const;


#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif
    
  protected:
  
    /** @brief Transforms a coordinate pair of image coordinates x, y in the 
        uncalibrated spherical image, to a calibrated pair of polar coordinates
        phi, theta.
        @author Birger Streckel */
    int TransfCoordImage2Sphere_ (const HomgPoint2D& Source,
                                 double &theta, double &phi) const;

    /** @brief Transforms a coordinate pair of image coordinates x, y in the 
        calibrated spherical image, to a calibrated pair of polar coordinates
        phi, theta.
        @author Birger Streckel */
    int TransfCoordCalibImage2Sphere_ (const HomgPoint2D& Source,
                                      double &theta, double &phi) const;

    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated 
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2Image_ (double theta, double phi,
                                 HomgPoint2D &Source) const;
    
    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated 
        spherical image.
        @author Birger Streckel */
    int TransfCoordRotateSphere2Image_ (double theta, double phi,
                                       double viewCenterX,
                                       double viewCenterY,
                                       HomgPoint2D &Source) const;
      
    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in an equidistant
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2CalibImage_ (double theta, double phi, 
                                      HomgPoint2D &Source) const;

    /** @brief theta and phi are periodic in 2*M_PI, if they are out of a 
        suitable range, try to find best equivalent angle by adding 2*l*M_PI
        @return 0 on success, <0 indicates that at least one angle is invalid
        @author koeser 10/2004 */
    int EnforceNormalRange_(double& theta, double& phi) const;



    // image radius if image is spherical
    double radius_;
    
    // Maximum image angle of the fisheye camera
    double MaxCamAngle_;

    // Spline interpolators for image distortion 
    // x : angle in an undistorted image
    // y : distorted angle
    // Example: x=pi, y =.8*pi means that the image at angle pi is 
    // distorted with factor 0.8
    Interpolator interpolatorAngle_;
    // back-transformation
    Interpolator interpolatorAngleBack_;
 
  
  };


 
} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

