#include "TFTensorEstimation.hh"
#include "TrifocalTensor.hh"
#include <Base/Geometry/HomgPoint2D.hh>
#include <MathAlgo/SVD.hh>
#include <Base/Math/Matrix.hh>
#include <iostream>

using namespace std;
using namespace BIAS;

#define USE_FOUR_EQUATIONS_PER_POINT
#define TFT_EST_CHECK_EQUATIONS
#define TFT_EST_USE_SVD_HACK


TFTensorEstimation::TFTensorEstimation()
{
}


TFTensorEstimation::~TFTensorEstimation(){}


int TFTensorEstimation::Compute( TrifocalTensor &T,
                                 const vector<HomgPoint2D> &x1,
                                 const vector<HomgPoint2D> &x2,
                                 const vector<HomgPoint2D> &x3,
                                 bool HartleyNormalization /* = true */ )
{
  BIASASSERT( x1.size()==x2.size() && x2.size()==x3.size() );
  if ( x1.size() < 7 )
    {
      BIASERR("Need at least 7 correspondences for computation!");
      return -1;
    }

  int n = x1.size();
  vector<HomgPoint2D> p1, p1norm;
  vector<HomgPoint2D> p2, p2norm;
  vector<HomgPoint2D> p3, p3norm;
  KMatrix K1, K2, K3;

  // original points shall not be touched
  p1 = x1;
  p2 = x2;
  p3 = x3;

  if (HartleyNormalization) // normalize points following Hartley-normalization
    {
      cout << "Performing Hartley normalization..." << endl;
      Norm_.Hartley(p1, p1norm, K1);
      Norm_.Hartley(p2, p2norm, K2);
      Norm_.Hartley(p3, p3norm, K3);
      p1 = p1norm;
      p2 = p2norm;
      p3 = p3norm;
      cout << "K1 = " << K1 << endl;
      cout << "K2 = " << K2 << endl;
      cout << "K3 = " << K3 << endl;
    }
  else
    {
      cerr << "Hartley normalization skipped." << endl;
    }



#ifdef USE_FOUR_EQUATIONS_PER_POINT
  Matrix<TRIFOCALTENSOR_TYPE> A(4*n, 27);
  for(int corr=0; corr<(int)p1.size(); corr++) {
    // first equation for i=1 and l=1 ///////////////////////////////////////
    /* T_1^{11} */ A[corr][0]    =   p1[corr][0];
    /* T_1^{12} */ A[corr][1]    = 0.0;
    /* T_1^{13} */ A[corr][2]    = - p1[corr][0] *               p3[corr][0];
    /* T_1^{21} */ A[corr][3]    = 0.0;
    /* T_1^{22} */ A[corr][4]    = 0.0;
    /* T_1^{23} */ A[corr][5]    = 0.0;
    /* T_1^{31} */ A[corr][6]    = - p1[corr][0] * p2[corr][0];
    /* T_1^{32} */ A[corr][7]    = 0.0;
    /* T_1^{33} */ A[corr][8]    =   p1[corr][0] * p2[corr][0] * p3[corr][0];
    
    /* T_2^{11} */ A[corr][9]    =   p1[corr][1];
    /* T_2^{12} */ A[corr][10]   = 0.0;
    /* T_2^{13} */ A[corr][11]   = - p1[corr][1] *               p3[corr][0];
    /* T_2^{21} */ A[corr][12]   = 0.0;
    /* T_2^{22} */ A[corr][13]   = 0.0;
    /* T_2^{23} */ A[corr][14]   = 0.0;
    /* T_2^{31} */ A[corr][15]   = - p1[corr][1] * p2[corr][0];
    /* T_2^{32} */ A[corr][16]   = 0.0;
    /* T_2^{33} */ A[corr][17]   =   p1[corr][1] * p2[corr][0] * p3[corr][0];

    /* T_3^{11} */ A[corr][18]   =   p1[corr][2];
    /* T_3^{12} */ A[corr][19]   = 0.0;
    /* T_3^{13} */ A[corr][20]   = - p1[corr][2] *               p3[corr][0];
    /* T_3^{21} */ A[corr][21]   = 0.0;
    /* T_3^{22} */ A[corr][22]   = 0.0;
    /* T_3^{23} */ A[corr][23]   = 0.0;
    /* T_3^{31} */ A[corr][24]   = - p1[corr][2] * p2[corr][0];
    /* T_3^{32} */ A[corr][25]   = 0.0;
    /* T_3^{33} */ A[corr][26]   =   p1[corr][2] * p2[corr][0] * p3[corr][0];
    
    // second equation for i=1 and l=2 //////////////////////////////////////
    /* T_1^{11} */ A[corr+n][0] = 0.0;
    /* T_1^{12} */ A[corr+n][1] =   p1[corr][0];
    /* T_1^{13} */ A[corr+n][2] = - p1[corr][0] *            p3[corr][1];
    /* T_1^{21} */ A[corr+n][3] = 0.0;
    /* T_1^{22} */ A[corr+n][4] = 0.0;
    /* T_1^{23} */ A[corr+n][5] = 0.0;
    /* T_1^{31} */ A[corr+n][6] = 0.0;
    /* T_1^{32} */ A[corr+n][7] = - p1[corr][0] * p2[corr][0];
    /* T_1^{33} */ A[corr+n][8] =   p1[corr][0] * p2[corr][0] * p3[corr][1];
    
    /* T_2^{11} */ A[corr+n][9] = 0.0;
    /* T_2^{12} */ A[corr+n][10]=   p1[corr][1];
    /* T_2^{13} */ A[corr+n][11]= - p1[corr][1] *               p3[corr][1];
    /* T_2^{21} */ A[corr+n][12]= 0.0;
    /* T_2^{22} */ A[corr+n][13]= 0.0;
    /* T_2^{23} */ A[corr+n][14]= 0.0;
    /* T_2^{31} */ A[corr+n][15]= 0.0;
    /* T_2^{32} */ A[corr+n][16]= - p1[corr][1] * p2[corr][0];
    /* T_2^{33} */ A[corr+n][17]=   p1[corr][1] * p2[corr][0] * p3[corr][1];
    
    /* T_3^{11} */ A[corr+n][18]= 0.0;
    /* T_3^{12} */ A[corr+n][19]=   p1[corr][2];
    /* T_3^{13} */ A[corr+n][20]= - p1[corr][2] *               p3[corr][1];
    /* T_3^{21} */ A[corr+n][21]= 0.0;
    /* T_3^{22} */ A[corr+n][22]= 0.0;
    /* T_3^{23} */ A[corr+n][23]= 0.0;
    /* T_3^{31} */ A[corr+n][24]= 0.0;
    /* T_3^{32} */ A[corr+n][25]= - p1[corr][2] * p2[corr][0];
    /* T_3^{33} */ A[corr+n][26]=   p1[corr][2] * p2[corr][0] * p3[corr][1];
    
    // third equation for i=2 and l=1 ///////////////////////////////////////
    /* T_1^{11} */ A[corr+2*n][0] = 0.0;
    /* T_1^{12} */ A[corr+2*n][1] = 0.0;
    /* T_1^{13} */ A[corr+2*n][2] = 0.0;
    /* T_1^{21} */ A[corr+2*n][3] =   p1[corr][0];
    /* T_1^{22} */ A[corr+2*n][4] = 0.0;
    /* T_1^{23} */ A[corr+2*n][5] = - p1[corr][0] *               p3[corr][0];
    /* T_1^{31} */ A[corr+2*n][6] = - p1[corr][0] * p2[corr][1];
    /* T_1^{32} */ A[corr+2*n][7] = 0.0;
    /* T_1^{33} */ A[corr+2*n][8] =   p1[corr][0] * p2[corr][1] * p3[corr][0];
    
    /* T_2^{11} */ A[corr+2*n][9] = 0.0;
    /* T_2^{12} */ A[corr+2*n][10]= 0.0;
    /* T_2^{13} */ A[corr+2*n][11]= 0.0;
    /* T_2^{21} */ A[corr+2*n][12]=   p1[corr][1];
    /* T_2^{22} */ A[corr+2*n][13]= 0.0;
    /* T_2^{23} */ A[corr+2*n][14]= - p1[corr][1] *               p3[corr][0];
    /* T_2^{31} */ A[corr+2*n][15]= - p1[corr][1] * p2[corr][1];
    /* T_2^{32} */ A[corr+2*n][16]= 0.0;
    /* T_2^{33} */ A[corr+2*n][17]=   p1[corr][1] * p2[corr][1] * p3[corr][0];
    
    /* T_3^{11} */ A[corr+2*n][18]= 0.0;
    /* T_3^{12} */ A[corr+2*n][19]= 0.0;
    /* T_3^{13} */ A[corr+2*n][20]= 0.0;
    /* T_3^{21} */ A[corr+2*n][21]=   p1[corr][2];
    /* T_3^{22} */ A[corr+2*n][22]= 0.0;
    /* T_3^{23} */ A[corr+2*n][23]= - p1[corr][2] *               p3[corr][0];
    /* T_3^{31} */ A[corr+2*n][24]= - p1[corr][2] * p2[corr][1];
    /* T_3^{32} */ A[corr+2*n][25]= 0.0;
    /* T_3^{33} */ A[corr+2*n][26]=   p1[corr][2] * p2[corr][1] * p3[corr][0];
    
    // fourth equation for i=2 and l=2 //////////////////////////////////
    /* T_1^{11} */ A[corr+3*n][0] = 0.0;
    /* T_1^{12} */ A[corr+3*n][1] = 0.0;
    /* T_1^{13} */ A[corr+3*n][2] = 0.0;
    /* T_1^{21} */ A[corr+3*n][3] = 0.0;
    /* T_1^{22} */ A[corr+3*n][4] =   p1[corr][0];
    /* T_1^{23} */ A[corr+3*n][5] = - p1[corr][0] *               p3[corr][1];
    /* T_1^{31} */ A[corr+3*n][6] = 0.0;
    /* T_1^{32} */ A[corr+3*n][7] = - p1[corr][0] * p2[corr][1];
    /* T_1^{33} */ A[corr+3*n][8] =   p1[corr][0] * p2[corr][1] * p3[corr][1];

    /* T_2^{11} */ A[corr+3*n][9] = 0.0;
    /* T_2^{12} */ A[corr+3*n][10]= 0.0;
    /* T_2^{13} */ A[corr+3*n][11]= 0.0;
    /* T_2^{21} */ A[corr+3*n][12]= 0.0;
    /* T_2^{22} */ A[corr+3*n][13]=   p1[corr][1];
    /* T_2^{23} */ A[corr+3*n][14]= - p1[corr][1] *               p3[corr][1];
    /* T_2^{31} */ A[corr+3*n][15]= 0.0;
    /* T_2^{32} */ A[corr+3*n][16]= - p1[corr][1] * p2[corr][1];
    /* T_2^{33} */ A[corr+3*n][17]=   p1[corr][1] * p2[corr][1] * p3[corr][1];
    
    /* T_3^{11} */ A[corr+3*n][18]= 0.0;
    /* T_3^{12} */ A[corr+3*n][19]= 0.0;
    /* T_3^{13} */ A[corr+3*n][20]= 0.0;
    /* T_3^{21} */ A[corr+3*n][21]= 0.0;
    /* T_3^{22} */ A[corr+3*n][22]=   p1[corr][2];
    /* T_3^{23} */ A[corr+3*n][23]= - p1[corr][2] *               p3[corr][1];
    /* T_3^{31} */ A[corr+3*n][24]= 0.0;
    /* T_3^{32} */ A[corr+3*n][25]= - p1[corr][2] * p2[corr][1];
    /* T_3^{33} */ A[corr+3*n][26]=   p1[corr][2] * p2[corr][1] * p3[corr][1];
  }
#else
  Matrix<TRIFOCALTENSOR_TYPE> A(p1.size(), 27);
  for (int i=0; i<(int)p1.size(); i++)
    {
      // T_1^{11}
      A[i][0]  = p1[i][0];
      // T_1^{12}
      A[i][1]  = p1[i][0];
      // T_1^{13}
      A[i][2]  = -p1[i][0] * p3[i][0] - p1[i][0] * p3[i][1];
      // T_1^{21}
      A[i][3]  = p1[i][0];
      // T_1^{22}
      A[i][4]  = p1[i][0];
      // T_1^{23}
      A[i][5]  = -p1[i][0] * p3[i][0] - p1[i][0] * p3[i][1];
      // T_1^{31}
      A[i][6]  = -p1[i][0] * p2[i][0] - p1[i][0] * p2[i][1];
      // T_1^{32}
      A[i][7]  = -p1[i][0] * p2[i][0] - p1[i][0] * p2[i][1];
      // T_1^{33}
      A[i][8]  =
        p1[i][0] * p2[i][0] * p3[i][0] +
        p1[i][0] * p2[i][0] * p3[i][1] +
        p1[i][0] * p2[i][1] * p3[i][0] +
        p1[i][0] * p2[i][1] * p3[i][1] ;
      // T_2^{11}
      A[i][9]  = p1[i][1];
      // T_2^{12}
      A[i][10] = p1[i][1];
      // T_2^{13}
      A[i][11] = -p1[i][1] * p3[i][0] - p1[i][1] * p3[i][1];
      // T_2^{21}
      A[i][12] = p1[i][1];
      // T_2^{22}
      A[i][13] = p1[i][1];
      // T_2^{23}
      A[i][14] = -p1[i][1] * p3[i][0] - p1[i][1] * p3[i][1];
      // T_2^{31}
      A[i][15] = -p1[i][1] * p2[i][0] - p1[i][1] * p2[i][1];
      // T_2^{32}
      A[i][16] = -p1[i][1] * p2[i][0] - p1[i][1] * p2[i][1];
      // T_2^{33}
      A[i][17] = 
        p1[i][1] * p2[i][0] * p3[i][0] +
        p1[i][1] * p2[i][0] * p3[i][1] +
        p1[i][1] * p2[i][1] * p3[i][0] +
        p1[i][1] * p2[i][1] * p3[i][1] ;
      // T_3^{11}
      A[i][18] = p1[i][2];
      // T_3^{12}
      A[i][19] = p1[i][2];
      // T_3^{13}
      A[i][20] = -p1[i][2] * p3[i][0] - p1[i][2] * p3[i][1];
      // T_3^{21}
      A[i][21] = p1[i][2];
      // T_3^{22}
      A[i][22] = p1[i][2];
      // T_3^{23}
      A[i][23] = -p1[i][2] * p3[i][0] - p1[i][2] * p3[i][1];
      // T_3^{31}
      A[i][24] = -p1[i][2] * p2[i][0] - p1[i][2] * p2[i][1];
      // T_3^{32}
      A[i][25] = -p1[i][2] * p2[i][0] - p1[i][2] * p2[i][1];
      // T_3^{33}
      A[i][26] = 
        p1[i][2] * p2[i][0] * p3[i][0] +
        p1[i][2] * p2[i][0] * p3[i][1] +
        p1[i][2] * p2[i][1] * p3[i][0] +
        p1[i][2] * p2[i][1] * p3[i][1] ;
    }
#endif

#ifdef TFT_EST_CHECK_EQUATIONS
  cout << "A = " << A << endl;
#endif

  SVD svda(A);
  cout << "LeftNullspaceDim  = " << svda.LeftNullspaceDim() << endl;
  cout << "RightNullspaceDim = " << svda.RightNullspaceDim() << endl;

#ifdef TFT_EST_USE_SVD_HACK
  Matrix<double> VT = svda.GetVT();
  Vector<TRIFOCALTENSOR_TYPE> Tvec(VT.num_cols());

  for (int col=0; col < VT.num_cols(); col++)
    {
      Tvec[col] = VT[ VT.num_cols() - 1 ][col];
    }
#else
  Vector<TRIFOCALTENSOR_TYPE> Tvec = svda.GetNullvector();
#endif

  cout << "Tvec = " << Tvec << endl;

#ifdef TFT_EST_CHECK_EQUATIONS
  cout << A * Tvec << endl;
#endif

  T.SetFromVector(Tvec);

#ifdef TFT_EST_CHECK_EQUATIONS
  double result;
  for (unsigned int i=0; i<p1.size(); i++)
    {
      cout << "\nChecking correspondence " << i << endl;
      result = 
          p1[i][0] * p2[i][0] * p3[i][0] * T(0,2,2)
        - p1[i][0] *            p3[i][0] * T(0,0,2)
        - p1[i][0] * p2[i][0] *            T(0,2,0)
        + p1[i][0] *                       T(0,0,0);
      result +=
          p1[i][1] * p2[i][0] * p3[i][0] * T(1,2,2)
        - p1[i][1] *            p3[i][0] * T(1,0,2)
        - p1[i][1] * p2[i][0] *            T(1,2,0)
        + p1[i][1] *                       T(1,0,0);
      result +=
          p1[i][2] * p2[i][0] * p3[i][0] * T(2,2,2)
        - p1[i][2] *            p3[i][0] * T(2,0,2)
        - p1[i][2] * p2[i][0] *            T(2,2,0)
        + p1[i][2] *                       T(2,0,0);
      cout <<"  result 1 is "<<result<<endl;

      result = 
          p1[i][0] * p2[i][0] * p3[i][1] * T(0,2,2)
        - p1[i][0] *            p3[i][1] * T(0,0,2)
        - p1[i][0] * p2[i][0] *            T(0,2,1)
        + p1[i][0] *                       T(0,0,1);
      result +=
          p1[i][1] * p2[i][0] * p3[i][1] * T(1,2,2)
        - p1[i][1] *            p3[i][1] * T(1,0,2)
        - p1[i][1] * p2[i][0] *            T(1,2,1)
        + p1[i][1] *                       T(1,0,1);
      result +=
          p1[i][2] * p2[i][0] * p3[i][1] * T(2,2,2)
        - p1[i][2] *            p3[i][1] * T(2,0,2)
        - p1[i][2] * p2[i][0] *            T(2,2,1)
        + p1[i][2] *                       T(2,0,1);
      cout <<"  result 2 is "<<result<<endl;

      result = 
          p1[i][0] * p2[i][1] * p3[i][0] * T(0,2,2)
        - p1[i][0] *            p3[i][0] * T(0,1,2)
        - p1[i][0] * p2[i][1] *            T(0,2,0)
        + p1[i][0] *                       T(0,1,0);
      result +=
          p1[i][1] * p2[i][1] * p3[i][0] * T(1,2,2)
        - p1[i][1] *            p3[i][0] * T(1,1,2)
        - p1[i][1] * p2[i][1] *            T(1,2,0)
        + p1[i][1] *                       T(1,1,0);
      result +=
          p1[i][2] * p2[i][1] * p3[i][0] * T(2,2,2)
        - p1[i][2] *            p3[i][0] * T(2,1,2)
        - p1[i][2] * p2[i][1] *            T(2,2,0)
        + p1[i][2] *                       T(2,1,0);
      cout <<"  result 3 is "<<result<<endl;

      result = 
          p1[i][0] * p2[i][1] * p3[i][1] * T(0,2,2)
        - p1[i][0] *            p3[i][1] * T(0,1,2)
        - p1[i][0] * p2[i][1] *            T(0,2,1)
        + p1[i][0] *                       T(0,1,1);
      result +=
          p1[i][1] * p2[i][1] * p3[i][1] * T(1,2,2)
        - p1[i][1] *            p3[i][1] * T(1,1,2)
        - p1[i][1] * p2[i][1] *            T(1,2,1)
        + p1[i][1] *                       T(1,1,1);
      result +=
          p1[i][2] * p2[i][1] * p3[i][1] * T(2,2,2)
        - p1[i][2] *            p3[i][1] * T(2,1,2)
        - p1[i][2] * p2[i][1] *            T(2,2,1)
        + p1[i][2] *                       T(2,1,1);
      cout <<"  result 4 is "<<result<<endl;
    }
#endif

  TrifocalTensor T2;
  AlgebraicMinimization(T,T2);

  if (HartleyNormalization)
    {
       T2.premultiply3(K3.Invert(),T);
       T. premultiply2(K2.Invert(),T2);
       T2.postmultiply1(K1,T);
    }
  else
    {
      T = T2;
    }

  return 0;
}

/* The algebraic minimization is described on page 384 of Hartley/Zisserman's
   book Multiple View Geometry in Computer Vision */
int TFTensorEstimation::AlgebraicMinimization(TrifocalTensor &Initial,
                                              TrifocalTensor &Result)
{
  cerr << "The algebraic minimization is not fully implemented yet.\n"
       << "So the result is the initial tensor!" << endl;

  Result = Initial;
  return -1;
  ////////////////////////////////////////////////////////////////////
/*
  cout << "Starting algebraic minimization..." << endl;
  HomgPoint2D e2 = Initial.GetEpipole21(true);
  HomgPoint2D e3 = Initial.GetEpipole31();

  cout << "Epipole in second image is " << e2 << endl;
  cout << "Epipole in third image is  " << e3 << endl;

  Matrix<TRIFOCALTENSOR_TYPE> E(27,18);
  E.SetZero();

  // build matrix E with t=Ea

  for( int i=0; i<3; i++)
    for( int j=0; j<3; j++)
      for( int k=0; k<3; k++)
        {
          E[9*i+3*j+k][3*i+j]   =  e3[k];
          E[9*i+3*j+k][9+3*i+k] = -e2[j];
        }

  cout << "E = " << E << endl;


  // hier fehlt noch was...


  cout << "Algebraic minimization finished." << endl;
  return 0;
  */
}



