#ifndef __DrawTextGL_hh__
#define __DrawTextGL_hh__

#include <bias_config.h>

#ifndef BIAS_HAVE_OPENGL
#  error Please recompile BIAS with USE_OPENGL to use this file. 
#endif // BIAS_HAVE_OPENGL

#ifndef BIAS_HAVE_FONTGL
#  error Please enable BIAS_HAVE_FONTGL to use this class (not completely tested on Linux) (JW)
#endif

// default font size, should be negative, e.g. -14 (pixel)
#define DrawTextGL_DEFAULT_FONTSIZE -14

// default font name
#ifdef WIN32
#  define DrawTextGL_DEFAULT_FONTNAME "Arial"
#else // WIN32
#  define DrawTextGL_DEFAULT_FONTNAME "-*-helvetica-bold-r-normal--24-*-*-*-p-*-iso8859-1"
#endif // WIN32


#include <string>
// BIAS
#include <Base/Common/W32Compat.hh>

namespace BIAS {

/** @brief Draw text to OpenGL as 2D bitmap font.
  Implemented as setting up a 2D display list for every standard character.
  - Creates font by given name and size "on the fly" for all installed fonts, e.g. True Type fonts or GLX fonts.
  - Does not depend on third party libraries (like glut,..)
  - Does not need to read a pre-rendered font from disk (like glf does) 
  - The font has to be available on each client machine at runtime, fallback is tried for unsupported font names. 

  Uses standard Win32 calls to render TTF fonts on Windows
  and GLX fonts on Linux. 

  Tested with MS Windows XP and (Suse 9.2) Linux (JW).

  See Gui/Examples/ExampleGLShared for an Example.

  See http://www.opengl.org/resources/features/fontsurvey/ for other possible implementations, in particular 3D "outlined" fonts. 
 	  @author Jan Woetzel */
class BIASGui_EXPORT DrawTextGL
{
public:
	/// destructor, kills display lists JW
	virtual ~DrawTextGL();

	/// ctor JW
	DrawTextGL();

	/// frees the display lists JW
	void DeleteFont();

	/** set the font up
	@return 0 on success, negativ eon error JW 
	@param theFontname name of teh fornt, e.g. "Arial"
	@param theFontname e.g. "Arial", "Times New Roman" or on Linux:  "-*-helvetica-bold-r-normal--24-*-*-*-p-*-iso8859-1");" 
	@param handleDC pointer to protected GDI device context,
	  on WIN32: a HDC, e.g.  for wxWindow use:  GetDC(HWND(this->GetHandle()))
	  on Linux: .. TODO ..
	@author Jan Woetzel */
	int InitFont(void* handleDC,
		const std::string & theFontname = std::string(DrawTextGL_DEFAULT_FONTNAME) ,
		const int & fontHeight = DrawTextGL_DEFAULT_FONTSIZE
		);

	/** print text on OpenGL screen using current RasterPosition , settings etc. JW */
	void Print(const std::string & text) const;

	/** print text on OpenGL screen at 2D pos (xpos,ypos) in windows (pixel) coordiantes. 
		@param winWidthFallback and @param winHeightFallback are used only as GL 1.1 fallback. JW */
	void PrintWinCoords(const std::string & text,
		const int & xpos, const int & ypos,
		const int & winWidthFallback=640, const int &winHeightFallback=480 ) const; 

	/** print text on OpenGL screen in 0..1 normalized coordinates
	with origin at lower left corner 
	with 0..1 normalizedx coordinates and origin in lower left corner:
  \code
	Y=1
	 ^
     |
     |
	0,0  ---> x=1
  \endcode
	JW */
	void PrintNormCoords(const std::string & text, 
		const float & xpos, const float & ypos) const;

	// use current MVP matrix and draw text in 3D position JW */
	void Print3D(const std::string & text, 
		const float & xpos, const float & ypos, const float & zpos) const;

	/** draw text in 3d at absolute position,
	thus ModelView matrix is set to identity (using push and pop).
	Projection matrix is kept JW */
	void Print3DAbsolute(const std::string & text, 
		const float & xpos, const float & ypos, const float & zpos) const;


  bool IsInitialized() const {return initialized;};

protected: 
	// --- data members: --------------------
	
	/// the name of our font, e.g. "Arial" or "-*-helvetica-bold-r-normal--24-*-*-*-p-*-iso8859-1");" 
	std::string fontname;

	/// true if we created teh font as display lists 
	bool initialized;

	/// inde xof our first display list: 
	unsigned int base;
};

} //namespace BIAS
#endif // __FontGL_HH__

