// Jan Woetzel

#include "ImageCanvasGLBase.hh"

#include <Base/Debug/Error.hh>
#include <Base/Debug/DebugSimple.hh>

#include <Base/Image/ImageIO.hh>
#include <Base/Math/Vector2.hh>

#include "GeometryGL.hh"

// -----------------------------------------
using namespace std;
using namespace BIAS;
// -----------------------------------------


ImageCanvasGLBase::~ImageCanvasGLBase()
{}


ImageCanvasGLBase::ImageCanvasGLBase(
                                     wxFrame* parent
                                     ,wxWindowID id 
                                     ,const wxPoint& pos 
                                     ,const wxSize& size 
                                     ,long style 
                                     ,const wxString& name 
                                     ,int* attribList
                                     ,const wxPalette& palette )
                                     : OpenGLCanvas(parent,id,pos,size,style,name,attribList,palette )
{
  //CALLINFO;
  InitMembers(); // TODO : really required?
}



ImageCanvasGLBase::ImageCanvasGLBase(
                                     wxFrame* parent
                                     ,wxGLContext* sharedContext 
                                     ,wxWindowID id 
                                     ,const wxPoint& pos 
                                     ,const wxSize& size 
                                     ,long style
                                     ,const wxString& name
                                     ,int* attribList 
                                     ,const wxPalette& palette )
                                     : OpenGLCanvas(parent,sharedContext,id,pos,size,style,name,attribList,palette)
{
  //CALLINFO;
  InitMembers();
}


void ImageCanvasGLBase::DisplayCalls()
{
  // draw a fullscreen quad 
  // textured with the image
  if ((texobj.widthBrutto>0) && (texobj.heightBrutto>0))
  {
    this->GetViewController().GetInternals().viewport.DisplayGL();
    GeometryGL::DisplayQuadTexCoord2DFull(
      this->texobj, 
      false /*additionalFlipY*/ );
  }
}


void ImageCanvasGLBase::InitMembers()
{
  // call the parent initializing function 
  OpenGLCanvas::InitMembers();

  // set our own additional members: 
  InterestPoint.Set(0.5, 0.5); 
  scalefactor=1.0f; 
}


void ImageCanvasGLBase::EnableMipMapping(const bool & avoidAliasingByMipmapping)
{
  if(avoidAliasingByMipmapping) {
    texobj.minificationMode=GL_LINEAR_MIPMAP_LINEAR; // trilinear mipmapping
  } else {
    texobj.minificationMode=GL_NEAREST; // accept aliasing 
  };
}


// unique id for OpenGL texture object
GLuint ImageCanvasGLBase::GetTexId() const
{ 
#ifdef BIAS_DEBUG
  if (texobj.id!=0) /* 0 marks unused, currently */
    BIASASSERT(glIsTexture(texobj.id));
#endif
  return texobj.id; 
}


// GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE_NV
GLenum ImageCanvasGLBase::GetTexTarget() const
{ 
  // must be pow2 in Base class which has no Cg. 
  BIASASSERT(texobj.target==GL_TEXTURE_2D);
  return texobj.target; 
}


// valid texobj for capabilities od this base class? 
bool ImageCanvasGLBase::ValidTexobj(const ImgObjGL & t) const
{
  if (!glIsTexture(t.id)) 
  {
    BIASERR("texture id not valid");
    return false;
  } else if (t.target!=GL_TEXTURE_2D) {
    BIASERR("texture target not valid. ImageCanvasBase can only handle GL_TEXTURE_2D");
    return false;
  } else if (!ImageBase::IsPowerOfTwoSize(t.widthBrutto,  t.heightBrutto)) {
    BIASERR("texture must have (brutto) pow2 size. please pad it yourself. npot is not supporte din Base class.");
    return false;
  } else if (t.widthBrutto<=1){
    BIASERR("texture widthBrutto too small");
    return false;
  } else if (t.heightBrutto<=1){
    BIASERR("texture heightBrutto too small");
    return false;
  };
  // OK
  return true;
}


int ImageCanvasGLBase::Set(const ImgObjGL & NewTexobj,
                           const std::string & name)
{
  this->name = name;
  this->SetCurrent();
  if (!ValidTexobj(NewTexobj)) {
    //BIASERR("texture id not valid");
    return -1;
  };

  // OK
  // clear old texobj, first: 
  if (this->texobj.p_valueImage_SelfAllocated) {
    delete this->texobj.p_valueImage;
    this->texobj.p_valueImage=NULL;
  }
  // copy new texobj
  this->texobj = NewTexobj;
  // we should not deallocate because we didn't allocate and copied instead
  this->texobj.p_valueImage_SelfAllocated=false;
  this->name = name;

  // redisplay because we changed the texture
  this->Display();

  return 0; // OK	
}


int ImageCanvasGLBase::Set(BIAS::ImageBase & img,
                           const std::string & name,
                           const bool & forcePow2tex,
                           const bool & flipY
                           )
{
  this->name = name;
  // void SetCurrent on creation: 
  if (GetParent()->IsShown()) this->SetCurrent();
  texobj.CreateGLTexture(img,
    forcePow2tex, // true
    flipY, // true 
    true /* ImmediateCopyData*/  
    );

  // redisplay because we changed the texture
  this->Display();

  return 0; // OK	
}


int ImageCanvasGLBase::SetRECT(BIAS::ImageBase & img,
                               const std::string & name,
                               const bool & flipY )
{
  this->GetTexobj().SetTargetRECT();
  return this->Set(img, 
    name, 
    false, /* forcePow2tex */
    flipY );
}
