/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "BackwardMapping.hh"
#include <Geometry/RMatrix.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <limits>
#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;

namespace BIAS {
 
// (1.0/log(2.0));
#define ONE_OVER_LOG2 1.442695

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
GetSourceCoordinates_(const HomgPoint2D& sink,
                      HomgPoint2D& source) const 
{
  BIASERR("you have to implement derived class function !");
  // for instance, if you simply want to magnify your source image by 2 do a:
  source = sink * 0.5;
  BIASABORT;
  return -1;
}

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
GetJacobian_(const HomgPoint2D& sink, Matrix2x2<double>& Jacobian) const 
{
  // numerical approximatioin of jacobian
  // compute mapped points for 4-neighborhood
  // this is actually a very good way for the jacobian approximation
  // since we use is exactly for the step size which we assumed for
  // computation

  HomgPoint2D source_left, source_right, source_up, source_down;
  HomgPoint2D locsink = sink;
  locsink[0] -= 0.5;
  if (GetSourceCoordinates_(locsink, source_left)) return -1;
  locsink[0] += 1.0;
  if (GetSourceCoordinates_(locsink, source_right)) return -1;
  locsink[0] -= 0.5;
  locsink[1] -= 0.5;
  if (GetSourceCoordinates_(locsink, source_up)) return -1;
  locsink[1] += 1.0;
  if (GetSourceCoordinates_(locsink, source_down)) return -1;

  // difference vectors are rows of jacobian
  Jacobian[0][0] = source_right[0] - source_left[0];
  Jacobian[1][0] = source_right[1] - source_left[1];
  Jacobian[0][1] = source_down[0]  - source_up[0];
  Jacobian[1][1] = source_down[1]  - source_up[1];
  return 0;
}


template <class InputStorageType, class OutputStorageType>
void BackwardMapping<InputStorageType, OutputStorageType>::
UpdatePyramidSize(const Image<OutputStorageType>& sink) {

  // compute maximum size of pyramid
  HomgPoint2D testpoint,tmp;
  double maxscale = 0.0, scale = 0.0;

  // map 4 corners and find maximum local scaling
  // ATTENTION: all points that are mapped to (x,y,0) in the source image
  // cause a warping of infinity. It might be incorrect assume that the corners
  // define the scale range !
  for (unsigned int c=0; c<4; c++) {
    switch (c) {
    case 0:testpoint = HomgPoint2D(0,0); break;
    case 1:testpoint = HomgPoint2D(0,sink.GetHeight()-1); break;
    case 2:testpoint = HomgPoint2D(sink.GetWidth()-1, 
             sink.GetHeight()-1); break;
    case 3:testpoint = HomgPoint2D(sink.GetWidth()-1, 0); break;
    }
    GetSourceCoordinates_(testpoint, tmp);
    // compute local sampling frequency for antialiasing
    Matrix2x2<double> Jacobian;  
    GetJacobian_(testpoint, Jacobian);

    // this is actually dx = NormL2( Jac * (1;0) )
    //              and dy = NormL2( Jac * (0;1) )
    // It tells how far we move in src if we move by one in sink.
    // This is important for complying with the sampling theorem.
    double 
      dx = sqrt(Jacobian[0][0]*Jacobian[0][0]+
      Jacobian[1][0]*Jacobian[1][0]),
      dy = sqrt(Jacobian[0][1]*Jacobian[0][1]+
      Jacobian[1][1]*Jacobian[1][1]);
    if (dx<JACOBI_THRESH) dx = JACOBI_THRESH;
    if (dy<JACOBI_THRESH) dy = JACOBI_THRESH;

    // use larger step to select pyramid level
    scale = (dx > dy)? log(dx)*ONE_OVER_LOG2 : log(dy)*ONE_OVER_LOG2;

    if (scale>maxscale) maxscale = scale;    
  }
  SetPyramidSize(int(maxscale+2.0));

  // restore automatic mode when this function is invoked
  autoPyramidSize_ = true;

}

template <class InputStorageType, class OutputStorageType>
void BackwardMapping<InputStorageType, OutputStorageType>::
SetPyramidSize(const int newsize) 
{
  int oldsize = pyramid_.size();
  pyramid_.SetPyramidSize(newsize);
  if (oldsize>0 && oldsize<newsize) {
    // have to build a bigger pyramid, simply adding levels would be sufficient
    // however this is not supported, so we build a completely new one
    // this should be fixed in PyramidImage somewhen !
    Image<InputStorageType> tmp;
    tmp.StealImage(*pyramid_[0]);
    BuildPyramid_(tmp, true);
  }
  // user is responsible now
  autoPyramidSize_ = false;
}


template <class InputStorageType, class OutputStorageType>
void BackwardMapping<InputStorageType, OutputStorageType>::
BuildPyramid_(const Image<InputStorageType>& source, 
              bool forceInit) 
{
  // if forceinit we use source to build the pyramid, otherwise
  // we check if there is already a pyramid
  if (forceInit) {
    if (!pyramid_.IsEmpty()) pyramid_.Release();
    pyramid_.Init(source);
  } else {
    if (pyramid_.IsEmpty()) pyramid_.Init(source);
  }
}

// #####################  dispatcher functions #########################
template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
Map(const Image<InputStorageType>& source, 
    Image<OutputStorageType>& sink,  
    InterpolationMethod interpolationQuality, bool newSink) 
{

#ifdef BIAS_DEBUG

  // sanity checks
  if (sink.IsEmpty()){
    BIASERR("sink is empty !");
    BIASABORT;
  }
  if (sink.GetChannelCount()!=source.GetChannelCount() ||
    sink.GetColorModel()!=source.GetColorModel() ||
    sink.GetStorageType()!=source.GetStorageType()) {
      BIASERR("different image types !");
      BIASABORT;
    }
    if (source.IsPlanar() && source.GetChannelCount()>1 ||
      sink.IsPlanar() && sink.GetChannelCount()>1) {
        BIASERR("not implemented for planar images !");
        BIASABORT;
      }
#endif
    width_=source.GetWidth();
    height_=source.GetHeight();
    if (newSink) {
      CalcCoordOffset_(sink, source);
      ChangeImgSize_(sink, source);
    } else {
      offsetX_=0; offsetY_=0;
    }
    // record interpolation mode since the simple map interface doesnt 
      interpolationType_ = interpolationQuality;
      if (interpolationQuality>=MapTrilinear) {
        // here we work with a pyramid
        // to avoid computational overhead, we save the pyramid and thus the 
        // source image. We dont need to calculate it every time

        // check if pyramid is large enough for current mapper
        if (autoPyramidSize_) UpdatePyramidSize(sink);
        BuildPyramid_(source);
        ROI_ = *source.GetROI();
        return MapTri_(sink);
      } else {
        ROI_ = *source.GetROI();
        return MapBi_(source, sink, interpolationQuality);
      }
}


template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
Map(Image<OutputStorageType>& sink) {
  if (interpolationType_ < MapTrilinear) {
    BIASERR("Call complete map function, no image caching necessary for"
      " simple mapper, dont have your input image any more !");
    BIASABORT;

  } 
  MapTri_(sink);
  return 0;
}

// #################### bi-Interpolation ################################

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
MapBi_(const Image<InputStorageType>& source,
       Image<OutputStorageType>& sink,  
       InterpolationMethod interpolationQuality) 
{
  HomgPoint2D p_sink(0,0), p_source(0,0);
  p_sink[2] = 1; 

  OutputStorageType** Data = sink.GetImageDataArray();
  double tmp = 0;
  const unsigned int ccs = sink.GetChannelCount();
  // calculate interesting region in sink
  int tlx,tly, brx1,bry1;
  GetBoundingBox(source.GetWidth(), source.GetHeight(),
                 sink.GetWidth(), sink.GetHeight(),
                 tlx, tly, brx1, bry1);
  tlx-=offsetX_;  brx1-=offsetX_; tly-=offsetY_; bry1-=offsetY_; 
  // clip interesting region with sink roi
  ClipBoundingBoxToROICorners_(sink, tlx, tly, brx1, bry1);

  const int brx = brx1, bry = bry1; 

  if (alphaImg_.GetSize()>0) {
    if ( (bry-tly)>(int)alphaImg_.GetHeight() ||
         (brx-tlx)>(int)alphaImg_.GetWidth() )
      BIASERR("alphaImg too small:"<<alphaImg_.GetWidth()<<"x"<<
              alphaImg_.GetHeight()<<"  minimum size:"<<
              (brx-tlx)<<"x"<<(bry-tly));
  }
  const bool alphaMap = (alphaImg_.GetSize()>0);  
  //cout<<"offsetX_:"<<offsetX_<<" offsetY_:"<<offsetY_<<endl;
  //cout<<"tl:"<<tlx<<","<<tly<<" br:"<<brx<<","<<bry<<endl;
  OutputStorageType value;
  // iterate over sink and calculate the corresponding pix in source  
  // check interpolation modes in speed order
  if (interpolationQuality==MapNearestNeighbor) { // nearest neighbor
    const InputStorageType** ppImageData 
      = (const InputStorageType**)source.GetImageDataArray();
    for (int y=tly; y<bry; y++) { 
      for (int x=tlx; x<brx; x++) {
        // image point in homogenous coords
        p_sink[0] = x + offsetX_;
        p_sink[1] = y + offsetY_;
        GetSourceCoordinates_(p_sink, p_source);
        // skip if pixel not in roi
        if (!ROI_.IsInROI(int(rint(p_source[0])),
                          int(rint(p_source[1])))) continue;
        // run over color channels
        for (unsigned int j=0; j<ccs; j++) {         
          value= Data[y][x*ccs+j];
          ClipValue_(ppImageData[(int)rint(p_source[1])]
                     [(int)rint(p_source[0])*ccs+j], value);          
          if (alphaMap) {
            Data[y][x*ccs+j]= 
              OutputStorageType(alphaImg_.GetImageDataArray()[y-tly][x-tlx] *  
                                float(Data[y][x*ccs+j]) +
                                (1.0-alphaImg_.GetImageDataArray()[y-tly][x-tlx]) *
                                float(value)
                                );
          } else
            Data[y][x*ccs+j]=value;
        }
      }      
    }
  } else if (interpolationQuality==MapBilinear) { // bilinear
    for (int y=tly; y<bry; y++) {
      for (int x=tlx; x<brx; x++) {
        // image point in homogenous coords
        p_sink[0] = x + offsetX_;
        p_sink[1] = y + offsetY_;
        GetSourceCoordinates_(p_sink, p_source);
        // skip if pixel not in roi
        if (!ROI_.IsInROI(p_source[0], p_source[1])) continue;
        for (unsigned int j=0; j<ccs; j++) {
           value= Data[y][x*ccs+j];
          ClipValue_(source.BilinearInterpolationRGBInterleaved(p_source[0],
                                                                p_source[1],j),
                     value);        
          if (alphaMap) {
            Data[y][x*ccs+j]= 
              OutputStorageType(alphaImg_.GetImageDataArray()[y-tly][x-tlx] *  
                                float(Data[y][x*ccs+j]) +
                                (1.0-alphaImg_.GetImageDataArray()[y-tly][x-tlx]) *
                                float(value)
                                );
          } else
            Data[y][x*ccs+j]=value;
        }
      }
    }
  } else if (interpolationQuality==MapBicubic) { // bicubic
    for (int y=tly; y<bry; y++) {
      for (int x=tlx; x<brx; x++) {
        // image point in homogenous coords
        p_sink[0] = x + offsetX_;
        p_sink[1] = y + offsetY_;;
        GetSourceCoordinates_(p_sink, p_source);
        // skip if pixel not in roi
        if (!ROI_.IsInROI(p_source[0], p_source[1])) continue;
        // run over color channels
        for (unsigned int j=0; j<ccs; j++) {
          if (GetImageValue_(p_source[0], p_source[1], source, j, tmp)==0) {
             value= Data[y][x*ccs+j];
            ClipValue_(tmp, value);      
            if (alphaMap) {
              Data[y][x*ccs+j]= 
                OutputStorageType(alphaImg_.GetImageDataArray()[y-tly][x-tlx] *  
                                  float(Data[y][x*ccs+j]) +
                                  (1.0-alphaImg_.GetImageDataArray()[y-tly][x-tlx]) *
                                  float(value)
                                  );
            } else
              Data[y][x*ccs+j]=value;
          }
        }
      }
    }
  }

  return 0;
}

// ########################## tri-Interpolatioon #########################

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
MapTri_(Image<OutputStorageType>& sink) {
  // sanity check (who knows ;-) )
  if (pyramid_.IsEmpty()) {
    BIASERR("Called function on uninitialized image !");
    sink.FillImageWithConstValue( (OutputStorageType)(255) );
    return -1;
  }
  // check if pyramid is still large enough, mapping might have changed,
  // e.g. other homography on same image !
  if (autoPyramidSize_) UpdatePyramidSize(sink);
  HomgPoint2D p_sink(0,0), p_source(0,0);

  OutputStorageType** Data = sink.GetImageDataArray();
  double tmp = 0, localscale = 0;
  const unsigned int ccs = sink.GetChannelCount();

  // calculate interesting region in sink by forward mapping of corners
  int tlx,tly, brx1, bry1;
  GetBoundingBox(pyramid_[0]->GetWidth(), pyramid_[0]->GetHeight(),
                 sink.GetWidth(), sink.GetHeight(),
                 tlx,tly,brx1,bry1);
  tlx-=offsetX_;  brx1-=offsetX_; tly-=offsetY_; bry1-=offsetY_;
  // clip interesint region with sink roi
  ClipBoundingBoxToROICorners_(sink, tlx, tly, brx1, bry1);
  const  int brx = brx1, bry = bry1; 

  if (alphaImg_.GetSize()>0) {
    if ( (bry-tly)>(int)alphaImg_.GetHeight() ||
         (brx-tlx)>(int)alphaImg_.GetWidth() )
      BIASERR("alphaImg too small:"<<alphaImg_.GetWidth()<<"x"<<
              alphaImg_.GetHeight()<<"  minimum size:"<<
              (brx-tlx)<<"x"<<(bry-tly));
  }
  const bool alphaMap = (alphaImg_.GetSize()>0);  

  OutputStorageType value;  
  // iterate over sink and calculate the corresponding pix in source  
  for ( int y=tly; y<bry; y++) {
    for ( int x=tlx; x<brx; x++) {
      // image point in homogenous coords
      p_sink[0] = x + offsetX_;
      p_sink[1] = y + offsetY_;
      // this is so slow we can make the decision in the innerst loop:
      if (interpolationType_ == MapTrilinear) {
        GetSourceCoordinates_(p_sink, p_source);
        // skip if pixel not in roi
        if (!ROI_.IsInROI(p_source[0], p_source[1])) continue;

        // compute local sampling frequency for antialiasing
        // this is only an approximation, more exact would be SVD:
        // sqrt(s[0]) from J^T * J = U * s * V^T
        Matrix2x2<double> Jacobian;  
        GetJacobian_(p_sink, Jacobian);

        double 
          dx = sqrt(Jacobian[0][0]*Jacobian[0][0]+
                    Jacobian[1][0]*Jacobian[1][0]),
          dy = sqrt(Jacobian[0][1]*Jacobian[0][1]+
                    Jacobian[1][1]*Jacobian[1][1]);
        if (dx<JACOBI_THRESH) dx = JACOBI_THRESH;
        if (dy<JACOBI_THRESH) dy = JACOBI_THRESH;

        // use larger step to select pyramid level
        localscale = (dx > dy)? log(dx)*ONE_OVER_LOG2 : log(dy)*ONE_OVER_LOG2;

        // clip scale to zero since we have no better resolution image than 0
        if (localscale<0.0) localscale = 0.0;        

        // compute color channels
        // this can be speeded up, all the weights are the same for all
        // channels, dont recompute them every time !
        // However for now this code does it ..
        for (unsigned int j=0; j<sink.GetChannelCount(); j++) {
          if (GetTrilinearImageValue_(p_source[0], p_source[1],
                                      localscale, j, tmp) == 0) {
             value= Data[y][x*ccs+j];
            ClipValue_(tmp, value);
            if (alphaMap) {
              Data[y][x*ccs+j]= 
                OutputStorageType(alphaImg_.GetImageDataArray()[y-tly][x-tlx] *  
                                  float(Data[y][x*ccs+j]) +
                                  (1.0-alphaImg_.GetImageDataArray()[y-tly][x-tlx]) *
                                  float(value)
                                  );
            } else
              Data[y][x*ccs+j]=value;
          }
        }
      } else {
        // anisotropic filtering !
        GetSourceCoordinates_(p_sink, p_source);

        // skip if pixel not in roi
        if (!ROI_.IsInROI(p_source[0], p_source[1])) continue; 

        // compute local sampling frequency matrix
        Matrix2x2<double> Jacobian;
        GetJacobian_(p_sink, Jacobian);        

        // compute color channels
        // this can be speeded up, all the weights are the same for all
        // channels, dont recompute them every time !
        // However for now this code does it ..
        for (unsigned int j=0; j<sink.GetChannelCount(); j++) {
          if (GetAnisotropicImageValue_(p_source[0], p_source[1],
                                        Jacobian, j, tmp) == 0) {
             value= Data[y][x*ccs+j];
            ClipValue_(tmp, value);
            if (alphaMap) {
              Data[y][x*ccs+j]= 
                OutputStorageType(alphaImg_.GetImageDataArray()[y-tly][x-tlx] *  
                                  float(Data[y][x*ccs+j]) +
                                  (1.0-alphaImg_.GetImageDataArray()[y-tly][x-tlx]) *
                                  float(value)
                                  );
            } else
              Data[y][x*ccs+j]=value;
          }
        }
      }
    }
  }
#ifdef BIAS_DEBUG
  if (!rangeChecked_) {
    rangeChecked_ = true;
    BIASERR("You may have got artefacts in your image because overflow "
            <<"checking of your data types combination "
            <<typeid(InputStorageType).name()<<" -> "
            <<typeid(InputStorageType).name()<<" is not yet implemented."
            <<" Please do it in CheckValueRange_().");
  }
  if (aliasing_) {
    aliasing_ = false;
    BIASERR("You probably got aliasing, because your pyramid was too small.");
  }
#endif

  return 0;
}


template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
GetImageValue_(const double& x, 
               const double& y,
               const Image<InputStorageType>& im,
               unsigned int channel,
               double& result) 
{
  // compute value only in image (no pyramid), first try bicubic
  if (x < (double)im.GetWidth()-2.0 && y < (double)im.GetHeight()-2.0 &&
    x >= 1.0 && y >= 1.0 ) {
      result = im.BicubicInterpolation(x, y, channel);

    } else if (x < (double)im.GetWidth()-1.0 && 
      y < (double)im.GetHeight()-1.0 &&
      x > 0.0 && y > 0.0 ) {
        // we are near the border, bilinear
        result = im.BilinearInterpolation(x, y, channel);
      } else {
        // we are AT the border, we could do nearest neighbor here or something 
        // else, but for now, skip this pixel
        return -1;
      }
      return 0;
}

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
GetTrilinearImageValue_(const double& xsource, 
                        const double& ysource, 
                        const double& scale, 
                        unsigned int channel,
                        double& thereturnvalue) 
{
  // compute a weight for the smaller and a weight for the larger pyramid level
  // then compute the colors at these levels and average them according
  // to their weights
  double value1 =  0, value2 = 0;
  double scale_error = scale - int(scale);
  int scale1 = int(scale);
  if (scale1>=(int)pyramid_.size()-1) {
    // we might get aliasing, pyramid is too small !!!
    scale1 = pyramid_.size()-1;
    scale_error = 0.0;
    aliasing_ = true;
  } else if (scale1<0) {
    // use max resolution image, dont need to upsample in pyramid
    scale1 = 0;
    scale_error = 0.0;
  }

  // compute coordinates in smaller image
  double x1 = xsource, y1 = ysource;
  const double offset = pyramid_.GetPositionOffset();
  for (int s=0; s<scale1; s++) {
    // -0.5 necessary because of offset in 2^n downsampling
    x1 = (x1 - offset) * 0.5; 
    y1 = (y1 - offset) * 0.5;
  }
  // if scale 1 fits perfectly (at lowest highest pyramid layer)
  // we dont need scale2, set it to the same value to avoid access violations
  const int scale2 = (scale_error==0.0)? scale1           : scale1 + 1;
  const double x2  = (scale_error==0.0)? x1               : (x1 - offset)*0.5; 
  const double y2  = (scale_error==0.0)? y1               : (y1 - offset)*0.5;
  

  double weight1 = 1.0 - scale_error, weight2 = scale_error;

  if (GetImageValue_(x1, y1, *pyramid_[scale1], channel, value1)!=0) weight1=0;

  // compute second scale only if necessary
  if (weight2>0.0) {
    if (GetImageValue_(x2, y2, *pyramid_[scale2], channel, value2)!=0) 
      weight2=0;
  }
  // check if we computed any value at all
  const double weightsum = weight1 + weight2;
  if ((weightsum)<=DBL_EPSILON) return  -1;

  // this is the tri in trilinear:
  thereturnvalue = (weight1 * value1 + weight2 * value2) / weightsum;

  return 0;
}

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
GetAnisotropicImageValue_(const double& xsource, 
                          const double& ysource, 
                          Matrix2x2<double>& Jacobian, 
                          unsigned int channel,
                          double& thereturnvalue) 
{
  // compute eigenvalues of jacobian, which determine the largest step sizes
  Vector2<double> ev1, ev2;
 
  // make postive definite jTj for gauss distribution
  Jacobian = Jacobian.Transpose() * Jacobian;
  
  // compute eigenvalues = squared step sizes in source
  double d1, d2;
  Jacobian.EigenvalueDecomposition(d1, ev1, d2, ev2);

  // remove square, require min eigenvalue, avoid pathological 0 case 
  // (should rarely happen for meaningful mappings)
  if (d1<JACOBI_THRESH) d1 = JACOBI_THRESH; else  d1 = sqrt(d1);
  if (d2<JACOBI_THRESH) d2 = JACOBI_THRESH; else  d2 = sqrt(d2);

  // compute pyramid levels  
  const double smallscale = (d1 < d2)? log(d1)*ONE_OVER_LOG2
    :log(d2)*ONE_OVER_LOG2;
  const double largescale = (d2 < d1)? log(d1)*ONE_OVER_LOG2
    :log(d2)*ONE_OVER_LOG2;
  
  // use trilinear if nearly isotropic
  if (largescale-smallscale<0.1) {
    return GetTrilinearImageValue_(xsource, ysource, smallscale, 
                                   channel, thereturnvalue);
  }
  
  if (largescale<=0.1) {
    // bicubic is sufficient since we are only enlarging
    return GetImageValue_(xsource, ysource, 
                          *pyramid_[0], channel, thereturnvalue);
  }

  // make small offset to account for ground gauss smoothing
  const double scale1 = (smallscale>0.3)? smallscale - 0.1 : 0.2;

  // now remove the chosen pyramid level from the jacobian
  // at that level, the jacobian is smaller
  Jacobian *=  1.0 / exp(8.0 * ONE_OVER_LOG2 * scale1);

  // select kernel size
  const int MaskHalfSize = (int)(floor(2*fabs(d2-d1)));

  Matrix<double> theMask(MaskHalfSize*2+1, MaskHalfSize*2+1);
  JacobianToMask_(Jacobian, theMask);
  double* pMaskData = theMask.GetData();
  double weightsum = 0, thevalue = 0;
  thereturnvalue = 0;
  for (int y=-MaskHalfSize; y<=MaskHalfSize; y++) {
    for (int x=-MaskHalfSize; x<=MaskHalfSize; x++) {
      if (GetTrilinearImageValue_(xsource+double(x), ysource+double(y),
                                  scale1, channel, thevalue)==0) {
        thereturnvalue += thevalue * *pMaskData;
        weightsum += *pMaskData++;
      }
    }
  }
  // check if more than half of the pixels needed for interpolation were there
  if (weightsum > 0.5) thereturnvalue /= weightsum;
  else return -1;

  return 0;
}


// ################## data type clipping: double -> char, ... ############

template <class InputStorageType, class OutputStorageType>
void BackwardMapping<InputStorageType, OutputStorageType>::  
ClipBoundingBoxToROICorners_(const Image<OutputStorageType>& sink, 
                             int& tlx, 
                             int& tly, 
                             int& brx, 
                             int& bry) const
{
  unsigned int minx, miny, maxx, maxy;
  sink.GetROICorners(minx, miny, maxx, maxy);
  if (tlx<(int)minx) tlx = minx;
  if (tly<(int)miny) tly = miny;
  if (brx>(int)maxx) brx = maxx;
  if (bry>(int)maxy) bry = maxy;
//   if (tlx<0) tlx=0;
//   if (tly<0) tlx=0;
//   if (brx<0) brx=0;
//   if (bry<0) bry=0;
}




// generic implementation tries to use limits
template <class InputStorageType, class OutputStorageType>
void BackwardMapping<InputStorageType, OutputStorageType>::
ClipValue_(const double& value, OutputStorageType& newvalue) 
{
#ifdef BIAS_DEBUG
  rangeChecked_ = false;
#endif
  
  if (value>=double(numeric_limits<OutputStorageType>::max()))
    newvalue = numeric_limits<OutputStorageType>::max();
  else if (value<=double(numeric_limits<OutputStorageType>::min()))
    newvalue = numeric_limits<OutputStorageType>::min();
  else newvalue = OutputStorageType(value);

}


// if this does not compile on windows, please enable this:
//#ifndef WIN32

template <>
void BackwardMapping<unsigned char, unsigned char>::
ClipValue_(const double& value, unsigned char& newvalue) 
{
  if (value<0.0) newvalue = 0;  
  else if (value>255.0) newvalue = 255;
  else newvalue =  ((unsigned char)rint(value));
}

template <>
void BackwardMapping<float, float>::
ClipValue_(const double& value, float& newvalue) 
{
  newvalue = float(value);
}

//#endif // ! WIN32

template <class InputStorageType, class OutputStorageType>
void BackwardMapping<InputStorageType, OutputStorageType>::
JacobianToMask_(const Matrix2x2<double>& Jacobian, 
                Matrix<double>& theMask) const 
{
  double theSum = 0;
  Matrix2x2<double> invJac;
  Jacobian.Invert(invJac);
  const double& a = invJac[0][0];
  const double& b = invJac[0][1];
  const double& c = invJac[1][0];
  const double& d = invJac[1][1];

  const int halfsize = theMask.num_cols()/2;
  double* pMaskData = theMask.GetData();
  for (int y=-halfsize; y<=halfsize; y++) {
    for (int x=-halfsize; x<=halfsize; x++) {
      // e^(- x^T  * invJac * x^T) is a gauss, unrolled:
      theSum += *pMaskData++ = exp(-(a*x*x+(b+c)*x*y+d*y*y));
    }
  }
  // Normalize to unity
  theMask *= 1.0 / theSum;
}

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
CalcCoordOffset_(const  Image<OutputStorageType> &sink, 
                 const Image<InputStorageType> &source) {
  //sink bounding box
  int xl=0;  int xr=0; int  yl=0; int yr=0; 
  GetBoundingBox(sink.GetWidth(), sink.GetHeight(),
                 source.GetWidth(), source.GetHeight(), xl, yl, xr, yr);
    
  // sink corner coordinates
  double xls=0;  double xrs=sink.GetWidth();
  double  yls=0; double yrs=sink.GetHeight();
  

  // result bounding box
  double xlr=0;  double xrr=0; double  ylr=0; double yrr=0;
  if (xl<xls) xlr=xl; else xlr=xls;
  if (yl<yls) ylr=yl; else ylr=yls;
  if (xr>xrs) xrr=xr; else xrr=xrs;
  if (yr>yrs) yrr=yr; else yrr=yrs;
  
  offsetX_= (int)ceil(xlr);
  offsetY_= (int)ceil(ylr);
  
  return 0; 
}

template <class InputStorageType, class OutputStorageType>
int BackwardMapping<InputStorageType, OutputStorageType>::
ChangeImgSize_ (Image<OutputStorageType> &sink,  
                const Image<InputStorageType> &source) {
  BIAS::Image<OutputStorageType> temp;
  int xl=0;  int xr=0; int  yl=0; int yr=0;
 
  GetBoundingBox(source.GetWidth(), source.GetHeight(), 
                 sink.GetWidth(), sink.GetHeight(), xl, yl, xr, yr);
       
  // sink corner coordinates
  double xls=0;  double xrs=sink.GetWidth();
  double  yls=0; double yrs=sink.GetHeight();
  
  // result bounding box
  double xlr=0;  double xrr=0; double  ylr=0; double yrr=0;
  if (xl<xls) xlr=xl; else xlr=xls;
  if (yl<yls) ylr=yl; else ylr=yls;
  if (xr>xrs) xrr=xr; else xrr=xrs;
  if (yr>yrs) yrr=yr; else yrr=yrs;
  
  int x_abs = (int)ceil(abs(xlr-xrr));
  int y_abs = (int)ceil(abs(ylr-yrr));
  
 
  
  temp.Init(x_abs,y_abs,3);

    
  int w = sink.GetWidth();
  int h = sink.GetHeight();

  
  OutputStorageType** data = sink.GetImageDataArray();
  OutputStorageType** datac = temp.GetImageDataArray();
  int xt = (int)abs(offsetX_);
  int yt = (int)abs(offsetY_);
  cout << "translation at tnc " << xt << " " << yt << "\n";    
  for (int hi =0; hi<h; hi++) {
    for (int wi=0; wi<w; wi++) {
      for (int c=0; c<3; c++) {
        datac[hi+yt][(wi+xt)*3+c]=data[hi][wi*3+c];
        //cout << hi << " " << wi << "\n";
        // cout <<  hi + 3*c +yt << " " << wi +xt << "\n";
      }
    }
  }
  sink=temp;

  return 0;
}

template class BackwardMapping<unsigned char, unsigned char>;
template class BackwardMapping<float, float>;
template class BackwardMapping<unsigned char, float>;

#if defined(BUILD_IMAGE_CHAR)
template class BackwardMapping<unsigned char, char>;
template class BackwardMapping<char, char>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class BackwardMapping<unsigned short,unsigned short>;
#endif

#if defined(BUILD_IMAGE_SHORT)
template class BackwardMapping<short, short>;
#endif

#if defined(BUILD_IMAGE_SHORT)&&defined(BUILD_IMAGE_USHORT)
template class BackwardMapping<unsigned short, short>;
#endif

#if defined(BUILD_IMAGE_INT)
template class BackwardMapping<int,int>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class BackwardMapping<unsigned short,float>;
#endif

#if defined(BUILD_IMAGE_USHORT) && defined(BUILD_IMAGE_INT)
template class BackwardMapping<unsigned short,int>;
#endif

#if defined(BUILD_IMAGE_DOUBLE)
template class BackwardMapping<double,double>;
#endif

} // namespace BIAS
