/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __COLORHISTOGRAM_HH__
#define __COLORHISTOGRAM_HH__
#include <Base/Common/BIASpragmaStart.hh>

#include <vector>
#include <Base/Image/Image.hh>
#include <Base/Math/Math.hh>



// start from big bits, because small bits are used in Image
#define D_COLOR_HIST_VEC 0x8000000

namespace BIAS{


    /**
    @class ColorHistogram
    @ingroup g_image
    This Class handels Histogram matters on Images.
    It represents an histogram and offers basic methods
    on histograms.
    @author Ingo Schiller
    Histogram is represented in HSL by an double vector + vector for L values 
    @status untested
    */

    template <class StorageType>
    class BIASImage_EXPORT ColorHistogram 
    {
    public:
        ///std Con-/Destructor 
        ColorHistogram();
        ~ColorHistogram();


        /**
        Deletes current Histogram information
        */
        int Clear();


        /**
        set the coords for the ROI
        */
        int SetROI(int ux, int uy, int lx, int ly);

        /**
        set the binsize (must be divideable by 2pown)
        */
        int SetBinSize(int binsize);

        /**
        set the binsize for different binSizes (must be divideable by 2pown)
        */
        int SetBinSize(int binSizeHue, int binSizeSat);


        /**
        set minsat_, which defines the minimum saturation
        */
        inline void SetMinSat(int minsat)
        {minSat_ = minsat;}

        /**
        sets Lambda, which is used to calculate the dist
        in CalcSimilarity
        */
        inline void SetLambda(int lambda)
        {lambda_=lambda;}
        /**
        sets expDist_ true or false, used to calculate the dist
        in CalcSimilarity
        */
        inline void SetExpDist(bool expdist)
        {expDist_ = expdist;}

        /**
        GetBinSize
        @return the binSize of Hue and light
        */
        inline int GetBinSize() const {return binSize_;} 

        /**
        GetBinSizeSat
        @return the binSize of Sat
        */
        inline int GetBinSizeSat() const {return binSizeSat_;} 

        /**
        returns the histogram with H&S values
        */
        std::vector<std::vector<double> > GetHist() const;

        /**
        returns the L-values of the HSLcolorspace that are <MINSAT in the ROI
        */
        std::vector<double> GetHistL() const;


        /**
        @returns the Height of the ROI
        */
        int GetHeightROI() const;

        /**
        @returns the width of the ROI
        */
        int GetWidthROI() const;

        /**
        @returns the coords of ROI in an array
        */
        void GetROI(std::vector<int>& roi) const;

        /** 
        calcluates the histogram from data and adds them to the internal data
        structures
        @author schiller 
        @param image = the image
        @param ux = upper left corner x of region of interest 
        @param uy = upper left corner y of region of interest
        @param lx = lower right corner x of region of interest 
        @param ly = lower right corner y of region of interest  
        */
        int GenerateHist(BIAS::Image<StorageType>& image, 
            int ux, int uy, int lx, int ly);

        /**
        calculates the Histogram of the image
        ATTENTION: ROI must be set in upper function SetROI if this function 
        is used
        @param image = the image
        @returns -1 if image does not exist
        */
        int GenerateHist(BIAS::Image<StorageType>& image);



        /** Updates the histogramm with values from hist, 
        where the factor gives the weighting of the new values from hist.
        @return 0 in case of no error
        @author Daniel Grest, March 2004
        */
        int Update(const BIAS::ColorHistogram<StorageType> &hist, 
            double factor=0.5);

        /**
        Calculates the Bhattacharyya similarity Coeiffient^2
        which is the distance or difference of 2 colorhistograms
        @param the chist to compare with
        @returns double value, which indicates the similarity. 1 if similar.
        @author schiller March 2004
        */
        double CalcSimilarity(const BIAS::ColorHistogram<StorageType> chist) const;

        /**
        Calculates the Bhattacharyya similarity Coeiffient^2
        which is the distance or difference of 2 colorhistograms.
        returns a weight, which is in [0..1] and is the percentage of the bins
        unequal zero.
        @param the chist to compare with
        @returns double value, which indicates the similarity. 1 if similar.
        @author grest May 2004
        */
        double CalcSimilarity(const BIAS::ColorHistogram<StorageType> chist, 
            double &weight) const;

        /** returns the percentage [0..1] of HS-bins in the histogram 
        that have values
        unequal to zero
        @author grest, Mar 2005
        */
        float GetSaturatedBins() const;


        /**   
        checks the image quality depending on the given Histogram 
        and increases ,decreases shutter and gain
        @param percent = percentage of pixels in the luminance vector required
        @param hist = the colorhistogram of the interesting part of the image
        @param shutter = shutter value
        @param gain = gain value
        @returns -1 if lum too high , 1 if lum too low , 0 if lum is good
        @author schiller
        */
        int CheckImageQuality(float percent,
            float &shutter,
            float &gain);



        /**
        same as above but luminance vector is created form
        all pixels in histogram
        @returns -1 if lum too high , 1 if lum too low , 0 if lum is good
        */
        int CheckImageQualityComplete(float percent,
            float &shutter,
            float &gain);



        /**
        same as above but checks similarity of luminance to a 
        given reference histogram
        @param cHist = the reference colorHistogram
        @returns -1 if lum too high , 1 if lum too low , 0 if lum is good
        */
        int CheckImageQualityRef(ColorHistogram<StorageType>& cHist,
            float percent,
            float &shutter,
            float &gain);

        /**
        dumps information on histogram
        @param os = the outstream, default std::cout
        @author Schiller
        */
        void Dump(std::ostream& os=std::cout);
        void DumpCSV(std::ostream& os=std::cout);

        /** returns the length of a dumped ColorHistogram
        this calculates of binsize, binsizesat (1 each)
        coords              (4)
        the matrix          (nrofbinssat * nrofbins)
        lvector             (nrofbins)
        luminance vector    (nrofbins)
        total pixelcount    (1)
        @author Ingo Schiller
        */
        int GetNumberOfEntries();

        /**
        reads a histogram from instream and fills the internal data 
        structures
        @param is = the instream  
        @author Schiller March 2004
        @returns 0 if ok, -1 if no file
        */
        int Read(char *filename);

        char* GetBinaryData(int &length);

        void ReadBinaryData(char *data);


    public:
        //vector which holds complete luminance information of histogram.
        std::vector<double> luminance_ ;


    protected:
        /**
        Normalises the histogram to be able to use it in the Bhatt... Difference
        */
        void Normalise_();

        /**
        deletes hist_, histl_...
        */
        int DeleteHist_();

        bool normalised_;

        unsigned short int histCount_;
        unsigned int binCount_; // number of bins in histogram
        BIAS::Image<StorageType> *image_;
        std::vector<std::vector<double> > hist_;// histogram
        std::vector<double> histl_; //the l-values below MINSAT

        int ux_,uy_,lx_,ly_;

        double minBinVal_, maxBinVal_; 

        //modifieable Parameters for GUI
        int binSize_;
        int binSizeSat_;
        int shift_; // used for shifting instead of dividing
        int shiftSat_;
        int shiftLum_;
        int minSat_,lambda_;
        bool expDist_;
    }; // class


} // namespace BIAS

#include <Base/Common/BIASpragmaEnd.hh>
#endif // __COLORHISTOGRAM_HH__
