/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/#ifndef __JPEGHANDLER_HH__
#define __JPEGHANDLER_HH__

/*
 * Include I/O handling and streams
 */
#include <stdio.h>
#include <sstream>
#include <string.h>

/* 
 *  include BIAS-image-libs
 */
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>

#ifdef WIN32
#include <windows.h>
#endif

using namespace std;
using namespace BIAS;

namespace BIAS {

  /**
     \class COmpressJpeg
     \author apetersen
     \brief Wrapper for fast libjpeg methods

     Class for mem-to-mem JPEG-compression. Compress a given BIAS::Image
     (unsigned char) to data-array in shared memory file or decompress a
     given JPEG-file located in user memory into a BIAS::Image (u. char)

     Author: A. Petersen, 15.8.`05
  */

  class JpegHandler {

  public:
    /** Standard constructor */
    JpegHandler();

    /** Standard destructor */
    virtual ~JpegHandler();

    /**
       (Re-)Initialize object with standard values and limit shared 
       memory area to 'size' (in Byte, default 1MB)

       !!! Init object BEVOR first use !!!

       (reinit only requiered for changing buffer size)
    */
    int Init(long size = 1048576);

    /**
       Releases allocated mem-structures. Use it to regain memory but 
       call Init again bevor using object afterwards!!!
    */
    void Release();

    /**
       Start compression for given BIAS::Image
    */
    int Compress(BIAS::Image<unsigned char> &srcImg, int quality=60);

    /**
       Start compression for image
    */
    int Compress(unsigned int width, unsigned int height, unsigned int channels,
                 unsigned char* data, int quality=60);

    /**
       Start decompression of previous given data (from ReadJPEG or 
       SetCompressedData), results will be stored in 'destImg'.
       Decompress will initialize 'destImg' properly if it isn't.
    */
    int Decompress(BIAS::Image<unsigned char> &destImg);

    /**
       Start decompression of previous given data (from ReadJPEG or 
       SetCompressedData), results will be stored in 'data','width',....
       New mem for 'data' will only be allocated if 'data == NULL', else 
       old allocated mem will be used (!!!so set data=NULL if not shure 
       about it's size!!!)
    */
    int Decompress(unsigned int &width, unsigned int &height, unsigned int &channels,
                   unsigned char* &data);
	
    /**
       Get pointer to memory where to find compressed data and 
       its length in byte
       NO copy will be used, so make sure it's still valid when 
       starting Decompress or reusing this!!!
    */
    inline void GetCompressedData(void* &data, long &length) {
      data = data_;
      length = dataLength_;
    }
    
    /**
       Copy memory-area in which compressed data can be found
       and set its length in byte.
    */
    inline void SetCompressedData(void* data, long length) {
      dataLength_ = length;
      memcpy(data_, data, dataLength_);
    }

    /**
       Write previously compressed data to file named filename
    */
    int WriteJPEG(const string &filename);

    /**
       Read JPEG-file named 'filename' and store its informations in
       compressed data field
    */
    int ReadJPEG(const string &filename);

  protected:
    /* Pointer to compression-buffer (compressed data)*/
    void* data_;

    /* Length in byte of compressed data */
    long dataLength_;

    /* Length in byte of compression-buffer */
    long buffsize_;

    /* Info if object has been correctly set */
    bool setup_;

  };

}

#endif //__JPEGHANDLER_HH__
