/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "MatlabTBUndistortion.hh"
#include <iostream>
#include <math.h>

#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;
  

MatlabTBUndistortion::MatlabTBUndistortion() {
  LuT_ = NULL;
  IsInitialized_=false;
}

MatlabTBUndistortion::MatlabTBUndistortion(const CameraParam& cp) {
  LuT_ = NULL;
  Init(cp);
}

MatlabTBUndistortion::~MatlabTBUndistortion(){
  if (LuT_ !=NULL)  delete[] LuT_;
}

int MatlabTBUndistortion::Init(BIAS::Projection &projection)
{
  CameraParam cp;
  ProjectionParametersPerspective* ppp;
  double k1,k2,k3,k4;
  unsigned int w,h;
  double princX, princY;
  double aspectratio;
  double focallength;

  if((dynamic_cast<ProjectionParametersPerspective *>(projection.GetParameters()))!=NULL)
    {
      ppp = dynamic_cast<ProjectionParametersPerspective *>(projection.GetParameters());
    }
  else{
    BIASERR("Not a perspective camera, return false"); 
    BIASABORT;
  }
  ppp->GetUndistortion(k1,k2,k3,k4);
  ppp->GetImageSize(w,h);
  ppp->GetPrincipal(princX, princY);
  ppp->GetFocalLength(focallength);
  aspectratio = ppp->GetAspectratio();
  //cellY = ppp->GetCellSizeY();
  std::vector<double> dummy;

  cp.InitPerspective("Perpspective Camera" ,0 ,"generic lens" , w, h, princX, princY, focallength, aspectratio, k1, k2, k3, k4,dummy ,dummy );


  Init(cp);

  return 0;

}

int MatlabTBUndistortion::Init(const CameraParam& cp)
{  
  if (cp.IsPerspective()) {
    CameraParam::Init(cp);
    if (zoomCamera_) {
      kc1Zoom_ =  knownparams_vect_[0].kc1;
      kc2Zoom_ =  knownparams_vect_[0].kc2;
      kc3Zoom_ =  knownparams_vect_[0].kc3;
      kc4Zoom_ =  knownparams_vect_[0].kc4;
    } 
    return 0;
  } else {
    BIASERR("Camera parameters for spherical camera given!");
    IsInitialized_=false;
    return -1;
  }
}


void MatlabTBUndistortion::SetZoom(const double abszoom){
  if (!IsInitialized_){
    BIASERR("Not inited!");
    return;
  }
  if (zoomCamera_) {
    zoom_=abszoom;      
    focallengthZoom_ = InterpolateFocalLength_();
    cout << "interpolating zoom:" << abszoom << " to: " <<
      focallengthZoom_ << endl;
  } else {
    BIASERR("No zoom camera!");
    return;
  }
}


void MatlabTBUndistortion::SetFocalLength(const double focallength){
  if (!IsInitialized_){
    BIASERR("Not inited!");
    return;
  }
  if (zoomCamera_) {
    focallengthZoom_=focallength;
    zoom_=InterpolateAbsZoom_();
  } else {
    BIASERR("No zoom camera!");
    return;
  }
}


void MatlabTBUndistortion::SetUnzoomed(){
  if (!IsInitialized_){
    BIASERR("Not inited!");
    return;
  }
  if (focallengthDef_==0) {
    kc1Zoom_=knownparams_vect_[0].kc1;
    kc2Zoom_=knownparams_vect_[0].kc2;
    kc3Zoom_=knownparams_vect_[0].kc3;
    kc4Zoom_=knownparams_vect_[0].kc4;
  } else {
    kc1Zoom_=kc1Def_;
    kc2Zoom_=kc2Def_;
    kc3Zoom_=kc3Def_;
    kc4Zoom_=kc4Def_;
  }
  CameraParam::SetUnzoomed();
}

void MatlabTBUndistortion::Undistort(BIAS::HomgPoint2D& point2d) const {
  if (!IsInitialized_){
    BIASERR("Not inited!");
    return;
  }
  Undistort(point2d[0],point2d[1]);
}

void MatlabTBUndistortion::Undistort(double& origx,double& origy) const {
  if (!IsInitialized_){
    BIASERR("Not inited!");
    return;
  }
  // do nothing in case of ideal lense
  if (kc1Def_==0 && kc2Def_==0 && kc3Def_==0 && kc4Def_==0) return;

// copy of distorted point
  double distx,disty;
  double x,y;
  double twoxy;
  
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  double kc1, kc2, kc3, kc4;
  double f=1;

  if (zoomCamera_) {
    kc1=kc1Zoom_;
    kc2=kc2Zoom_;
    kc3=kc3Zoom_;
    kc4=kc4Zoom_;
    f=focallengthZoom_;
  } else {
    kc1=kc1Def_;
    kc2=kc2Def_;
    kc3=kc3Def_;
    kc4=kc4Def_;
    f=focallengthDef_;
  }
     
  //  cout << "Undistorting Point (" << origx << "/" << origy << ")";  

  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = origx = (origx-principalX_)/f;
  y = origy = (origy-principalY_)/(aspectratio_*f);
  
  // 5 iterations are enough
  for (unsigned int i=0;i<5;i++){
    
    //cout << "Iteration: " << i << ", X:" << x << ", Y:" << y << endl;
    
    twoxy = 2 * x*y;
    rsquared = x*x + y*y;
    
    // radial distortion:
    raddist = 1 + kc1*rsquared + kc2*rsquared*rsquared;
    // tangential distortion:
    tandistx = kc3*twoxy + kc4 * (rsquared + 2*x*x);
    tandisty = kc4*twoxy + kc3 * (rsquared + 2*y*y);
    
    // apply distortion:
    distx = raddist * x + tandistx;
    disty = raddist * y + tandisty;
    
    // subtract distortion from distorted coordinates
    x = origx - (distx-x);
    y = origy - (disty-y);
    
  }

  // invert Bougouet-normalization
  origx = x*f + principalX_;
  origy = y*aspectratio_*f + principalY_;
  //cout << " to (" << origx << "/" << origy << ")" << endl;  
}

void MatlabTBUndistortion::Distort(BIAS::HomgPoint2D& point2d) const {
  if (!IsInitialized_){
    BIASERR("Not inited!");
    return;
  }
  Distort(point2d[0],point2d[1]);
}

void MatlabTBUndistortion::Distort(double& origx,double& origy) const {
  if (!IsInitialized_){
    BIASERR("Not inited!");
    return;
  }
  // copy of distorted point
  double distx,disty;
  double x,y;
  double twoxy;
  
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  double kc1, kc2, kc3, kc4;
  double f=1;

  if (zoomCamera_) {
    kc1=kc1Zoom_;
    kc2=kc2Zoom_;
    kc3=kc3Zoom_;
    kc4=kc4Zoom_;
    f=focallengthZoom_;
  } else {
    kc1=kc1Def_;
    kc2=kc2Def_;
    kc3=kc3Def_;
    kc4=kc4Def_;
    f=focallengthDef_;
  }

  //cout << "Distorting Point (" << origx << "/" << origy << ")";  

  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = (origx-principalX_)/f;
  y = (origy-principalY_)/(aspectratio_*f);
      
  twoxy = 2 * x*y;
  rsquared = x*x + y*y;
    
  // radial distortion:
  raddist = 1 + kc1*rsquared + kc2*rsquared*rsquared;
  // tangential distortion:
  tandistx = kc3*twoxy + kc4 * (rsquared + 2*x*x);
  tandisty = kc4*twoxy + kc3 * (rsquared + 2*y*y);
  
  // apply distortion:
  distx = raddist * x + tandistx;
  disty = raddist * y + tandisty;
   
  // invert Bougouet-normalization
  origx = distx*f + principalX_;
  origy = disty*aspectratio_*f + principalY_;
  //cout << " to (" << origx << "/" << origy << ")" << endl;  
}


void MatlabTBUndistortion::InterpolateDistortionFromIndex_(unsigned int i,
                                                          double relnorm){
  //todo: brown: adequate methode finden (lin. interp. ist
  //schlechter)
  
  kc1Zoom_=relnorm*(knownparams_vect_[i].kc1-knownparams_vect_[i-1].kc1)+
    knownparams_vect_[i-1].kc1;

  kc2Zoom_=relnorm*(knownparams_vect_[i].kc2-knownparams_vect_[i-1].kc2)+
    knownparams_vect_[i-1].kc2;
  
  kc3Zoom_=relnorm*(knownparams_vect_[i].kc3-knownparams_vect_[i-1].kc3)+
    knownparams_vect_[i-1].kc3;
  
  kc4Zoom_=relnorm*(knownparams_vect_[i].kc4-knownparams_vect_[i-1].kc4)+
    knownparams_vect_[i-1].kc4;
  
}

double MatlabTBUndistortion::InterpolateFocalLength_(){
  
  double relnorm;
  
  bool found;
  unsigned int i;
  found=false;

  if (knownparams_vect_[0].zoom<zoom_){ // zoom is higher than the 
    // known values:
    zoom_=knownparams_vect_[0].zoom;  // set to highest available
  }
  
  for (i=0;i<knownparams_vect_.size();i++){
    if (knownparams_vect_[i].zoom<=zoom_){
	    found=true;
	    break;
    }
  }
  
  if (!found){ // zoom is lower than the known values
    focallengthZoom_=knownparams_vect_[i-1].focallength;
    return knownparams_vect_[i-1].focallength;
  }
  
  
  if (knownparams_vect_[i].zoom==zoom_){
	// interpolate distortion parameters
    InterpolateDistortionFromIndex_(i+1,0);
    return knownparams_vect_[i].focallength;
  } else {
    relnorm=(zoom_ - knownparams_vect_[i].zoom) /
      (knownparams_vect_[i-1].zoom - knownparams_vect_[i].zoom);
    relnorm=1.0-relnorm;
    // interpolate distortion parameters
    InterpolateDistortionFromIndex_(i,relnorm);
  }
  return CameraParam::InterpolateFocalLength_();
}



double MatlabTBUndistortion::InterpolateAbsZoom_(){
  
  double relnorm;
  
  bool found;
  unsigned int i;
  found=false;
  
  if (knownparams_vect_[0].focallength<focallengthZoom_){
    focallengthZoom_=knownparams_vect_[0].focallength;
  }
  
  for (i=0;i<knownparams_vect_.size();i++){
    if (knownparams_vect_[i].focallength<=focallengthZoom_){
	    found=true;
	    break;
    }
  }
  if (!found){
    return -1;
  }
  
  if (knownparams_vect_[i].focallength==focallengthZoom_){
    // interpolate distortion parameters
    InterpolateDistortionFromIndex_(i+1,0);
  } else { 
    relnorm=(focallengthZoom_ - knownparams_vect_[i].focallength) /
      (knownparams_vect_[i-1].focallength - knownparams_vect_[i].focallength);
    relnorm=1.0-relnorm;
    // interpolate distortion parameters
    InterpolateDistortionFromIndex_(i,relnorm);
  }
  
  return CameraParam::InterpolateAbsZoom_();
}



int MatlabTBUndistortion::
SetupProjectionLUT(const unsigned int& srcwidth,const 
                   unsigned int& srcheight,
                   const unsigned int &ChanCnt,
                   const unsigned int& SizeX, const unsigned int& SizeY)
{
  if (!IsInitialized_){
#ifdef BIAS_DEBUG
    BIASERR("Not inited!");
#endif
    return -1;
  }
  double xDst, yDst;

  CutOutWidth_ = SizeX;
  CutOutHeight_ = SizeY;
  
  if (LuT_ !=NULL)  delete[] LuT_;
  LuT_ = new LuTEntry[SizeX * SizeY];
  LuTEntry *currLE = LuT_;

  // go through pixels in dest image (backward mapping)
  for (unsigned int y=0; y<SizeY; y++) {
    for (unsigned int x=0; x<SizeX; x++) {
      xDst = x;
      yDst = y;
      Distort(xDst,yDst);

      // xSrc,ySrc are the coordinates in the original image typically 
      // between four pixels, so find the neighbours for bilinear 
      // interpolation
      unsigned int x_floor = (unsigned int)floor(xDst);
      unsigned int y_floor = (unsigned int)floor(yDst);
      unsigned int x_ceil = x_floor+1;
      unsigned int y_ceil = y_floor+1;
      if (xDst<0) {
        x_floor = 0;
        x_ceil = 0;
      }
      if (xDst>SizeX-1) {
        x_floor = SizeX-1;
        x_ceil = SizeX-1;
      }
      if (yDst<0) {
        y_floor = 0;
        y_ceil = 0;
      } 
      if (yDst>SizeY-1) {
        y_floor = SizeY-1;
        y_ceil = SizeY-1;
      }
      currLE->weight[0] = (unsigned int)rint((xDst - x_floor)*256);
      currLE->weight[1] = (unsigned int)rint((yDst - y_floor)*256);
      // calculate the offsetes in the image data array for all four corners
      // but only for the first channel of ChanCnt
      currLE->offset[0] = (y_floor * srcwidth + x_floor) * ChanCnt;
      currLE->offset[1] = (y_ceil * srcwidth + x_floor) * ChanCnt;
      currLE->offset[2] = (y_ceil * srcwidth + x_ceil) * ChanCnt;
      currLE->offset[3] = (y_floor * srcwidth + x_ceil) * ChanCnt;
      currLE++;
    }
  }
  return 0;
}


int MatlabTBUndistortion::
UndistortFast(const Image<unsigned char>& src,Image<unsigned char>& dst)
{
  unsigned char *srcp, *dstp;
  srcp = src.GetImageData();

#ifdef BIAS_DEBUG
  if (LuT_ ==NULL){
    BIASERR("LUT not initialized!");
    return -1;
  }
#endif
 
  if (dst.IsEmpty()) {
    dst.Init((int)(src.GetWidth()), (int)(src.GetHeight()), 
             src.GetChannelCount());
  }

  dstp = dst.GetImageData();

  if (src.GetChannelCount() ==1) {
    for ( unsigned int i =0; i<dst.GetPixelCount(); i++ ) {
      unsigned char ul = *(srcp + LuT_[i].offset[0]); //ida[y_floor][x_floor];
      unsigned char ll = *(srcp + LuT_[i].offset[1]);// ida[y_ceil][x_floor];
      unsigned char lr = *(srcp + LuT_[i].offset[2]);// ida[y_ceil][x_ceil];   
      unsigned char ur = *(srcp + LuT_[i].offset[3]); // ida[y_floor][x_ceil];
      unsigned int dx = LuT_[i].weight[0];
      unsigned int dy = LuT_[i].weight[1];
      *dstp =  (unsigned char)(((ul<<16) + (dy<<8)*(ll - ul) + 
                                (dx<<8)*(ur - ul) + 
                                dy*dx*(ul - ll - ur + lr))>>16);
      dstp++;
    }
  }
  else {
    unsigned int cc = src.GetChannelCount();
    for ( unsigned int i =0; i<dst.GetPixelCount(); i++ ) {
      for (unsigned int c=0; c<cc; c++) {
        unsigned char ul = *(srcp + LuT_[i].offset[0] + c); 
        unsigned char ll = *(srcp + LuT_[i].offset[1] + c);
        unsigned char lr = *(srcp + LuT_[i].offset[2] + c);
        unsigned char ur = *(srcp + LuT_[i].offset[3] + c); 
        unsigned int dx = LuT_[i].weight[0];
        unsigned int dy = LuT_[i].weight[1];
        *dstp =  (unsigned char)(((ul<<16) + (dy<<8)*(ll - ul) + 
                                  (dx<<8)*(ur - ul) + 
                                  dy*dx*(ul - ll - ur + lr))>>16);
        dstp++;
      }
    }
  }
  return 0;
}
