/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __MatlabTBUndistortion_hh__
#define __MatlabTBUndistortion_hh__

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Image/Image.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Geometry/Projection.hh>
#include <Image/CameraParam.hh>

#include <vector>
#include <string>

#define UNDIST_EPS 1e-9

namespace BIAS{

  class BIASImage_EXPORT MatlabTBUndistortion : public BIAS::CameraParam{

    //    static std::string CameraModelName[];
    // static std::string LensModelName[];

  public:
    MatlabTBUndistortion();
    MatlabTBUndistortion(const CameraParam &cp);

    ~MatlabTBUndistortion();

    // @brief default initialization
    int Init(const CameraParam &cp);
    int Init(BIAS::Projection &projection);

    /** @brief Set the current camera zoom, the focal length will be 
        calculated.
        @author buck */
    void SetZoom(const double abszoom);

    /** @brief get the current camera focal length
        @author buck */
    void SetFocalLength(const double f);

    /** @brief Set the zoom and focal length of a zoom camera to the 
        default settings.
        @author streckel 06/2004*/
    void SetUnzoomed();

   /** @brief Using the Matlab camera calibration toolbox parameters 
        for an undistortion of distorted coordinates - correct implementation
        (Error in the Christian Buck implementation: Iteration does 
        not converge?!)
        @author streckel 06/2004 */
    void Undistort(BIAS::HomgPoint2D& point2d) const;
    void Undistort(double& origx,double& origy) const;

    /** @brief Using the Matlab camera calibration toolbox parameters 
        for a distortion of undistorted coordinates. Inverse to the 
        undistort function.
        @author streckel 06/2004 */
    void Distort(BIAS::HomgPoint2D& point2d) const;
    void Distort(double& origx,double& origy) const;

    int SetupProjectionLUT(const unsigned int& srcwidth,const 
                           unsigned int& srcheight,
                           const unsigned int &ChanCnt,
                           const unsigned int& SizeX, 
                           const unsigned int& SizeY);

   /** @brief Undistortion of an image via a Lookup table
       corrects only lens distortion and not aspect ratio
       @author streckel 06/2004 */
    int UndistortFast(const Image<unsigned char>& src,
		      Image<unsigned char>& dst);

  protected:
	   
    // array of camera calibration constants for zoom cameras
    std::vector<CPDiscreteParam> knownparams_vect_;
    
    // calibration settings for the current zoom setting of a zoom camera
    // zoom and focallengthZoom are dependend:
    // if zoom is set, focallengthZoom will be calculated and vice versa
    double kc1Zoom_,kc2Zoom_,kc3Zoom_,kc4Zoom_;
        
    // illumination vignette correction
    // x is distance in pixel from principal point
    // y is illumination correction factor
    std::vector<double> IlluCorrX_;
    std::vector<double> IlluCorrY_;

    void InterpolateDistortionFromIndex_(unsigned int i,double relnorm);
    double InterpolateFocalLength_(); // w/ given abszoom
    double InterpolateAbsZoom_(); // w/ given focal length

    // structure for Look-up-table for bilinear interpolation 
    // during backward warping
    struct LuTEntry {
      unsigned int offset[4];
      unsigned int  weight[2];
    };
    LuTEntry *LuT_;
    unsigned int CutOutWidth_, CutOutHeight_;
  };

}

#include <Base/Common/BIASpragmaEnd.hh>

#endif
