/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Image/PMDImageProc.hh>
#include <Geometry/RMatrix.hh>
#include <Image/HomographyMapping.hh>
#include <Base/Image/ImageIO.hh>
#include <bias_config.h>
#include <stdio.h>

using namespace BIAS;
using namespace std;

#define UINT16_MAX (65535U)

// depth range in m
#define DEPTH_MAX 50
#define DEPTH_MIN 0
// modulation coefficient range
#define MODCOEFF_MAX 100
#define MODCOEFF_MIN -100


int PMDImageProc::Write(const std::string& FileName, 
                        Image<unsigned char> &Img2D, 
                        Image<float> &ImgDepth, 
                        Image<float> &ImgModCoeff, 
                        PMDImageMetaData &MetaData2D,
                        PMDImageMetaData &MetaDataDepth) {
  
  uint8* data2D=NULL;
  uint16* dataDepth=NULL;
  uint16* dataModCoeff=NULL;

  MetaData2D.width = Img2D.GetWidth();
  MetaData2D.height = Img2D.GetHeight();
  MetaData2D.channels = Img2D.GetChannelCount();

  if (!Img2D.IsEmpty()) {
    if (Img2D.GetStorageType()!=ImageBase::ST_unsignedchar) {
      BIASERR("The RGB or grey image has the wrong storage type, has to be" << 
              "ST_unsignedchar");
      return -1;    
    }
    data2D = (uint8*) Img2D.GetImageData();
  }

  if (!ImgDepth.IsEmpty()) {

    MetaDataDepth.width = ImgDepth.GetWidth();
    MetaDataDepth.height = ImgDepth.GetHeight();
    MetaDataDepth.channels = ImgDepth.GetChannelCount();
    if (ImgDepth.GetStorageType()!=ImageBase::ST_float) {
      BIASERR("The depth image has the wrong storage type, has to be" << 
              "ST_float");
      return -1;    
    }

    unsigned int i;
    dataDepth = new uint16[MetaDataDepth.width*MetaDataDepth.height];
    uint16* pDataDepth = dataDepth;
    float* pDataDepthf = ImgDepth.GetImageData();
    for (i=0; i<MetaDataDepth.width*MetaDataDepth.height; 
         i++, pDataDepthf++, pDataDepth++) {
      *pDataDepth = DepthConvert_(*pDataDepthf);
    }

    if (!ImgModCoeff.IsEmpty()) {
      if (ImgModCoeff.GetStorageType()!=ImageBase::ST_float) {
        BIASERR("The modulation coefficients image has the wrong storage" <<
                "type, has to be ST_float");
        return -1;    
      }
      if (MetaDataDepth.width != ImgModCoeff.GetWidth() ||
          MetaDataDepth.height != ImgModCoeff.GetHeight() ||
          MetaDataDepth.channels != ImgModCoeff.GetChannelCount()) {
        BIASERR("The modulation coefficients image must have same size " <<
                "and channels as the depth image");
        return -1;    
      }
  
      dataModCoeff = new uint16[MetaDataDepth.width*MetaDataDepth.height];
      uint16* pDataModCoeff = dataModCoeff;
      float* pDataModCoefff = ImgModCoeff.GetImageData();
      for (i=0; i<MetaDataDepth.width*MetaDataDepth.height; 
           i++, pDataModCoefff++, pDataModCoeff++) {
        *pDataModCoeff = ModCoeffConvert_(*pDataModCoefff);
      }
    }
  }
  PMDImageIO::Write(FileName, data2D, dataDepth, dataModCoeff,
                    MetaData2D, MetaDataDepth);
 
  if (dataDepth!=NULL) delete [] dataDepth;
  if (dataModCoeff!=NULL) delete [] dataModCoeff;

  return 0;
  
}


int PMDImageProc::Read(const std::string& FileName, 
                       Image<unsigned char> &Img2D, 
                       Image<float> &ImgDepth, 
                       Image<float> &ImgModCoeff, 
                       PMDImageMetaData &MetaData2D, 
                       PMDImageMetaData &MetaDataDepth) {

  uint8* data2d=NULL;
  uint16* dataDepth=NULL;
  uint16* dataModCoeff=NULL;
  int ret=0, ret2=0, ret3=0;
  
  PMDImageIO::Read(FileName, data2d, dataDepth, dataModCoeff,
                   MetaData2D, MetaDataDepth);
  if (data2d!=NULL) {
    if (!Img2D.IsEmpty()) Img2D.Release();
    ret = Img2D.InitWithForeignData(MetaData2D.width, MetaData2D.height,
                                    MetaData2D.channels, 
                                    (unsigned char*) data2d);
    if (ret!=0) BIASERR("Error reading 2D image" << ret);
    
    if (MetaData2D.channels==3) {
      Img2D.SetColorModel(ImageBase::CM_RGB);
    } else {
      Img2D.SetColorModel(ImageBase::CM_Grey);
    }
    
  }

  if (dataDepth!=NULL) {
    float *dataDepthf = new float[MetaDataDepth.width*MetaDataDepth.height];
    float *pDataDepthf = dataDepthf;
    uint16* pDataDepth = dataDepth;
    for (unsigned int i=0; i<MetaDataDepth.width*MetaDataDepth.height; 
         i++, pDataDepthf++, pDataDepth++) {
      *pDataDepthf = DepthConvert_(*pDataDepth);
    }
    if (!ImgDepth.IsEmpty()) ImgDepth.Release();
    ret2 = ImgDepth.InitWithForeignData(MetaDataDepth.width, 
                                        MetaDataDepth.height,
                                        MetaDataDepth.channels, 
                                        dataDepthf);
    delete [] dataDepth;
    if (ret2!=0) BIASERR("Error reading Depth image" << ret2);
  }
  
  if (dataModCoeff!=NULL) {
    float *dataModCoefff = new float[MetaDataDepth.width*MetaDataDepth.height];
    float *pDataModCoefff = dataModCoefff;
    uint16* pDataModCoeff = dataModCoeff;
    for (unsigned int i=0; i<MetaDataDepth.width*MetaDataDepth.height; 
         i++, pDataModCoefff++, pDataModCoeff++) {
      *pDataModCoefff = ModCoeffConvert_(*pDataModCoeff);
    }
    if (!ImgModCoeff.IsEmpty()) ImgModCoeff.Release();
    ret3 = ImgModCoeff.InitWithForeignData(MetaDataDepth.width, 
                                           MetaDataDepth.height,
                                           MetaDataDepth.channels, 
                                           dataModCoefff);
    delete [] dataModCoeff;
    if (ret3!=0) BIASERR("Error reading Modulation Coefficients image" 
                         << ret3);
  }
  return ret+ret2+ret3;
}


int PMDImageProc::FitDepthTo2DImage(Image<float> &ImgDepth, 
                                    const PMDImageMetaData &MetaData2D, 
                                    PMDImageMetaData &MetaDataDepth) {

  HomographyMapping<float, float> hm;
  hm.SetHomography(HFromMeta_(MetaData2D, MetaDataDepth));
  BIAS::ImageIO::Write("depthBefore.mip",ImgDepth);
  Image<float> mapto(MetaData2D.width, MetaData2D.height, 1);
  int ret = hm.Map(ImgDepth, mapto);
  BIAS::ImageIO::Write("depthAfter.mip",mapto);
  ImgDepth = mapto;
  return ret;
}


HMatrix PMDImageProc::HFromMeta_(const PMDImageMetaData &MetaData2D, 
                                 const PMDImageMetaData &MetaDataDepth) {

  if (!Equal(MetaData2D.CameraCenter[0], MetaDataDepth.CameraCenter[0]) ||
      !Equal(MetaData2D.CameraCenter[1], MetaDataDepth.CameraCenter[1]) ||
      !Equal(MetaData2D.CameraCenter[2], MetaDataDepth.CameraCenter[2])) {
    BIASERR("Can not Fit Depth Image to 2D Iamge, centers are not equal!!");
  }

  KMatrix K2D, KDepth;
  K2D = KFromMeta(MetaData2D);
  KDepth =  KFromMeta(MetaDataDepth);
  
  Vector3<double> rot2D(MetaData2D.CameraRotation[0],
                        MetaData2D.CameraRotation[1],
                        MetaData2D.CameraRotation[2]);
  Vector3<double> rotDepth(MetaDataDepth.CameraRotation[0],
                           MetaDataDepth.CameraRotation[1],
                           MetaDataDepth.CameraRotation[2]);
  RMatrix Rdiff, RDepth, R2D;
  RDepth.SetFromAxisAngle(rotDepth);
  R2D.SetFromAxisAngle(rot2D);
  Rdiff = R2D*RDepth.Invert();
  HMatrix H = KDepth*Rdiff*K2D.Invert();
  return H;

}


KMatrix PMDImageProc::KFromMeta(const PMDImageMetaData &MetaData) {
  KMatrix K;
  K.SetIdentity();
  K[0][0] = MetaData.FocallengthX;
  K[1][1] = MetaData.AspectRatio * MetaData.FocallengthX;
  K[0][2] = MetaData.PrincipalX;
  K[1][2] = MetaData.PrincipalY;
  return K;
} 


Pose PMDImageProc::PoseFromMeta(const PMDImageMetaData &MetaData) {
  Pose pose;
  RMatrix R;
  R.SetFromAxisAngle(Vector3<double>(MetaData.CameraRotation));
  Quaternion<double> Q;
  R.GetQuaternion(Q);
  pose.SetQ(Q);
  pose.SetC(Vector3<double>(MetaData.CameraCenter));
  return pose;
}


double PMDImageProc::CalcSigmaDepth(Image<float> &ImgDepth, 
                                    unsigned int x, unsigned int y) 
{ 
  if (x<1 || x>ImgDepth.GetWidth()-2 ||
      y<1 || y>ImgDepth.GetHeight()-2) {
    return -1;
  }

  // Surrounding min and max depth in m
  double minDepth = 1000;
  double maxDepth = 0;

  for (int dx=-1; dx<=1; dx++) {
    for (int dy=-1; dy<=1; dy++) {
      if (ImgDepth.PixelValue(x+dx, y+dy)<minDepth) 
        minDepth = ImgDepth.PixelValue(x+dx, y+dy);
      if (ImgDepth.PixelValue(x+dx, y+dy)>maxDepth) 
        maxDepth = ImgDepth.PixelValue(x+dx, y+dy);
    }
  }

  // Standard deviation for a certain depth in meter
  double minDev= 0.002734*minDepth*minDepth + 
    0.00286723*minDepth - 0.0004229692;
  double maxDev= 0.002734*maxDepth*maxDepth + 
    0.00286723*maxDepth - 0.0004229692;

  //depth value is in this range
  minDepth -= minDev;
  maxDepth += maxDev;
  return (maxDepth-minDepth)/2;
}


uint16 PMDImageProc::DepthConvert_(float depth) {
  return Convert_(depth, DEPTH_MIN, DEPTH_MAX);
}

float PMDImageProc::DepthConvert_(uint16 depth) {
  return Convert_(depth, DEPTH_MIN, DEPTH_MAX);
}

uint16 PMDImageProc::ModCoeffConvert_(float modcoeff) {
  return Convert_(modcoeff, MODCOEFF_MIN, MODCOEFF_MAX);
}

float PMDImageProc::ModCoeffConvert_(uint16 modcoeff) {
  return Convert_(modcoeff, MODCOEFF_MIN, MODCOEFF_MAX);
}

uint16 PMDImageProc::Convert_(float val, float lower, float upper) {
  float step=(upper-lower)/UINT16_MAX;
  if (val < lower) return 0;
  if (val > upper) return UINT16_MAX;
  return (uint16)((val-lower)/step);
}

float PMDImageProc::Convert_(uint16 val, float lower, float upper) {
  float step=(upper-lower)/UINT16_MAX;
  return val*step+lower;
}


