/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BIAS_PMDIMAGEPROC_H__
#define __BIAS_PMDIMAGEPROC_H__

#include <bias_config.h>
#include <Base/Image/Image.hh>

#define __USE_BIAS__
#include <Base/Image/PMDImageIO.hh>  // only after #define __USE_BIAS__ 

#include <Base/Geometry/KMatrix.hh>
#include <Geometry/HMatrix.hh>
#include <Geometry/Pose.hh>
#include <Base/Common/CompareFloatingPoint.hh>

#ifndef BIAS_HAVE_TIFF
#  error Please recompile BIAS with USE_TIFF
#endif // BIAS_HAVE_TIFF

namespace BIAS {


  class BIASImage_EXPORT PMDImageProc
  {
  public:
    /** @brief Write a PMD-data 2d/3d-tiff image to disk
        @param FileName Output filename, should have the extension .tif
        @param img the 2d-image data to save
        @param depthimg the depth data to save
        @param MetaData2D the Metadata of the 2D-image to save
        @param MetaDataDepth the Metadata of the depth image to use for padding
        @author Birger Streckel **/
    static int Write(const std::string& FileName, 
                     Image<unsigned char> &Img2D, 
                     Image<float> &ImgDepth, 
                     Image<float> &ImgModCoeff, 
                     PMDImageMetaData &MetaData2D,
                     PMDImageMetaData &MetaDataDepth);
    
    /** @brief Read a PMD-data 2d/3d-tiff image from disk
        @param FileName complete name of file to be loaded
        @param img to receive the 2d-image data from the file
        @param depthimg to receive the depth-image data from the file
        @param MetaData to receive the Metadata from the file
        @author Birger Streckel **/
    static int Read(const std::string& FileName, 
                    Image<unsigned char> &Img2D, 
                    Image<float> &ImgDepth, 
                    Image<float> &ImgModCoeff, 
                    PMDImageMetaData &MetaData2D, 
                    PMDImageMetaData &MetaDataDepth);

    /** @brief convert a depth image such that the depth and the grey value 
        at the same pixel position belong to the same ray. The transformaton 
        information is taken from the metadata. MetaDataDepth is changed
        for the converted depth image and is identical to MetaData2D afterwards
        @param ImgDepth is the image to convert. This image has to contain 
        depth informaiton and is changed in size and contents.
        @param MetaData2D this is the metadata for the 2D-image.
        @param MetaDataDepth the metadata for the depth image. Is adapted 
        to reflect the image transformation.
        @author Birger Streckel **/
    static int FitDepthTo2DImage(Image<float> &ImgDepth, 
                                 const PMDImageMetaData &MetaData2D, 
                                 PMDImageMetaData &MetaDataDepth);

    static KMatrix KFromMeta(const PMDImageMetaData &MetaData);
    static Pose PoseFromMeta(const PMDImageMetaData &MetaData);
   
    static double CalcSigmaDepth(Image<float> &ImgDepth, 
                                 unsigned int x, unsigned int y);

  protected:

    /* @brief Create a H-Matrix that converts the depth image to the 2D view
     */
    static HMatrix HFromMeta_(const PMDImageMetaData &MetaData2D, 
                              const PMDImageMetaData &MetaDataDepth);

    
    static uint16 DepthConvert_(float depth);
    static float DepthConvert_(uint16 depth);
    static uint16 ModCoeffConvert_(float modcoeff);
    static float ModCoeffConvert_(uint16 modcoeff);
    static uint16 Convert_(float val, float lower, float upper);
    static float Convert_(uint16 val, float lower, float upper);

  };

}
#endif
