/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __SphericalUndistortion_hh__
#define __SphericalUndistortion_hh__

#include <bias_config.h>

#include <Base/Common/W32Compat.hh>

#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>
#include <Base/Math/Vector.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Debug/Debug.hh>

#include <MathAlgo/Interpolator.hh>
#include <Image/CameraParam.hh>

namespace BIAS{
  /** @class SphericalUndistortion
      @ingroup g_filter
      @brief transforms sphere images coming from a fisheye-lens according to 
      different projection models. 

      @author streckel
    
      Input is an image from a fisheye lens looking circular 
      with a black border. 5 types of projection are supported 
      (see: M.M.Fleck "Perspective Projection: The wrong imaging model")

      The algorithm expects a 3rd order calibration polynomial to get a 
      linear relation between viewing angle and distance to circle center.

      The supported projections are:
      PERSPECTIVE      : Perspective projection      - p=k*tan(theta)
      STEREOGRAPHIC    : Stereographic projection    - p=k*tan(theta/2)
      EQUIDISTANT      : Equidistant projection      - p=k*theta
      SINE_LAW         : sine-law projection         - p=k*sin(theta)
      EQUI_SOLID_ANGLE : equi-solid-angle projection - p=k*sin(theta/2)
      p     : projected distance from circle center
      theta : angle between projected point and optical axis
  */

#define DEBUG_SPHEREPROJ_AVG 1
#define DEBUG_SPHEREPROJ_CIRCLE 2

  class BIASImage_EXPORT SphericalUndistortion : 
    public BIAS::CameraParam, public BIAS::Debug{

  public:

    /** @brief Projection-type constants
        @author Birger Streckel */
    enum{PERSPECTIVE, STEREOGRAPHIC, EQUIDISTANT, SINE_LAW, EQUI_SOLID_ANGLE};
  
    /** @brief default constructor
        @author Birger Streckel */
    SphericalUndistortion();

    /** @brief copy constructor
        @author Birger Streckel */
    SphericalUndistortion(const SphericalUndistortion& sp);

    /** @brief common constructor - initialization with CameraParam-object!
        @author Birger Streckel */
    SphericalUndistortion(const CameraParam& cp);

    /** @brief filter initialization with calibration curve measurements
        @author Birger Streckel */
    SphericalUndistortion(const double& centerX, const double& centerY, 
                          const double& radius,
                          const std::vector<double> &AngleCorrX, 
                          const std::vector<double> &AngleCorrY,
                          const std::vector<double> &IlluCorrX, 
                          const std::vector<double> &IlluCorrY);

    ~SphericalUndistortion();
  
    /** @brief Initialization by setting all necessary parameters 
        (center, radius, Vectors of Angle and Vignette correction 
        measurement points) from the outside */
    int Init(const double& centerX, const double& centerY, 
             const double& radius,
             const std::vector<double> &AngleCorrX, 
             const std::vector<double> &AngleCorrY,
             const std::vector<double> &IlluCorrX, 
             const std::vector<double> &IlluCorrY);

    /** @brief Initialization with CameraParam-object!
        @author Birger Streckel */
   int Init(const CameraParam& cp);
    
#ifdef BIAS_HAVE_XML2
    // @brief initialization with camera data file
    int Init(const std::string &filename);
    // @brief initialization with camera node from pre-parsed data file
    int Init(XMLIO &myXML, xmlNodePtr &cameraNode);
#endif //BIAS_HAVE_XML2
    
   /** @brief Performs a cylingric projection of a spherical image given 
        a projection-type and a zoom factor.

        @param src source image 
        @param dst (output) uninitialized destination image
    */
    template <class StorageType>
    int ProjectImageCylindric(const Image<StorageType>& src, 
                              Image<StorageType>& dst,
                              double viewCenterX,
                              double viewCenterY,
                              double zoomFactor);
    
      /** @brief Performs a projection of a spherical image coordinate
        to a projected and calibrated coordinate
        given a projection-type and a K-matrix.
        @author Birger Streckel */
    int ProjectCoord(const int projectionType, const KMatrix& kMatrix,
                     const HomgPoint2D& Source, HomgPoint2D &Dest) const;
  
    /** @brief Performs a projection of a projected and calibrated coordinate
        to a spherical image coordinate
        given a projection-type and a  K-matrix.
        @author Birger Streckel */
    int ProjectCoordBack(const int projectionType, const KMatrix& kMatrix,
                         const HomgPoint2D& Source, HomgPoint2D &Dest) const; 

    /** @brief Transforms a coordinate pair of image coordinates x, y in the 
        uncalibrated spherical image, to a calibrated pair of polar coordinates
        phi, theta.
        @author Birger Streckel */
    int TransfCoordImage2Sphere (const HomgPoint2D& Source,
                                 double &theta, double &phi) const;
    int TransfCoordImage2Sphere (const double& SourceX, const double& SourceY, 
                                 double &theta, double &phi) const;

    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated 
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2Image (const double& theta, const double& phi,
                                 HomgPoint2D &Source) const;
    int TransfCoordSphere2Image (const double& theta, const double& phi,
                                 double &SourceX, double &SourceY) const;
    
    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated 
        spherical image.
        @author Birger Streckel */
    int TransfCoordRotateSphere2Image (double theta, double phi,
                                       const double& viewCenterX,
                                       const double& viewCenterY,
                                       HomgPoint2D &Source) const;
    int TransfCoordRotateSphere2Image (double theta, double phi,
                                       const double& viewCenterX, 
                                       const double& viewCenterY,
                                       double &SourceX, double &SourceY) const;
      
    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in an equidistant
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2CalibImage (const double& theta, const double& phi, 
                                      HomgPoint2D &Source) const;
    int TransfCoordSphere2CalibImage (const double& theta, const double& phi,
                                      double &SourceX, double &SourceY) const;

 /** @brief Get the Cameras KMatrix corresponding 	 
         @author streckel */ 	 
     KMatrix GetKMatrix(); 	 
 
    //int TransfRectImage2Sphere (double DestX, double DestY, int imgWidth,
		//		 double zoomFactor,double &theta, double &phi);
    /** @brief Returns the principal point of the camera image.
        @author Birger Streckel */
    int GetSphereCenter(double& centerX, double& centerY) const;
    int GetSphereCenter(double& centerX, double& centerY, double& radius)const;

    /** @brief Returns the radius in pixels of the camera image circle.
        @author Birger Streckel */
    double GetSphereRadius() const;
     
    /** @brief Returns the maximum fov-angle of the fisheye optics.
        @author Birger Streckel */
    double GetMaxCamAngle() const;
     
    /** @brief Draw circles of constant angle steps into the uncalibrated 
        spherical image using the angle calibration.
        @author Birger Streckel */
    template <class StorageType>
    int DrawCircles(Image<StorageType>& src, const double& DegreeStep);
    
    /** @brief Corrects the illumination vignette of an uncalibrated
        (source-)image.
        @author Birger Streckel */
    template <class StorageType>
    int CorrectVignette(Image<StorageType>& src);

    /** @brief theta and phi are periodic in 2*M_PI, if they are out of a 
        suitable range, try to find best equivalent angle by adding 2*l*M_PI
        @return 0 on success, <0 indicates that at least one angle is invalid
        @author koeser 10/2004 */
    int EnforceNormalRange(double& theta, double& phi) const;

    /** Fast spehrical undistortion via LookUpTable. This function creates 
        a table, use ProjectCutOut to apply it.
        @param srcwidth, srcheight size of the sperical source image.
        @param ChanCnt channel count.
        @param SizeX, SizeY Is the size of the resulting perspective image.
        Since there should be no zoom this determines also the 
        perspective cutouts field of view.
        @param CenterX, CenterY point in sphere image to which the
        tangential mapping plane is attached. This point is mapped to 
        the center of the resulting image.
        @param projectionType normally PERSPECTIVE.
        @param KMatDst Returns the KMatrix of the perspective image.
        @author evers 2004 */
        
    int SetupProjectionLUT(const unsigned int& srcwidth,
			   const unsigned int& srcheight,
			   const unsigned int& ChanCnt,
			   const unsigned int& SizeX, const unsigned int& SizeY,
			   const double& CenterX, const double& CenterY, 
			   const int projectionType, KMatrix& KMatDst);
    template <class StorageType>
    int ProjectCutOut(const Image<StorageType>& src,
 		      Image<StorageType>& dst);

    /** Applies the LookUpTable from SetupProjectionLUT to the src-image.
        If dst is not set, this is done internally.
        @author evers 2004 */

    int ProjectCutOut(const Image<unsigned char>& src,
		      Image<unsigned char>& dst);

    /** Creates a LUT for conversion of angles PhiX, PhiY, (0,0) is camera
        center, to sphere image coordinates.
        @param srcwidth, srcheight size of the sperical source image.
        @param ChanCnt channel count.
        @param stepwidth of PhiX, PhiY in rad
        @author streckel 01/2005
     */
    int SetupLutPxPySphere1(const unsigned int srcwidth,
                           const unsigned int srcheight,
                           const unsigned int ChanCnt, 
                           const double stepwidth);
    int SetupLutPxPySphere2(const unsigned int srcwidth,
                           const unsigned int srcheight,
                           const unsigned int ChanCnt, 
                           const double stepwidth);

    /** Creates a LUT for conversion of angles PhiX, PhiY, (0,0) is camera
        center, to sphere image coordinates.
        @param ChanCnt channel count.
        @param SizeX, SizeY Is the size of the resulting perspective image.
        Since there should be no zoom this determines also the 
        perspective cutouts field of view.
        @param projectionType normally PERSPECTIVE.
        @param KMatDst Returns the KMatrix of the perspective image.
        @param stepwidth of PhiX, PhiY in rad
        @author streckel 01/2005
     */
    int SetupLutPerspPxPy1(const unsigned int SizeX, const unsigned int SizeY,
                           const int projectionType,
                           KMatrix& KMatDst);
    int SetupLutPerspPxPy2(const unsigned int SizeX, const unsigned int SizeY,
                           const int projectionType,
                           KMatrix& KMatDst);
    
    /** Applies the LookUpTables from SetupLutPxPySphere and SetupLutPerspPxPy
        to the src-image. Using 2 LUTs allows an active rotation of the 
        perspective camera in the fisheye image.
        If dst is not set, this is done internally.
        @author streckel 01/2005
    */
    template <class StorageType>
    int ProjectCutOut1(const Image<StorageType>& src,
                       Image<StorageType>& dst,
                       const double stepwidth,
                       const double& yaw, const double& pitch, 
                       const double& roll);
    template <class StorageType>
    int ProjectCutOut2(const Image<StorageType>& src,
                       Image<StorageType>& dst,
                       const double stepwidth,
                       const double& yaw, const double& pitch, 
                       const double& roll);
    int ProjectCutOut1(const Image<unsigned char>& src,
                       Image<unsigned char>& dst,
                       const double stepwidth,
                       const double& yaw, const double& pitch, 
                       const double& roll);
    int ProjectCutOut2(const Image<unsigned char>& src,
                       Image<unsigned char>& dst,
                       const double stepwidth,
                       const double& yaw, const double& pitch, 
                       const double& roll);
    
  protected:

    // DLL JW
    BIAS_EXPORT_STL_VECTOR( BIASImage_EXPORT, double)

    // Spline interpolators for the measurement interpolation
    Interpolator interpolatorAngle_;
    Interpolator interpolatorAngleBack_;
    Interpolator interpolatorIllu_;
    
    // structure for Look-up-table for bilinear interpolation 
    // during backward warping
    struct LuTEntry {
      unsigned int offset[4];
      unsigned int weight[2];
    };
    LuTEntry *LuT_;
    unsigned int CutOutWidth_, CutOutHeight_;
    
    // LUT for active rotation compensation
    LuTEntry *LuTPxPySphere_;
    struct LuTpxpyEntry {
      float PhiX;
      float PhiY;
    };
    LuTpxpyEntry *LuTPerspPxPy_;
  };

} // namespace 


#endif // __SphericalUndistortion_hh__
