/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __RegionMatcher_hh__
#define __RegionMatcher_hh__

#include <bias_config.h> 

#include <Base/Debug/Debug.hh>
#include <Base/Debug/Error.hh>
#include <MathAlgo/Lapack.hh>
#include <Base/Math/Matrix2x2.hh>
#include <Base/Math/Vector2.hh>

#define D_REGION_MATCHER_SEARCHNCC   0x0001
#define D_REGION_MATCHER_KLT         0x0002
#define D_REGION_MATCHER_LINE        0x0004
#define D_REGION_MATCHER_PARABOLA    0x0008
#define D_REGION_MATCHER_COLORNCC    0x0010
#define D_REGION_MATCHER_COLORCCV    0x0020
#define D_REGION_MATCHER_KLT_EDGE    0x0040
#define D_REGION_MATCHER_SEARCH_EVEN 0x0080
#define D_REGION_MATCHER_SEARCH_ODD  0x0100

// data type for internal calculations in KLT
#define KLT_TYPE double
//#define KLT_TYPE float

namespace BIAS{

  /** some fast SAD functions via defines
      @author woelk */
#define absdiff(a, b) ((a>b)?(a-b):(b-a))
#define SAD3x3(x1, y1, x2, y2, ida1, ida2, re)\
  unsigned x1m=x1-1, x1p=x1+1, y1m=y1-1, y1p=y1+1;\
  unsigned x2m=x2-1, x2p=x2+1, y2m=y2-1, y2p=y2+1;\
  re = (unsigned)absdiff(ida1[y1m][x1m], ida2[y2m][x2m]) + \
       (unsigned)absdiff(ida1[y1m][x1], ida2[y2m][x2]) + \
       (unsigned)absdiff(ida1[y1m][x1p], ida2[y2m][x2p]) + \
       (unsigned)absdiff(ida1[y1][x1m], ida2[y2][x2m]) + \
       (unsigned)absdiff(ida1[y1][x1], ida2[y2][x2]) + \
       (unsigned)absdiff(ida1[y1][x1p], ida2[y2][x2p]) + \
       (unsigned)absdiff(ida1[y1p][x1m], ida2[y2p][x2m]) + \
       (unsigned)absdiff(ida1[y1p][x1], ida2[y2p][x2]) + \
       (unsigned)absdiff(ida1[y1p][x1p], ida2[y2p][x2p]);

#define SAD5x5(x1, y1, x2, y2, ida1, ida2, re)\
  unsigned x1m=x1-1, x1p=x1+1, y1m=y1-1, y1p=y1+1, x1m2=x1-2, x1p2=x2+2, y1m2=y2-2, y1p2=y2+2;\
  unsigned x2m=x2-1, x2p=x2+1, y2m=y2-1, y2p=y2+1, x2m2=x1-2, x2p2=x2+2, y2m2=y2-2, y2p2=y2+2;\
  re = (unsigned)absdiff(ida1[y1m2][x1m2], ida2[y2m2][x2m2]) + \
       (unsigned)absdiff(ida1[y1m2][x1m],  ida2[y2m2][x2m]) + \
       (unsigned)absdiff(ida1[y1m2][x1],   ida2[y2m2][x2]) + \
       (unsigned)absdiff(ida1[y1m2][x1p],  ida2[y2m2][x2p]) + \
       (unsigned)absdiff(ida1[y1m2][x1p2], ida2[y2m2][x2p2]) + \
       (unsigned)absdiff(ida1[y1m][x1m2], ida2[y2m][x2m2]) + \
       (unsigned)absdiff(ida1[y1m][x1m],  ida2[y2m][x2m]) + \
       (unsigned)absdiff(ida1[y1m][x1],   ida2[y2m][x2]) + \
       (unsigned)absdiff(ida1[y1m][x1p],  ida2[y2m][x2p]) + \
       (unsigned)absdiff(ida1[y1m][x1p2], ida2[y2m][x2p2]) + \
       (unsigned)absdiff(ida1[y1][x1m2], ida2[y2][x2m2]) + \
       (unsigned)absdiff(ida1[y1][x1m],  ida2[y2][x2m]) + \
       (unsigned)absdiff(ida1[y1][x1],   ida2[y2][x2]) + \
       (unsigned)absdiff(ida1[y1][x1p],  ida2[y2][x2p]) + \
       (unsigned)absdiff(ida1[y1][x1p2], ida2[y2][x2p2]) + \
       (unsigned)absdiff(ida1[y1p][x1m2], ida2[y2p][x2m2]) + \
       (unsigned)absdiff(ida1[y1p][x1m],  ida2[y2p][x2m]) + \
       (unsigned)absdiff(ida1[y1p][x1],   ida2[y2p][x2]) + \
       (unsigned)absdiff(ida1[y1p][x1p],  ida2[y2p][x2p]) + \
       (unsigned)absdiff(ida1[y1p][x1p2], ida2[y2p][x2p2]) + \
       (unsigned)absdiff(ida1[y1p2][x1m2], ida2[y2p2][x2m2]) + \
       (unsigned)absdiff(ida1[y1p2][x1m],  ida2[y2p2][x2m]) + \
       (unsigned)absdiff(ida1[y1p2][x1],   ida2[y2p2][x2]) + \
       (unsigned)absdiff(ida1[y1p2][x1p],  ida2[y2p2][x2p]) + \
       (unsigned)absdiff(ida1[y1p2][x1p2], ida2[y2p2][x2p2]);

#define NCC5x5(x1, y1, x2, y2, ida1, ida2, re)\
  unsigned x1m=x1-1, x1p=x1+1, y1m=y1-1, y1p=y1+1, x1m2=x1-2, x1p2=x2+2, y1m2=y2-2, y1p2=y2+2;\
  unsigned x2m=x2-1, x2p=x2+1, y2m=y2-1, y2p=y2+1, x2m2=x1-2, x2p2=x2+2, y2m2=y2-2, y2p2=y2+2;\
  int im1, im2; unsigned long S11=0, S12=0, S22=0, N=0, S1=0, S2=0;\
  im1=ida1[y1m2][x1m2]; im2=ida2[y2m2][x2m2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m2][x1m];  im2=ida2[y2m2][x2m]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m2][x1];   im2=ida2[y2m2][x2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m2][x1p];  im2=ida2[y2m2][x2p]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m2][x1p2]; im2=ida2[y2m2][x2p2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m][x1m2];  im2=ida2[y2m][x2m2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m][x1m];   im2=ida2[y2m][x2m]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m][x1];    im2=ida2[y2m][x2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m][x1p];   im2=ida2[y2m][x2p]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1m][x1p2];  im2=ida2[y2m][x2p2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1][x1m2];   im2=ida2[y2][x2m2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1][x1m];    im2=ida2[y2][x2m]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1][x1];     im2=ida2[y2][x2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1][x1p];    im2=ida2[y2][x2p]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1][x1p2];   im2=ida2[y2][x2p2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p][x1m2];  im2=ida2[y2p][x2m2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p][x1m];   im2=ida2[y2p][x2m]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p][x1];    im2=ida2[y2p][x2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p][x1p];   im2=ida2[y2p][x2p]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p][x1p2];  im2=ida2[y2p][x2p2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p2][x1m2]; im2=ida2[y2p2][x2m2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p2][x1m];  im2=ida2[y2p2][x2m]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p2][x1];   im2=ida2[y2p2][x2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p2][x1p];  im2=ida2[y2p2][x2p]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  im1=ida1[y1p2][x1p2]; im2=ida2[y2p2][x2p2]; N++; S1+=im1; S2+=im2; S11+=im1*im1; S12+=im1*im2; S22+=im2*im2; \
  re =((double)N*(double)S12 - (double)S1*(double)S2)/ \
    sqrt(((double)N*(double)S11 - (double)S1*(double)S1) * \
         ((double)N*(double)S22 - (double)S2*(double)S2)); \
  if (re!=re) re=0.0;

  /** @class RegionMatcher
      @ingroup corrG
      @brief basic functions for CornerMatcher
      
      class that holds sum of square diffeences (ssd),  
      normalized cross correlation (ncc) and KLT and parabola interpolation
      for a convenient interface working with HomgPoint2D
      doing boundary checks use CornerMatcher
      @todo check if all klt code has been moved to new filter2 structure
      @author fw 
  */
  class BIASMatcher2D_EXPORT RegionMatcher : public Debug
  {
  public: 
    RegionMatcher();
    ~RegionMatcher();

    /** fast sum of square differences (SSD) sim. measurment
        between p1 and p2 without boundary checking
        specialized version for unsigned char exist
        for halfwindowsizes <= 127 if sizeof(unsigned char)=1 and 
        sizeof(unsigned long int)=4
        @author woelk ? untested */
    template <class StorageType>
    void SSD(unsigned int p1[2], unsigned int p2[2], StorageType **ida1, 
             StorageType **ida2, unsigned int halfwinsize, double& result);

    /** fast sum of absolut differences (SAD) sim. measurement
        between p1 and p2 without boundary checking
        specialized version for unsigned char exist only 
        for halfwindowsizes < 2050
        @author woelk 02 2003 untested */
    template <class StorageType>
    void SAD(unsigned int p1[2], unsigned int p2[2], StorageType **ida1, 
             StorageType **ida2, unsigned int halfwinsize, double& result);

    /** 'normalized' version of SAD, returns 1.0 incase of no differences
        0.0 in case of max difference
        specialized version for unsigned char exist only 
        for halfwindowsizes < 2050
        @author woelk 05 2003 untested */
    template <class StorageType>
    void SADN(unsigned int p1[2], unsigned int p2[2], StorageType **ida1, 
              StorageType **ida2, unsigned int halfwinsize, double& result);
    
    /** fast ncc between p1 and p2 without boundary checking
        Faster specialisation for unsigned char exists
        because of integer arithmetic
        Specialisation for unsigned char does work without overflow with 
        halfwindowsizes <= 127 if sizeof(unsigned char)=1 and 
        sizeof(unsigned long int)=4 
        @author woelk */
    template <class StorageType>
    void NCC(unsigned int p1[2], unsigned int p2[2], StorageType **ida1, 
             StorageType **ida2, unsigned int halfwinsize, double& result);

    /** fast interpolated ncc 
        @author woelk */
    template <class StorageType>
    void NCC(double p1[2], double p2[2], StorageType **ida1, 
             StorageType **ida2, unsigned halfwinsize, double &result);

    /** fast ncc between p1 in ida1 and p2 in ida2
        the correlation is calculated in a window of width=2*hww+1 
        and height = 2*hwh+1
        only those pixels within the window are used, where the according roi
        is not zero 
        @author woelk 09 2003 , tested */
    template <class StorageType>
    void NCC(unsigned p1[2], unsigned p2[2], StorageType **ida1, 
             StorageType **roi1, StorageType **ida2, StorageType **roi2,
             unsigned hww, unsigned hwh, double& result);

   
    /** ColorNCC for interleaved HSV images
        @author Daniel Grest, March 2003 */
    template <class StorageType>
    void ColorNCC(unsigned int p1[2], unsigned int p2[2], 
		    StorageType **ida1, StorageType **ida2, 
		    unsigned int halfwinsize, double &result);
  
    
    /** computes the Cross-CoVariance ,which is equal to NCC
        but does no variance(contrast) normailzation      
        @author Daniel Grest, April 2003 */
    template <class StorageType>
    void ColorCCV(unsigned int p1[2], unsigned int p2[2], 
                  StorageType **ida1, StorageType **ida2, 
                  unsigned int halfwinsize, double &result);

    
    /** ColorNCC for interleaved hsL images
        @author Daniel Grest, June 2003 */
    template <class StorageType>
    void CNCC(unsigned int p1[2], unsigned int p2[2], 
              StorageType **ida1, StorageType **ida2, 
              unsigned int halfwinsize, double &result);

    /** Searches in a window around p2 [p2[i]+-halfsearchwinsize] for a 
        point with correlation > mincorrelation. 
        If point with correlation > mincorrelation is found, the search is 
        aborted and resultpoint is set. 
        If no point with correlation > mincorrelation is found, a negative 
        value is returned. resultpoint is set to the point with biggest 
        correlation within search window, result is set tot this correlation.
        Searching follows an outbound spiral around p2, either until
        searchwinsize is completly searched or until a point with 
        correlation greater than mincorrelation is found
        If whole search area should be used, use mincorrelation>=1.0
        Searchwindow is always odd-sized (p2+/-halfsearchwinsize)
        returns 0 on success, negative value if no point found with 
        correlation > mincorrelation
        !!! warning, no boundary checking is performed, 
        i.e. p1[i]/p2[i] +-(halfnccwinsize+halfsearchwinsize) must lie
        within StorageType-array
        input:  p1, p2, ida1, ida2, halfnccwinsize, halfsearchwinsize, 
        mincorrelation
        output: resultpoint, result 
        */
    template <class StorageType>
    int NCCSearchOdd(unsigned int p1[2], unsigned int p2[2], 
                     StorageType **ida1, StorageType **ida2, 
                     unsigned int halfnccwinsize, 
                     unsigned int halfsearchwinsize, 
                     unsigned int resultpoint[2], 
                     double mincorrelation, double& result);

    /** As above but middle point of searchwindow is p1[0]+0.5, p1[1]+0.5
        Searchwindow is always even-sized floor(p2[i]+0.5+-halfsearchwinsize)
        */
    template <class StorageType>
    int NCCSearchEven(unsigned int p1[2], unsigned int p2[2], 
                      StorageType **ida1, StorageType **ida2, 
                      unsigned int halfnccwinsize,
                      unsigned int halfsearchwinsize, 
                      unsigned int resultpoint[2], 
                      double mincorrelation, double& result);

//     /** find correspondence to p1 from ida1 in ida2 
//         at aproximate starting position p2 using KLT algorithm: 
        
//         Solve G * d = e with

//         G = | sum(gx*gx)  sum(gx*gy) |   , and
//         | sum(gx*gy)  sum(gy*gy) | 
        
//         e = | sum(gx*(I2-I1)) sum(gy*(I2-I1)) |^T
        
//         sums are over the window, gx is x-gradient and gy the y-gradient in 
//         first image, I1 and I2 are the grey values in the first 
//         resp. the second image  
//         iterate numiterations times 
//         !!! no initial boundary checks is performed in this algorithm !!!
//         returns 0 on success
//         returns -1 if point slids out of [0, with][0, height] modulo halfwinsize
//         output: result - image coo of p1 in im2
//                 error - norm(e)
//         input : p1 - location of feature point in image1
//                 p2 - aproximate location of same feature point in new image 2
//                      could be set to p1 featurepoint is not moving a lot
//                 ida1 - pointer to the image data (row major) of first image
//                 ida2 - pointer to the image data (row major) of new image
//                 gradx - pointer to the x-gradient data of image1
//                 grady - pointer to the y-gradient data of image1
//                 halfwinsize - the window to use 
//                 numiterations - number of iterations to do
//                 width - width of StorageType Arrays
//                 height - height of StorageType Array  ida1[height-1][with-1]
//                          is the maximal allowed acces to the arrays
//         @author fw */
//     template <class StorageType>
//     int KLT(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//             StorageType **ida2, StorageType **gradx, StorageType **grady, 
//             unsigned int width, unsigned int height, unsigned int halfwinsize,
//             unsigned int numiterations, KLT_TYPE result[2], KLT_TYPE& error);
    
//     /** same as KLT above but do it until either maxiterations is reached
//         or the error is dropping below maxerror */
//     template <class StorageType>
//     int KLT(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//             StorageType **ida2, StorageType **gradx, StorageType **grady, 
//             unsigned int width, unsigned int height, unsigned int halfwinsize,
//             unsigned int maxiterations, KLT_TYPE maxerror, KLT_TYPE result[2], 
//             KLT_TYPE& error);
    
//     /** as above but uses the averaged gradient of the 2 images */
//     template <class StorageType>
//     int KLT2(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//              StorageType **ida2, StorageType **gradx1, StorageType **grady1, 
//              StorageType **gradx2, StorageType **grady2, 
//              unsigned int width, unsigned int height, unsigned int halfwinsize,
//              unsigned int maxiterations, KLT_TYPE maxerror, KLT_TYPE result[2],
//              KLT_TYPE& error);

//     /** same as above but also returns the number of iterations used */
//     template <class StorageType>
//     int KLT2(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//              StorageType **ida2, StorageType **gradx1, StorageType **grady1, 
//              StorageType **gradx2, StorageType **grady2, 
//              unsigned int width, unsigned int height, unsigned int halfwinsize,
//              unsigned int maxiterations, KLT_TYPE maxerror, KLT_TYPE result[2],
//              KLT_TYPE& error, unsigned &iter);

//     /** as above but uses a 5x5 window
//         @author woelk 08 2003 */
//     template <class StorageType>
//     int KLT5x5(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//                StorageType **ida2, StorageType **gradx1, StorageType **grady1, 
//                StorageType **gradx2, StorageType **grady2, 
//                unsigned width, unsigned height, unsigned maxiterations, 
//                KLT_TYPE maxerror, KLT_TYPE result[2], KLT_TYPE& error);

//     /** searches only perpendicular to the dominating edge, i.e.
//         parallel to the dominating gradient
//         the search direction is discretized in 4 different directions 
//         the discretization-error of direction is max 22 deg which 
//         accounts to 7% of length error
//         decides which KLT_? to use
//         @author woelk 08 2003    */
    
//     template <class StorageType>
//     int KLTEdge(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//                 StorageType **ida2, StorageType **gradx, StorageType **grady, 
//                 StorageType **gradNWSE, StorageType **gradNESW, 
//                 unsigned width, unsigned height, unsigned maxiterations, 
//                 KLT_TYPE maxerror, KLT_TYPE result[2], KLT_TYPE& error);

//     template <class StorageType>
//     int KLTEdge_NS(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//                    StorageType **ida2, StorageType **grad,  
//                    unsigned width, unsigned height, unsigned maxiterations, 
//                    KLT_TYPE maxerror, KLT_TYPE result[2], KLT_TYPE& error);

//     template <class StorageType>
//     int KLTEdge_WE(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//                    StorageType **ida2, StorageType **grad,  
//                    unsigned width, unsigned height, unsigned maxiterations, 
//                    KLT_TYPE maxerror, KLT_TYPE result[2], KLT_TYPE& error);

//     template <class StorageType>
//     int KLTEdge_NESW(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//                      StorageType **ida2, StorageType **grad,  
//                      unsigned width, unsigned height, unsigned maxiterations, 
//                      KLT_TYPE maxerror, KLT_TYPE result[2], KLT_TYPE& error);

//     template <class StorageType>
//     int KLTEdge_NWSE(KLT_TYPE p1[2], KLT_TYPE p2[2], StorageType **ida1, 
//                      StorageType **ida2, StorageType **grad,  
//                      unsigned width, unsigned height, unsigned maxiterations, 
//                      KLT_TYPE maxerror, KLT_TYPE result[2], KLT_TYPE& error);


    /** searches along line from start2 to end2 within +/- epsilon pixels
        away from the line in ida2 for a match to p1 
        using ncc with halfwinsize 
        boundary checking is not done, i.e. start2 and end2 
        must lie more than halfwindowsize+epsilon away from the 
        image border 
        p1 must lie more than halfwinsize away from image border
        only for those points in im2 a correlation is calculated for which
        the gradient is > grad1[p1[1]][p1[0]]*gradientscale 
        input: ida1 - pointer to the image data array of im1, row mayor
        

        output: result - image coo of point with greatest corelation
        correlation - the correlation betwen p1 and result
        returns: 0   on success
        -1  no matching point found
        <-1 error */
    template <class StorageType>
    int LineMatcherNCC(unsigned int p1[2], unsigned int start2[2], 
                       unsigned int end2[2], 
                       StorageType **ida1, StorageType **ida2,
                       StorageType **grad1, StorageType **grad2, 
                       unsigned int halfwinsize, unsigned int epsilon,
                       double gradientscale, unsigned int result[2], 
                       double& correlation);

    /** as above but uses SAD */
    template <class StorageType>
    int LineMatcherSAD(unsigned int p1[2], unsigned int start2[2], 
                       unsigned int end2[2], 
                       StorageType **ida1, StorageType **ida2,
                       StorageType **grad1, StorageType **grad2, 
                       unsigned int halfwinsize, unsigned int epsilon,
                       double gradientscale, unsigned int result[2], 
                       double& sad);

    /** function to achieve sub-pixel accuracy if p1 and p2
        are corresponding image points from ida1 resp ida2 
        exact parabola approximation with 3 values is used 
        no boundary checks, p1 and p2 must be further than 
        halfwinsize+1 from array border
        correlation between p1 and p2 must be a local discrete maximum
        i.e. NCC(p1, {p2[0]+/-1, p2[1]+/-1})<NCC(p1, p2)
        @author woelk 03 2003 */
    template <class StorageType>
    int ParabolaNCC(unsigned int p1[2], unsigned int p2[2],
                    StorageType **ida1, StorageType **ida2,
                    unsigned int halfwinsize, KLT_TYPE result[2]);


    /** function to achieve sub-pixel accuracy if p1 and p2
        are corresponding image points from ida1 resp ida2 
        over-determined parabola approximation with 5 values is used 
        no boundary checks, p1 and p2 must be further than 
        halfwinsize+2 from array border
        correlation between p1 and p2 must be a local discrete maximum
        i.e. NCC(p1, {p2[0]+/-2, p2[1]+/-2})<NCC(p1, p2)
        @author woelk 04 2003 */
    template <class StorageType>
    int ParabolaNCC5(unsigned int p1[2], unsigned int p2[2],
                    StorageType **ida1, StorageType **ida2,
                    unsigned int halfwinsize, KLT_TYPE result[2]);

    /*
    template <class StorageType>
    int GetTrackableRegions(StorageType **gx, StorageType **gy,
                            unsigned int halfwinsize, StorageType **tmap,
                            StorageType &min, StorageType&max);
    */

  protected:
    template <class StorageType>
    KLT_TYPE _Bilinear(StorageType **ida, KLT_TYPE x, KLT_TYPE y);

    KLT_TYPE _Bilinear(unsigned char **ida, 
				      KLT_TYPE x, KLT_TYPE y);

    /** bilinear interpolation, also moved gradient image data from [0:255]
        to [-127, 128] befor interpolating */
//     template <class StorageType>
//     KLT_TYPE _BilinearGrad(StorageType **ida, KLT_TYPE x, KLT_TYPE y);

    template <class StorageType>
    void _BilinearRegion(StorageType **ida, KLT_TYPE x, KLT_TYPE y, 
                         unsigned hws, Matrix<KLT_TYPE>& res);

    void _BilinearRegion(unsigned char **ida, KLT_TYPE x, KLT_TYPE y, 
                         unsigned hws, Matrix<KLT_TYPE>& res);

    /** bilinear interpolation, also moved gradient image data from [0:255]
        to [-127, 128] befor interpolating */
//     template <class StorageType>
//     void _BilinearRegionGrad(StorageType **ida, KLT_TYPE x, KLT_TYPE y, 
//                             unsigned hws, Matrix<KLT_TYPE>& res);

    inline
    KLT_TYPE _ParabolaApprox(KLT_TYPE x1, KLT_TYPE x2, KLT_TYPE x3)
    { KLT_TYPE a, p;
      a = (x1-2.0*x2+x3);
      p = (x1-x3);
      return (a!=0.0) ? (KLT_TYPE) p/(2.0*a) : (KLT_TYPE) 0.0; }

    inline
    KLT_TYPE _ParabolaApprox(KLT_TYPE x1, KLT_TYPE x2, KLT_TYPE x3, 
                             KLT_TYPE x4, KLT_TYPE x5)
    { Matrix<double> A(5, 3, "4 -2 1 1 -1 1 0 0 1 1 1 1 4 2 1");
      Vector<double> b(5);
      b[0]=x1; b[1]=x2; b[2]=x3; b[3]=x4; b[4]=x5;
      Vector<double> res=Lapack_LLS_QR_linear_solve(A, b);
      return (res[0]!=0.0) ? (KLT_TYPE) res[1]/(2.0*res[0]) : (KLT_TYPE) 0.0; }

    // TODO DLL JW
    //BIAS_EXPORT_STL_VECTOR
    //class BIASMatcher2D_EXPORT BIAS::Matrix2x2<KLT_TYPE>;
    //class BIASMatcher2D_EXPORT BIAS::Vector2<KLT_TYPE>;
    //class BIASMatcher2D_EXPORT BIAS::Matrix<KLT_TYPE>;
    //template class BIASMatcher2D_EXPORT BIAS::Matrix2x2<double>;
    //template class BIASMatcher2D_EXPORT BIAS::Vector2<double>;
    //template class BIASMatcher2D_EXPORT BIAS::Matrix<double>;


    // internal vars for KLT
    Matrix2x2<KLT_TYPE> _G, _Ginv;
    Vector2<KLT_TYPE> _e, _d, _disp;
    Matrix<KLT_TYPE> _gx, _gy, _bl1, _bl2, _gx1, _gy1, _gx2, _gy2, _gsx, _gsy;
    Matrix<KLT_TYPE> _bl1_5, _bl2_5, _gx1_5, _gy1_5, _gx2_5, _gy2_5, _gsx_5, 
      _gsy_5;
    unsigned _klthws, _winsize;
    // internals vars for interpolated NCC
    Matrix<KLT_TYPE> _iim1, _iim2;
    unsigned _ncchws, _nccwinsize;
    
  };
} // namespace
#endif // __RegionMatcher_hh__
