/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Tracker.hh"
#include <Base/Image/ImageIO.hh>

  //#define TIMING
#ifdef TIMING
#include <Base/Debug/TimeMeasure.hh>
#endif

using namespace BIAS;
using namespace std;

#ifdef BIAS_DEBUG
double totalitersum=0.0;
#endif

#define DFLT_PYSIZE 4


//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////

template <class StorageType, class CalculationType>
Tracker<StorageType, CalculationType>::
Tracker(Param& para, const std::string& postfix)
  : Debug(), tb_(para, "-Tracker"+postfix), 
    gsobel3x3_(para, "-Tracker"+postfix), ggauss_(para, "-Tracker"+postfix),
    ggaussasymmetric_(para, "-Tracker"+postfix), 
    gauss_(para, "-Tracker"+postfix), binomial_(para, "-Tracker"+postfix), 
    mean_(para, "-Tracker"+postfix)
{
  _ParamsAllocated = false;
  grad_=NULL;
  lowpass_=NULL;
  UseCoarserPointIfLost_=NULL;
  int grp=para.GetFreeGroupID();
  AddParameter_(para, grp, "-Tracker"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp, "Tracker");
  ParameterChanged();
}

template <class StorageType, class CalculationType>
Tracker<StorageType, CalculationType>::
Tracker(Param& para, const int grp, const std::string& postfix)
  : Debug(), tb_(para, grp, "-Tracker"+postfix), 
    gsobel3x3_(para, grp, "-Tracker"+postfix), 
    ggauss_(para, grp, "-Tracker"+postfix),
    ggaussasymmetric_(para, grp, "-Tracker"+postfix), 
    gauss_(para, grp, "-Tracker"+postfix), 
    binomial_(para, grp, "-Tracker"+postfix), 
    mean_(para, grp, "-Tracker"+postfix)
{
  _ParamsAllocated = false;
  grad_=NULL;
  lowpass_=NULL;
  UseCoarserPointIfLost_=NULL;
  AddParameter_(para, grp, "-Tracker"+postfix);
  ParameterChanged();
}

template <class StorageType, class CalculationType>
Tracker<StorageType, CalculationType>::
Tracker()
  : Debug(), tb_(),
    gsobel3x3_(), ggauss_(),
    ggaussasymmetric_(), 
    gauss_(), binomial_(), mean_()
{
  grad_=NULL;
  lowpass_=NULL;
  GradientType_= new int;
  *GradientType_ = GT_Sobel3x3; // sobel 3x3
  MeanType_= new int;
  *MeanType_ = LPT_Mean;    // 

  PyIndex_= new int;
  *PyIndex_ = 0;
  
  MaxIter_= new Vector<int>;
  MaxIter_->newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) (*MaxIter_)[i]=5;

  HalfWinSize_= new Vector<int>;
  HalfWinSize_->newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) (*HalfWinSize_)[i]=3;
    
   
  MaxError_ = new Vector<KLT_TYPE>;
  MaxError_->newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) (*MaxError_)[i]=0.1;

  UseCoarserPointIfLost_ = new bool(false);

  // use parameter from SimpleKLT as default
  _HalfWinSize = tb_.GetHalfWinSizePointer();
  _MaxIterations  = tb_.GetMaxIterationsPointer();
  _MaxError = tb_.GetMaxErrorPointer();

  ///////////////////////////////////////////////////////////////////////
  //  now get the parameter from mean gauss
  ///////////////////////////////////////////////////////////////////////
  _GaussSigma = gauss_.GetSigmaPointer();
  _GaussRatio = gauss_.GetRatioPointer();

  ///////////////////////////////////////////////////////////////////////
  //  now get the parameter from mean average
  ///////////////////////////////////////////////////////////////////////

  _MeanSize= mean_.GetSizePointer();
  _ParamsAllocated = true;
  ParameterChanged();
}


template <class StorageType, class CalculationType>
Tracker<StorageType, CalculationType>::
~Tracker()
{
  if (_ParamsAllocated){
    delete GradientType_;
    delete MeanType_;
    delete PyIndex_;
    delete MaxIter_;
    delete HalfWinSize_;
    delete MaxError_;
    delete UseCoarserPointIfLost_;
  }
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ReplaceLastImages(PyramidImage<CalculationType> *pim, 
                  PyramidImage<CalculationType> *gradx,
                  PyramidImage<CalculationType> *grady)
{
  LastImage_ = pim;
  LastGradX_ = gradx;
  LastGradY_ = grady;
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ParameterChanged()
{
  /*AddDebugLevel(D_TRACKER_INIT);
  AddDebugLevel(D_TRACKER_TRACK);
  AddDebugLevel(D_TRACKER_PARAM);
  AddDebugLevel(D_TRACKER_WRITE_IM);
  AddDebugLevel(D_TRACKER_FILL_UP);
  AddDebugLevel(D_TRACKER_IMAGES);
  AddDebugLevel(D_TRACKER_NUMITER);
  AddDebugLevel(D_TRACKER_ERROR);*/

  ClearPoints();
  LastImage_=NULL;
  LastGradX_=LastGradY_=NULL;
  
  //tb_.AddDebugLevel(D_TRACKERB_KLT);
  //tb_.AddDebugLevel(D_TRACKERB_BILINEAR);
  //tb_.AddDebugLevel(D_TRACKERB_INPUT);
  //tb_.AddDebugLevel(D_TRACKERB_INIT);
  //tb_.AddDebugLevel(D_TRACKERB_REJECT);

  // choose the gradient class
  switch (*GradientType_){
  case GT_Sobel3x3:
    grad_ = &gsobel3x3_;
    BIASCDOUT(D_TRACKER_PARAM, "using gradient Sobel3x3\n");
    break;
  case GT_Gauss:
    grad_ = &ggauss_; 
    BIASCDOUT(D_TRACKER_PARAM, "using gradient Gauss\n");
    break;
  case GT_GaussAsymmetric:
    grad_ = &ggaussasymmetric_;
    BIASCDOUT(D_TRACKER_PARAM, "using gradient GaussAsymmetric\n");
    break;
  default:
    BIASERR("invalid gradient type");
    grad_ = NULL;
    break;
  }
  // choose the mean class
  switch (*MeanType_){
  case LPT_Gauss:
    lowpass_ = &gauss_;
    BIASCDOUT(D_TRACKER_PARAM ,"using Gauss low pass filter\n");
    break;
  case LPT_Binomial:
    lowpass_ = &binomial_;
    BIASCDOUT(D_TRACKER_PARAM ,"using Binomial low pass filter\n");
    break;
  case LPT_Mean:
    lowpass_ = &mean_;
    BIASCDOUT(D_TRACKER_PARAM , "using mean low pass filter \n");
    break;
  default:
    BIASERR("invalid low pass filter type");
    lowpass_ = NULL;
    break;
  }
  
  // compute the minimum border distance
  int lpx, lpy, gx, gy;
  lowpass_->GetBorders(lpx, lpy);
  grad_->GetBorders(gx, gy);
  int mbd = max(lpx, lpy) + max(gx, gy) + (*HalfWinSize_)[*PyIndex_];
  MinBorderDist_ = mbd;
  /*cout << "MinBorderDist = "<<MinBorderDist_<<endl;
    cout << "lowpass = "<<max(lpx, lpy)<<endl;
    cout << "grad = "<<max(gx, gy)<<endl;
    cout << "hws = "<<(*HalfWinSize_)[*PyIndex_]<<endl;*/
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ClearPoints()
{
  Points_.clear();
  PredictedPoints_.clear();
  NumIter_.clear();
  DisplInLastIter_.clear();
  ResiduiMAD_.clear();
  ResiduiMSD_.clear();
  res_.clear();                  // rabe, 11.11.2004
  Cov_.clear();
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ResizeInternals_(const int pysize)
{
  if (Points_.size() != (unsigned)pysize){
    Points_.resize(pysize);
    PredictedPoints_.resize(pysize);
    NumIter_.resize(pysize);
    DisplInLastIter_.resize(pysize);
    ResiduiMAD_.resize(pysize);
    ResiduiMSD_.resize(pysize);
    res_.resize(pysize);
    Cov_.resize(pysize);
    AffineResults_.resize(pysize);
  }
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ResizeInternals_(const int pysize, const int num_points)
{
  ResizeInternals_(pysize);
  
  // dherzog: replaced '<numnew' by '!=numnew', because Track() asserts sizes
  if ((int)Points_[*PyIndex_].size() != num_points){
    HomgPoint2D pz(42.0, 42.0, 1.0);
    p2tracked_.resize(num_points, pz);
    AffinePrediction_.resize(num_points);
    for (int p = *PyIndex_; p < pysize; p++){
      res_[p].resize(num_points);
      Points_[p].resize(num_points);
      PredictedPoints_[p].resize(num_points);
      NumIter_[p].resize(num_points);
      DisplInLastIter_[p].resize(num_points);
      ResiduiMAD_[p].resize(num_points);
      ResiduiMSD_[p].resize(num_points);
      Cov_[p].resize(num_points);
      AffineResults_[p].resize(num_points);
    }
  }
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ReserveInternals_(const int pysize, const int newsize)
{
  int pyindex = *PyIndex_;
  for (int p=pyindex; p<pysize; p++){
    res_[p].reserve(newsize);
    Points_[p].reserve(newsize);
    PredictedPoints_[p].reserve(newsize);
    NumIter_[p].reserve(newsize);
    DisplInLastIter_[p].reserve(newsize);
    ResiduiMAD_[p].reserve(newsize);
    ResiduiMSD_[p].reserve(newsize);
    Cov_[p].reserve(newsize);
    AffineResults_[p].resize(newsize);
  }
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
Reset() 
{
  ClearPoints();
  LastImage_=NULL;
  LastGradX_=LastGradY_=NULL;
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
DetermineROI_(int &tlx, int &tly, int &brx, int &bry) const 
{
  BIASASSERT(LastImage_ && LastGradX_ && LastGradY_);
  (*LastImage_)[*PyIndex_]->GetROI()->GetCorners(tlx, tly, brx, bry);
  int gxtlx, gxtly, gxbrx, gxbry;
  (*LastGradX_)[*PyIndex_]->GetROI()->GetCorners(gxtlx, gxtly, gxbrx, gxbry);
  int gytlx, gytly, gybrx, gybry;
  (*LastGradY_)[*PyIndex_]->GetROI()->GetCorners(gytlx, gytly, gybrx, gybry);
  
  tlx = max(tlx, max(gxtlx, gytlx));
  tly = max(tly, max(gxtly, gytly));
  brx = min(brx, min(gxbrx, gybrx));
  bry = min(bry, min(gxbry, gybry));
  
  tlx += (*HalfWinSize_)[*PyIndex_];
  tly += (*HalfWinSize_)[*PyIndex_];
  brx -= (*HalfWinSize_)[*PyIndex_];
  bry -= (*HalfWinSize_)[*PyIndex_];

  // heuristic, ROIs are mostly symmetric
  //MinBorderDist_ = max(tlx, tly);
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
AddPoints(const vector<HomgPoint2D>& points, const vector<float>& qual)
{
  assert("deprecated, use ReplaceAllPoints instead" && false);
  //AddDebugLevel(D_TRACKER_INIT);
  BIASASSERT(LastImage_!=NULL);
  BIASASSERT(points.size()==qual.size());
  const int pysize = LastImage_->size();
  ResizeInternals_(pysize);

  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();

  int newsize=Points_[0].size()+points.size();
  vector<HomgPoint2D>::const_iterator it;
  vector<float>::const_iterator qit;
  const int pyindex=*PyIndex_;
  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();
  BIASCDOUT(D_TRACKER_INIT, "min distance to image border "
            <<MinBorderDist_<<endl);
  int tlx, tly, brx, bry;
  (*LastImage_)[pyindex]->GetROI()->GetCorners(tlx, tly, brx, bry);
  const double minx=MinBorderDist_+tlx;
  const double miny=MinBorderDist_+tly;
  const double maxx=brx-MinBorderDist_;
  const double maxy=bry-MinBorderDist_;
    
  ReserveInternals_(pysize, newsize);

  for (it=points.begin(), qit=qual.begin(); it!=points.end(); it++, qit++){
    p2d=*it;
    p2d.Homogenize();
    if (!(p2d[0]>minx && p2d[1]>miny && p2d[0]<maxx && p2d[1]<maxy)){
      // don't track point too close to image border
      p2d[2] = 0.0;
    }
    BIASCDOUT(D_TRACKER_INIT, "add point ");
    for (int p=pyindex; p<pysize; p++){
      Points_[p].push_back(p2d);
      BIASCDOUT(D_TRACKER_INIT, p2d<<"  ");
      NumIter_[p].push_back(0);
      DisplInLastIter_[p].push_back(-1.0);
      ResiduiMAD_[p].push_back(-1.0);
      ResiduiMSD_[p].push_back(-1.0);
      p2d[0]-=offset;
      p2d[1]-=offset;
      p2d[0]*=factor;
      p2d[1]*=factor;
      res_[p].push_back((int)rint(*qit));
      Cov_[p].push_back(zero2x2);
    }
    BIASCDOUT(D_TRACKER_INIT, endl);
  }
  p2d.Set(0.0, 0.0, 1.0);
  newsize=Points_[pyindex].size();
  for (int p=0; p<pysize; p++)
    PredictedPoints_[p].resize(newsize, p2d);

  p2tracked_.resize(newsize, p2d);
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
AddNumPoints(const vector<HomgPoint2D>& points, const vector<float>& qual,
             const int nfeat)
{
  assert("deprecated, use ReplaceAllPoints instead" && false);
  //AddDebugLevel(D_TRACKER_INIT);
  BIASASSERT(LastImage_!=NULL);
  BIASASSERT(points.size()==qual.size());
  const int pysize = LastImage_->size();
  ResizeInternals_(pysize);

  const int pyindex=*PyIndex_;
  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();
  int newsize=Points_[pyindex].size()+points.size();
  vector<HomgPoint2D>::const_iterator it;
  vector<float>::const_iterator qit;
  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();
  BIASCDOUT(D_TRACKER_INIT, "min distance to image border "
            <<MinBorderDist_<<endl);
  int tlx, tly, brx, bry;
  (*LastImage_)[pyindex]->GetROI()->GetCorners(tlx, tly, brx, bry);
  const double minx=MinBorderDist_+tlx;
  const double miny=MinBorderDist_+tly;
  const double maxx=brx-MinBorderDist_;
  const double maxy=bry-MinBorderDist_;
    
  BIASCDOUT(D_TRACKER_INIT, "roi: ("<<minx<<", "<<miny<<") <--> ("<<maxx
            <<", "<<maxy<<")  "<<MinBorderDist_<<"\n");

  ReserveInternals_(pysize, newsize);

  int n=0, num=Points_[pyindex].size();
  const int psize=points.size();

  while (n<psize && num<nfeat){
    p2d=points[n];
    p2d.Homogenize();
    if (p2d[0]>minx && p2d[1]>miny && p2d[0]<maxx && p2d[1]<maxy){
      BIASCDOUT(D_TRACKER_INIT, "add point ");
      for (int p=pyindex; p<pysize; p++){
        Points_[p].push_back(p2d);
        BIASCDOUT(D_TRACKER_INIT, p2d<<"  ");
        NumIter_[p].push_back(0);
        DisplInLastIter_[p].push_back(-1.0);
        ResiduiMAD_[p].push_back(-1.0);
        ResiduiMSD_[p].push_back(-1.0);
        p2d[0]-=offset;
        p2d[1]-=offset;
        p2d[0]*=factor;
        p2d[1]*=factor;
        res_[p].push_back((int)rint(qual[n]));
        Cov_[p].push_back(zero2x2);
      }
      BIASCDOUT(D_TRACKER_INIT, endl);
      num++;
    } else {
      BIASCDOUT(D_TRACKER_INIT,"Skipped border point "<<p2d);
    }
    n++;
  }
  p2d.Set(0.0, 0.0, 0.0);
  newsize=Points_[pyindex].size();
#ifdef BIAS_DEBUG  
  if (num!=nfeat){
    BIASERR("could not add enough points, maybe there were not enough input "
            <<"points "<<psize<<", have "<<num<<" of "<<nfeat 
            <<"\n try decreasing MinDistance in CornerDetector or \n"
            <<" decreasing MinCornerness in CornerDetector\n\n\n");
  }
#endif
  // not enough input point, add points at infinity
  for (;num<nfeat; num++){
    for (int p=pyindex; p<pysize; p++){
      Points_[p].push_back(p2d);
      BIASCDOUT(D_TRACKER_INIT, p2d<<"  ");
      NumIter_[p].push_back(0);
      DisplInLastIter_[p].push_back(-1.0);
      ResiduiMAD_[p].push_back(-1.0);
      ResiduiMSD_[p].push_back(-1.0);
      res_[p].push_back((int)rint(qual[n]));
      Cov_[p].push_back(zero2x2);
    }
    BIASCDOUT(D_TRACKER_INIT, endl);
  }
    
  p2d.Set(0.0, 0.0, 1.0);
  for (int p=0; p<pysize; p++)
    PredictedPoints_[p].resize(nfeat, p2d);
  p2tracked_.resize(nfeat, p2d);

  return (newsize==nfeat)?0:-1;
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
FillUpPoints(const vector<HomgPoint2D>& points,
             const vector<float>& qual)
{
  //AddDebugLevel(D_TRACKER_FILL_UP);
  BIASASSERT(LastImage_!=NULL);
  BIASASSERT(Points_.size()>0);
  const int pysize=(int)LastImage_->size();
  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();
  int numnew=points.size();
  const int size=Points_[0].size();
  const int pyindex=*PyIndex_;
  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();
  BIASCDOUT(D_TRACKER_INIT, "min distance to image border "
            <<MinBorderDist_<<endl);
  int tlx, tly, brx, bry;
  (*LastImage_)[pyindex]->GetROI()->GetCorners(tlx, tly, brx, bry);
  const double minx=MinBorderDist_+tlx;
  const double miny=MinBorderDist_+tly;
  const double maxx=brx-MinBorderDist_;
  const double maxy=bry-MinBorderDist_;

#ifdef BIAS_DEBUG
  BIASCDOUT(D_TRACKER_FILL_UP, "minx" << minx << "miny" << miny <<
            "maxx" << maxx << "maxy" << maxy << endl);
#endif
  int n=0, num=0;
  while (n<numnew && (points[n].IsAtInfinity() || 
                      !(points[n][0]>minx && points[n][1]>miny && 
                        points[n][0]<maxx && points[n][1]<maxy))) 
    n++;
#ifdef BIAS_DEBUG
  if (n<numnew)
    BIASCDOUT(D_TRACKER_FILL_UP, "next new point is "<<points[n]<<endl);
#endif
  for (int i=0; i<size; i++){
    if (n<numnew){
      if (Points_[pyindex][i].IsAtInfinity()){
        p2d=points[n];
        p2d.Homogenize();
        BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<i<<" : replacing "
                  <<Points_[pyindex][i]<<" with "<<p2d<<endl);
        for (int p=pyindex; p<pysize; p++){
          Points_[p][i]=p2d;
          NumIter_[p][i]=-1;
          DisplInLastIter_[p][i]=-1.0;
          ResiduiMAD_[p][i]=-1.0;
          ResiduiMSD_[p][i]=-1.0;
          p2d[0]-=offset;
          p2d[1]-=offset;
          p2d[0]*=factor;
          p2d[1]*=factor;
          res_[p][i]=(int)rint(qual[n]);
          Cov_[p][i]=zero2x2;
        }
        n++; num++;
        while (n<numnew && ( points[n].IsAtInfinity() || 
                             !(points[n][0]>minx && points[n][1]>miny && 
                               points[n][0]<maxx && points[n][1]<maxy)))
          n++;
#ifdef BIAS_DEBUG
        if (n<numnew)
          BIASCDOUT(D_TRACKER_FILL_UP, "next new point is "<<points[n]<<endl);
#endif
      } else {
#ifdef BIAS_DEBUG
        BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<i<<" : "<<Points_[pyindex][i]
                  <<" is still valid:  ");
        for (int p=pyindex+1; p<pysize; p++){
          BIASCDOUT(D_TRACKER_FILL_UP, p<<": "<<Points_[p][i]);
        }
        BIASCDOUT(D_TRACKER_FILL_UP, endl);
#endif
      } 
    } else {
      num=-num;
      BIASCDOUT(D_TRACKER_FILL_UP, "not enought points in points "
                <<numnew<<endl);
      break;
    }
  }
  return num;
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
ReplacePoints(const vector<HomgPoint2D>& points,
              const vector<float>& qual)
{
  assert("deprecated, uese ReplaceAllPoints instead" && false);
  //AddDebugLevel(D_TRACKER_FILL_UP);
  BIASASSERT(LastImage_!=NULL);
  BIASASSERT(Points_.size()>0);
  const int pysize=(int)LastImage_->size();
  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();
  const int numnew=points.size();
#ifdef BIAS_DEBUG
  const int size=Points_[0].size();
  BIASASSERT(size==numnew);
#endif
  const int pyindex=*PyIndex_;
  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();
  
  int n=0, num=0;
  while (n<numnew){
    while (n<numnew && points[n].IsAtInfinity()) n++;
    if (n<numnew){
#ifdef BIAS_DEBUG
      BIASCDOUT(D_TRACKER_FILL_UP, "next new point is "<<points[n]<<endl);
#endif
      p2d=points[n];
      p2d.Homogenize();
      BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<n<<" : replacing "
                <<Points_[pyindex][n]<<" with "<<p2d<<endl);
      for (int p=pyindex; p<pysize; p++){
        Points_[p][n]=p2d;
        NumIter_[p][n]=-1;
        DisplInLastIter_[p][n]=-1.0;
        ResiduiMAD_[p][n]=-1.0;
        ResiduiMSD_[p][n]=-1.0;
        p2d[0]-=offset;
        p2d[1]-=offset;
        p2d[0]*=factor;
        p2d[1]*=factor;
        res_[p][n]=(int)rint(qual[n]);
        Cov_[p][n]=zero2x2;
      }
      n++; num++;
    }
  }
  BIASCDOUT(D_TRACKER_FILL_UP, "replaced "<<num<<" points"<<endl);
  return n;
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
ReplaceAllPoints(const vector<HomgPoint2D>& points,
                 const vector<float>& qual)
{
  //AddDebugLevel(D_TRACKER_FILL_UP);
  BIASASSERT(LastImage_!=NULL);
  const int numnew=points.size();
  const int pysize=(int)LastImage_->size();
  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();
  const int pyindex=*PyIndex_;
  // dherzog: replaced '<numnew' by '!=numnew', because Track() asserts sizes
  if ((int)Points_.size() != pysize || (int)Points_[pyindex].size()!=numnew) {
    ClearPoints();
    ResizeInternals_(pysize, numnew);
  }

  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();

  for (int n=0; n<numnew; n++){
    p2d = points[n];
    p2d.Homogenize();
    BIASASSERT(!p2d.IsAtInfinity());
    BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<n<<" : replacing "
              <<Points_[pyindex][n]);
    for (int p=pyindex; p<pysize; p++){
      res_[p][n] = (int)rint(qual[n]);
      Points_[p][n] = p2d;
      PredictedPoints_[p][n] = p2d;
      BIASASSERT(!Points_[p][n].IsAtInfinity());
      BIASASSERT(!PredictedPoints_[p][n].IsAtInfinity());
      NumIter_[p][n] = -1;
      DisplInLastIter_[p][n] = -1.0;
      ResiduiMAD_[p][n] = -1.0;
      ResiduiMSD_[p][n] = -1.0;
      Cov_[p][n] = zero2x2;

      p2d[0] -= offset;
      p2d[1] -= offset;
      p2d[0] *= factor;
      p2d[1] *= factor;
    }
    BIASCDOUT(D_TRACKER_FILL_UP, " with "<<Points_[pyindex][n]<<endl);
  }
  BIASCDOUT(D_TRACKER_FILL_UP, "replaced "<<numnew<<" points"<<endl);

  PadPoints_(numnew);

  return numnew;
}


#ifdef BIAS_DEBUG
static int Tracker_GlobalImageCounter=0;
#endif

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
PreparePyramide(const Image<StorageType>& im, 
                PyramidImage<CalculationType>& pim,
                PyramidImage<CalculationType>& gx, 
                PyramidImage<CalculationType>& gy)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  int fres=0, gres=0;
#ifdef BIAS_DEBUG
  BIASASSERT(im.GetChannelCount()==1);
  BIASASSERT(im.GetColorModel()==ImageBase::CM_Grey);
#endif

  if (pim.size()==0 || !im.SamePixelAndChannelCount(*pim[0])){
    if (!pim.IsEmpty()) pim.Release();
    pim.Init(im.GetWidth(), im.GetHeight(), 1);
    pim.SetZero(); 
  }
  if (gx.size()==0 || !im.SamePixelAndChannelCount(*gx[0])){
    if (!gx.IsEmpty()) gx.Release();
    gx.Init(im.GetWidth(), im.GetHeight(), 1);
    gx.SetZero();
  }    
  if (gy.size()==0 || !im.SamePixelAndChannelCount(*gy[0])){
    if (!gy.IsEmpty()) gy.Release();
    gy.Init(im.GetWidth(), im.GetHeight(), 1);
    gy.SetZero();
  }

#ifdef TIMING
  timer.Stop();
  cerr << "initializing pyramid images took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif
  
#ifdef BIAS_DEBUG
  ostringstream os;
  if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
    os.str("");
    os << "input-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
       <<".mip";
    Image<StorageType> tmp_im=im;
    if (ImageIO::Write(os.str(), tmp_im)!=0){
      BIASERR("error writing "<<os.str());
    }
  }
#endif

  //lowpass_->AddDebugLevel(D_CONV_KERNEL);
  //lowpass_->AddDebugLevel(D_FILTERBASE_CALLSTACK);
  //lowpass_->AddDebugLevel(D_WRITE_IMAGES);
  if((fres=lowpass_->Filter(im, *pim[0]))!=0){
    BIASERR("filtering failed "<<fres);
  }
  //  lowpass_->PrintKernel();

#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
    os.str("");
    os << "pyim-lowpass-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
       << "-pyind"<<setw(1)<<setfill('0')<<0<<".mip";
    if (ImageIO::Write(os.str(), *pim[0])!=0){
      BIASERR("error writing "<<os.str());
    }
  }
#endif

  //lowpass_->RemoveDebugLevel(D_CONV_KERNEL);
#ifdef TIMING
  timer.Stop();
  cerr << "lowpass  took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif

  pim.Downsample();

#ifdef TIMING
  timer.Stop();
  cerr << "downsampling  took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif

  const int pysize=pim.size(); // comes from parameter
  for (int p=0; p<pysize; p++){
    if((gres=grad_->Filter(*pim[p], *gx[p], *gy[p]))!=0){
      BIASERR("gradient failed "<<gres);
      break;
    }
#ifdef BIAS_DEBUG
    if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
      os.str("");
      os << "pyim-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
         << "-pyind"<<setw(1)<<setfill('0')<<p<<".mip";
      if (ImageIO::Write(os.str(), *pim[p])!=0){
        BIASERR("error writing "<<os.str());
      }
      os.str("");
      os << "pygx-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
         << "-pyind"<<setw(1)<<setfill('0')<<p<<".mip";
      if (ImageIO::Write(os.str(), *gx[p])!=0){
        BIASERR("error writing "<<os.str());
      }
      os.str("");
      os << "pygy-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
         << "-pyind"<<setw(1)<<setfill('0')<<p<<".mip";
      if (ImageIO::Write(os.str(), *gy[p])!=0){
        BIASERR("error writing "<<os.str());
      }
    }
#endif
  }
#ifdef BIAS_DEBUG
  Tracker_GlobalImageCounter++;
#endif
#ifdef TIMING
  timer.Stop();
  cerr << "gradients  took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif
  return (fres==0 && gres==0)?0:-1;
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
Track(PyramidImage<CalculationType>& pim, 
      PyramidImage<CalculationType>& gradx,
      PyramidImage<CalculationType>& grady)
{
  return Track(pim, gradx, grady, Points_[*PyIndex_]);
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
Track(PyramidImage<CalculationType>& pim, 
      PyramidImage<CalculationType>& gradx,
      PyramidImage<CalculationType>& grady,
      const vector<HomgPoint2D>& prediction,
      const vector<Matrix2x2<KLT_TYPE> >& AffinePred)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  int num=0;
  const int pysize=(int)pim.size();
  const double factor=pim.GetRescaleFactor();
  const double offset=pim.GetPositionOffset();
  const Matrix2x2<double> K(factor, 0.0, 0.0, factor);
  //tb_.AddDebugLevel(D_TRACKERB_GKLT);
  //tb_.AddDebugLevel(D_TRACKERB_INIT);

  BIASASSERT(*PyIndex_>=0 && *PyIndex_<pysize);

  if (LastImage_ && LastGradX_ && LastGradY_){
#ifdef BIAS_DEBUG
    if ((&pim)==LastImage_ || (&gradx)==LastGradX_ || (&grady)==LastGradY_ ){
      BIASERR("do not remove image between successive calls of Track\n"
              <<(int)(&pim)<<"!="<<(int)(LastImage_)<<","<<(int)(&gradx)<<"!="
              <<(int)(LastGradX_)<<","<<(int)(&grady)<<"!="<<(int)(LastGradY_));
    }
    BIASASSERT((&pim)!=LastImage_ && (&gradx)!=LastGradX_ && 
               (&grady)!=LastGradY_ );
    if (DebugLevelIsSet(D_TRACKER_IMAGES)){
      pim.WriteImages("act-py"); LastImage_->WriteImages("last-py");
      gradx.WriteImages("act-gx"); LastGradX_->WriteImages("last-gx");
      grady.WriteImages("act-gy"); LastGradY_->WriteImages("last-gy");
    }
#endif    

    const int pyindex=*PyIndex_;
    const int nump=Points_[pyindex].size();

    // fill PredictedPoints_ and Points_
    PredictedPoints_[pyindex]=prediction;

    const double ifac=1.0/factor;
    for (int p=pyindex+1; p<pysize; p++){
      for (int i=0; i<nump; i++){
        Points_[p][i][0]=(Points_[p-1][i][0]-offset)*ifac;
        Points_[p][i][1]=(Points_[p-1][i][1]-offset)*ifac;
        PredictedPoints_[p][i][0]=(PredictedPoints_[p-1][i][0]-offset)*ifac;
        PredictedPoints_[p][i][1]=(PredictedPoints_[p-1][i][1]-offset)*ifac;
      }
    }
    const bool Affine = (*tb_.GetTrackerType()!=TT_Simple) &&
      ((*tb_.GetTrackerType()!=TT_Weighted));
    if (Affine) {
      //cout<<"affine mode !"<<nump<<" "<< AffinePred.size()  <<endl;
      if ((int)AffinePred.size()==nump) {
        // copy from vector
        AffinePrediction_ = AffinePred;
      } else {
        // have no prediction, start with identity
        AffinePrediction_.resize(nump);
        for (int i=0; i<nump; i++) {
          AffinePrediction_[i].SetIdentity();
        }
      }
    }

#ifdef TIMING
    timer.Stop();
    cerr << "initializing "<<nump<<" points took "<<timer.GetRealTime()<<" us\n";
    timer.Reset();
    timer.Start();
#endif

    // now do the actual tracking, starting on smalles pyramid level
    for (int p=pysize-1;p>=pyindex; p--){
      BIASCDOUT(D_TRACKER_TRACK, "---- pyramid level "<<p<<" ----\n");

      *_HalfWinSize=(*HalfWinSize_)[p];
      *_MaxIterations=(*MaxIter_)[p];
      *_MaxError=(*MaxError_)[p];

      tb_.Init(*(*LastImage_)[p], *pim[p], *(*LastGradX_)[p],
                *(*LastGradY_)[p], *gradx[p], *grady[p]);
#ifdef TIMING
      timer.Stop();
      cerr << "initializing tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

      tb_.Track(Points_[p], PredictedPoints_[p], p2tracked_, 
                DisplInLastIter_[p], NumIter_[p], ResiduiMAD_[p], 
                ResiduiMSD_[p], res_[p],Cov_[p], AffinePrediction_);
      
      if (Affine) {
        tb_.GetAffineResults(AffineResults_[p]);
        // update prediction with last estimate
        AffinePrediction_ = AffineResults_[p];
      }

#ifdef TIMING
      timer.Stop();
      cerr << "tracking tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif
      
      // check for result
      for (int i=0; i<nump; i++){
        BIASCDOUT(D_TRACKER_TRACK, setw(3)<<i<<"  ("<<Points_[p][i][0]<<", "
                  <<Points_[p][i][1]<<")  <-->  ("<<PredictedPoints_[p][i][0]
                  <<", "<<PredictedPoints_[p][i][1]<<") ");
        if (res_[p][i]>=0){
          Points_[p][i]=p2tracked_[i];
          BIASCDOUT(D_TRACKER_TRACK, " -> ("<<p2tracked_[i][0]<<", "
                    <<p2tracked_[i][1]<<") in "<<NumIter_[p][i]
                    <<" iterations with err "<<DisplInLastIter_[p][i]
                    <<"   residuumMAD "<<ResiduiMAD_[p][i]
                    <<"   residuumMSD "<<ResiduiMSD_[p][i]<<endl);
          if (p==pyindex) {
            num++;
          } else {
            PredictedPoints_[p-1][i][0]=Points_[p][i][0]*factor+offset;
            PredictedPoints_[p-1][i][1]=Points_[p][i][1]*factor+offset;
          }
        } else { // if (res_[p][i]>=0){
          BIASCDOUT((D_TRACKER_ERROR | D_TRACKER_TRACK), " error tracking: "
                    <<Points_[p][i]<<" on py level "<<p<<" result code "
                    <<res_[p][i]<<endl);
          // cerr <<" error tracking: "<<res_[p][i]<<endl;
          // cerr << setw(3)<<i<<"  ("<<Points_[p][i][0]<<", "
          //      <<Points_[p][i][1]<<")  <-->  ("<<PredictedPoints_[p][i][0]
          //      <<", "<<PredictedPoints_[p][i][1]<<") "<<endl;
          if (!UseCoarserPointIfLost_) {
            // manifest the failure in homogenous coordinate
            // do manifest it in all pyramid levels smaller than actual
            // the bigger pyramid levels are taken care of by code below
            for (int k=p; k<pysize; k++)
              Points_[k][i][2]=0.0;
          } else {          
            // try to propagate results from last successfull track
            if (p!=(pysize-1) && res_[p+1][i]>=0){  
              if (p!=pyindex) {
                PredictedPoints_[p-1][i][0]=Points_[p+1][i][0]*factor+offset;
                PredictedPoints_[p-1][i][1]=Points_[p+1][i][1]*factor+offset; 
              }
              if (res_[p+1][i]!=0){
                res_[p][i]=res_[p+1][i];
                NumIter_[p][i]=NumIter_[p+1][i];
                DisplInLastIter_[p][i]=DisplInLastIter_[p+1][i]*factor;
                ResiduiMAD_[p][i]=ResiduiMAD_[p+1][i];
                ResiduiMSD_[p][i]=ResiduiMSD_[p+1][i];
                // if a random variable is transformed by a linear function with 
                // matrix F, the covariance transformed as J*cov*J^T, where J is
                // the jacobian of F. Here the Jacobian is a matrix where the 
                // only entries on the main diagonal are the factors 
                Cov_[p][i]=(factor*factor)*Cov_[p+1][i];
              }
            }
          } 
        } // if (res_[p][i]>=0){
      } 

#ifdef TIMING
      timer.Stop();
      cerr << "checking results of tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

#ifdef BIAS_DEBUG
      if (DebugLevelIsSet(D_TRACKER_NUMITER)){
        //check number of iterations
        double itersum=0.0;
        for (unsigned h=0; h<NumIter_[p].size(); h++)
          itersum+=NumIter_[p][h];
        //cout << p << " average number of iterations "
        //     <<itersum/(double)NumIter_[p].size()<<endl;
        totalitersum+=itersum;
        BIASCDOUT(D_TRACKER_NUMITER, "totalitersum "<<totalitersum<<endl);
      }
#endif      
    }

    //DumpResults(); cerr << endl << endl;

#ifdef TIMING
    timer.Stop();
    cerr << "propagating result codes through pyramide took "<<timer.GetRealTime()<<" us\n";
    timer.Reset();
    timer.Start();
#endif


#ifdef BIAS_DEBUG
    if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
      Image<CalculationType> wim;
      Image<CalculationType> gim;
      pim.GetSingleImage(wim);
      ImageIO::Write("py2.mip", wim);
      LastImage_->GetSingleImage(wim);
      ImageIO::Write("py1.mip", wim);
      gradx.GetSingleImage(gim);
      ImageIO::Write("gradx2.mip", gim);
      LastGradX_->GetSingleImage(gim);
      ImageIO::Write("gradx1.mip", gim);
      grady.GetSingleImage(gim);
      ImageIO::Write("grady2.mip", gim);
      LastGradY_->GetSingleImage(gim);
      ImageIO::Write("grady1.mip", gim);
    }
#endif
   

  } else {
    BIASCDOUT(D_TRACKER_TRACK, "initialization call\n");
  }

  

  LastImage_=&pim;
  LastGradX_=&gradx;
  LastGradY_=&grady;

#ifdef TIMING
  timer.Stop();
  cerr << "updateing iternals  took "<<timer.GetRealTime()<<" us\n";
#endif

#ifdef BIAS_DEBUG
  //WriteTrackLog_(num);
#endif

  return num;
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
WriteTrackLog_(int num)
{
  static bool first = true;
  static int img_num = 0;
  const int min_err_code = -7;
  static ofstream os("tracker.log");  
  if (first){
    // open the file
    if (!os.good()){
      BIASERR("error opening \"tracker.log\"");
      BIASABORT;
    }
    // write header
    os << "# reasons for lost tracks\n" 
       << "#" << setw(4)<<"no"
       << setw(7) << "num"
       << setw(7) << "lost"
       << setw(7) << "ignrd";
    for (int r = 0; r >= min_err_code; r--){ 
      os << setw(7) << r;
    }
    os << endl;
    first = false;
  }

  const int pyindex = *PyIndex_;
  if (!Points_.empty()){
    const int nump = Points_[pyindex].size();
    // count the different failure reasons
    // and compute mean and variance of covariances
    vector<int> reason(-min_err_code + 1, 0);
    int reas;
    for (int i=0; i<nump; i++){
      reas = res_[pyindex][i];
      if (reas < 0){
        if (reas<min_err_code) {
          BIASABORT;
        }
        reason[-reas]++;
      } else {
        reason[0]++;
      }
    }
  

    os << setw(5) << img_num
       << setw(7) << nump
       << setw(7) << nump - num - reason[6]
       << setw(7) << reason[6];
    for (int r = 0; r <= -min_err_code; r++){ 
      os << setw(7) << reason[r];
    }
    os << endl << flush;

    img_num++;
  }
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
WriteTrackFile(deque<vector<HomgPoint2D> >& p, deque<vector<int> >& results, 
               const string& name)
{
  const int width=20;
  int i, k;
  ofstream os(name.c_str());
  if (!os){
    BIASERR("error opening "<<name);
    return -1;
  }

  os << "Feel free to place comments here.\n\n";

  os << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";
  os << "!!! Warning:  This is a KLT data file.  Do not modify below this line !!!\n\n";

  os << "------------------------------\n";
  os << "KLT Feature Table\n";
  os << "------------------------------\n\n";

  os << "nFrames = "<<p.size()<<", nFeatures = "<<p[0].size()<<"\n\n";  
  
  os << "feature |          frame\n";
  os << "        |";
  for (i = 0 ; i < (int)p.size() ; i++) 
    os << setw(width)<<i;
  os << "\n";
  os << "--------+";
  for (i = 0 ; i < (int)p.size() ; i++) 
    for (k=0; k<width; k++) os << "-";
  os << "\n";
  if (!os){
    BIASERR("error writing "<<name);
    return -1;
  }
  os.close();
  

  FILE *fp=fopen(name.c_str(), "ab");
  for (k=0; k<(int)p[0].size(); k++){ // k = feature
    fprintf(fp, "%7d | ", k);
    for (i=0; i<(int)p.size(); i++){ // i = image
      //      cout << p[i][k] << endl;
      //cout << residui[i][k] << endl;
      if (!p[i][k].IsAtInfinity()){
        fprintf(fp, "(%7.3f,%7.3f)=%5d ", p[i][k][0], p[i][k][1], 
               results[i][k]); 
      } else {
        fprintf(fp, "(%7.3f,%7.3f)=%5d ", -1.0, -1.0, results[i][k]);
      }
    }
    fprintf(fp, "\n");
  }
 
  fclose(fp);
  return 0;
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
AddParameter_(Param& para, const int grp, const std::string& postfix)
{

  BIASCDOUT(D_TRACKER_PARAM, "Tracker<StorageType, CalculationType>: using "
            <<"group ID "<<grp<<"\n");

  GradientType_= AddGradientType(para, "GradientType"+postfix, grp);

  MeanType_= AddLowPassType(para, "LowPassType"+postfix, grp);
  
  if (para.CheckParam("PyIndex"+postfix)){ 
    PyIndex_=para.GetParamInt("PyIndex"+postfix);
  } else { 
    PyIndex_=para.AddParamInt("PyIndex"+postfix, 
                              "biggest pyramid image used for tracking", 
                              0, 0, INT_MAX, 0, grp); 
  }

  Vector<int> idflt;
  idflt.newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) idflt[i]=5;
  if (para.CheckParam("MaxIter"+postfix)){ 
    MaxIter_=para.GetParamVecInt("MaxIter"+postfix);
  } else {
    MaxIter_=para.AddParamVecInt("MaxIter"+postfix, 
                                 "number max iterations per pyramid level", 
                                 idflt, 0, grp); 
  }

  for (int i=0; i<DFLT_PYSIZE; i++) idflt[i]=3;
  if (para.CheckParam("HalfWinSize"+postfix)){ 
    HalfWinSize_=para.GetParamVecInt("HalfWinSize"+postfix);
  } else {
    HalfWinSize_=para.AddParamVecInt("HalfWinSize"+postfix, 
                                     "half window sizes per pyramid level", 
                                     idflt, 0, grp); 
  }
  
  Vector<KLT_TYPE> ddflt;
  ddflt.newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) ddflt[i]=0.1;
  if (para.CheckParam("MaxError"+postfix)){ 
    MaxError_=para.GetParamVecDbl("MaxError"+postfix);
  } else {
    MaxError_=para.AddParamVecDbl("MaxError"+postfix, 
                                  "iteration error stopping criterion "
                                  "per pyramid level", 
                                  ddflt, 0, grp); 
  }

  

  if (para.CheckParam("UseCoarserPointIfLost"+postfix)){ 
    UseCoarserPointIfLost_ = para.GetParamBool("UseCoarserPointIfLost" 
                                               + postfix);
  } else { 
    UseCoarserPointIfLost_ = para.AddParamBool("UseCoarserPointIfLost"+postfix,
                                               "if point untrackable in high "
                                               " res, use coarse result", 
                                               false, 0, grp); 
  }


  ///////////////////////////////////////////////////////////////////////
  //  now get the parameter from tracker base
  ///////////////////////////////////////////////////////////////////////
  
  if (para.CheckParam("HalfWinSize-TrackerBase"+postfix)){ 
    _HalfWinSize=para.GetParamInt("HalfWinSize-TrackerBase"+postfix);
  } else { 
    BIASERR("something funny happend, maybe TrackerBase is not initialized");
  }

  if (para.CheckParam("MaxIterations-TrackerBase"+postfix)){ 
    _MaxIterations=para.GetParamInt("MaxIterations-TrackerBase"+postfix);
  } else { 
    BIASERR("something funny happend, maybe TrackerBase is not initialized");
  }

  if (para.CheckParam("MaxError-TrackerBase"+postfix)){ 
    _MaxError=para.GetParamDouble("MaxError-TrackerBase"+postfix);
  } else { 
    BIASERR("something funny happend, maybe TrackerBase is not initialized");
  }

  ///////////////////////////////////////////////////////////////////////
  //  now get the parameter from mean gauss
  ///////////////////////////////////////////////////////////////////////
  
  if (para.CheckParam("Sigma-Gauss"+postfix)){ 
    _GaussSigma=para.GetParamDouble("Sigma-Gauss"+postfix);
  } else { 
    BIASERR("something funny happend, maybe Gauss is not initialized");
  }

  if (para.CheckParam("Ratio-Gauss"+postfix)){ 
    _GaussRatio=para.GetParamDouble("Ratio-Gauss"+postfix);
  } else { 
    BIASERR("something funny happend, maybe Gauss is not initialized");
  }

  ///////////////////////////////////////////////////////////////////////
  //  now get the parameter from mean average
  ///////////////////////////////////////////////////////////////////////

  if (para.CheckParam("Size-Mean"+postfix)){ 
    _MeanSize=para.GetParamInt("Size-Mean"+postfix);
  } else { 
    BIASERR("something funny happend, maybe Mean is not initialized");
  }
#undef DFLT_PYSIZE 
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
PadPoints_(const int start)
{
  const int pysize = (int)LastImage_->size();
  const int pyindex = *PyIndex_;
  const int num = static_cast<int>(Points_[pyindex].size());
  for (int i=start; i<num; i++){
    for (int p=pyindex; p<pysize; p++){
      Points_[p][i][2] = 0.0;
    }
  }
}


//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

template class Tracker<unsigned char, float>; 
template class Tracker<float, float>; 

// fill in instances as required
#ifdef BUILD_IMAGE_INT
template class Tracker<unsigned char, int>; 
#endif
#ifdef BUILD_IMAGE_CHAR
template class Tracker<unsigned char, char>; 
#endif
#ifdef BUILD_IMAGE_SHORT
#endif
#ifdef BUILD_IMAGE_USHORT
template class Tracker<unsigned short, float>; 
#ifdef BUILD_IMAGE_INT
template class Tracker<unsigned short, int>; 
#endif
#endif
#ifdef BUILD_IMAGE_UINT
#endif
#ifdef BUILD_IMAGE_DOUBLE
#endif

