/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef __TrackerBaseAffine_hh__
#define __TrackerBaseAffine_hh__

namespace BIAS {

  /** @class TrackerBase
      @ingroup g_tracker
      @brief  
      @author woelk/kollmann 06/2006 */
  template <class StorageType>
  class BIASMatcher2D_EXPORT TrackerBaseAffine 
    : public TrackerBaseInterface<StorageType>
  {
  public :
    TrackerBaseAffine();


    /** @brief get results of affine tracking */
    inline void GetAffineResults(std::vector<Matrix2x2<KLT_TYPE> > & Results) {
      Results = _AffineResults;
    }

    /** @brief get result of affine tracking */
    inline void GetAffineResult(Matrix2x2<KLT_TYPE> & Result) {
      BIASASSERT(_AffineResults.size()>0);
      Result = _AffineResults[0];
    }

    inline void SetMaxAffineChange(const KLT_TYPE maxerr) { 
      *_MaxAffineChange=maxerr; }
  protected:
     /// iteration stops if position refinement is less than *_Maxerror
    KLT_TYPE *_MaxAffineChange; 
    /// results after affine tracking
    std::vector<Matrix2x2<KLT_TYPE> > _AffineResults;

    /** p1 has length 2, 2d coo
        p2 has length 2 or 6, 2d coo and affine parameters
        result has length 6, 2d coo and affine parameters
        @author woelk/koeser/kollmann 06/2006*/
    virtual int Track_(Vector<KLT_TYPE>& p1, Vector<KLT_TYPE>& p2,
                       Vector<KLT_TYPE>& result, KLT_TYPE& error, 
                       int &iter) = 0;

    /** @brief track using affine warp
        @param p1 point in image 1
        @param p2 prediction for point in image 2
        @param p2A prediction for affine warp
        @param result result after tracking
        @param resultA affine warp after tracking
        @param error maxnorm of last iteration step
        @param iter number of iterations done
        @param Cov resulting covariance
        @param sad sum of absolute differences
        @param ssd sum of squared differences
        @param Only4 only estimate A, not dx,dy
        @param ComputeCovariance, if true Cov is computed using svd

        This implementation is based on Thorsten Thormaelen's version,
        see cpp file.

        @todo really compute cov with svd ???, conceptual problem: where to 
        compute residui, where covariance ???

        Not evaluated, tested or optimized by now !
        @author koeser 06/2005
    */
    int TrackAffine_(KLT_TYPE p1[2], KLT_TYPE p2[2], 
                     const Matrix2x2<KLT_TYPE>& p2A,
                     KLT_TYPE result[2],  Matrix2x2<KLT_TYPE>& resultA,
                     KLT_TYPE& error, int &iter, Matrix<KLT_TYPE>& Cov,
                     KLT_TYPE& mad, KLT_TYPE& msd,
                     bool Only4 = false, bool ComputeCovariance = true);

     // ------------ affine worker functions ----------------------------
    void _am_getGradientWinAffine(const KLT_TYPE& x, const KLT_TYPE& y,
                                  const KLT_TYPE& Axx, const KLT_TYPE& Ayx,
                                  const KLT_TYPE& Axy, const KLT_TYPE& Ayy,   
                                  const int halfwidth, const int halfheight, 
                                  KLT_TYPE* out_gradx,     
                                  KLT_TYPE* out_grady);   

    void _am_computeIntensityDifferenceAffine(Image<StorageType>& img1,
                                              Image<StorageType>& img2,
                                              const KLT_TYPE& x1, 
                                              const KLT_TYPE& y1,    
                                              const KLT_TYPE& x2, 
                                              const KLT_TYPE& y2,     
                                              const KLT_TYPE& Axx, 
                                              const KLT_TYPE& Ayx, 
                                              const KLT_TYPE& Axy, 
                                              const KLT_TYPE& Ayy,   
                                              const int halfwidth,
                                              const int halfheight, 
                                              KLT_TYPE* imgdiff);  

  }; 
}

void _am_compute4by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw, 
                                const int hh,
                                BIAS::Vector<KLT_TYPE>& e);

void _am_compute4by4GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw,
                                   const int hh,
                                   BIAS::Matrix<KLT_TYPE>& T);

int  _am_gauss_jordan_elimination(BIAS::Matrix<KLT_TYPE>& a,
			     BIAS::Vector<KLT_TYPE>& b);
void _am_compute6by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw,
                                const int hh,
                                BIAS::Vector<KLT_TYPE>& e);

void _am_compute6by6GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw, 
                                   const int hh,
                                   BIAS::Matrix<KLT_TYPE>& T);


#endif // __TrackerBaseAffine_hh__
