/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TrackerBaseInterface__hh__
#define __TrackerBaseInterface__hh__

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Debug/Debug.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Math/Matrix2x2.hh>
#include <Base/Image/Image.hh>
#include <Utils/Param.hh>
#include <MathAlgo/Median1D.hh>

#include <vector>

#define KLT_TYPE double

#define D_TRACKERB_INIT        0x00000001
#define D_TRACKERB_KLT         0x00000002
#define D_TRACKERB_BILINEAR    0x00000004
#define D_TRACKERB_INPUT       0x00000008
#define D_TRACKERB_REJECT      0x00000010
#define D_TRACKERB_AFFINE      0x00000020

#define TRACKER_SUCCESS         0 
#define TRACKER_NOSTRUCTURE    -1 
#define TRACKER_OUTOFIMAGE     -2
#define TRACKER_TOOMANYITER    -3
#define TRACKER_OUTOFROI       -4 
#define TRACKER_MEANABSDIF     -5
#define TRACKER_IGNOREPOINT    -6 
#define TRACKER_X84            -7 
#define TRACKER_ERRORINCREASED  1



namespace BIAS {

  enum ERejectionType {RT_Invalid=-1, RT_MAD, RT_X86};

  std::ostream& operator<<(std::ostream& os, const enum ERejectionType& rt);

  /** @class TrackerBaseInterface
      @ingroup g_tracker
      @brief  

      Base class for the different tracking algorithms, defining the interfaces
      for the tracking functions. Also stores some basic options and datas, 
      i.e. the pointer to the image datas. See ExampleTrackerBaseInterface for brief 
      exemplary usage.

      \todo Implement DLL - add lost of missing explicit instantiations (JW)

      @author woelk 09/2004 */
  template <class StorageType>
  class BIASMatcher2D_EXPORT TrackerBaseInterface : public Debug
  {
  public :
    TrackerBaseInterface();

    TrackerBaseInterface(Param& para, const std::string& postfix="");

    TrackerBaseInterface(Param& para, const int grp, const std::string& postfix);

    ~TrackerBaseInterface();

    /** @brief Track from im1 to im2, 
        gradx1/2: horizontal gradient imgs
        grady1/2: vertical gradient imgs
        The ROI must be set correctly in gradient images (representing the 
        gradient filter mask size). 
        For good results im1 and im2 should be lowpass filtered images. */
    void Init(Image<StorageType>& im1, Image<StorageType>& im2, 
              Image<StorageType>& gradx1, Image<StorageType>& grady1, 
              Image<StorageType>& gradx2, Image<StorageType>& grady2);
    
    /** Sets _ComputeFilteredImages to true as well as 
        _SobelBinomial and _MasksSeparable accordingly
        @author kollmann 06/2006 */
    void Init(Image<StorageType>& im1, Image<StorageType>& im2,
              FilterMask& lowpass_mask, FilterMask& gradx_mask, 
              FilterMask& grady_mask);

    /** Calculates correspondence from image1 to image2. The two images and 
        corresponding gradients must specified using the function Init()
        before calling this function.
        
        The algorithm searches for best correspondence to point p1 from image1 
        in image2. The search starts at p2 in image 2.
        Upon completition 
        - result      : contains the best match, 
        - error       : contains the change in displacement calculated in 
        the last iteration, 
        - iter        : contains then number of iterations used,
        - residuumMAD : contains the mean absolute grey value difference 
        between p1 in image 1 and p2tracked in image 2
        over the tracking support window,
        - residuumMSD : contains the squared grey value difference between p1 
        in image 1 and p2tracked in image 2 summed over the 
        tracking support window,
        - cov         : contains an approximation to the covarinace matrix
        image1 and result in image2 over the tracking window ( = sum of
        absolute difference divided by number of pixels in tracking value ).
            
        All homogenous points must be homogenized (i.e. w=1.0).
        
        The algorithms either terminates if the displacement resulting from
        one iteration drops below *_Maxerror or the maximum number of 
        iterations *_MaxIterations is reached.

        @return   see defines
        - 0 on success (error< maxerror)
        - -1 when no spatial structure is present
        - -2 when point slid out of image
        - -3 when maxiter is reached and error is above maxerror
        - -4 when a point lies outside of valid region in images
        - -5 when point was rejected by MAD criterion
        - -6 when input point was at infinity
        - -7 when point is rejected by X86
        - 1 when error increased from last iteration 
        @author woelk */
    int Track(HomgPoint2D& p1, HomgPoint2D& p2, HomgPoint2D& p2tracked, 
              KLT_TYPE& error, int &iter, KLT_TYPE& residuumMAD,
              KLT_TYPE& residuumMSD, Matrix<KLT_TYPE>& cov,
              const Matrix2x2<KLT_TYPE> & AffinePred =
              Matrix2x2<KLT_TYPE>());

    /** Same as Track(), but a correspondence is searched for every point 
        in the vector p1. */
    void Track(std::vector<HomgPoint2D>& p1, 
               std::vector<HomgPoint2D>& p2, 
               std::vector<HomgPoint2D>& p2tracked, 
               std::vector<KLT_TYPE>& error, 
               std::vector<int>& numiter,
               std::vector<KLT_TYPE>& residuiMAD, 
               std::vector<KLT_TYPE>& residuiMSD, 
               std::vector<int>& res,
               std::vector<Matrix<KLT_TYPE> >& cov,
               const std::vector<Matrix2x2<KLT_TYPE> >& AffinePred =
               std::vector<Matrix2x2<KLT_TYPE> >());

    ////////////////////////////////////////////////////////////////////////

    inline void SetMaxError(const KLT_TYPE maxerr) { *_MaxError=maxerr; }
    
    inline void SetMaxResiduumMAD(const KLT_TYPE maxres) 
    { *_MaxResiduumMAD=maxres; }

    inline void SetMaxIterations(const int maxiter){ *_MaxIterations=maxiter; }
    
    inline void SetHalfWinSize(const int hws) { *_HalfWinSize=hws; }
    
    /// these are needed in the highlevel Tracker class
    inline int *GetMaxIterationsPointer() {return _MaxIterations;};
    inline int *GetHalfWinSizePointer() {return _HalfWinSize;};
    inline KLT_TYPE *GetMaxErrorPointer() {return _MaxError;};

    inline int *GetRejectionType() { return _RejectionType; };
    inline void SetRejectionType(const int rejection_type) 
    { *_RejectionType = rejection_type; };

  protected:
    bool _ParamsAllocated;
    int *_MaxIterations; /// iteration stops after _MaxIterations
    int *_HalfWinSize; /// use support window of size 2*hws+1
    /// iteration stops if position refinement is less than *_MaxError
    KLT_TYPE *_MaxError; 

    /// outlier rejection via MAD 
    KLT_TYPE *_MaxResiduumMAD; 
    /// The rejection type: valid RT_MAD and RT_X86
    int *_RejectionType;

    ///////////////////////////////////////////////////////////////////////////
    /// the image data arrays
    StorageType **_ida1, **_ida2;
    /// gradients image data arrays
    StorageType **_gradx1, **_grady1, **_gradx2, **_grady2;
    /// the image borders modulo filter and support windows or ROI
    /// for image 1 and image 2
    int _maxx1, _maxy1, _minx1, _miny1;
    int _maxx2, _maxy2, _minx2, _miny2;
    ///remember last support window size (for recalculation of gaussian weight)
    int _LastHalfWinSize, _WinSize;
    /// used in iteration
    Matrix2x2<KLT_TYPE> _G, _Ginv;
    Vector2<KLT_TYPE> _e, _d, _disp;
    Matrix<KLT_TYPE> _gx, _gy, _bl1, _bl2, _gx1, _gy1, _gx2, _gy2, _gsx, _gsy;
    KLT_TYPE _gxx, _gxy, _gyy, _idet;
    Matrix<KLT_TYPE> _weight; // the weight matrix used in TrackWeighted_

    /// used in RejectX86_
    Median1D<KLT_TYPE> _Median; // median calculation class
    std::vector<KLT_TYPE> _all_ssd_residui; // sum of sq. grey value diff
    /// sum of sq. grey value diff only for points where tracking succeeded
    std::vector<KLT_TYPE> _ssd_residui; 
    KLT_TYPE _median_ssd_residui; // median of _ssd_residui
    /// _diff_from_median = fabs(_ssd_residui - _median_ssd_residui)
    std::vector<KLT_TYPE> _diff_from_median;
    /// same as _diff_from_median but also for points where tracking was not
    /// sucessfull
    std::vector<KLT_TYPE> _all_diff_from_median;
    /// the rejection threshold calculated by X86 
    /// _threshold = 5.2 * median(_diff_from_median)
    KLT_TYPE _threshold; 

    /// pointers to images
    Image<StorageType> *_im1, *_im2, 
      *_gradim1x, *_gradim2x, *_gradim1y, *_gradim2y;

    /// filtered images are computed when set to true
    bool _ComputeFilteredImages; 
    /// uses specialized filter implementation with soebel 3x3 
    /// and 5x5 binomial filter mask
    bool _SobelBinomial;
    /// filter masks can be applied in separated manner when set to true
    bool _MasksSeparable;
    FilterMask _LowpassMask, _GradXMask, _GradYMask;
    

    ///////////////////////////////////////////////////////////////////
    //  initialization
    ///////////////////////////////////////////////////////////////////

    /** @brief set member pointers to NULL */
    void InitMembers_();

    /// minx/y maxx/y defines a ROI
    void Init_(StorageType **ida1, StorageType **ida2, 
               StorageType **gradx1, StorageType **grady1, 
               StorageType **gradx2, StorageType **grady2, 
               unsigned minx1, unsigned miny1, unsigned maxx1, unsigned maxy1,
               unsigned minx2, unsigned miny2, unsigned maxx2, unsigned maxy2);

    /** brief
        kollmann 06/2006 */
    void Init_(StorageType **ida1, StorageType **ida2, 
               unsigned minx1, unsigned miny1, unsigned maxx1, unsigned maxy1,
               unsigned minx2, unsigned miny2, unsigned maxx2, unsigned maxy2);

    ///////////////////////////////////////////////////////////////////
    //  tracking
    ///////////////////////////////////////////////////////////////////

    /** p1 has length 2, 2d coo
        p2 has length 2 or 6, 2d coo and affine parameters
        result has length 2 or 6, 2d coo and affine parameters
        @author woelk / kollmann 06/2006 */
    virtual int Track_(Vector<KLT_TYPE>& p1, Vector<KLT_TYPE>& p2,
                       Vector<KLT_TYPE>& result, KLT_TYPE& error, 
                       int &iter) = 0;

    ///////////////////////////////////////////////////////////////////
    //  calculation of cov, sad and ssd
    ///////////////////////////////////////////////////////////////////

    /** Uses _bl1, _bl2, _gxx, _gxy and _gyy to calculate the residui
        and covariance matrix and must hence be called immediatly after the 
        Track_() function call!!
        @author woelk */
    int EvaluateResult_(int& res,KLT_TYPE& mad,KLT_TYPE& msd, KLT_TYPE cov[4]);

    ///////////////////////////////////////////////////////////////////
    //  outlier rejection
    ///////////////////////////////////////////////////////////////////

    /** 
        @author woelk */
    int RejectMAD_(std::vector<KLT_TYPE>& sad, std::vector<int>& res);

    /** 
        @author woelk */
    int RejectX86_(std::vector<KLT_TYPE>& ssd, std::vector<int>& res);

    ///////////////////////////////////////////////////////////////////
    //  helper functions
    ///////////////////////////////////////////////////////////////////
    
    /** resizes the internal matrices _gx1, _gy1, _bl1, _gx2, _gy2 and _bl2
        @author woelk */
    inline void Resize_(const int halfwinsize);

    /** 
        @author woelk */
    void AddParam_(Param& para, const int grp, const std::string& postfix);
    
    ////////////////////////////////////////////////////////////////////
    //  bilinear interpolation functions 
    ////////////////////////////////////////////////////////////////////
    
    /// variables needed in MyBilienarRegion?_
    unsigned rf, rc, cf, cc, r, c, x_floor, y_floor, x_floor_end, y_floor_end;
    /// variables needed in MyBilienarRegion?_
    KLT_TYPE ul, ur, ll, lr, brdx, brdy, brdxy;
    
    /** dispatcher function for BilinearRegion*1_() functions
        @author woelk 06/2006 */
    inline void BilinearRegion1_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** dispatcher function for BilinearRegion*2_() functions 
        @author woelk 06/2006 */
    inline void BilinearRegion2_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx1, _gy1 and _bl1 by interpolating 
        _gradx1, _grady1 and _ida1 at positions between (x-hws, y-hws) and
        (x+hws, y+hws).
        @author woelk */
    inline void BilinearRegionPrefiltered1_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx2, _gy2 and _bl2 by interpolating 
        _gradx2, _grady2 and _ida2 at positions between (x-hws, y-hws) and
        (x+hws, y+hws).
        @author woelk */
    inline void BilinearRegionPrefiltered2_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx1, _gy1 and _bl1 by applying 
        _LowpassMask.GetKernel?(), _GradXMask.GetKernel?() and 
        _GradXMask.GetKernel?() at positions between (x-hws, y-hws) and
        (x+hws, y+hws) and interpolating afterwards .
        @author kollmann 06/2006 */
    inline void BilinearRegionFilterMask1_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx2, _gy2 and _bl2 by applying 
        _LowpassMask.GetKernel?(), _GradXMask.GetKernel?() and 
        _GradXMask.GetKernel?() at positions between (x-hws, y-hws) and
        (x+hws, y+hws) and interpolating afterwards .
        @author kollmann 06/2006  */
    inline void BilinearRegionFilterMask2_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx1, _gy1 and _bl1 by applying 
        _LowpassMask.GetSep??(), _GradXMask.GetSep??() and 
        _GradXMask.GetSep??() at positions between (x-hws, y-hws) and
        (x+hws, y+hws) and interpolating afterwards .
        @author kollmann 06/2006  */
    inline void BilinearRegionFilterVec1_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx2, _gy2 and _bl2 by applying 
        _LowpassMask.GetSep??(), _GradXMask.GetSep??() and 
        _GradXMask.GetSep??() at positions between (x-hws, y-hws) and
        (x+hws, y+hws) and interpolating afterwards .
        @author  */
    inline void BilinearRegionFilterVec2_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx1, _gy1 and _bl1 by applying 
        lowpass: [1 2 1], [1 2 1]^T , gradx: [1 0 -1], [1 2 1]^T and 
        grady:  [1 2 1], [1 0 -1]^T
        at positions between (x-hws, y-hws) and (x+hws, y+hws) and 
        interpolating afterwards .
        @author kollmann 06/2006  */
    inline void BilinearRegionFilterBinomial3Sobel3x3_1_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx2, _gy2 and _bl2 by applying 
        lowpass: [1 2 1], [1 2 1]^T , gradx: [1 0 -1], [1 2 1]^T and 
        grady:  [1 2 1], [1 0 -1]^T at positions between (x-hws, y-hws) and
        (x+hws, y+hws) and interpolating afterwards .
        @author kollmann 06/2006  */
    inline void BilinearRegionFilterBinomial3Sobel3x3_2_(KLT_TYPE x, KLT_TYPE y, int hws);


  }; // class

  ////////////////////////////////////////////////////////////////////
  //  inline functions
  ////////////////////////////////////////////////////////////////////

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  BilinearRegion1_(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    if (_ComputeFilteredImages){
      if (_SobelBinomial){
        BilinearRegionFilterSobel3x3Binomial1_(x, y, hws);
      } else {
        if (_MasksSeparabel){
          BilinearRegionFilterVec1_(x, y, hws);
        } else {
          BilinearRegionFilterMask1_(x, y, hws);
        }
      }
    } else {
      BilinearRegionPrefiltered1_(x, y, hws);
    }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  BilinearRegion2_(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    if (_ComputeFilteredImages){
      if (_SobelBinomial){
        BilinearRegionFilterSobel3x3Binomial2_(x, y, hws);
      } else {
        if (_MasksSeparabel){
          BilinearRegionFilterVec2_(x, y, hws);
        } else {
          BilinearRegionFilterMask2_(x, y, hws);
        }
      }
    } else {
      BilinearRegionPrefiltered2_(x, y, hws);
    }
  }
  
  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  BilinearRegionPrefiltered1_(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    x_floor = (unsigned int)floor(x);
    y_floor = (unsigned int)floor(y);
    x_floor_end = x_floor+hws+1;
    y_floor_end = y_floor+hws+1;
    brdy = y - (KLT_TYPE)(y_floor);
    brdx = x - (KLT_TYPE)(x_floor);
    brdxy = brdx*brdy;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++){
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++){
        ul = (KLT_TYPE)this->_ida1[rf][cf];
        ur = (KLT_TYPE)this->_ida1[rf][cc];
        ll = (KLT_TYPE)this->_ida1[rc][cf];
        lr = (KLT_TYPE)this->_ida1[rc][cc];
        _bl1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_gradx1[rf][cf]);
        ur = (KLT_TYPE)(this->_gradx1[rf][cc]);
        ll = (KLT_TYPE)(this->_gradx1[rc][cf]);
        lr = (KLT_TYPE)(this->_gradx1[rc][cc]);
        _gx1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_grady1[rf][cf]);
        ur = (KLT_TYPE)(this->_grady1[rf][cc]);
        ll = (KLT_TYPE)(this->_grady1[rc][cf]);
        lr = (KLT_TYPE)(this->_grady1[rc][cc]);
        _gy1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  BilinearRegionPrefiltered2_(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    x_floor = (unsigned int)floor(x);
    y_floor = (unsigned int)floor(y);
    x_floor_end = x_floor+hws+1;
    y_floor_end = y_floor+hws+1;
    brdy = y - (KLT_TYPE)(y_floor);
    brdx = x - (KLT_TYPE)(x_floor);
    brdxy = brdx*brdy;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++){
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++){
        ul = (KLT_TYPE)this->_ida2[rf][cf];
        ur = (KLT_TYPE)this->_ida2[rf][cc];
        ll = (KLT_TYPE)this->_ida2[rc][cf];
        lr = (KLT_TYPE)this->_ida2[rc][cc];
        _bl2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_gradx2[rf][cf]);
        ur = (KLT_TYPE)(this->_gradx2[rf][cc]);
        ll = (KLT_TYPE)(this->_gradx2[rc][cf]);
        lr = (KLT_TYPE)(this->_gradx2[rc][cc]);
        _gx2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_grady2[rf][cf]);
        ur = (KLT_TYPE)(this->_grady2[rf][cc]);
        ll = (KLT_TYPE)(this->_grady2[rc][cf]);
        lr = (KLT_TYPE)(this->_grady2[rc][cc]);
        _gy2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  Resize_(const int halfwinsize)
  {
    _LastHalfWinSize=halfwinsize;
    _WinSize=(halfwinsize<<1)+1;
    _gx1.newsize(_WinSize, _WinSize);
    _gy1.newsize(_WinSize, _WinSize);
    _gx2.newsize(_WinSize, _WinSize);
    _gy2.newsize(_WinSize, _WinSize);
    _gsx.newsize(_WinSize, _WinSize);
    _gsy.newsize(_WinSize, _WinSize);
    _bl1.newsize(_WinSize, _WinSize);
    _bl2.newsize(_WinSize, _WinSize);
  
  }


} // namespace



#include <Base/Common/BIASpragmaEnd.hh>

#endif // __TrackerBaseInterface__hh__
