/* This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <MathAlgo/UnscentedTransform.hh>
#include <MathAlgo/SVD.hh>
#include <MathAlgo/Lapack.hh>
#include <Base/Common/CompareFloatingPoint.hh>
#include <Base/Debug/Exception.hh>
#include <Base/Math/Matrix.hh>

using namespace BIAS;
using namespace std;

UnscentedTransform::
UnscentedTransform()
  : Debug(), Alpha_(0.001), Beta_(2.0), Kappa_(0.0)
{}


UnscentedTransform::
~UnscentedTransform()
{}


void UnscentedTransform::
Transform(const Vector<double>& src_mean, const Matrix<double>& src_cov,
          Vector<double>& dst_mean, Matrix<double>& dst_cov) const
{
  
  if (src_cov.IsZero(numeric_limits<double>::epsilon())){
    Transform_(src_mean, dst_mean);
    const int dim = dst_mean.Size();
    dst_cov.newsize(dim, dim);
    dst_cov.SetZero();
  } else {
    static vector<WeightedSigmaPoint> sigma_points;
    // compute the sigma points
    ComputeSigmaPoints_(src_mean, src_cov, sigma_points);

    // check result dimension
    Transform_(src_mean, dst_mean);
    const int dim = dst_mean.Size();
    // initialize result
    dst_mean.SetZero();
    dst_cov.newsize(dim, dim);
    dst_cov.SetZero();

    static Vector<double> diff, src;
    vector<WeightedSigmaPoint>::iterator it;
    for (it=sigma_points.begin(); it!=sigma_points.end(); it++){
      // transform the sigma points
      src = it->point;
      BIASCDOUT(D_UT_TRANSFORM, "src: "<<src<<endl);
      Transform_(src, it->point);
      BIASCDOUT(D_UT_TRANSFORM, "dst: "<<it->point<<endl);
      // compute weighted mean
      dst_mean += (it->point * it->weight_mean);
      BIASCDOUT(D_UT_TRANSFORM, "mean: "<<dst_mean<<endl);
    }
  
    for (it=sigma_points.begin(); it!=sigma_points.end(); it++){
      // compute the covariance
      diff = it->point - dst_mean;
      BIASCDOUT(D_UT_TRANSFORM, "diff: "<<diff<<endl);
      BIASCDOUT(D_UT_TRANSFORM, "weight "<< it->weight_cov
                << " * outer product: " <<  diff.OuterProduct(diff) << endl);
      dst_cov += it->weight_cov * diff.OuterProduct(diff);
      BIASCDOUT(D_UT_TRANSFORM, "cov: "<<dst_cov<<endl);
    }
  }
}


void UnscentedTransform::
ComputeSigmaPoints_(const Vector<double>& src_mean, 
                    const Matrix<double>& src_cov,
                    std::vector<WeightedSigmaPoint>& sigma_points) const
{
  const int dim = src_mean.Size();
  const double dbl_dim = static_cast<double>(dim);
  const double alpha_square = Alpha_ * Alpha_;
  const double lambda = alpha_square * ( dbl_dim + Kappa_ ) - dbl_dim; 
  const double dim_plus_lambda = dbl_dim + lambda;
  //cout << "dim_plus_lambda="<<dim_plus_lambda<<endl;
  BIASCDOUT(D_UT_PARAM, "Alpha: "<<Alpha_<<"  Beta : "<<Beta_
            <<"  Kappa: "<<Kappa_<<endl);
  sigma_points.resize(2 * dim + 1);

  static SVD svd;
  static Matrix<double> sqrt_cov;

  // numerically very unstable without normalization !!!
#define UT_NORMALIZE
#ifdef UT_NORMALIZE
  double factor = 0.0;
  for (int i=0; i<dim; i++)
    factor += src_cov[i][i];
  if (Less(fabs(factor), numeric_limits<double>::epsilon())){
    BEXCEPTION("cannot normalize zero covariance matrix "<<src_cov);
  }
  
  Matrix<double> normalized_cov = src_cov / factor;
  sqrt_cov = svd.SqrtT(normalized_cov);
  factor *= dim_plus_lambda;
  factor = sqrt(factor);
#else 
  sqrt_cov = svd.SqrtT(dim_plus_lambda * src_cov);
#endif  

  static Vector<double> vec;

  sigma_points[0].point = src_mean;
  sigma_points[0].weight_mean = lambda / dim_plus_lambda;
  sigma_points[0].weight_cov = 
    lambda / dim_plus_lambda + 1.0 - alpha_square + Beta_;
  int ind;
  for (int i=0; i<dim; i++){
    ind = i+1;

    vec = sqrt_cov.GetCol(i);
#ifdef UT_NORMALIZE
    vec *= factor;
#endif
    sigma_points[ind].point = src_mean + vec;
    sigma_points[ind+dim].point = src_mean - vec;
    sigma_points[ind].weight_mean = 
      sigma_points[ind+dim].weight_mean = 
      sigma_points[ind].weight_cov = 
      sigma_points[ind+dim].weight_cov = 
      1.0 / ( 2.0 * dim_plus_lambda );
  }
#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_UT_SIGMA_POINTS)){
    for (int k=0; k<2; k++){
      cout << "sp["<<k<<"] = "<<sigma_points[k].point<<" W_m = "
           <<sigma_points[k].weight_mean << endl<<" W_c = "
           <<sigma_points[k].weight_cov << endl;
    }
  }
#endif
}
