/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Image/Camera.hh>
#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageAttributes.hh>
#include <Geometry/PMatrix.hh>
#include <bias_config.h>
#include <stdio.h>

#ifdef   WIN32
#  include "Base/Common/getopt_W32.h"
#else  //not WIN32
#  include <getopt.h>
#endif //WIN32

#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;

void usage()
{
  cout << "Usage: biasconvert [options] infile outfile"<<endl;
  cout <<" Options: "<<endl
    <<" -g         : convert to grey values"<<endl
    //        <<" -d <int>   : downsample by power of 2"<<endl
    <<" -h         : show this help"<<endl
    <<" -p <file>  : add P-matrix"<<endl
    <<" -e Mode    : Mode=1 : compute K from EXIF data"<<endl
    <<" -v         : be verbose"<<endl
    <<" -c         : clip values to [0,255], only effective if -m or -o"
    << endl
    <<" -q quality : quality for lossy formats  [0;1] best=1"<<endl
    <<" -m <float> : modulate (scale) all pixel values with <float> (def:1)"
    <<endl
    <<" -z <float> : new value for zero (useful for inverting depth maps)"
    <<endl
    <<" -o <float> : offset (shift) all pixel values with <float> (def: 0)"
    <<endl
    <<" -M : use max dynamics after scaling, shifting and setting zero"
       <<endl
    <<" if neither -m nor -o is given float images are scaled "
    <<"and shifted automatically"<<endl
    <<endl;
}




int DetermineFiletype(const string &fn)
{
  return BIAS::ImageIO::GetFileFormat(fn);
}



void AdoptPMatrix(PMatrix &P,float scalex,float scaley)
{
  PMatrix *newP=NULL;
  Vector3<double> C;
  KMatrix K;
  Matrix3x3<double> R;

  if (P[0][0] != 0) {
    P.GetC(C);
    P.GetK(K);
    P.GetR(R);

    K[0][0] *= scalex;
    K[1][1] *= scaley;
    K[0][2] *= scalex;
    K[1][2] *= scaley;
    // the skew
    K[0][1] *= scaley;
    newP = new PMatrix(K,R,C);
    P = *newP;
  }
}



int ReadPMatrix(PMatrixBase &pmat,const string &filename)
{
  int row,col;
  ifstream pmatfile(filename.c_str());
  if (!pmatfile){
    cerr << "unable to open " << filename<< endl; 
    perror("");
    return -1;
  }
  pmatfile >> row >> col;
  if (!pmatfile){
    cerr << "error reading " << filename << endl;
    return -1;
  }
  if (row!=3 || col!=4){
    cerr << "P Matrix should have size 3x4, not: " <<row<<"x"<<col << endl;
    return -1;
  }


  for (int r=0; r < row; r++)
    for (int c=0; c< col; c++)
      pmatfile >> pmat[r][c];
  return 0;

}



/** @class biasconvert
@ingroup g_tools
@brief convert different image types
@author evers */
int main (int argc, char *argv[])
{
  //    pic.SetDebugLevel(D_CAM_IO);
  bool togrey = false;
  bool verbose = false;
  int EXIFMode = 0;
  string pmatfile="";
  string ifile,ofile;
  //long int sec,usec;
  //int downbypoweroftwo =0;
  //float scalex=1.0;
  //float scaley=1.0;
  float scale = 1.0;
  float shift = 0.0;
  float newzero = 0.0;
  bool replacezero = false;
  bool scaleshift = false;
  bool clip = false;
  bool maxDynamics = false;
  int imagequality = BIAS_DEFAULT_IMAGE_QUALITY;

  ImageBase ipic;
  Camera<unsigned char> ucocam;
  Camera<float> focam;
#ifdef BUILD_IMAGE_USHORT
  Camera<unsigned short> usocam;
#endif
  ImageBase::EStorageType OutST=ImageBase::ST_invalid;

  ImageIO::TFileFormat ofiletype;

  Camera<unsigned char> opic;
  PMatrix P_;


  if (argc <3) {
    usage();
    //exit(1);
    return 0; // for Dart testing
  }

  // scan options
  char c=0; 
  double quality=0;
  while (c>=0) {
    c = getopt(argc,argv,"Me:ghd:p:q:vm:o:cz:");
    switch (c) {
    case 'h':
    case ':':
    case '?':
      usage();
      exit(1);
      break;
    case 'e':// try to read exif data and make a K from it
      EXIFMode = atoi(optarg);
      if (EXIFMode>2) EXIFMode=2;
      if (EXIFMode<0) EXIFMode=0;
      break;
    case 'g':
      togrey = true;
      break;
      //     case 'd':
      //       downbypoweroftwo = atoi(optarg);
      //       scalex = 1.0/(float)(1<< downbypoweroftwo);
      //       scaley = 1.0/(float)(1<< downbypoweroftwo);
      //      break;
    case 'p':
      pmatfile=optarg;
      break;
    case 'q':
      quality = 100.0*float(atof(optarg));
      if (quality<0.0) quality=0.0;
      if (quality>100.0) quality=100.0;
      imagequality = (int)quality;
      break;
    case 'v':
      verbose = true;
      break;
    case 'c':
      clip = true;
      break;
    case 'o':
      shift = float(atof(optarg));
      scaleshift = true;
      break;
    case 'm':
      scale = float(atof(optarg));
      scaleshift = true;
      break;
    case 'z':
      replacezero = true;
      newzero =  float(atof(optarg));
      break;
    case 'M':
      maxDynamics = true;
      break;	
    }
  }

  // after getopt() all recognized options are before non-options
  ifile = argv[argc-2];
  ofile = argv[argc-1];

  ofiletype = (ImageIO::TFileFormat)DetermineFiletype(ofile);
  if ((ofiletype ==-1) || (ofiletype==ImageIO::FF_unknown))
  {
    BIASERR("unknown FileFormat"<<ofile);
    exit(1);
  }

  if (ofiletype == ImageIO::FF_pgm) 
    togrey = true;

  if (verbose) {
    cout <<"load "<<ifile<<", "<<flush;
  }

  if (ImageIO::Load(ifile,ipic)<0)
  {
    BIASERR("could not load"<<ofile);
    exit(1);
  }

  KMatrix KExif;
  if (EXIFMode>0) {
    //cout<<"reading exif tags"<<endl;
    ImageAttributes Att;
    Att.InitExifTags();
    if (Att.Read(ifile)!=BIAS_IMAGEATTRIBUTES_SUCCESS) {
      cerr<<"error reading exif tags ..."<<endl;
      EXIFMode = 0;
    } else {
      //Att.Print();
      KExif.SetIdentity();
      bool success = true;
      success = success && (Att.GetFocalLengthXPixel(KExif[0][0])>=0);
      success = success && (Att.GetFocalLengthYPixel(KExif[1][1])>=0);
      success = success && (Att.GetPrincipalPoint(KExif[0][2],
        KExif[1][2] )>=0);
      if (success) {
        cout<<"Found a K="<<KExif<<endl;
      } else {
        cout<<"Error, not enough EXIF data to get K:"<<KExif<<endl;
        EXIFMode = 0;
      }
    }
  }

  if (ipic.GetStorageType() == ImageBase::ST_float) {
    if (verbose) cout <<"storagetype is float"<<endl;
    Camera<float> fcam(ipic);
    fcam.ParseMetaData();
    P_ = fcam.GetP();

    if (ofiletype != ImageIO::FF_mip) {
      if (scaleshift){ 

        if (verbose) {
          float min,max;
          fcam.GetMinMaxPixelValue(min,max);
          cout << "min = " << min << ", max = " << max << endl;
        }

        float mean =0.0; 
        float *fcamptr=fcam.GetImageData();
        float num_elem=0;
        //float max=(255-shift)/scale;
        //float min=-shift/scale;
        for (unsigned int i = 0; i < fcam.GetWidth()*fcam.GetHeight(); i++){
          if (*fcamptr != -1.0){
            mean += *fcamptr;
            num_elem++;
          }
          fcamptr++;
        }

        mean /= num_elem;
        fcamptr=fcam.GetImageData();
        float variance=0.0;
        float scaledmean=0.0;
        for (unsigned int i = 0; i < fcam.GetWidth()*fcam.GetHeight();i++){
          if (*fcamptr != -1.0)
            variance += (*fcamptr-mean)*(*fcamptr-mean);
        }           

        fcam.ScaleShift(scale,shift);
        fcamptr=fcam.GetImageData();
        if (clip)           
          for (unsigned int i = 0; i < fcam.GetWidth()*fcam.GetHeight();i++){
            *fcamptr = (*fcamptr < 255) ? *fcamptr : 255 ;
            *fcamptr = (*fcamptr > 0) ? *fcamptr : 0 ;
            scaledmean += *fcamptr;
            fcamptr++;
          }

          if (verbose){
            variance/= num_elem;
            cout << "mean = " << mean  << ", scaledmean=" 
              << scaledmean/num_elem
              << ", variance = " 
              << variance <<  ", fill rate = " 
              <<  num_elem/fcam.GetWidth()/fcam.GetHeight() << endl;
          } 
      } else {
        if (verbose) {
          float min,max;
          fcam.GetMinMaxPixelValue(min,max);
          cout << "min = " << min << ", max = " << max << endl;
        }
        fcam.ScaleShiftBetween(0,255);
      }
      //ucocam = fcam;
      //OutST = ImageBase::ST_float;
      OutST = ImageBase::ST_unsignedchar;
      ImageConvert::ConvertST(fcam, ucocam, OutST );
      // copy meta data to ucocam
      *ucocam.GetMetaData()=*fcam.GetMetaData();
      ucocam.ParseMetaData();
    } else {
      focam = fcam;
      if (scaleshift) {
        focam.ScaleShift(scale,shift);
        if (replacezero) {/*
          if (newzero==0.0) {
            float min, max;
            // automatic value range detection
            focam.GetMinMaxPixelValue(min,max);
            newzero = min - (max-min / 20.0);
	    }*/
	  float *focamptr = focam.GetImageData(), 
            *fcamptr = fcam.GetImageData();
          for (unsigned int i = 0; i < focam.GetWidth()*focam.GetHeight();i++){
            if (*fcamptr++ == 255) {
              *focamptr = newzero; 
              //cout<<"0";
            }
            focamptr++;
          }
        }
      }
      OutST = ImageBase::ST_float; 
    }   
    
  } else if (ipic.GetStorageType() == ImageBase::ST_unsignedchar)  {
#ifdef BIAS_DEBUG
    if (verbose) {
      if (Debug::GetGlobalDebugLevel()<D_CONVERT)
        Debug::SetGlobalDebugLevel(D_CONVERT);
    }
#endif
    if (verbose) cout <<"storagetype is unsigned char"<<endl;
    Camera<unsigned char> uccam(ipic);
    OutST = ImageBase::ST_unsignedchar;       
    uccam.ParseMetaData();
    P_ = uccam.GetP();

    // all meta datas are preserved now (felix)
    if (verbose) cout <<"P: "<<P_<<endl;

    // meta datas are lost here
    if (togrey) ImageConvert::ToGrey(uccam, ucocam);
    else ImageConvert::ToRGB(uccam, ucocam);

    // copy meta data to ucocam
    *ucocam.GetMetaData()=*uccam.GetMetaData();
    ucocam.ParseMetaData();
  } 
#ifdef BUILD_IMAGE_USHORT
  else if (ipic.GetStorageType() == ImageBase::ST_unsignedshortint)  {
#  ifdef BIAS_DEBUG
    if (verbose) {
      if (Debug::GetGlobalDebugLevel()<D_CONVERT)
        Debug::SetGlobalDebugLevel(D_CONVERT);
    }
#  endif
    if (verbose) cout <<"storagetype is unsigned int"<<endl;
    Camera<unsigned short> uscam(ipic);
    Camera<unsigned char> tmpcam;
    OutST = ImageBase::ST_unsignedchar;       
    uscam.ParseMetaData();
    P_ = uscam.GetP();
    // uccam.GetTime(sec,usec);
    // all meta datas are preserved now (felix)
    if (verbose) cout <<"P: "<<P_<<endl;

    // now scale and shift if necessary
    if (scaleshift){
      Image<float> foim;
      ImageConvert::ConvertST(uscam, foim, ImageBase::ST_float);
      foim.ScaleShift(scale, shift);
      if (clip){
        foim.BelowThresholdToValue(0.0, 0.0);
        foim.AboveThresholdToValue(255.0, 255.0);
      }
      ImageConvert::ConvertST(foim, tmpcam, OutST);
    } else {
      ImageConvert::ConvertST(uscam, tmpcam, OutST);
    }

    // meta datas are lost here
    if (togrey) ImageConvert::ToGrey(tmpcam, ucocam);
    else ImageConvert::ToRGB(tmpcam, ucocam);

    // copy meta data to ucocam
    *ucocam.GetMetaData()=*uscam.GetMetaData();
    ucocam.ParseMetaData();
  }
#endif

  if (pmatfile != "") {
    if (verbose) cout <<"read "<<pmatfile<<", ";
    if (ReadPMatrix(P_,pmatfile) < 0)
      exit(1);
    ucocam.SetP(P_);
    focam.SetP(P_);
  }

  //   if (downbypoweroftwo >0) {
  //     if (verbose) cout <<"rescale, "<<flush;
  //     Filter f;
  //     Camera<unsigned char> tmppic;
  //     f.DownsampleByPowerOfTwo(opic,tmppic,downbypoweroftwo);
  //     opic = tmppic;
  //     AdoptPMatrix(P_, scalex,scaley);
  //   }


  if (OutST == ImageBase::ST_unsignedchar) {
    if (EXIFMode>0) ucocam.SetK(KExif);
    ucocam.UpdateMetaData();
    if (verbose) cout <<"write unsigned char "<<ofile<<", "<<flush;
    if (ofiletype == ImageIO::FF_mip) {
    if(maxDynamics)
      ucocam.ScaleShiftBetween(0,255);
      ImageIO::Write(ofile, ucocam);
    } else {
      //     if (ofiletype !=ImageIO::FF_ppm && ofiletype !=ImageIO::FF_pgm) {
      int endofprefix = ofile.rfind('.');
      ofile = ofile.substr(0,endofprefix);
      ImageIO::ExportImage(ofile,ucocam,ofiletype,false,imagequality);
    }
  } else if (OutST == ImageBase::ST_float) {
    // all meta datas are preserved now (felix)
    if (EXIFMode>0) focam.SetK(KExif);
    focam.UpdateMetaData();
    if(maxDynamics)
      focam.ScaleShiftBetween(0,255);
    ImageIO::Write(ofile, focam);
    if (verbose) cout <<"write float "<<ofile<<", "<<flush;
  } 
#ifdef BUILD_IMAGE_USHORT
  else if (OutST == ImageBase::ST_unsignedshortint) {
    if (EXIFMode>0) usocam.SetK(KExif);
    usocam.UpdateMetaData();
    if(maxDynamics)
      ucocam.ScaleShiftBetween(0,255);
    ImageIO::Write(ofile, usocam);
    if (verbose) cout <<"write unsigned short "<<ofile<<", "<<flush;
  }
#endif

  if (verbose) cout <<"done"<<endl;


  return 0;

}

