/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <wtypes.h>
#include <1394Camera.h>
#include "VideoSource_DCAM.hh"

using namespace BIAS;

using namespace std;

// tested with Scorpion, Dragonfly

VideoSource_DCAM::VideoSource_DCAM(int card): VideoSource(card)
{  
	// check if a camera is connected.
	int res = camera.CheckLink();
	if (res == CAM_SUCCESS)
	{
		BIASDOUT(D_DCAM,"A Camera was found and selected.");
		SelectedCam_=camera.GetNode();
	} 
	else 
	{
		BIASERR("No Camera was found.");
	}

  ForceF7_ = false;
  DirectF7_ = false;
  UserTopLeft_ = false;
  UseBinning_ = false;
  ColorMode_ = ImageBase::CM_UYVY422;
	DcamFps_ = 3;
	Width_ = 0;
	Height_ = 0;
	Left_ = 0;
	Top_ = 0;
  BytesPerPacket_ = 0;
  ColorChannels_ = 0;
}

VideoSource_DCAM::~VideoSource_DCAM()
{
}

int VideoSource_DCAM::OpenDevice(int camid)
{
  BIASDOUT(D_DCAM,"VideoSource_DCAM::OpenDevice("<<camid<<")");

  int select = camera.SelectCamera(camid);
  
  // is there a camera with id: camid?
  if (select == CAM_SUCCESS)
  {
    SelectedCam_ = select;
    BIASDOUT(D_DCAM,"VideoSource_DCAM::OpenDevice("<<camid<<") successful");
  } else {
    BIASERR("No Device found");   
    return -1; 
  }

  BIASDOUT(D_DCAM,"Setting up camera "<<SelectedCam_<<"");  

  // QueryCam(); is done by the camera

  return InitCam();
}

int VideoSource_DCAM::OpenDevice()
{
  return OpenDevice(0);
}

int VideoSource_DCAM::CloseDevice()
{
  return 0;
}

int VideoSource_DCAM::InitCam()
{	
	//Just for testing the code, tested with Scorpion, Dragonfly
	//Width_ = 400;
	//Height_ = 200;  
  //Top_ = 0;
  //Left_ = 0;
  //UserTopLeft_ = true;

	if(camera.InitCamera()==CAM_SUCCESS)
	{
    BIASDOUT(D_DCAM,"VideoSource_DCAM::InitCamera() successful");
	}
	else
	{
    BIASERR("VideoSource_DCAM::InitCamera() failed");
		return -1;
	}	

	// Look up camera Model and Vendor
	CamModel_ = GetModelForEuid(camera.m_UniqueID);

  if (Width_ != 0 || Height_ != 0) // UseBinning_ not implemented.
	{

		// try to set user definitions
		if ( UserSizeToMode() < 0) 
		{
			BIASERR("UserSizeToMode() failed");
			return -1;
    }
	}
  else
	{
		// Select best mode, format and framerate for camera.
		SelectDefault();
	}

	// set camera to parameters.
	if (camera.SetVideoFormat(DcamFormat_) == CAM_SUCCESS && 
		  camera.SetVideoMode(DcamMode_)		 == CAM_SUCCESS)
	{
		if (DcamFormat_ != 7)
		{
			camera.SetVideoFrameRate(DcamFps_);
		}
		else
		{				
			camera.m_controlSize.SetColorCode(0);
			camera.m_controlSize.SetSize(Width_,Height_);
	
			// Maximum bpp is selected.
			if (BytesPerPacket_ == 0) 
			{
				CalcBytesPerPacket();
			} 			

			camera.m_controlSize.SetBytesPerPacket(BytesPerPacket_);
			camera.m_controlSize.SetPosition(Left_,Top_);
		}

    CompleteInitialized_ = true;
		return 0;
	}

	// Initialization failed.
	BIASERR("Can not set DCAM Camera to requested parameters.");
	return -1;
}

int VideoSource_DCAM::GrabSingle(Camera <unsigned char> &image)
{

  BIASDOUT(D_DCAM,"VideoSource_DCAM::GrabSingle("<<SelectedCam_<<")");

	if (Grabbing_) 
	{
		BIASERR("Camera is already grabbing.");
	}

  Grabbing_ = true;

	// Start Image Capturing
	int res =	camera.StartImageCapture();

  if (res == CAM_SUCCESS) 
	{
    BIASDOUT(D_DCAM,"VideoSource_DCAM::StartImageCapture successful");
	}
	else
	{
    BIASERR("VideoSource_DCAM::StartImageCapture failed");		
	}

	// try to capture image.
	res = camera.CaptureImage();

	if (res == CAM_SUCCESS)
	{
    BIASDOUT(D_DCAM,"VideoSource_DCAM::CaptureImage successful");
	}
	else
	{
    BIASERR("VideoSource_DCAM::CaptureImage failed");		
	}

	// initialized with Width_, Height_
  image.CopyIn_NoInit(camera.m_pData);        
  
  // Stop Grabbing
	camera.StopImageCapture();
	Grabbing_=false;

  return 0;
}

int VideoSource_DCAM::QueryCam()
{  
  // is done automatically by the camera after initialisation
  return 0;
}

int VideoSource_DCAM::ShowCapabilities()
{  
  cout <<"Camera: "<<SelectedCam_<<endl;

  cout <<"Supported Formats: ";
  for (int i=0; i<8; i++)
  {
    if (camera.m_bxAvailableFormats[i]) 
    {
      cout <<"Format"<<i<<" ";
    }
  }
  cout <<endl;

	// Get supported modes for available formats.
  cout <<"Suported Modes:"<<endl;
  for (int i=0; i<8; i++)
  {
    if (camera.m_bxAvailableFormats[i])
    {
      cout <<"Format"<<i<<": ";
      for (int j=0; j<8; j++)
      {
        if (camera.m_bxAvailableModes[i][j])
        {
          cout <<"Mode"<<j<<" ";
        }      
      cout <<endl;      
			}
    }
  }
  
  // Only for Formats 0..2 is a framerate meaningful
  cout <<"Supported Framerates:"<<endl;
  for (int format=0; format < 3; format++)
  {  
    cout <<" Format"<<format<<": ";
    for (int mode=0; mode<8; mode++)
    {
      if (camera.m_bxAvailableModes[format][mode])
      {
        cout <<"Mode"<<i<<":  ";
        for (int f=0; f<8; f++)
        {
          if (camera.m_videoFlags[format][mode][f]) 
          {
            cout <<1.875 * (0x1<<f)<<"  ";
          }
        }
      }
    }    
    cout <<endl;
  }
  return 0;
}

#define USE_NATIVE_XGA

int VideoSource_DCAM::UserSizeToMode()
{
  BIASDOUT(D_DCAM,"VideoSource_DCAM::UserSizeToMode("<<SelectedCam_<<"): "<<Width_<<"x"<<Height_<<" BytesPerPixel_: "<<BytesPerPixel_<<" ColorMode "<<ColorMode_);

	if (BytesPerPixel_ == 1.0)
	{
    if (ColorMode_ == ImageBase::CM_UYVY422) {
      ColorChannels_ = 2;
      if (!ForceF7_) {
				if (Width_ == 1600 && Height_ == 1200) 
				{	
					DcamFormat_ = 2;
					DcamMode_   = 3;
				}
				else if (Width_ == 1280 && Height_ == 960) 
				{
					DcamFormat_ = 2;
					DcamMode_   = 0;
				}
				else if (Width_ == 1024 && Height_ == 768) 
				{
					DcamFormat_ = 1;
					DcamMode_   = 3;
				}
				else if (Width_ == 800 && Height_ == 600) 
				{
					DcamFormat_ = 1;
					DcamMode_   = 0;
				}
				else if (Width_ == 640 && Height_ == 480) 
				{
					DcamFormat_ = 0;
					DcamMode_   = 3;
				}
				else if (Width_ == 320 && Height_ == 240) 
				{
					DcamFormat_ = 0;
					DcamMode_   = 1;
				}
				else  // Format 7 required
				{
					if (UserSizeToFormat7_() < 0)
					{
						BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
						return -1;
					}
				}
			}
      else
			{
				if (UserSizeToFormat7_() < 0)
				{
					BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
					return -1;
				}
			}
    }
    else if (ColorMode_ == ImageBase::CM_Grey ||ColorMode_ == ImageBase::CM_Bayer_RGGB
																						  || ColorMode_ == ImageBase::CM_Bayer_GBRG
																							|| ColorMode_ == ImageBase::CM_Bayer_GRBG
																							|| ColorMode_ == ImageBase::CM_Bayer_BGGR )
		{ 
      ColorChannels_ = 1;
      if (!ForceF7_ ) {
				if (Width_ == 1600 && Height_ == 1200)
				{
				  DcamFormat_ = 2;
					DcamMode_   = 5;	
				}
				else if (Width_ == 1280 && Height_ == 960)
				{
					DcamFormat_ = 2;
					DcamMode_   = 2;	
				}	
				else if (Width_ == 1024 && Height_ == 768)
				{
				  DcamFormat_ = 1;
					DcamMode_   = 5;	
				}
				else if (Width_ == 640 && Height_ == 480 && CamModel_!=CAMERA_PTGREY_DRAGONFLY_HIRES)
				{
				  DcamFormat_ = 0;
					DcamMode_   = 5;	
				}
				else
				{
					if (UserSizeToFormat7_() < 0)
					{
						BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
						return -1;
					}
				}
      }
      else
			{
				if (UserSizeToFormat7_() < 0)
				{
					BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
					return -1;
				}
			}
    }
    else 
		{ 
      BIASERR("VideoSource_DCAM::UserSizeToMode(): Unsupported ColorModel_"
              <<ColorMode_);
      return -1;
    }
	}
	else
	{
		BIASDOUT(D_DCAM,"VideoSource_DCAM::UserSizeToMode(): 16 Bit not implemented.");	
		return -1;
	}
	return 0;
}

 
int VideoSource_DCAM::UserSizeToFormat7_()
{

	// Set Camera to Format7 to inquire unitwidth and height.
	camera.SetVideoFormat(7);
	camera.SetVideoMode(0);
	camera.m_controlSize.Inquire();

	UnitWidth_  = camera.m_controlSize.m_unitH;
	UnitHeight_ = camera.m_controlSize.m_unitV;

	MaxWidth_  = camera.m_controlSize.m_maxH;
	MaxHeight_ = camera.m_controlSize.m_maxV;

  if ((unsigned int)Width_ % UnitWidth_ >0)
	{
    BIASERR("VideoSource_DCAM::UserSizeToMode(): width of "<<Width_
	    <<" not multiple of: "<<UnitWidth_);
    return -1;
  }
  if ((unsigned int)Height_ % UnitHeight_ >0)
	{
    BIASERR("VideoSource_DCAM::UserSizeToMode(): height of "<<Height_
	    <<" not multiple of: "<<UnitHeight_);
    return -1;
  }

  if (!UseBinning_) 
	{
    DcamMode_   = 0;
    DcamFormat_ = 7;

    if (!UserTopLeft_) 
		{
      Left_ = (MaxWidth_ - Width_) / 2;
      Top_  = (MaxHeight_ - Height_) / 2;
    }
    BIASDOUT(D_DCAM_F7,"Using format7: "<<Width_<<"x"<<Height_<<"+"<<Left_
	     <<"+"<<Top_);
  }
  else 
	{ // Binning for PtGrey Dragonfly, not implemented yet.
    BIASDOUT(D_DCAM_F7,"Using format7 Mode 1 (Binning): "<<Width_<<"x"<<Height_
	     <<"+"<<Left_<<"+"<<Top_);

    DcamMode_      = 1; 
    DcamFormat_    = 7;
    ColorChannels_ = 1;
  }

  return 0;
}

int VideoSource_DCAM::SelectDefault()
{

	// Get best Mode for Camera
  for (int i=0; i<3; i++)
  {
    if (camera.m_bxAvailableFormats[i])
    {
      for (int j=0; j<6; j++)					// ToDo: No 16-Bits Modes allowed.
      {
        if (camera.m_bxAvailableModes[i][j])
        {
					DcamFormat_ = i;
					DcamMode_   = j;
        }         
			}
    }
  }

	// Select best Framerate for selected format and mode
	for (int k=0; k<8; k++)
	{
    if(camera.m_videoFlags[DcamFormat_][DcamMode_][k]) {
			DcamFps_ = k;
    }
	}

  switch (DcamFormat_) {
  case 0:  // Format 0
    switch (DcamMode_) {
    case 0:  // Mode 0
      ColorChannels_ = 2;
      Width_ = 160; Height_ = 120; ColorMode_ = ImageBase::CM_YUV444;
      break;
    case 1:  // F0 Mode 1
      ColorChannels_ = 2;
      Width_ = 320; Height_ = 240; ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 2:  // F0 Mode 2
      ColorChannels_ = 2;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_YUV411;
      break;
    case 3:  // F0 Mode 3
      ColorChannels_ = 2;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4:  // F0 Mode 4
      ColorChannels_ = 3;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5:  // F0 Mode 5
      ColorChannels_ = 1;
      Width_ = 640; Height_ = 480;
      break;
    case 6:     // F0 Mode 6   
      ColorChannels_ = 1;
      Width_ = 640; Height_ = 480;
      break;
    }
    
    break;
  case 1:   // Format 1
    switch (DcamMode_) {
    case 0: // F1 Mode 0
      ColorChannels_ = 2;
      Width_ = 800; Height_ = 600;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 1: // F1 Mode 1
      ColorChannels_ = 3;
      Width_ = 800; Height_ = 600;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 2: // F1 Mode 2
      ColorChannels_ = 1;
      Width_ = 800; Height_ = 600;  
      break;
    case 3:   // F1 Mode 3
      ColorChannels_ = 2;
      Width_ = 1024; Height_ = 768;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4: // F1 Mode 4
      ColorChannels_ = 3;
      Width_ = 1024; Height_ = 768;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5: // F1 Mode 5
      ColorChannels_ = 1;
      Width_ = 1024; Height_ = 768;  
      break;
    case 6: // F1 Mode 6
      ColorChannels_ = 1;
      Width_ = 800; Height_ = 600;  
      break;
    case 7: // F1 Mode 
      Width_ = 1024; Height_ = 768;  
      ColorChannels_ = 1;
      BytesPerPixel_ = 2.0;
      break;
    }
    
    break;
  case 2: // Format 2
    switch (DcamMode_) {
    case 0:
      ColorChannels_ = 2;
      Width_ = 1280; Height_ = 960;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 1:
      ColorChannels_ = 3;
      Width_ = 1280; Height_ = 960;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 2:
      ColorChannels_ = 1;
      Width_ = 1280; Height_ = 960; 
      break;
    case 3:
      ColorChannels_ = 2;
      Width_ = 1600; Height_ = 1200;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4:
      ColorChannels_ = 3;
      Width_ = 1600; Height_ = 1200;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5:
      ColorChannels_ = 1;
      Width_ = 1600; Height_ = 1200;  
      break;
    case 6:
      ColorChannels_ = 1;
      Width_ = 1280; Height_ = 960;  
      break;
    case 7:
      ColorChannels_ = 1;
      Width_ = 1600; Height_ = 1200;  
      break;
    }
    break;
  }

	return 0;
}

#ifdef BUILD_IMAGE_USHORT
int VideoSource_DCAM::GrabSingle(BIAS::Camera <unsigned short> &image)
{
  return 0;
}
#endif  

CameraModel VideoSource_DCAM::GetModelForEuid(LARGE_INTEGER /*euid*/) 
{

	// Look up for Model and Vendor
  char* search_vendor = "Point Grey Research";
	char* search_model = "Dragonfly";

	if (strstr(camera.m_nameVendor, search_vendor) != NULL)
	{
	  if (strstr(camera.m_nameModel, search_model) != NULL)
	  {
  	  unsigned long r1040;
	    camera.ReadQuadlet(0x1040, &r1040);  
      // little endian Red: 52 Green: 47 Blue: 42 Monochrome 59

      switch(r1040)
      { 	  
        case 0x52474742 :
          ColorMode_ = ImageBase::CM_Bayer_RGGB;	  
          break;
          
        case 0x47425247 :
          ColorMode_ = ImageBase::CM_Bayer_GBRG;	  
          break;
          
        case 0x47524247 :
          ColorMode_ = ImageBase::CM_Bayer_GRBG;	    
          break;
          
        case 0x42474752 :
          ColorMode_ = ImageBase::CM_Bayer_BGGR;	    
          break;       
          
        default :
					ColorMode_ = ImageBase::CM_Grey;
          break;                                                       
      }
  	  return CAMERA_PTGREY_DRAGONFLY_COLOR;
	  }
		else if(strstr(camera.m_nameModel, "Scorpion") != NULL)
		{
			ColorMode_ = ImageBase::CM_Bayer_RGGB;
			return CAMERA_PTGREY_SCORPION_COLOR_1600;
		}
	}
  // other camera	
  return CAMERA_UNKNOWN;	
}

void VideoSource_DCAM::ActivateScorpionTimestamp()
{
}

int VideoSource_DCAM::PreGrab()
{
  return 0;
}

int VideoSource_DCAM::PostGrab()
{
  return 0;
}

unsigned int VideoSource_DCAM::CalcBytesPerPacket()
{
	camera.m_controlSize.Status();
	BytesPerPacket_ = camera.m_controlSize.m_bytesPacketMax;
	return 0;
}

int  VideoSource_DCAM::ApplyBytesPerPacket()
{
	return 0;
}

int VideoSource_DCAM::StopIso() 
{
	return 0;
}

int VideoSource_DCAM::StartIso() 
{
	return 0;
}

float VideoSource_DCAM::GetShutter()
{  
  camera.m_controlShutter.Inquire();
	return camera.m_controlShutter.m_value1;
}

float VideoSource_DCAM::GetGain()
{  
  camera.m_controlGain.Inquire();
  return camera.m_controlGain.m_value1;
}

void VideoSource_DCAM::PrintFormatAndMode_()
{
  BIASDOUT(D_DCAM,"Format: <<DcamFormat_<< Mode: <<DcamMode_<< ");
}

int VideoSource_DCAM::SetGain(float gain)
{
  switch(CamModel_)
  {
    case CAMERA_PTGREY_DRAGONFLY_COLOR:

      camera.m_controlSize.Inquire();
      camera.m_controlGain.Inquire();

      // Calculation from Technical Reference Manual for 640x480 Dragonfly
      if (camera.m_controlSize.m_maxV == 640)
      {
        if(gain < 512)
        {
          gain = -6,0404 + (20*log(658+gain)*(658-gain));
        }
        else
        {
          gain = -6,0404 + (0,0354*gain);
        }
      }
      // The camera is the 1024x768 Dragonfly
      else
      {
        if(gain < 512)
        {
          gain = -9,40 + (20*log(658+gain)*(658-gain));
        }
        else
        {
          gain = 9,40 + (0,0354*gain);
        }    
      }
      break;

    default :
      BIASERR("VideoSource_DCAM::SetGain: Not implemented for this camera.");
      return -1;
      break;
  }

  // check if gain value can be set
  if ((int)gain < camera.m_controlGain.m_min || (int)gain > camera.m_controlGain.m_max)
  {
    BIASERR("VideoSource_DCAM::SetGain(): Value out of range.")
    return -1;
  }
  else if (!camera.m_controlGain.m_manual)
  {
    BIASERR("VideoSource_DCAM::SetGain(): Feature is not set to be controlled manually.")
    return -1;
  }
  else
  {
    camera.SetGain((int)gain);  
  }
	return 0;
}

int VideoSource_DCAM::SetWhiteBalance(float rvalue,float bvalue)
{
  camera.m_controlWhiteBalance.Inquire();

  // check if whiteBalance values can be set.
  if (rvalue < camera.m_controlWhiteBalance.m_min || rvalue > camera.m_controlWhiteBalance.m_max ||
      bvalue < camera.m_controlWhiteBalance.m_min || bvalue > camera.m_controlWhiteBalance.m_max)
  {
    BIASERR("VideoSource_DCAM::SetWhiteBalance(): Value out of range.")
    return -1;
  }
  else if (!camera.m_controlWhiteBalance.m_manual)
  {
    BIASERR("VideoSource_DCAM::SetWhiteBalance(): Feature is not set to be controlled manually.")
    return -1;
  }
  else
  {
    // conversion from float to int, maybe value must be computed.
    camera.SetWhiteBalance((int)rvalue, (int)bvalue);  
  }
	return 0;
}

int VideoSource_DCAM::OnePushAuto( int feature)
{
  int res = 0;

  switch(feature)
  {
    case FEATURE_BRIGHTNESS:
      res = camera.m_controlBrightness.SetOnePush();
      break;
    case FEATURE_EXPOSURE:
      res = camera.m_controlAutoExposure.SetOnePush();
      break;
    case FEATURE_SHARPNESS :
      res = camera.m_controlSharpness.SetOnePush();
      break;
    case FEATURE_WHITE_BALANCE :
      res = camera.m_controlWhiteBalance.SetOnePush();
      break;
    case FEATURE_HUE:
      res = camera.m_controlHue.SetOnePush();
      break;
    case FEATURE_SATURATION:
      res = camera.m_controlSaturation.SetOnePush();
      break;
    case FEATURE_GAMMA:
      res = camera.m_controlGamma.SetOnePush();
      break;
    case FEATURE_SHUTTER:
      res = camera.m_controlShutter.SetOnePush();
      break;
    case FEATURE_GAIN:
      res = camera.m_controlGain.SetOnePush();
      break;
    case FEATURE_IRIS:
      res = camera.m_controlIris.SetOnePush();
      break;
    case FEATURE_FOCUS:
      res = camera.m_controlFocus.SetOnePush();
      break;
    case FEATURE_TRIGGER:
      break;

    case FEATURE_ZOOM:
      res = camera.m_controlZoom.SetOnePush();
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::SetAutoMode(): Unkown Feature");
      return -1;
  }

  if (res != CAM_SUCCESS)
  {
    BIASERR("VideoSource_DCAM:SetOnePush(): failed")
    return -1;
  }

	return 0;
}

int VideoSource_DCAM::SetShutter(float exptime)
{
  camera.m_controlShutter.Inquire();

  switch(CamModel_)
  {
    case CAMERA_PTGREY_DRAGONFLY_COLOR:
      camera.m_controlSize.Inquire();

      // Calculation from Technical Reference Manual for 640x480 Dragonfly
      if (camera.m_controlSize.m_maxV == 640) 
      {
        int fps = camera.GetVideoFrameRate();
        exptime = (exptime*30)/(16000*(pow(2,fps)*1,875));
      }
      // Camera is the Dragonfly 1024x768
      else
      {
        int fps = camera.GetVideoFrameRate();
        exptime = (exptime*15)/(12000*(pow(2,fps)*1,875));        
      }
      break;

    default :
      BIASERR("VideoSource_DCAM::SetShutter: Not implemented for this camera.");
      return -1;
      break;
  }

  // check if shutter value can be set
  if ((int)exptime < camera.m_controlShutter.m_min || (int)exptime > camera.m_controlShutter.m_max)
  {
    BIASERR("VideoSource_DCAM::SetShutter(): Value out of range.")
    return -1;
  }
  else if (!camera.m_controlShutter.m_manual)
  {
    BIASERR("VideoSource_DCAM::SetShutter(): Feature is not set to be controlled manually.")
    return -1;
  }
  else
  {
    camera.SetShutter((int)exptime);  
  }
	return 0;
}

int VideoSource_DCAM::SetLeftTop(unsigned int left, unsigned int top)
{
  int res;

  if (left<=(MaxWidth_ - Width_) && top <=(MaxHeight_ - Height_)) {
    Top_ = top;
    Left_ = left;
    UserTopLeft_ = true;
  }
  else {
    BIASERR("VideoSource_DCAM::SetLeftTop() values too large:"<<left<<"x"<<top)
    return -1;
  }		    
  
  // try to set defined position.
  res = camera.m_controlSize.SetPosition(left,top);

  if (res == CAM_SUCCESS)
  {
    BIASDOUT(D_DCAM,"VideoSource_DCAM::SetLeftTop: successful");
    return 0;
  }
  else
  {
    BIASERR("VideoSource_DCAM::SetLeftTop: failed");
    return -1;
  }
}

unsigned int VideoSource_DCAM::GetFeature(int feature)
{
  int value = 0;

  switch(feature)
  {
    case FEATURE_BRIGHTNESS:
      camera.m_controlBrightness.Inquire();
      value = camera.m_controlBrightness.m_value1;
      break;
    case FEATURE_EXPOSURE:
      camera.m_controlAutoExposure.Inquire();
      value = camera.m_controlAutoExposure.m_value1;
      break;
    case FEATURE_SHARPNESS :
      camera.m_controlSharpness.Inquire();
      value = camera.m_controlSharpness.m_value1;
      break;
    case FEATURE_WHITE_BALANCE :
      camera.m_controlWhiteBalance.Inquire();
      value = camera.m_controlWhiteBalance.m_value1;
      break;
    case FEATURE_HUE:
      camera.m_controlHue.Inquire();
      value = camera.m_controlHue.m_value1;
      break;
    case FEATURE_SATURATION:
      camera.m_controlSaturation.Inquire();
      value = camera.m_controlSaturation.m_value1;
      break;
    case FEATURE_GAMMA:
      camera.m_controlGamma.Inquire();
      value = camera.m_controlGamma.m_value1;
      break;
    case FEATURE_SHUTTER:
      camera.m_controlShutter.Inquire();
      value = camera.m_controlShutter.m_value1;
      break;
    case FEATURE_GAIN:
      camera.m_controlGain.Inquire();
      value = camera.m_controlGain.m_value1;
      break;
    case FEATURE_IRIS:
      camera.m_controlIris.Inquire();
      value = camera.m_controlIris.m_value1;
      break;
    case FEATURE_FOCUS:
      camera.m_controlFocus.Inquire();
      value = camera.m_controlFocus.m_value1;
      break;
    case FEATURE_TRIGGER:
      break;

    case FEATURE_ZOOM:
      camera.m_controlZoom.Inquire();
      value = camera.m_controlZoom.m_value1;
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::GetFeature(): Unkown Feature");
      break;
  }
  return value;
}


int VideoSource_DCAM::SetFeature(int feature, unsigned int value)
{
  switch(feature)
  {
    case FEATURE_BRIGHTNESS:
      camera.SetBrightness(value);
      break;
    case FEATURE_EXPOSURE:
      camera.SetAutoExposure(value);
      break;
    case FEATURE_SHARPNESS :
      camera.SetSharpness(value);
      break;
    case FEATURE_WHITE_BALANCE :
      camera.SetWhiteBalance(value,value);
      break;
    case FEATURE_HUE:
      camera.SetHue(value);
      break;
    case FEATURE_SATURATION:
      camera.SetSaturation(value);
      break;
    case FEATURE_GAMMA:
      camera.SetGamma(value);
      break;
    case FEATURE_SHUTTER:
      camera.SetShutter(value);
      break;
    case FEATURE_GAIN:
      camera.SetGain(value);
      break;
    case FEATURE_IRIS:
      camera.SetIris(value);
      break;
    case FEATURE_FOCUS:
      camera.SetFocus(value);
      break;
    case FEATURE_TRIGGER:
      break;

    case FEATURE_ZOOM:
      camera.SetZoom(value);
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::SetFeature(): Unkown Feature");
      return -1;
  }
	return 0;
}

int VideoSource_DCAM::AutoMode( int feature, bool on)
{
  int res = 0;

  switch(feature)
  {
    case FEATURE_BRIGHTNESS:
      res = camera.m_controlBrightness.SetAutoMode(on);
      break;
    case FEATURE_EXPOSURE:
      res = camera.m_controlAutoExposure.SetAutoMode(on);
      break;
    case FEATURE_SHARPNESS :
      res = camera.m_controlSharpness.SetAutoMode(on);
      break;
    case FEATURE_WHITE_BALANCE :
      res = camera.m_controlWhiteBalance.SetAutoMode(on);
      break;
    case FEATURE_HUE:
      res = camera.m_controlHue.SetAutoMode(on);
      break;
    case FEATURE_SATURATION:
      res = camera.m_controlSaturation.SetAutoMode(on);
      break;
    case FEATURE_GAMMA:
      res = camera.m_controlGamma.SetAutoMode(on);
      break;
    case FEATURE_SHUTTER:
      res = camera.m_controlShutter.SetAutoMode(on);
      break;
    case FEATURE_GAIN:
      res = camera.m_controlGain.SetAutoMode(on);
      break;
    case FEATURE_IRIS:
      res = camera.m_controlIris.SetAutoMode(on);
      break;
    case FEATURE_FOCUS:
      res = camera.m_controlFocus.SetAutoMode(on);
      break;
    case FEATURE_TRIGGER:
      break;

    case FEATURE_ZOOM:
      res = camera.m_controlZoom.SetAutoMode(on);
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::SetAutoMode(): Unkown Feature");
      return -1;
  }

  if (res != CAM_SUCCESS)
  {
    BIASERR("VideoSource_DCAM:AutoMode(): failed")
  }

	return 0;
}
