/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/videodev.h>
#include <errno.h>


#include <Image/Camera.hh>
#include "VideoSource_V4L.hh"

using namespace BIAS;
using namespace std;

VideoSource_V4L::VideoSource_V4L(): VideoSource()
{
  ColorMode_ = ImageBase::CM_BGR;
  DefaultDevice_ = "/dev/video";
  ActiveFrame = 0;
  SetDebugLevel(0);
  GrabbingContinous_ = false;
  // try PAL as default
  Width_=768;
  Height_=576;
  ColorChannels_=3;
  DeviceChannel_=-1;
  newFrame_=false; 
  pthread_mutex_init(&ActiveFrameMutex_,NULL);
  numBuffers_=4;
  bufferLastCaptured_=0;
  waitForNew_=true;
}

VideoSource_V4L::~VideoSource_V4L()
{
}



int VideoSource_V4L::ShowCapabilities()
{  
  return ShowCapabilities(DefaultDevice_);
  
}


void printFormatInfo_(struct v4l2_pix_format &format)
{  
  cout<<"width:           "<<format.width            <<endl;
  cout<<"height:          "<<format.height           <<endl;
  cout<<"pixelformat:     ";
  __u32 pf = format.pixelformat; 
  cout << char((pf & 0x000000FF)>>0)  << char((pf & 0x0000FF00)>>8) <<  
    char((pf & 0x00FF0000)>>16) << char((pf & 0xFF000000)>>24)<<endl;
  cout<<"Interlace Mode:  ";
  if (format.field ==V4L2_FIELD_NONE) 
    cout<<"non interlaced"<<endl;
  else if (format.field ==V4L2_FIELD_INTERLACED)
    cout<<"interlaced"<<endl;
  else cout<<format.field<<endl;
  
}


int VideoSource_V4L::ShowCapabilities(const char *device)
{
  if ((fd = open(device, O_RDWR))<0) {
    BIASERR("VideoSource_V4L::ShowCapabilities():");
    perror(device);
    return -1;
  }

  if (ioctl(fd,VIDIOC_QUERYCAP,&capability2)<0) {
    perror("ioctl VIDIO_QUERYCAP");
    exit(1);
  }   

  cout << device << ":\n";
  cout << "driver:       " << capability2.driver     << "\n";
  cout << "Card:         " << capability2.card       <<endl;
  cout << "Bus:          " << capability2.bus_info   <<endl;
  cout<<  "Version:      " << capability2.version    <<endl;
  cout<<  "Capabilities: "<<endl; 
  if (capability2.capabilities & V4L2_CAP_VIDEO_CAPTURE) 
    cout<<"The device supports the video capture interface."<<endl;
  if (capability2.capabilities & V4L2_CAP_VIDEO_OUTPUT)
    cout<<"The device supports the video output interface."<<endl;
  if (capability2.capabilities &V4L2_CAP_VIDEO_OVERLAY)
    cout<<"The device supports the video overlay interface. Overlay "
      "typically stores captured images directly in the video memory "
      "of a graphics card, with support for clipping."<<endl;
  if (capability2.capabilities & V4L2_CAP_VBI_CAPTURE) 
    cout<<"The device supports the VBI capture interface"<<endl;  
  if (capability2.capabilities & V4L2_CAP_TUNER) 
    cout<<"The device has a tuner of some form"<<endl;
  if (capability2.capabilities & V4L2_CAP_AUDIO)
    cout<<"The device has audio inputs or outputs."<<endl;
  
  if (capability2.capabilities &V4L2_CAP_READWRITE)
    cout<<"The device supports the read() and/or write() I/O methods."<<endl;
  if (capability2.capabilities & V4L2_CAP_ASYNCIO)
    cout<<"The device supports the asynchronous I/O methods."<<endl;
  if (capability2.capabilities & V4L2_CAP_STREAMING)
    cout<<"The device supports the streaming I/O method."<<endl;

  // if CAPTURE device print info about all inputs
  int index=0, current=-1;
  cout<<endl;
  struct v4l2_input input;
  memset (&input, 0, sizeof (input));
  input.index = index;
  if (ioctl (fd,VIDIOC_G_INPUT, &current)!=0)
    perror("couldn't get current active input number");
  
  while (ioctl (fd, VIDIOC_ENUMINPUT, &input)==0) {
    cout<<"Input nr: "<<index<<endl;   
    cout<<"Name:      "<<input.name       <<endl;
    if (input.type == V4L2_INPUT_TYPE_CAMERA)
      cout<<"Type: CAMERA"<<endl;
    if (input.type == V4L2_INPUT_TYPE_TUNER)
      cout<<"Type: TUNER"<<endl;
    cout<<"Supported standards:";           
    if (input.std & V4L2_STD_PAL )
      cout<<"PAL ";
    if (input.std & V4L2_STD_NTSC )
      cout<<"NTSC ";
    if (input.std & V4L2_STD_SECAM )
      cout<<"SECAM ";
    cout<<endl;
    if (index==current) {
      cout<<"this is the current active input"<<endl;
      if  (input.status & V4L2_IN_ST_NO_POWER) 
        cout<<"Status: No Power"<<endl;
      if  (input.status & V4L2_IN_ST_NO_SIGNAL) 
        cout<<"Status: No Signal"<<endl;
      if  (input.status & V4L2_IN_ST_NO_COLOR) 
        cout<<"Status: No Color"<<endl;
      if (input.status == 0)
        cout<<"Status: POWER ON, SIGNAL and COLOR"<<endl;
    }
    cout<<endl;
    index++;
    input.index = index;
    usleep(10000); //wait 10ms 
  }
  if (index==0) {
    cerr<<"No Input found. Error:";
    if (errno==EBUSY) cout<<"EBUSY"<<endl;
    if (errno==EINVAL) cout<<"EINVAL"<<endl;
    else cout<<errno<<endl;
  }

  memset (&format2_, 0, sizeof (format2_));  
  format2_.type =V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if (ioctl(fd,VIDIOC_G_FMT,&format2_)!=0)
    perror("couldn't get format \n");
  else {
    printFormatInfo_(format2_.fmt.pix);
  }

  close(fd);
  return 0;
  
}



int VideoSource_V4L::OpenDevice()
{
  return OpenDevice(DefaultDevice_);
}


int VideoSource_V4L::OpenDevice(const char *device)
{
  if ((fd = open(device, O_RDWR))<0) {
    perror(device);
    return -1;
  }

  if (ioctl(fd,VIDIOC_QUERYCAP,&capability2)<0) {
    perror("ioctl VIDIO_QUERYCAP");
    exit(1);
  }   

  // at first try to select the input/channel where a signal is
  // only if none ist set by the user

  int index =0;

  if (DeviceChannel_<0) {
    struct v4l2_input input;   
    int res;
    res=ioctl (fd,VIDIOC_S_INPUT, &index);
    if (res!=0)
      cerr<<"VIDIOC_S_INPUT: couldn't set active input:"<<index<<endl;    
    bool found =false;
    while (!found && res==0) {
      // check wether there is a signal and color
      memset (&input, 0, sizeof (input));
      input.index = index;      
      usleep(50000);
      if (ioctl (fd, VIDIOC_ENUMINPUT, &input)==0) {
        if (input.status==0) {
          found=true; // input with a valid signal found
          cout<<"valid signal on input channel:"<<index<<" found."<<endl;
        }
      }
      if (!found) {
        usleep(50000);
        index++;
        res=ioctl (fd,VIDIOC_S_INPUT, &index);
        if (res!=0)
          cerr<<"VIDIOC_S_INPUT: couldn't set active input:"<<index<<endl;    
      }
    }
    if (!found) {
      BIASERR("no valid input signal found on all "<<index<<" channels");    
      return -2;
    }
  }
  

  // set PAL if possible 
  v4l2_std_id palId=V4L2_STD_PAL;
  if (ioctl(fd,VIDIOC_S_STD,&palId)!=0) {
    perror("couldn't set PAL\n");      
    return -3;
  }

  //// get the current image format, then alter it and set the edited
  memset (&format2_, 0, sizeof (format2_));  
  format2_.type =V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if (ioctl(fd,VIDIOC_G_FMT,&format2_)!=0) {
    perror("couldn't get format \n");      
    return -3;
  }
  struct v4l2_format format2Copy;
  memcpy(&format2Copy, &format2_,  sizeof (format2_));

   switch (ColorMode_ )
    {
    case ImageBase::CM_RGB:
      BIASDOUT(D_V4L, "Grabbing RGB which is ");
      format2Copy.fmt.pix.pixelformat=v4l2_fourcc('R','G','B','3');
      ColorChannels_ = 3;
      BytesPerPixel_=3;
      //      BIASWARN("the Sony PTZ do not support RGB, use to BGR");
      break;
    case ImageBase::CM_BGR:
      BIASDOUT(D_V4L, "Grabbing BGR");
      format2Copy.fmt.pix.pixelformat=v4l2_fourcc('B','G','R','3');
      ColorChannels_ = 3;
      BytesPerPixel_=3;
      break;
    case ImageBase::CM_YUYV422:
      BIASDOUT(D_V4L,  "Grabbing YUV");
      format2Copy.fmt.pix.pixelformat=V4L2_PIX_FMT_YUV422P;
      ColorChannels_ = 2;
      BytesPerPixel_=2;
      break;
    case ImageBase::CM_Grey:
      BIASDOUT(D_V4L, "Grabbing Grey");
      format2Copy.fmt.pix.pixelformat=V4L2_PIX_FMT_GREY;
      ColorChannels_ = 1;
      BytesPerPixel_=1;
      break;
    case ImageBase::CM_RGBA:
      BIASWARN("the Sony PTZ do not support RGBA, use BGRA");
      break;
    case ImageBase::CM_BGRA:
      BIASDOUT(D_V4L, "Grabbing BGRA");
      format2Copy.fmt.pix.pixelformat=v4l2_fourcc('B','G','R','4');
      ColorChannels_ = 4;
      BytesPerPixel_=4;
      break;
    default:
      BIASERR("VideoSource_V4L::OpenDevice(): Unsupported colormode");
      return -5;
    }
   
   format2Copy.fmt.pix.width=Width_;
   format2Copy.fmt.pix.height=Height_;
   //   format2_.fmt.pix.field=V4L2_FIELD_NONE; // try non-interlaced

   if (ioctl(fd,VIDIOC_S_FMT,&format2Copy)!=0) {
     perror("couldn't set format: \n");      
     printFormatInfo_(format2Copy.fmt.pix);
     return -5;
   }
   Width_=format2Copy.fmt.pix.width;
   Height_=format2Copy.fmt.pix.height;

  //// get buffers for capturing
  memset (&reqBuf, 0, sizeof (reqBuf));
  reqBuf.count= numBuffers_;
  reqBuf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
  reqBuf.memory=V4L2_MEMORY_MMAP;
  if (-1 == ioctl (fd, VIDIOC_REQBUFS, &reqBuf)) {
    if (errno == EINVAL)
      printf ("Video capturing or mmap-streaming is not supported\n");
    else
      perror ("VIDIOC_REQBUFS");
    return -4;
  }
  // mmap the buffer
  for (unsigned int i = 0; i < reqBuf.count; i++) {    
    memset (&buffer_, 0, sizeof (buffer_));
    buffer_.type = reqBuf.type;
    buffer_.memory = V4L2_MEMORY_MMAP;
    buffer_.index = i;    
    if (-1 == ioctl (fd, VIDIOC_QUERYBUF, &buffer_)) {
      perror ("VIDIOC_QUERYBUF");
      exit (EXIT_FAILURE);
    }    
    bufLengths.push_back(buffer_.length); /* remember for munmap() */    
    bufStarts.push_back((char*)mmap (NULL, buffer_.length,
                              PROT_READ | PROT_WRITE, /* required */
                              MAP_SHARED,             /* recommended */
                              fd, buffer_.m.offset));
    
    if (bufStarts[i] == MAP_FAILED) {    
      perror ("mmap");
      exit (EXIT_FAILURE);
    }
  }

  // start capturing into the buffers
  if (-1 == ioctl (fd, VIDIOC_STREAMON, & reqBuf.type)) {
    perror ("VIDIOC_STREAMON");
    exit (EXIT_FAILURE);
  }
  CompleteInitialized_ = true;
  return 0;
}

int VideoSource_V4L::CloseDevice()
{
  if (Grabbing_) {
    BIASERR("VideoSource_V4L::CloseDevice(): Grabbing is running, but closing anyway");    
  }
  
  Grabbing_ = false;
  PostGrab();
  
  CompleteInitialized_ = false;

  if (-1 == ioctl (fd, VIDIOC_STREAMOFF, &reqBuf.type)) {
    perror ("VIDIOC_STREAMON");
    return -1;
  }
  for (unsigned int i = 0; i < bufLengths.size(); i++)
    munmap (bufStarts[i], bufLengths[i]);  

  close(fd);
  return 0;
  
}



int VideoSource_V4L::GrabSingle(Camera <unsigned char> &image)
{

  BIASDOUT(D_V4L,"VideoSource_V4L::GrabSingle()");
  assert(! image.IsEmpty());
  
  if (!Active_) {
    BIASERR("VideoSource_V4L::GrabSingle(): VideoSource Object not active, use PreGrab()");
    return -1;
  }
  
  if (Grabbing_) {
    BIASERR("VideoSource_V4L::GrabSingle(): VideoSource Object is already grabbing()");
    return -1;
  }
    
  Grabbing_ = true;

  if (GrabbingContinous_ ) { 
    if (waitForNew_)
      while (!newFrame_) usleep(5);
    // locking not necessary because more than 2 buffers are used fro grabbing
    //pthread_mutex_lock(&ActiveFrameMutex_); 
    gettimeofday(&GrabTime_,NULL);
    image.SetTime(GrabTime_.tv_sec,GrabTime_.tv_usec);
    newFrame_=false;
    image.CopyIn_NoInit((unsigned char*)bufStarts[bufferLastCaptured_]);

    //    pthread_mutex_unlock(&ActiveFrameMutex_); 
  } else { 
    buffer_.index=0; // always use first buffer, if not in continous mode
    if (ioctl(fd, VIDIOC_QBUF, &buffer_)!=0) {
      perror("ioctl: freeing the buffer memory\n");
      return -3;
    }
    errno=0;
    // wait for a new frame
    while ( (errno!=EINVAL) &&
            (ioctl(fd, VIDIOC_DQBUF, &buffer_)!=0) ) {
      if (errno==EINVAL) {
        perror("ioctl: waiting for a new frame \n");
        return -4;
      } 
    }  
    gettimeofday(&GrabTime_,NULL);
    image.SetTime(GrabTime_.tv_sec,GrabTime_.tv_usec);
    image.CopyIn_NoInit( (unsigned char*)bufStarts[0]);
  }

  Grabbing_ = false;

  image.InvalidateUID();
  BIASDOUT(D_V4L,"leaving GrabSingle()");

  return 0;
}

int VideoSource_V4L::StartGrabContinous(){
  GrabbingContinous_ = true;
  return pthread_mutex_init(&ActiveFrameMutex_, NULL) || 
    pthread_cond_init(&ThreadCondition_, NULL) || 
    pthread_create(&ContinousThread_, NULL, GrabContThread_ , this);
}


int VideoSource_V4L::StopGrabContinous(){
  int res = 0;

  GrabbingContinous_ = false;

  if ( (res = pthread_join(ContinousThread_, NULL)) != 0){
    perror("pthread_join");
    BIASERR("VideoSource_V4L::StopGrabContinous()  pthread_join unsucessfull "<< res);
    res--;
  }
  pthread_mutex_unlock(&ActiveFrameMutex_);
  if (pthread_mutex_destroy(&ActiveFrameMutex_) != 0){
    perror("pthread_mutex_destroy");
    BIASERR("VideoSource_V4L::StopGrabContinous()  pthread_mutex_destroy unsucessfull");
    res--;
  }
  if (pthread_cond_destroy(&ThreadCondition_) != 0){
    perror("pthread_cond_destroy");
    BIASERR("VideoSource_V4L::StopGrabContinous()  pthread_cond_destroy unsucessfull");
    res--;
  }

  return res;
}





void *VideoSource_V4L::GrabContThread_(void * arg){
  VideoSource_V4L *obj = (VideoSource_V4L *) arg;


  // free all buffers such that capturing may start
  for (unsigned int i=0;i<obj->numBuffers_;i++) {
    obj->buffer_.index=i;
    if (ioctl(obj->fd, VIDIOC_QBUF, & obj->buffer_)!=0) {
      perror("ioctl: initial freeing of all buffer memories\n");
      return (void *)NULL;
    }
    usleep(10);
  }


  while (obj->GrabbingContinous_){
    errno=0;
    // wait for a new frame    
    obj->buffer_.index=obj->bufferLastCaptured_+1;
    if (obj->buffer_.index >= obj->numBuffers_)
      obj->buffer_.index=0;
    while ( (errno!=EINVAL) &&
            (ioctl(obj->fd, VIDIOC_DQBUF, & obj->buffer_)!=0) ) {
      if (errno==EINVAL) {
        perror("ioctl: waiting for a new frame \n");        
        return (void *)NULL;
      }
      usleep(20);
    }    
  
    pthread_mutex_lock(&obj->ActiveFrameMutex_);   
    obj->bufferLastCaptured_++; // switch between 0 and 1 with xor
    if (obj->bufferLastCaptured_ >= (int)obj->numBuffers_)
      obj->bufferLastCaptured_=0;
    obj->newFrame_=true; 
    pthread_mutex_unlock(&obj->ActiveFrameMutex_);    
    
    obj->buffer_.index=obj->bufferLastCaptured_+1;
    if (obj->buffer_.index >= obj->numBuffers_)
      obj->buffer_.index=0;
    if (ioctl(obj->fd, VIDIOC_QBUF, & obj->buffer_)!=0) {
      //      perror("ioctl: freeing the buffer memory\n");
      //      return (void *)NULL;
    }
    usleep(10);
 
  }

  pthread_exit(NULL);
}





